/*
 *	CHR vector font stuff.
 *	Original copyright: Copyright (c) 1988,89 Borland International
 *	Fixes: Vladimir A. Smirnov, 2000
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#define CHR_PREFIX_SIZE		0x80
#define CHR_MAJOR_VERSION	1
#define CHR_MINOR_VERSION	0
#define CHR_SIGNATURE		'+'

/*
 *	Opcodes of stroke
 */
#define	CHR_END_OF_CHAR			0
#define	CHR_DO_SCAN				1
#define	CHR_MOVE				2
#define	CHR_DRAW				3

#pragma pack(push)
#pragma pack(1)

typedef struct	chr_font_prefix_s
{
	char		sig;			/* SIGNATURE byte				*/
	short		nchrs;			/* number of characters in file	*/
	char		reserved1;		/* Currently undefined			*/
	char		first;			/* first character in file		*/
	short		cdefs;			/* offset to char definitions	*/
	char		scan_flag;		/* True if set is scanable		*/
	char		org_to_cap;		/* Height from origin to top of capitol */
	char		org_to_base;	/* Height from origin to baseline	*/
	char		org_to_dec;		/* Height from origin to bot of decender*/
	char		fntname[4];		/* Four character name of font	*/
	char		reserved2;		/* Currently undefined			*/
}	chr_font_prefix_t, *chr_font_prefix_p;

typedef struct chr_font_header_s
{
	unsigned short	header_size;		/* Version 2.0 Header Format */
	unsigned char	font_name[4];		/* Font Internal Name		*/
	unsigned short	font_size;			/* Size in byte of file 	*/
	unsigned char font_major, font_minor; /* Driver Version Information */
	unsigned char min_major, min_minor;	/* BGI Revision Information	*/
}	chr_font_header_t, *chr_font_header_p;

typedef struct chr_font_stroke_s
{
	char		opcode;			/* Stroke opcode byte			*/
	short		x;				/* Relative offset x direction	*/
	short		y;				/* Relative offset y direction	*/
}	chr_stroke_t, *shr_stroke_p;

#pragma pack(pop)


FILE	*ffile, *OFile;

char	    *Font;				/* Pointer to font storage	*/
char		Prefix[CHR_PREFIX_SIZE];	/* File Prefix Holder		*/
chr_font_prefix_t Header; 		/* File Data Header		*/

short		Offset[256];		/* Font data offsets		*/
char		Char_Width[256];	/* Character Width Table	*/
short		Stroke_Count[256];	/* Stroke Count Table		*/
chr_stroke_t   *Strokes[256];	/* Stroke Data Base			*/

char  *OpName[] = {
  "End    ",
  "Do Scan",
  "Move To",
  "Line To"
  };

char help[] = "Dump Font File Copyright (c) 1987,1989 Borland International, Inc.\n\n"
	      "Usage is:  DFONT [font file name] \n";

void dvalue( char *str, unsigned short value );

int unpack( char *buf, short index, chr_stroke_t **new );

int decode( unsigned short *iptr, short *x, short *y );

void main( int argc, char **argv )
{
  long length, current, base;
  char *cptr;
  chr_font_header_t *fptr;
  short last_chr, i, j;
  chr_stroke_t *sptr;

  if( argc < 2 || argc > 3 ){		/* Invalid 3 of arguments	*/
    fprintf( stderr, help);
    exit( 1 );				/* Exit the program		*/
    }

  ffile = fopen( argv[1], "rb" );       /* Open the input font file     */
  if(!ffile )
  {
    fprintf( stderr, "\nFontDisplay: Can not open input file %s.\n", argv[1] );
    exit( 1 );				/* Exit the program		*/
  }

  if( 3 == argc ){			/* Is there an output file?	*/
    OFile = fopen( argv[2], "rb" );       /* Open the input font file     */
    if( NULL == OFile ){		  /* Can not open font file	  */
      fprintf( stderr, "\nFontDisplay: Can not open output file %s.\n", argv[2] );
      exit( 1 );			  /* Exit the program		  */
      }
    }
  else	OFile = stdout; 		/* Write it to screen		*/

/*	Read in and display the file prefix record.			*/

  base = ftell( ffile );		/* Remember the address of table*/
  fread(Prefix, CHR_PREFIX_SIZE, 1, ffile); /* Read in the file prefix	*/

  fprintf( stdout, "Prefix Record   File Base: %lx\n\n", base );

  cptr = Prefix;			/* Begin at base of prefix	*/
  while( 0x1a != *cptr ) ++cptr;	/* Move to EOF in prefix	*/
  *cptr = '\0';                         /* Cut prefix at EOF            */
  fptr = (chr_font_header_t*)(cptr+1);		/* Point at Font Header Record	*/

  fprintf( stdout, "Text:\n\n%s\n", Prefix );
  dvalue( "Prefix Size", fptr->header_size );
  fprintf( stdout, "Prefix Name: %-.4s\n", fptr->font_name );
  dvalue( "Font Size", fptr->font_size );
  fprintf( stdout, "Revision:    %d.%d\n", fptr->font_major, fptr->font_minor );
  fprintf( stdout, "BGI Version: %d.%d\n", fptr->min_major, fptr->min_minor );

/*	Read in and display the font header record.			*/

  base = ftell( ffile );		/* Remember the address of table*/
  fread(&Header, sizeof(chr_font_header_t), 1, ffile);  /* Read in the file prefix */

  fprintf( stdout, "\nHeader Record   File Base: %lx\n\n", base );

  fprintf( stdout, "Signature:    %c\n", Header.sig );
  fprintf( stdout, "# Characters: %d.\n", Header.nchrs );
  fprintf( stdout, "First Char:   %02x\n", Header.first );
  fprintf( stdout, "Definition Offset:   %04x\n", Header.cdefs );
  fprintf( stdout, "Scanable Font:       %s\n", Header.scan_flag ? "Yes" : "No" );
  fprintf( stdout, "Origin to Cap  Height: %d\n", Header.org_to_cap );
  fprintf( stdout, "Origin to Base Height: %d\n", Header.org_to_base );
  fprintf( stdout, "Origin to Dec  Height: %d\n", Header.org_to_dec );
  fprintf( stdout, "Header Name: %-.4s\n\n", Header.fntname );

/*	Read file offset table into memory.				*/

  base = ftell( ffile );		/* Remember the address of table*/
  fread( &Offset[Header.first], Header.nchrs, sizeof(int), ffile );

/*	Display the offset table					*/

  last_chr = Header.first + Header.nchrs;

  fprintf( stdout, "Offset Table    File Base: %lx\n", base );

  for( i=Header.first ; i<last_chr ; ++i ){
    if( !((i+3) % 4) ) fprintf( stdout, "\n" );
    fprintf( stdout, " %c (%02x) : %04x ",
		    isprint(i) ? i : '.', i, Offset[i] );
    }

/*	Load the character width table into memory.			*/

  base = ftell( ffile );
  fread( &Char_Width[Header.first], Header.nchrs, sizeof(char), ffile );

/*	Determine the length of the stroke database.			*/

  current = ftell( ffile );		/* Current file location	*/
  fseek( ffile, 0, SEEK_END );		/* Go to the end of the file	*/
  length = ftell( ffile );		/* Get the file length		*/
  fseek( ffile, current, SEEK_SET );	/* Restore old file location	*/

/*	Load the stroke database.					*/

  Font = malloc( (int) length );	/* Create space for font data	*/
  if( NULL == Font ){			/* Was there enough memory	*/
    fprintf( stderr, "FontDisplay: Not Enough Memory to load Font.\n\n" );
    exit( 1 );
    }

  fread( Font, (int)length, 1 , ffile ); /* Load the stroke data	*/

/*	Font is now loaded, display the internal data			*/

  fprintf( stdout, "\n\nWidth Table File Base:  %lx\n", base );
  fprintf( stdout, "Stroke Table File Base: %lx\n", current );
  fprintf( stdout, "\n\nStroke Information\n\n" );

  for( i=Header.first ; i<last_chr ; ++i ){
    if( !Offset[i] && i!=Header.first ) continue;
    printf( "Char %02x (%c)    ", i, i );
    Stroke_Count[i] = unpack( Font, Offset[i], &Strokes[i] );
    printf( "Offset: %04x   Width: %-5d   Stroke Count: %d\n",
      Offset[i], Char_Width[i], Stroke_Count[i] );
    sptr = Strokes[i];
    for( j=0 ; j<Stroke_Count[i] ; ++j, ++sptr ){
      printf( "  %3d : OpCode: %s (%d)   X: %4d   Y: %4d\n",
	j, OpName[sptr->opcode], sptr->opcode, sptr->x, sptr->y );
      }
    }

}

void dvalue( char *str, unsigned short value )
{

  fprintf( OFile, "%-50s  %04xh  (%6u. )\n", str, value, value );

}

/*	UNPACK: This routine decodes the file format into a more	*/
/*	reasonable internal structure					*/

int unpack( char *buf, short index, chr_stroke_t **new )
{
  unsigned short *pb;
  chr_stroke_t *po;
  short	  num_ops = 0;
  short	  jx, jy, opcode, i, opc;

  pb = (unsigned short*)(buf + index);	/* Reset pointer to buffer	*/

  for(;;)
  {			/* For each byte in buffer	*/
    num_ops += 1;			/* Count the operation		*/
    opcode = decode( pb++, &jx, &jy );	/* Decode the data record	*/
    if( opcode == CHR_END_OF_CHAR ) break;	/* Exit loop at end of char	*/
  }

  po = *new = calloc( num_ops, sizeof(chr_stroke_t) );

  if( !po ){				/* Out of memory loading font	*/
    fprintf( stderr, "DMPFNT: Out of memory decoding font\n" );
    exit( 100 );
    }

  pb = (unsigned short *)(buf + index);	/* Reset pointer to buffer	*/

  for( i=0 ; i<num_ops ; ++i ){ 	/* For each opcode in buffer	*/
    opc = decode(pb++, &po->x, &po->y); /* Decode the data field	*/
    po->opcode = opc;			/* Save the opcode		*/
    po++;
    }

  return( num_ops );			/* return OPS count		*/

}



/*	DECODE: This routine decodes a single word in file to a 	*/
/*	stroke record.							*/

int decode( unsigned short *iptr, short *x, short *y )
{
  struct DECODE {
    signed   short xoff  : 7;
    unsigned short flag1 : 1;
    signed   short yoff  : 7;
    unsigned short flag2 : 1;
  } cword;

  cword = *(struct DECODE *)iptr;

  *x = cword.xoff;
  *y = cword.yoff;

  return( (cword.flag1 << 1) + cword.flag2 );

}

