pbyte V_GLOBAL ppmLoad(char *name,int *width,int *height)
{
	file_c f;
	int w,h,s,i;
	char head[4];
	vec_t peak;
	pbyte pixels,ptr;
	char buffer[512];
	if(!name)
	{
		conlPrintf(LOG_ERROR,"ppmLoad([NULL],%X,%X): incorrect parameter(s).\n",
			width,height);
		return NULL;
	}
	if(!f.openread(name))
	{
		conlPrintf(LOG_ERROR,"ppmLoad('%s'): file not found.\n",name);
		return NULL;
	}
	strncpy(head,ppm_token(f,buffer),3);
	w=atoi(ppm_token(f,buffer));
	h=atoi(ppm_token(f,buffer));
	peak=(vec_t)atof(ppm_token(f,buffer));
	if((strcmp(head,"P3") && strcmp(head,"P6")) || !w || !h || peak<0.0 || peak>255.0)
	{
		conlPrintf(LOG_ERROR,"ppmLoad('%s'): isn't a ppm file.\n",name);
		return NULL;
	}
	if(*width) *width=w;
	if(*height) *height=h;
	s=w*h*4;
	if(head[1]=='3')
	{
		// ASCII data
		pixels=(pbyte)malloc(s);
		ptr=pixels;
		s>>=2;
		if(peak!=255.0)
		{
			peak=(vec_t)256.0/peak;
			for(i=0;i<s;i++)
			{
				ptr[0]=scalebyte(atoi(ppm_token(f,buffer)),peak);
				ptr[1]=scalebyte(atoi(ppm_token(f,buffer)),peak);
				ptr[2]=scalebyte(atoi(ppm_token(f,buffer)),peak);
				ptr[3]=GetThreadData()->constant_alpha;
				ptr+=4;
			}
		}
		else
		{
			for(i=0;i<s;i++)
			{
				ptr[0]=(byte)atoi(ppm_token(f,buffer));
				ptr[1]=(byte)atoi(ppm_token(f,buffer));
				ptr[2]=(byte)atoi(ppm_token(f,buffer));
				ptr[3]=GetThreadData()->constant_alpha;
				ptr+=4;
			}
		}
	} 
	else
	{
		// RAW data
		pixels=(pbyte)malloc(s);
		ptr=pixels;
		s>>=2;
		int c;
		while(((c=f.getc())==0x0D)||(c==0x0A));
		f.seek(-1,SEEK_CUR);
		if(peak!=255.0)
		{
			peak=(vec_t)256.0/peak;
			for(i=0;i<s;i++)
			{
				ptr[0]=scalebyte(f.getc(),peak);
				ptr[1]=scalebyte(f.getc(),peak);
				ptr[2]=scalebyte(f.getc(),peak);
				ptr[3]=GetThreadData()->constant_alpha;
				ptr+=4;
			}
		}
		else
		{
			for(i=0;i<s;i++)
			{
				ptr[0]=f.getc();
				ptr[1]=f.getc();
				ptr[2]=f.getc();
				ptr[3]=GetThreadData()->constant_alpha;
				ptr+=4;
			}
		}
	}
	f.close();
	return pixels;
}    
		
void V_GLOBAL ppmSave(char *name,int width,int height,byte *pixels)
{
	int i,s;
	file_c f;
	byte *ptr=pixels;
	char buf[20];
	s=width*height;
	if(!name || !width || !height || !pixels)
	{
		conlPrintf(LOG_ERROR,"ppmSave(%s,%d,%d,%X): incorrect parameter(s).\n",
			name ? name : "[NULL]",width,height,pixels);
		return;
	}
	if(!f.openwrite(name))
	{
		conlPrintf(LOG_ERROR,"ppmSave('%s'): can't create file.\n",name);
		return;
	}
	f.puts("P6");
	f.puts(itoa(width,buf,10));
	f.puts(itoa(height,buf,10));
	f.puts("255");
	for(i=0;i<s;i++)
	{
		f.write(ptr,3);
		ptr+=4;
	}
	f.close();
}

pbyte V_GLOBAL rgbaLoad(char *name,int *width,int *height)
{
	int w,h;
	file_c f;
	unsigned char *pixels;
	if(!name)
	{
		conlPrintf(LOG_ERROR,"rgbaLoad([NULL],%X,%X): incorrect parameter(s).\n",
			width,height);
		return NULL;
	}
	if(!f.openread(name))
	{
		conlPrintf(LOG_ERROR,"rgbaLoad('%s'): file not found.\n",name);
		return NULL;
	}
	f.read(&w,4);
	f.read(&h,4);
	if(width) *width=w;
	if(height) *height=h;
	pixels=(byte*)malloc(w*h*4);
	f.read(pixels,w*h*4);
	f.close();
	return pixels;
}

void V_GLOBAL rgbaSave(char *name,int width,int height,byte *pixels)
{
	file_c f;
	if(!name || !width || !height || !pixels)
	{
		conlPrintf(LOG_ERROR,"rgbaSave(%s,%d,%d,%X): incorrect parameter(s).\n",
			name ? name : "[NULL]",width,height,pixels);
		return;
	}
	if(!f.openwrite(name))
	{
		conlPrintf(LOG_ERROR,"rgbaSave('%s'): can't create file.\n",name);
		return;
	}
	f.write_32(width);
	f.write_32(height);
	f.write(pixels,width*height*4);
	f.close();
}
	
pbyte V_GLOBAL rgbaLoadChannel(char *name,int *width,int *height)
{
	int w,h;
	file_c f;
	unsigned char *ch_data;
	if(!name)
	{
		conlPrintf(LOG_ERROR,"rgbaLoadChannel([NULL],%X,%X): incorrect parameter(s).\n",
			width,height);
		return NULL;
	}
	if(!f.openread(name))
	{
		conlPrintf(LOG_ERROR,"rgbaLoadChannel('%s'): file not found.\n",name);
		return NULL;
	}
	f.read(&w,4);
	f.read(&h,4);
	if(width) *width=w;
	if(height) *height=h;
	ch_data=(byte*)malloc(w*h);
	f.read(ch_data,w*h);
	f.close();
	return ch_data;
}

void V_GLOBAL rgbaSaveChannel(char *name,int width,int height,byte *ch_data)
{
	file_c f;
	if(!name || !width || !height || !ch_data)
	{
		conlPrintf(LOG_ERROR,"rgbaSaveChannel(%s,%d,%d,%X): incorrect "
			"parameter(s).\n",name ? name:"[NULL]",width,height,ch_data);
		return;
	}
	if(!f.openwrite(name))
	{
		conlPrintf(LOG_ERROR,"rgbaSaveChannel('%s'): can't create file.\n",name);
		return;
	}
	f.write_32(width);
	f.write_32(height);
	f.write(ch_data,width*height);
	f.close();
}

void V_GLOBAL rgbaSplitChannels(int width,int height,byte *pixels,
	byte *r,byte *g,byte *b, byte *a)
{
	int i,n;
	byte *ptr;
	n=width*height;
	if(!n || !pixels)
	{
		conlPrintf(LOG_ERROR,
			"rgbaSplitChannels(%d,%d,%X,%X,%X,%X,%X): incorrect parameter(s).\n",
			width,height,pixels,r,g,b,a);
		return;
	}
	if(r)
	{
		ptr=pixels;
		for(i=0;i<n;i++,ptr+=4) r[i]=*ptr;
	}
	if(g)
	{
		ptr=pixels+1;
		for(i=0;i<n;i++,ptr+=4) g[i]=*ptr;
	}
	if(b)
	{
		ptr=pixels+2;
		for(i=0;i<n;i++,ptr+=4) b[i]=*ptr;
	}
	if(a)
	{
		ptr=pixels+3;
		for(i=0;i<n;i++,ptr+=4) a[i]=*ptr;
	}
}

void V_GLOBAL rgbaCombineChannels(int width,int height,byte *pixels,byte *r,
	byte *g,byte *b, byte *a,
	vec_t sr,vec_t sg,vec_t sb,vec_t sa)
{
	int i,n;
	byte *ptr;
	n=width*height;
	if(!n || !pixels)
	{
		conlPrintf(LOG_ERROR,
			"rgbaCombineChannels(%d,%d,%X,%X,%X,%X,%X): incorrect parameter(s).\n",
			width,height,pixels,r,g,b,a);
		return;
	}
	if(sr==1.0 && sg==1.0 && sb==1.0 && sa==1.0)
	{
		if(r)
		{
			ptr=pixels;
			for(i=0;i<n;i++,ptr+=4) *ptr=r[i];
		}
		if(g)
		{
			ptr=pixels+1;
			for(i=0;i<n;i++,ptr+=4) *ptr=g[i];
		}
		if(b)
		{
			ptr=pixels+2;
			for(i=0;i<n;i++,ptr+=4) *ptr=b[i];
		}
		if(a)
		{
			ptr=pixels+3;
			for(i=0;i<n;i++,ptr+=4) *ptr=a[i];
		}
	}
	else
	{
		if(r)
		{
			ptr=pixels;
			for(i=0;i<n;i++,ptr+=4) *ptr=saturate(sr*(vec_t)r[i]);
		}
		if(g)
		{
			ptr=pixels+1;
			for(i=0;i<n;i++,ptr+=4) *ptr=saturate(sg*(vec_t)g[i]);
		}
		if(b)
		{
			ptr=pixels+2;
			for(i=0;i<n;i++,ptr+=4) *ptr=saturate(sb*(vec_t)b[i]);
		}
		if(a)
		{
			ptr=pixels+3;
			for(i=0;i<n;i++,ptr+=4) *ptr=saturate(sa*(vec_t)a[i]);
		}
	}
}  

vboolean V_GLOBAL mipLoadLevel(miphead_t *mip,char *rgbafile,int level)
{
	int w,h;
	int tw,th;
	byte *tex;
	if(!rgbafile || !mip || level<0 || level>MAX_MIP_LEVELS)
	{
		conlPrintf(LOG_ERROR,"mipLoadLevel(%X,%s,%d): incorrect parameter(s).\n",
			mip,rgbafile ? rgbafile : "[NULL]",level);
		return false;
	}
	tex=rgbaLoad(rgbafile,&w,&h);
	if(!tex) return false;
	tw=mip->width>>level;
	th=mip->height>>level;
	if(w!=tw || h!=th)
	{
		conlPrintf(LOG_ERROR,"mipLoadLevel(): target is %d X %d, source %d X %d.\n",
			tw,th,w,h);
		return false;
	}
	v_copydw(mip->level[level],tex,w*h);
	rgbaFree(tex);
	return true;
}

void V_GLOBAL mipSaveLevel(miphead_t *mip,char *rgbafile,int level)
{
	if(!mip || !rgbafile || level<0 || level>MAX_MIP_LEVELS)
	{
		conlPrintf(LOG_ERROR,"mipSaveLevel(%X,%s,%d): incorrect parameter(s).\n",
			mip,rgbafile ? rgbafile : "[NULL]",level);
		return;
	}
	rgbaSave(rgbafile,mip->width>>level,mip->height>>level,mip->level[level]);
}

/*
u8_t	V_LOCAL
ClampFloatToByte(vec_t v)
{
	if(v>255)
		v=255;
	else
		if(v<0)
			v=0;
	return (u8_t)v;
}

u8_t	V_LOCAL
ScaleIntToByte(u32_t in,vec_t scale)
{
	return ClampFloatToByte(in*scale);
}
*/

