#include <windows.h>
#include <v_lib.h>

void		atomInit();
void		atomShutdown();
vboolean	atomAdd(char *key,u32_t value);
vboolean	atomGet(char *key,u32_t *dest);
vboolean	atomDel(char *key);

typedef struct atom_s
{
	u32_t	value;
	char	key[256];	/* Variable - sized */
}	atom_t, *atom_p;

#define		ATOM_TIMEOUT	INFINITE
HANDLE		at_lock=NULL;
vp_stack_t	atom_table;

void		atomInit()
{
	at_lock=CreateMutex(NULL,TRUE,NULL);
	vpsClear(&atom_table);
	ReleaseMutex(at_lock);
}

void		atomShutdown()
{
	unsigned i, num;
	HANDLE lock;
	atom_t *atom;
	while(WaitForSingleObject(at_lock,1000)!=WAIT_OBJECT_0)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_WARNING,"atomStudown(): waiting for "
			"completion in another threads...\n");
#endif	/* CHECKED_BUILD */
	}
	lock=at_lock;
	at_lock=NULL;
	num=vpsCount(&atom_table);
	for(i=0;i<num;i++)
	{
		atom=(atom_t*)vpsGet(&atom_table,i);
		if(atom)
		{
#ifdef			CHECKED_BUILD
			conlPrintf(LOG_WARNING,
				"atomStudown(): pair '%s'-'%X' removed.\n",
				atom->key,atom->value);
#endif			/* CHECKED_BUILD */
			free(atom);
		}
	}
	vpsClear(&atom_table);
	ReleaseMutex(lock);
	CloseHandle(lock);
}

#define		atomAquireMutex()\
			(WaitForSingleObject(at_lock,ATOM_TIMEOUT)==WAIT_OBJECT_0)

vboolean	atomAdd(char *key,u32_t value)
{
	unsigned i, num, pos;
	atom_t *atom;
	if(!key || !key[0])
	{
		/* invalid parameter */
		return false;
	}
	if(!atomAquireMutex())
	{
		/* can't lock atom table; op. pending (or, may be, not initialized ?) */
		return false;
	}
	num=vpsCount(&atom_table);
	pos=num;
	for(i=0;i<num;i++)
	{
		atom=(atom_t*)vpsGet(&atom_table,i);
		if(atom)
		{
			if(!strcmp(key,atom->key))
			{
				/* already exists */
				ReleaseMutex(at_lock);
				return false;
			}
		}
		else
			pos=i;
	}
	/* ok, key is uniqie */
	i=sizeof(u32_t)+strlen(key);
	while(i & 3)
		i++;
	atom=(atom_t*)malloc(i);
	atom->value=value;
	strcpy(atom->key,key);
	if(pos<num)
		vpsSet(&atom_table,pos,atom);
	else
		vpsPush(&atom_table,atom);
	ReleaseMutex(at_lock);
	return true;
}

vboolean	atomGet(char *key,u32_t *dest)
{
	unsigned i, num;
	atom_t *atom;
	if(!key || !key[0])
	{
		return false;
	}
	if(!atomAquireMutex())
	{
		/* can't lock atom table; op. pending (or, may be, not initialized ?) */
		return false;
	}
	num=vpsCount(&atom_table);
	for(i=0;i<num;i++)
	{
		atom=(atom_t*)vpsGet(&atom_table,i);
		if(atom && !strcmp(key,atom->key))
		{
			if(dest)
				*dest=atom->value;
			ReleaseMutex(at_lock);
			return true;
		}
	}
	ReleaseMutex(at_lock);
	return false;
}

vboolean	atomDel(char *key)
{
	unsigned i, num;
	atom_t *atom;
	if(!key || !key[0])
	{
		return false;
	}
	if(!atomAquireMutex())
	{
		/* can't lock atom table; op. pending (or, may be, not initialized ?) */
		return false;
	}
	num=vpsCount(&atom_table);
	for(i=0;i<num;i++)
	{
		atom=(atom_t*)vpsGet(&atom_table,i);
		if(atom && !strcmp(key,atom->key))
		{
			free(atom);
			vpsSet(&atom_table,i,NULL);
			ReleaseMutex(at_lock);
			return true;
		}
	}
	ReleaseMutex(at_lock);
	return false;
}
