/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../include/v_lib.h"
#include	"../source/v_mt.h"
#include	"v_os.h"

#ifdef	__WIN32__

/*	Maximal number of IPC channels */
#define MAX_IPC_CHANNELS	256
/*	Significant length of IPC channel name */
#define	MAX_IPC_NAME		_MAX_PATH

/*
 *	Bidirectional pipe
 */
typedef struct	duplex_pipe_s
{
	u8_fifo_t			recv_fifo;
	CRITICAL_SECTION	recv_access;
	u8_fifo_t			send_fifo;
	CRITICAL_SECTION	send_access;
	HANDLE				send_event;
}	duplex_pipe_t, *duplex_pipe_p;

/*
 *	Local channel data
 */
typedef	struct	ipc_local_s
{
	char		name[MAX_IPC_NAME];
	HANDLE		recv_event;
	vp_stack_t	pipes;
	/*	Server - side callbacks */
	void		(V_CALL *on_client_connect)(ipc_handle_t server_handle,u32_t cl_id);
	void		(V_CALL *on_client_disconnect)(ipc_handle_t server_handle,u32_t cl_id);
}	ipc_local_t, *ipc_local_p;


/*
 *	Static data accessible by all threads (IPC handles CAN be shared).
 */
static	ipc_local_p			ipc_channels[MAX_IPC_CHANNELS];
static	HANDLE				ipc_access=NULL;

/*
 *	Finds unused pos in ipc_channels.
 *	Must be called AFTER setting ipc_access to non-signaled state.
 */
u32_t	V_LOCAL
ipc_local_get_free_slot()
{
	u32_t slot;
	/* Slot #0 reserved for error detection */
	for(slot=1;slot<MAX_IPC_CHANNELS;slot++)
		if(!ipc_channels[slot])
			return slot;
	return 0;
}
	
/*
 *	Finds existing channel in ipc_channels with given name.
 *	Must be called AFTER setting ipc_access to non-signaled state.
 *	Assumes that 'name' is valid name (not NULL).
 */
u32_t		V_LOCAL
ipc_local_find_channel(char *name)
{
	u32_t slot;
	ipc_local_t *ptr;
	/* Slot #0 reserved for error detection */
	for(slot=1;slot<MAX_IPC_CHANNELS;slot++)
	{
		ptr=ipc_channels[slot];
		if(ptr && !strncmp(ptr->name,name,MAX_IPC_NAME-1))
			return slot;
	}
	return 0;
}

/*
 *	Disconnects given client 'cl_slot' from server 'slot'.
 *	Assumes that both 'slot' and 'cl_slot' are valid.
 */
void	V_LOCAL
ipc_local_disconnect_client(u32_t slot,u32_t cl_slot)
{
	ipc_local_t *obj;
	ipc_handle_t ipch;
	obj=ipc_channels[slot];
	if(obj->on_client_disconnect)
	{
		ipch=slot | (((cl_slot+1) << 16) & 0xFFFF0000);
		obj->on_client_disconnect(ipch,cl_slot);
	}
	u8fClear(&(obj->send_fifo[cl_slot]));
	CloseHandle(obj->send_access[cl_slot]);
	/* Release all waiting threads */
	SetEvent(obj->send_event[cl_slot]);
	CloseHandle(obj->send_event[cl_slot]);
	obj->send_access[cl_slot]=NULL;
	obj->send_event[cl_slot]=NULL;
	obj->num_clients--;
}

/*
 *	Initializes IPC subsystem.
 */
void	V_GLOBAL
ipcInit(void)
{
	if(ipc_access==NULL)
	{
		ipc_access=CreateSemaphore(NULL,1,1,NULL);
		v_zero(ipc_channels,MAX_IPC_CHANNELS*sizeof(ipc_local_p));
	}
}

/*
 *	Performs cleanup if no used slots.
 */
void	V_GLOBAL
ipcShutdown(void)
{
	u32_t slot, count;
	if(ipc_access)
	{
		/* Close storage access semaphore */
		WaitForSingleObject(ipc_access,INFINITE);
		/* Count used slots */
		for(count=1,slot=1;slot<MAX_IPC_CHANNELS;slot++)
			if(ipc_channels[slot])
				count++;
		if(!count)
		{
			/* There are no channels in use - perform cleanup */
			CloseHandle(ipc_access);
			ipc_access=NULL;
		}
	}
}

/*
 *	ipcServerEndpoint()
 */
 
ipc_handle_t	V_GLOBAL
ipcServerEndpoint(char *channel_name)
{
	u32_t slot;
	ipc_local_t *obj;
	if(!channel_name || channel_name[0]==0)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcServerEndpoint('%s'): invalid channel name.\n",
			dbs(channel_name));
#endif	/* CHECKED_BUILD */
		return 0;
	}
	if(!ipc_access)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcServerEndpoint('%s'): IPC subsystem not initialized.\n",
			channel_name);
#endif	/* CHECKED_BUILD */
		return 0;
	}
	WaitForSingleObject(ipc_access,INFINITE);
	slot=ipc_local_find_channel(channel_name);
	if(slot)
	{
		/* Already exists */
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcServerEndpoint('%s'): channel already exists.\n",
			channel_name);
#endif	/* CHECKED_BUILD */
		ReleaseSemaphore(ipc_access,1,NULL);
		return 0;
	}
	slot=ipc_local_get_free_slot();
	if(slot==-1)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcServerEndpoint('%s'): no more place fro channels.\n",
			channel_name);
#endif	/* CHECKED_BUILD */
		ReleaseSemaphore(ipc_access,1,NULL);
		return 0;
	}
	obj=(ipc_local_t*)malloc(sizeof(ipc_local_t));
	v_zero(obj,sizeof(ipc_local_t));
	ipc_channels[slot]=obj;
	strCopy(obj->name,channel_name,MAX_IPC_NAME);
	u8fInit(&(obj->recv_fifo));
	obj->num_clients=0;
	obj->recv_access=CreateSemaphore(NULL,1,1,NULL);
	obj->recv_event=CreateEvent(NULL,TRUE,FALSE,NULL);
	ReleaseSemaphore(ipc_access,1,NULL);
	return slot;
}

ipc_handle_t	V_GLOBAL
ipcClientEndpoint(char *channel_name)
{
	u32_t slot, cl_slot;
	ipc_local_t *obj;
	ipc_handle_t ipch;
	if(!channel_name || channel_name[0]==0)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcClientEndpoint('%s'): invalid channel name.\n",
			dbs(channel_name));
#endif	/* CHECKED_BUILD */
		return 0;
	}
	if(!ipc_access)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcClientEndpoint('%s'): IPC subsystem not initialized.\n",
			channel_name);
#endif	/* CHECKED_BUILD */
		return 0;
	}
	WaitForSingleObject(ipc_access,INFINITE);
	slot=ipc_local_find_channel(channel_name);
	if(!slot)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcClientEndpoint('%s'): no such channel.\n",channel_name);
#endif	/* CHECKED_BUILD */
		ReleaseSemaphore(ipc_access,1,NULL);
		return 0;
	}
	obj=ipc_channels[slot];
	/* we are sure that obj!=NULL - see ipc_local_find_channel */
	if(obj->num_clients==MAX_IPC_CLIENTS)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcClientEndpoint('%s'): MAX_CLIENTS already connected.\n",
			channel_name);
#endif	/* CHECKED_BUILD */
		ReleaseSemaphore(ipc_access,1,NULL);
		return 0;
	}
	/* Find unused client slot */
	for(cl_slot=0;obj->send_access[cl_slot];cl_slot++);
	obj->send_access[cl_slot]=CreateSemaphore(NULL,1,1,NULL);
	obj->send_event[cl_slot]=CreateEvent(NULL,TRUE,FALSE,NULL);
	u8fInit(&(obj->send_fifo[cl_slot]));
	obj->num_clients++;
	ReleaseSemaphore(ipc_access,1,NULL);
	/* Hi-word contains number of client+1 */
	ipch=slot | (((cl_slot+1) << 16) & 0xFFFF0000);
	if(obj->on_client_connect)
		obj->on_client_connect(ipch,cl_slot);
	return ipch;
}

/*
 *	Input:
 *		caller: caller's name.
 *	Output:
 *		true: handle is valid (i.e. in bounds and both server and client
 *		exists). In such case server and (client ID)+1 stored in
 *		slot and cl_slot.
 *		false: handle is invalid or IPC subsystem not initialized.
 */

#ifdef	CHECKED_BUILD
vboolean	V_LOCAL
ipc_local_validate_handle(char *caller,ipc_handle_t ipch,u32_t *slot,
	u32_t *cl_slot)
#else	/* CHECKED_BUILD */
vboolean	V_LOCAL
ipc_local_validate_handle_quiet(ipc_handle_t ipch,u32_t *slot,
	u32_t *cl_slot)
#define		ipc_local_validate_handle(api,h,s,c)\
			ipc_local_validate_handle_quiet(h,s,c)
#endif	/* CHECKED_BUILD */
{
	int srv, cl;
	/* Server: 1,2, ..., IPC_MAX_CHANNELS-1 */
	srv=ipch & 0xFFFF;
	/* Client+1: can be 0 (server handle),1, ..., IPC_MAX_CLIENTS-1 */
	cl=(ipch >> 16) & 0xFFFF;
	if(!ipc_access)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"%s(%X): IPC subsystem not initialized.\n",
			caller,ipch);
#endif	/* CHECKED_BUILD */
		return false;
	}
	if(!srv || srv>=MAX_IPC_CHANNELS || !ipc_channels[srv])
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"%s(%X): invalid server ID '%d'.\n",
			caller,ipch,srv);
#endif	/* CHECKED_BUILD */
		return false;
	}
	if(cl>=MAX_IPC_CLIENTS || (cl && !ipc_channels[srv]->send_access[cl-1]))
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"%s(%X): invalid client ID '%d'.\n",
			caller,ipch,cl);
#endif	/* CHECKED_BUILD */
		return false;
	}
	if(slot)
		*slot=srv;
	if(cl_slot)
		*cl_slot=cl;
	return true;
}

void			V_GLOBAL
ipcClose(ipc_handle_t ipch)
{
	u32_t slot, cl_slot;
	ipc_local_t *obj;
	if(!ipc_local_validate_handle("ipcClose",ipch,&slot,&cl_slot))
		return;
	/* See note at top of file */
	/* BUGBUG */
	timeDelay(10);
	
	WaitForSingleObject(ipc_access,INFINITE);
	if(cl_slot)
	{
		/* Client endpoint */
		ipc_local_disconnect_client(slot,cl_slot-1);
	}
	else
	{
		/* Server endpoint - disconnect all clients and cleanup */
		obj=ipc_channels[slot];
		for(cl_slot=0;cl_slot<MAX_IPC_CLIENTS;cl_slot++)
			if(obj->send_access[cl_slot])
				ipc_local_disconnect_client(slot,cl_slot);
		CloseHandle(obj->recv_access);
		/* Release all waiting threads */
		SetEvent(obj->recv_event);
		CloseHandle(obj->recv_event);
		u8fClear(&(obj->recv_fifo));
		free(obj);
		ipc_channels[slot]=NULL;
	}
	ReleaseSemaphore(ipc_access,1,NULL);
}

u32_t	V_GLOBAL
ipcSendCount(ipc_handle_t ipch)
{
	u32_t slot, cl_slot, total;
	ipc_local_t *obj;
	if(!ipc_local_validate_handle("ipcSendCount",ipch,&slot,&cl_slot))
		return 0;
	obj=ipc_channels[slot];
	if(cl_slot)
	{
		/* Client side */
		total=u8fCount(&(obj->recv_fifo));
	}
	else
	{
		/* Server side */
		total=0;
		for(cl_slot=0;cl_slot<MAX_IPC_CLIENTS;cl_slot++)
		{
			if(!obj->send_access[cl_slot])
				continue;
			total+=u8fCount(&(obj->send_fifo[cl_slot]));
		}
	}
	return total;
}

u32_t		V_GLOBAL
ipcRecvCount(ipc_handle_t ipch)
{
	u32_t slot, cl_slot;
	ipc_local_t *obj;
	if(!ipc_local_validate_handle("ipcRecvCount",ipch,&slot,&cl_slot))
		return 0;
	obj=ipc_channels[slot];
	if(cl_slot)
		return u8fCount(&(obj->send_fifo[cl_slot-1]));
	else
		return u8fCount(&(obj->recv_fifo));
}

void		V_GLOBAL
ipcSendTo(ipc_handle_t ipch,u32_t cl_id,void *src,u32_t num_bytes)
{
	u32_t slot, cl_slot;
	ipc_local_t *obj;
	if(!ipc_local_validate_handle("ipcSendTo",ipch,&slot,&cl_slot))
		return;
	if(cl_slot)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcSendTo(%X,%d): call must not be used by clients.\n",
			ipch,cl_id);
#endif	/* CHECKED_BUILD */
		return;
	}
	if(!src || !num_bytes)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcSendTo(%X,%d,%X,%d): incorrect source / len.\n",
			ipch,cl_id,src,num_bytes);
#endif	/* CHECKED_BUILD */
		return;
	}
	obj=ipc_channels[slot];
	if(cl_id>=(MAX_IPC_CLIENTS-1) || !obj->send_access[cl_id])
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcSendTo(%X,%d,%X,%d): client %d not exists.\n",
			ipch,cl_id,src,num_bytes,cl_id);
#endif	/* CHECKED_BUILD */
		return;
	}
	WaitForSingleObject(obj->send_access[cl_id],INFINITE);
	u8fPushBuffer(&(obj->send_fifo[cl_id]),src,num_bytes);
	ReleaseSemaphore(obj->send_access[cl_id],1,NULL);
	SetEvent(obj->send_event[cl_id]);
}

void			V_GLOBAL
ipcSend(ipc_handle_t ipch,void *src,u32_t num_bytes)
{
	u32_t slot, cl_slot;
	ipc_local_t *obj;
	if(!ipc_local_validate_handle("ipcSend",ipch,&slot,&cl_slot))
		return;
	if(!src || !num_bytes)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcSend(%X,%X,%d): incorrect source / len.\n",
			ipch,src,num_bytes);
#endif	/* CHECKED_BUILD */
		return;
	}
	obj=ipc_channels[slot];
	if(cl_slot)
	{
		/* client - side */
		WaitForSingleObject(obj->recv_access,INFINITE);
		u8fPushBuffer(&(obj->recv_fifo),src,num_bytes);
		ReleaseSemaphore(obj->recv_access,1,NULL);
		SetEvent(obj->recv_event);
	}
	else
	{
		/* server - side: broadcast message to all clients */
		for(cl_slot=0;cl_slot<MAX_IPC_CLIENTS;cl_slot++)
		{
			if(!obj->send_access[cl_slot])
				continue;
			WaitForSingleObject(obj->send_access[cl_slot],INFINITE);
			u8fPushBuffer(&(obj->send_fifo[cl_slot]),src,num_bytes);
			ReleaseSemaphore(obj->send_access[cl_slot],1,NULL);
			SetEvent(obj->send_event[cl_slot]);
		}
	}
}

vboolean		V_GLOBAL
ipcRecv(ipc_handle_t ipch,void *dst,u32_t num_bytes)
{
	ipc_local_t *obj;
	u32_t slot, cl_slot;
	if(!ipc_local_validate_handle("ipcRecv",ipch,&slot,&cl_slot))
		return false;
	if(!dst || !num_bytes)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcRecv(%X,%X,%d): NULL dest / zero length.\n",
			ipch,dst,num_bytes);
#endif	/* CHECKED_BUILD */
		return false;
	}
	obj=ipc_channels[slot];
	if(cl_slot)
	{
		/* client - side */
		cl_slot--;
		WaitForSingleObject(obj->send_access[cl_slot],INFINITE);
		if(u8fCount(&(obj->send_fifo[cl_slot]))>=num_bytes)
			u8fPopBuffer(&(obj->send_fifo[cl_slot]),dst,num_bytes);
		else
			num_bytes=0;
		ReleaseSemaphore(obj->send_access[cl_slot],1,NULL);
		/* Reset event if no more data */
		if(!u8fCount(&(obj->send_fifo[cl_slot])))
			ResetEvent(obj->send_event[cl_slot]);
	}
	else
	{
		/* server - side */
		WaitForSingleObject(obj->recv_access,INFINITE);
		if(u8fCount(&(obj->recv_fifo))>=num_bytes)
			u8fPopBuffer(&(obj->recv_fifo),dst,num_bytes);
		else
			num_bytes=0;
		ReleaseSemaphore(obj->recv_access,1,NULL);
		/* Reset event if no more data */
		if(!u8fCount(&(obj->recv_fifo)))
			ResetEvent(obj->recv_event);
	}
	return num_bytes ? true : false;
}

vboolean		V_GLOBAL
ipcPurge(ipc_handle_t ipch,vboolean purge_input,vboolean purge_output)
{
	ipc_local_t *obj;
	u32_t slot, cl_slot;
	if(!ipc_local_validate_handle("ipcPurge",ipch,&slot,&cl_slot))
		return false;
	obj=ipc_channels[slot];
	if(cl_slot)
	{
		/* client - side */
		cl_slot--;
		if(purge_input)
		{
			/* Clear client input queue */
			WaitForSingleObject(obj->send_access[cl_slot],INFINITE);
			u8fPurge(&(obj->send_fifo[cl_slot]));
			ReleaseSemaphore(obj->send_access[cl_slot],1,NULL);
			ResetEvent(obj->send_event[cl_slot]);
		}
		if(purge_output)
		{
			/* We can't do it: output queue shared by all clients. */
#ifdef		CHECHED_BUILD
			conlPrintf(LOG_WARNING,
				"ipcPurge(%X,%d,%d): shared queue can only be purged from"
				" server side.\n",ipch,purge_input,purge_output);
#endif		/* CHECKED_BUILD */
			return false;
		}
	}
	else
	{
		/* server - side */
		if(purge_input)
		{
			/* Clear input queue */
			WaitForSingleObject(obj->recv_access,INFINITE);
			u8fPurge(&(obj->recv_fifo));
			ReleaseSemaphore(obj->recv_access,1,NULL);
			ResetEvent(obj->recv_event);
		}
		if(purge_output)
		{
			/* Clear ALL output queues. */
			for(cl_slot=0;cl_slot<MAX_IPC_CLIENTS;cl_slot++)
			{
				if(!obj->send_access[cl_slot])
					continue;
				WaitForSingleObject(obj->send_access[cl_slot],INFINITE);
				u8fPurge(&(obj->send_fifo[cl_slot]));
				ReleaseSemaphore(obj->send_access[cl_slot],1,NULL);
				ResetEvent(obj->send_event[cl_slot]);
			}
		}
	}
	return true;
}

int		V_GLOBAL
ipcGets(ipc_handle_t ipch,char *dst,int dst_len)
{
	int len;
	char ch;
	if(!ipch || !dst || !dst_len)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"ipcGets(%X,%X,%d): incorrect parameter(s).\n",
			ipch,dst,dst_len);
#endif	/* CHECKED_BUILD */
		return 0;
	}
	*dst=0;
	len=(int)ipcRecvCount(ipch);
	if(!len) /* nothing to receive */
		return 0;
	if(len>=dst_len)
		len=dst_len-1;
	dst_len=len;
	/* Skip leading EOLs */
	ipcRecv(ipch,&ch,1);
	while(len && (ch==0x0A || ch==0x0D))
	{
		len--;
		ipcRecv(ipch,&ch,1);
	}
	if(!len)	/* only whitespace in input buffer */
	{
		*dst=0;
		return 0;
	}
	while(len)
	{
		*dst=ch;
		dst++;
		len--;
		ipcRecv(ipch,&ch,1);
		if(ch==0x0D || ch==0x0A)
		{
			*dst=0;
			return dst_len-len-1;
		}
	}
	*dst=0;
	return dst_len-1;
}

void	V_GLOBAL
ipcConnectCallback(ipc_handle_t ipch,
	void (V_CALL *connect_callback)(ipc_handle_t ipch,u32_t cl_id))
{
	u32_t slot, cl_slot;
	if(!ipc_local_validate_handle("ipcConnectCallback",ipch,&slot,&cl_slot))
		return;
	ipc_channels[slot]->on_client_connect=connect_callback;
}

void	V_GLOBAL
ipcDisconnectCallback(ipc_handle_t ipch,
	void (V_CALL *disconnect_callback)(ipc_handle_t ipch,u32_t cl_id))
{
	u32_t slot, cl_slot;
	if(!ipc_local_validate_handle("ipcDisconnectCallback",ipch,&slot,&cl_slot))
		return;
	ipc_channels[slot]->on_client_disconnect=disconnect_callback;
}

vboolean	V_GLOBAL
ipcWait(ipc_handle_t ipch,u32_t timeout)
{
	u32_t slot, cl_slot;
	ipc_local_t *obj;
	DWORD retval;
	if(!ipc_local_validate_handle("ipcWait",ipch,&slot,&cl_slot))
		return false;
	obj=ipc_channels[slot];
	/* If data available, return immediately */
	if(cl_slot)
	{
		/* client - side */
		if(u8fCount(&(obj->send_fifo[cl_slot-1])))
			return true;
		retval=WaitForSingleObject(obj->send_event[cl_slot-1],timeout);
	}
	else
	{
		/* server - side */
		if(u8fCount(&(obj->recv_fifo)))
			return true;
		retval=WaitForSingleObject(obj->recv_event,timeout);
	}
	return retval==WAIT_TIMEOUT ? false : true;
}

vboolean	V_GLOBAL
ipcGetServerName(ipc_handle_t ipch,char *dest)
{
	u32_t slot, cl_slot;
	if(!dest)
		return false;
	if(ipc_local_validate_handle("ipcGetServerName",ipch,&slot,&cl_slot))
	{
		strCopy(dest,ipc_channels[slot]->name,_MAX_PATH);
		return true;
	}
	return false;
}

vboolean	V_GLOBAL
ipcGetClientId(ipc_handle_t ipch,u32_t *id)
{
	u32_t slot, cl_slot;
	if(!id)
		return false;
	if(ipc_local_validate_handle("ipcGetClientId",ipch,&slot,&cl_slot))
	{
		*id=cl_slot;
		return true;
	}
	return false;
}

#elif	defined(__UNIX__)	/* __WIN32__ */
#error	TODO: Implement UNIX IPC.
#elif	defined(__DOS32__)	/* defined(__UNIX__) */
#error	TODO: Implement DOS IPC.
#endif	/* defined(__DOS32__) */

/*</pre></strong></body></html>*/
/*EOF*/
