/* Copyright 1998 by Scott Franke [druid-]
    sfranke@scf.usc.edu
    http://www-scf.usc.edu/~sfranke/glj
*/

#include <windows.h>
#include <mmsystem.h>
#include <winuser.h>
#include <gl\gl.h>
#include <gl\glu.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#pragma   hdrstop

#include "console.h"
#include "tex.h"

#ifndef M_PI
#define M_PI 3.13159
#endif

static color_t color_white = {255, 255, 255};
static color_t color_grey  = {180, 180, 180};



	//
	//  Clear the console text buffer
	//
void GLConClear( console_t *con)
{
	int i, j;
	if( !con )
		return;

	for(i=0;i<CONSOLE_MAX_ROW;i++)
	{
		for(j=0;j<60;j++)
			con->buffer[i].text[j] = '\0';
		con->buffer[i].color = color_white;
	}
	con->newline = 0;
	con->viewbottom = 0;
}

	//
	//  Clear the console text buffer
	//
void GLConCmdClear( console_t *con)
{
	int i, j;

	if( !con )
		return;

	for(i=0;i<30;i++)
		for(j=0;j<60;j++)
			con->cmdhist[i][j] = '\0';
	con->newcmd = 0;
	con->numcmdhist = 0;	
	con->curcmd = 0;
}


void GLConCmdexit( console_t *con, char *param )
{
	con->active = 0;
	mode = MODE_PLAYBACK;
}
void GLConCmdquit( console_t *con, char *param )
{
	done = TRUE;
	PostQuitMessage( 0 );
}	
void GLConCmddestroy( console_t *con, char *param )
{
	GLConDestroy( con );
	mode = MODE_PLAYBACK;
}
void GLConCmdecho( console_t *con, char *param )
{
	GLConOut( con, param );
}
void GLConCmdcls( console_t *con, char *param )
{
	GLConClear( con );
}
void GLConCmdExecFile( console_t *con, char *param )
{
	GLConExecFile( con, param );
}
void GLConCmdcmdlist( console_t *con, char *param )
{
	int i;
	cmd_t *pcmd;
	if( !param )
	{
		for( i=0;i<256;i++ )
		{
			pcmd = con->cmds[i];
			while( pcmd )
			{
				GLConOut( con, pcmd->fullcmd );
				pcmd = pcmd->next;
			}
		}
	}
	else
	{
		i = param[0];
		if( (i >= 'A') && (i<='Z') )
			i -= 'A' - 'a';
		pcmd = con->cmds[i];
		while( pcmd )
		{
			if( strncmp(pcmd->fullcmd, param, strlen(param)) == 0 )
				GLConOut( con, pcmd->fullcmd );
			pcmd = pcmd->next;
		}
	}
}
void GLConCmdcmdhist( console_t *con, char *param )
{
	int i, index;
	index = con->newcmd - con->numcmdhist;
	for(i=0;i<con->numcmdhist;i++)
	{
		if( index > 30 )
			index -= 30;
		GLConOut( con, con->cmdhist[index] );
		index++;
	}
}

	

	//
	//  Create a console
	//		initialize the data, add the basic commands
	//
console_t *GLConCreateConsole( float version, char *font, char *logo )
{
	int i;
	console_t *con = ( console_t *)malloc(sizeof(console_t) );

	con->creating = 1;
	con->destroying = 0;
	con->active = 1;
	con->col = 1;
	con->drawortho = TRUE;

	con->font_w = 50;
	con->font_h = 20;

	sprintf( con->ver, "%.3f", version );

	for( i=0;i<256;i++)
		con->cmds[i] = NULL;

	glGenTextures( 1, &con->font_id );
	glGenTextures( 1, &con->logo_id );

	if( font )
		GLLoadBMP8( font, con->font_id, TEX_ALPHA_COLORKEY_0 );
	if( logo )
		GLLoadBMP8( logo, con->logo_id, TEX_ALPHA_COLORKEY_0 );
	else
		con->logo_id = 0;

	GLConAddCmd( con, "exit", GLConCmdexit );
	GLConAddCmd( con, "quit", GLConCmdquit );
	GLConAddCmd( con, "destroy", GLConCmddestroy );
	GLConAddCmd( con, "echo", GLConCmdecho );
	GLConAddCmd( con, "cmdlist", GLConCmdcmdlist );
	GLConAddCmd( con, "cls", GLConCmdcls );
	GLConAddCmd( con, "cmdhist", GLConCmdcmdhist );
	GLConAddCmd( con, "exec", GLConCmdExecFile );

	GLConClear( con );
	GLConCmdClear( con );
	GLConAdvance( con );
	con->buffer[con->newline].text[0] = ']';
	con->buffer[con->newline].text[1] = '_';
	con->buffer[con->newline].text[2] = '\0';

	return con;
}

	//
	//  Destroy the console
	//
void GLConDestroy( console_t *con )
{
	int i;
	cmd_t *pcmd;
	glDeleteTextures( 1, &con->font_id );
	con->destroying = 1;
		// clean up command list
	for( i=0;i<256;i++)
	{
		if( con->cmds[i] )
		{
			pcmd = con->cmds[i];
			while( pcmd != NULL )
			{
				cmd_t *next = pcmd->next;
				free( pcmd );
				pcmd = next;
			}
		}
	}
}

	//
	//  Write a character to the console
	//
void GLConWriteChar( console_t *con, int row, int col, char letter )
{
	int cl, rw;
	float fw, fh;
	float l,t,r,b;
	//GLfloat box[4];
	GLfloat poly[4];

	/*box[0] = (GLfloat) ((((letter - 33) % 16) * 16)) / 256.0f;
	box[1] = (GLfloat) 1- ((((letter - 33) / 16) * 24)) / 256.0f;
	box[2] = (box[0]*256.0f + 16) / 256.0f;
	box[3] = (box[1]*256.0f - 24) / 256.0f;
	*/
  
  cl = letter & 7;
  rw = 15 - (letter >> 4);
  l = cl * 8.0 / 128.0;
  r = l+7.0/128.0;
  t = rw * 8.0 / 128.0;
  b = t+7.0/128.0;
  
	/*if(box[2] > 1.0f)
		box[2] = 1.0f;*/
  

	fw = 0.9f / con->font_w;
	fh = 0.4f / con->font_h;

	poly[0] = fw + fw * (col);
	poly[1] = 0.08f + fh * (row);
	poly[2] = poly[0] + fw;
	poly[3] = poly[1] + fh;

	glBegin(GL_QUADS);
		//glTexCoord2f( box[0], box[1] );
		glTexCoord2f(l,b);
		glVertex2f( poly[0], poly[3] );
		//glTexCoord2f( box[0], box[3] );
		glTexCoord2f(l,t);
		glVertex2f( poly[0], poly[1] );
		//glTexCoord2f( box[2], box[3] );
		glTexCoord2f(r,t);
		glVertex2f( poly[2], poly[1] );
		//glTexCoord2f( box[2], box[1] );
		glTexCoord2f(r,b);
		glVertex2f( poly[2], poly[3] );
	glEnd();
}

	//
	//  Write a string to the console
	//
void GLConWriteString( console_t *con, int row, int col, char *string )
{
	int i;

	i=0;

	while(string[i]!='\0')
	{
		if( (string[i] >= 33) && (string[i] < 127) )
			GLConWriteChar( con, row, col + i, string[i] );
		i++;
	}

}

	//
	//  Advance the console one line
	//
void GLConAdvance( console_t *con )
{
	con->newline--;
	if( con->newline < 0 )
		con->newline += CONSOLE_MAX_ROW;
	con->buffer[con->newline].text[0] = '\0';
	con->viewbottom = con->newline;
}

	//
	//  Scroll the console down
	//
void GLConScroll( console_t *con, DWORD param )
{
	switch( param )
	{
	case VK_PAGEUP:
		con->viewbottom++;
		if( con->viewbottom >= CONSOLE_MAX_ROW )
			con->viewbottom -= CONSOLE_MAX_ROW;
		if( con->viewbottom == con->newline )
			con->viewbottom--;
		break;
	case VK_PAGEDOWN:
		con->viewbottom--;
		if( con->viewbottom < 0 )
			con->viewbottom += CONSOLE_MAX_ROW;
		if( con->viewbottom == con->newline-1 )
			con->viewbottom = con->newline;
		break;
	}

}

	//
	//  Write a string to the console
	//
void GLConOut( console_t *con, char *string )
{
	GLConAdvance( con );
	if( string )
		strcpy( con->buffer[con->newline].text,string );
}

	//
	//  Write a string to the console and put it in the foreground
	//
void GLConAlert( console_t *con, char *string )
{
	GLConAdvance( con );
	strcpy( con->buffer[con->newline].text,string );
	mode = MODE_CONSOLE;
	con->active = 1;
	con->viewbottom = con->newline;
}

	//
	//  Add a command
	//
void GLConAddCmd( console_t *con, char *fullcmd, void (*function)(console_t *, char *) )
{
	int hashindex;
	cmd_t *pcmd;

	if( !fullcmd )
		return;
	if( !function )
		return;

	hashindex = fullcmd[0];
	pcmd = con->cmds[hashindex];

	if( pcmd == NULL )
	{
		con->cmds[hashindex] = (cmd_t *) malloc( sizeof(cmd_t) );
		pcmd = con->cmds[hashindex];
	}
	else
	{
		while( pcmd->next != NULL )
		{
			pcmd = pcmd->next;
		}
		pcmd->next = (cmd_t *) malloc( sizeof(cmd_t) );
		pcmd = pcmd->next;
	}
	strcpy( pcmd->fullcmd, fullcmd );
	pcmd->function = function;
	pcmd->next = NULL;
}

	//
	//  Try to execute a command
	//
int GLConExecCmd( console_t *con, char *cmd, char *param )
{
	cmd_t *pcmd;
	int hashindex = cmd[0];

	if( !cmd )
		return 0;
	pcmd = con->cmds[hashindex];

	while( pcmd != NULL )
	{
		if( _stricmp( cmd, pcmd->fullcmd ) == 0 )
		{
			pcmd->function( con, param );
			return 1;
		}
		else
			pcmd = pcmd->next;
	}

	return 0;
}

	//
	//	Load a file and execute it's contents
	//		as if they were typed in at the console
	//
void GLConExecFile( console_t *con, char *filename )
{
	FILE *fp;
	char buf[80], *param;
	int ret = 1;

	if (!filename)
	{
		GLConOut( con, "No file specified" );
		return;
	}

	fp = fopen( filename, "r" );
	if( !fp )
	{
		sprintf( buf, "Unable to open file: %s", filename );
		GLConAlert( con, buf );
		return;
	}
	while( 1 )
	{
		ret = (int) fgets( buf, 80, fp );
		if( !ret )
			break;
		strtok( buf, " \n\f" );
		param = buf + strlen(buf) + 1;
		if( param[0] == 0 )
			param = NULL;
		if( ! GLConExecCmd( con, buf, param ) )
		{
			GLConOut( con, "Invalid command in file" );
		}
	}
	if( fp )
		fclose( fp );
}

	//
	//  Process an instruction
	//		called after the user hits enter
	//
void GLConProcess( console_t *con )
{
	char cmd[60];
	char *cmdparam;
	char temp;
	strcpy( cmd, &con->buffer[con->newline].text[1]);
	strcpy( con->cmdhist[con->newcmd], cmd );
	con->newcmd = con->newcmd++ % 30;
	con->curcmd = 0;
	con->numcmdhist++;
	if( con->numcmdhist > 30 )
		con->numcmdhist = 30;

	con->buffer[con->newline].color = color_grey;

	strtok( cmd, " " );
	temp = *( &con->buffer[con->newline].text[1] + strlen(cmd) );
	if( temp == 0 )
		cmdparam = NULL;
	else
		cmdparam = cmd + strlen(cmd) + 1;

	if( !GLConExecCmd( con, cmd, cmdparam ) )
	{
		GLConOut( con, "Invalid command" );
	}
	return;
}

	// Called from winmain
	//		up and down arrow
	//		for command history
	//
void GLConCmdHistory( console_t *con, DWORD param)
{
	switch( param )
	{
	case VK_UP:
		con->curcmd++;
		if( con->curcmd > con->numcmdhist )
			con->curcmd = con->numcmdhist;
		break;
	case VK_DOWN:
		con->curcmd--;
		if( con->curcmd <= 0 )
			con->curcmd = 0;
		break;
	}
	if( con->curcmd != -1 )
	{
		char buf[60];
		int cmdnum;
		buf[0] = ']';
		buf[1] = '\0';
		cmdnum = con->newcmd - con->curcmd;
		if( cmdnum < 0 )
			cmdnum += 30;
		strcat( buf, con->cmdhist[cmdnum] );
		strcat( buf, "_" );
		strcpy( con->buffer[con->newline].text, buf );
		con->col = strlen( buf )-1;
	}

}

	//
	// Called from winmain.
	//		reads in the character and processes it
	//
void GLConIn( console_t *con, long wParam )
{
	int i;

	if( !con->active )
		return;

	switch(wParam)
	{
	case '\b':
		if( con->col>1 )
		{
			con->buffer[con->newline].text[con->col] = '\0';
			con->buffer[con->newline].text[con->col-1] = '_';
			con->col--;
		}
		break;
	case '\n':
	case '\r':
		con->buffer[con->newline].text[con->col] = '\0';
		GLConProcess( con );
		GLConAdvance( con );
		con->col = 1;
		con->buffer[con->newline].text[0] = ']';
		con->buffer[con->newline].text[1] = '_';
		con->buffer[con->newline].text[2] = '\0';
		break;
	case '\t':
		con->buffer[con->newline].text[con->col]=' ';
		for(i=0;i<5;i++)
			con->buffer[con->newline].text[++con->col]=' ';
		con->col++;
		con->buffer[con->newline].text[con->col] = '_';
		break;
	case '\x1B':							// Escape
		con->col=0;
		con->buffer[con->newline].text[con->col]='_';
		break;
	default:
		if(con->col > 60)
			break;
		con->buffer[con->newline].text[con->col] = (char) wParam;
		con->col++;
		con->buffer[con->newline].text[con->col] = '_';
		con->buffer[con->newline].text[con->col+1] = '\0';
		break;
	}

}

	//
	// render the console (only orthographically now)
	//
void GLConDrawConsole( console_t *con, float dTime )
{
	int i, temp;

	if( !con )
		return;
	if( !con->active )
		return;

	if( con->drawortho )
	{
		glMatrixMode( GL_PROJECTION );
		glLoadIdentity( );
		glOrtho( 0.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f );
		glMatrixMode( GL_MODELVIEW );
	}
	else
	{
		glTranslatef( -0.5f, -0.5f, 0.0f );
	}
	glDepthFunc( GL_ALWAYS );

	glDisable( GL_TEXTURE_2D );
	glBegin( GL_QUADS );
			/* Border */
		glColor4f( 0.2f,0.2f,0.6f,0.3f);
		glVertex3f( 0.0f, 0.52f, -0.5f );
		glVertex3f( 0.0f, 0.05f, -0.5f );
		glVertex3f( 1.0f, 0.05f, -0.5f );
		glVertex3f( 1.0f, 0.52f, -0.5f );
			/* Terminal Area */
		glColor4f( 0.0f,0.0f,0.0f,0.6f);
		glVertex3f( 0.0f, 0.51f, -0.5f );
		glVertex3f( 0.0f, 0.06f, -0.5f );
		glVertex3f( 1.0f, 0.06f, -0.5f );
		glVertex3f( 1.0f, 0.51f, -0.5f );
	glEnd( );

		// draw the logo
	if( con->logo_id )
	{
		glColor4f( 1.0f,1.0f,1.0f,0.8f);
		glBindTexture( GL_TEXTURE_2D, con->logo_id );
		glEnable( GL_TEXTURE_2D );
		glBegin( GL_QUADS );
			glTexCoord2f( 0.0f, 1.0f );
			glVertex3f( 0.73f, 0.52f, -0.5f );
			glTexCoord2f( 0.0f, 0.0f );
			glVertex3f( 0.73f, 0.4f, -0.5f );
			glTexCoord2f( 1.0f, 0.0f );
			glVertex3f( 0.99f, 0.4f, -0.5f );
			glTexCoord2f( 1.0f, 1.0f );
			glVertex3f( 0.99f, 0.52f, -0.5f );
		glEnd ();
	}
	else
		glEnable( GL_TEXTURE_2D );

	glBindTexture( GL_TEXTURE_2D, con->font_id );
	temp = con->viewbottom;
	for(i=0;i<con->font_h;i++)
	{
		glColor4f(  (float)con->buffer[temp].color.r/255,
					(float)con->buffer[temp].color.g/255,
					(float)con->buffer[temp].color.b/255, 0.8f );
		GLConWriteString( con, i, 0, con->buffer[temp].text );
		temp = (temp + 1) % CONSOLE_MAX_ROW;
	}

	if( con->viewbottom != con->newline )
		GLConWriteChar( con, 0, -1, '|' );

	glColor4f( 0.0f,1.0f,0.0f,0.5f);
	GLConWriteString( con, con->font_h-3, con->font_w-2, con->ver);

	if( con->creating )
	{
		float val1, val2;
		glDisable( GL_TEXTURE_2D );
		glBlendFunc( GL_ONE, GL_ONE );
		glBegin( GL_QUADS );
			val1 = (float) cos (M_PI*0.5*con->creating/256)/4;
			val2 = (float) cos (M_PI*0.5*con->creating/256)/2;
			glColor3f( val2, val2, val2 );
			glVertex3f( 0.0f, 0.51f, -0.5f );
			glVertex3f( 1.0f, 0.51f, -0.5f );
			glColor3f( val1, val1, val1 );
			glVertex3f( 1.0f, 0.57f, -0.5f );
			glVertex3f( 0.0f, 0.57f, -0.5f );
			
			glVertex3f( 0.0f, 0.0f, -0.5f );
			glVertex3f( 1.0f, 0.0f, -0.5f );
			glColor3f( val2, val2, val2 );
			glVertex3f( 1.0f, 0.06f, -0.5f );
			glVertex3f( 0.0f, 0.06f, -0.5f );

			glVertex3f( 0.0f, 0.51f, -0.5f );
			glVertex3f( 0.0f, 0.06f, -0.5f );
			glVertex3f( 1.0f, 0.06f, -0.5f );
			glVertex3f( 1.0f, 0.51f, -0.5f );
		glEnd( );

		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		con->creating += (int) dTime/16;
		if( con->creating >= 256 )
			con->creating = 0;
	}

	/* Back to normal viewing */
	glMatrixMode( GL_PROJECTION );
	glDepthFunc( GL_LEQUAL );
	glLoadIdentity( );
	glFrustum( -4.0f, 4.0f, -2.704225f, 2.704225f, 4.0f, 4096.0f );
	glMatrixMode( GL_MODELVIEW );
}