#include      "u_gl.h"

/*=======================================================================
*                       OpenGL.
* :       ugl.c
* :     1.0 (10.07.99)
* : Misrosoft C  Watcom C.
* :  Win32 (9x & NT)
* :         OpenGL.
* :       ..
* :
* 10.07.99 -  .                                          */

/*
==================================
    PreloadLayer
==================================
*/

static GLIMLayer *PreloadLayer()
{
  GLIMLayer *layer;
  layer=(GLIMLayer *)malloc(sizeof(GLIMLayer));
  memset(layer,0,sizeof(GLIMLayer));
  return layer;
}

#ifdef PROCADDR
#undef PROCADDR
#endif
#define PROCADDR(name)\
  layer->name = (pgl##name)GetProcAddress(layer->hRenderer,"gl"#name);\
  if(!layer->name) goto Failed;

/*
==================================
    PreloadRenderer - ,     OpenGL
==================================
*/

static uboolean PreloadRenderer(GLIMLayer *layer,char *dllname)
{
  //  
  layer->hRenderer = LoadLibrary(dllname);
  if(!layer->hRenderer) return false;
  PROCADDR(AlphaFunc);
  PROCADDR(Begin);
  PROCADDR(BlendFunc);
  PROCADDR(Clear);
  PROCADDR(ClearColor);
  PROCADDR(Color3f);
  PROCADDR(Color3ubv);
  PROCADDR(Color4f);
  PROCADDR(Color4ubv);
  PROCADDR(CullFace);
  PROCADDR(DepthFunc);
  PROCADDR(DepthMask);
  PROCADDR(DepthRange);
  PROCADDR(Disable);
  PROCADDR(DrawBuffer);
  PROCADDR(Enable);
  PROCADDR(End);
  PROCADDR(Finish);
  PROCADDR(Frustum);
  PROCADDR(GetFloatv);
  PROCADDR(GetString);
  PROCADDR(Hint);
  PROCADDR(LoadIdentity);
  PROCADDR(LoadMatrixf);
  PROCADDR(MatrixMode);
  PROCADDR(Ortho);
  PROCADDR(PolygonMode);
  PROCADDR(PopMatrix);
  PROCADDR(PushMatrix);
  PROCADDR(ReadBuffer);
  PROCADDR(ReadPixels);
  PROCADDR(Rotatef);
  PROCADDR(Scalef);
  PROCADDR(ShadeModel);
  PROCADDR(TexCoord2f);
  PROCADDR(TexEnvf);
  PROCADDR(TexImage2D);
  PROCADDR(TexParameterf);
  PROCADDR(Translatef);
  PROCADDR(Vertex2f);
  PROCADDR(Vertex3f);
  PROCADDR(Vertex3fv);
  PROCADDR(Viewport);
  PROCADDR(Fogf);
  PROCADDR(Fogfv);
  PROCADDR(Fogi);
  PROCADDR(Fogiv);
  PROCADDR(GenTextures);
  PROCADDR(BindTexture);
  PROCADDR(DeleteTextures);
  PROCADDR(EnableClientState);
  PROCADDR(DisableClientState);
  PROCADDR(DrawArrays);
  PROCADDR(ColorPointer);
  PROCADDR(TexCoordPointer);
  PROCADDR(VertexPointer);
  PROCADDR(ClearDepth);
  return true;
  Failed:;
  return false;
}

#undef PROCADDR

/*
==================================
        
==================================
*/

unsigned char threeto8[8] = {
    0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377
};

unsigned char twoto8[4] = {
    0, 0x55, 0xaa, 0xff
};

unsigned char oneto8[2] = {
    0, 255
};

static int defaultOverride[13] = {
    0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91
};

static PALETTEENTRY defaultPalEntry[20] = {
    { 0,   0,   0,    0 },
    { 0x80,0,   0,    0 },
    { 0,   0x80,0,    0 },
    { 0x80,0x80,0,    0 },
    { 0,   0,   0x80, 0 },
    { 0x80,0,   0x80, 0 },
    { 0,   0x80,0x80, 0 },
    { 0xC0,0xC0,0xC0, 0 },

    { 192, 220, 192,  0 },
    { 166, 202, 240,  0 },
    { 255, 251, 240,  0 },
    { 160, 160, 164,  0 },

    { 0x80,0x80,0x80, 0 },
    { 0xFF,0,   0,    0 },
    { 0,   0xFF,0,    0 },
    { 0xFF,0xFF,0,    0 },
    { 0,   0,   0xFF, 0 },
    { 0xFF,0,   0xFF, 0 },
    { 0,   0xFF,0xFF, 0 },
    { 0xFF,0xFF,0xFF, 0 }
};

static unsigned char ComponentFromIndex(int i, UINT nbits, UINT shift)
{
    unsigned char val;

    val = (unsigned char) (i >> shift);
    switch (nbits) {

    case 1:
        val &= 0x1;
        return oneto8[val];

    case 2:
        val &= 0x3;
        return twoto8[val];

    case 3:
        val &= 0x7;
        return threeto8[val];

    default:
        return 0;
    }
}

static void CreateRGBPalette(GLIMLayer *layer)
{
    PIXELFORMATDESCRIPTOR pfd;
    LOGPALETTE *pPal;
    int n, i;

    n = layer->pGetPixelFormat(layer->pwglGetCurrentDC());
    layer->pDescribePixelFormat(layer->pwglGetCurrentDC(), n, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
    
    {
			int genericFormat = (pfd.dwFlags & PFD_GENERIC_FORMAT) != 0;
			int genericAccel = (pfd.dwFlags & PFD_GENERIC_ACCELERATED) != 0;
			/* Determine hardware driver status by the combination of these
			 * two flags.  Both FALSE means ICD, both TRUE means MCD.
			 */
			if(genericFormat && genericAccel)
			  layer->RenderType = MCD;
			else
			  if((!genericFormat) && (!genericAccel))
			    layer->RenderType = ICD;
			  else
			    layer->RenderType = Software;
    }
    
    if (pfd.dwFlags & PFD_NEED_PALETTE) {
        n = 1 << pfd.cColorBits;
        pPal = (PLOGPALETTE)malloc(sizeof(LOGPALETTE) +
                n * sizeof(PALETTEENTRY));
        pPal->palVersion = 0x300;
        pPal->palNumEntries = n;
        for (i=0; i<n; i++) {
            pPal->palPalEntry[i].peRed =
                    ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
            pPal->palPalEntry[i].peGreen =
                    ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
            pPal->palPalEntry[i].peBlue =
                    ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
            pPal->palPalEntry[i].peFlags = 0;
        }

        /* fix up the palette to include the default GDI palette */
        if ((pfd.cColorBits == 8)                           &&
            (pfd.cRedBits   == 3) && (pfd.cRedShift   == 0) &&
            (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) &&
            (pfd.cBlueBits  == 2) && (pfd.cBlueShift  == 6)
           ) {
            for (i = 1 ; i <= 12 ; i++)
                pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
        }

        layer->ghPalette = CreatePalette(pPal);
        free(pPal);

        layer->ghpalOld = SelectPalette(layer->pwglGetCurrentDC(), layer->ghPalette, FALSE);
        n = RealizePalette(layer->pwglGetCurrentDC());
    }
}

static uboolean bSetupPixelFormat(GLIMLayer *layer,HWND hWnd)
{
  HDC dc;
  static PIXELFORMATDESCRIPTOR 
  pfd = {
	sizeof(PIXELFORMATDESCRIPTOR),	// size of this pfd
	1,				// version number
	PFD_DRAW_TO_WINDOW |		// support window
	  PFD_SUPPORT_OPENGL |		// support OpenGL
	  PFD_DOUBLEBUFFER,		// double buffered
	PFD_TYPE_RGBA,			// RGBA type
	32,				// 32-bit color depth
	0, 0, 0, 0, 0, 0,		// color bits ignored
	0,				// no alpha buffer
	0,				// shift bit ignored
	0,				// no accumulation buffer
	0, 0, 0, 0, 			// accum bits ignored
	32,				// 32-bit z-buffer	
	0,				// no stencil buffer
	0,				// no auxiliary buffer
	PFD_MAIN_PLANE,			// main layer
	0,				// reserved
	0, 0, 0				// layer masks ignored
  };
  int pixelformat;
  dc=GetDC(hWnd);
    if ( (pixelformat = layer->pChoosePixelFormat(dc, &pfd)) == 0 )
    {
        MessageBox(hWnd, "ChoosePixelFormat failed", "Error", MB_OK);
        ReleaseDC(hWnd,dc);
        return false;
        
    }

    if (layer->pSetPixelFormat(dc, pixelformat, &pfd) == FALSE)
    {
        MessageBox(hWnd, "SetPixelFormat failed", "Error", MB_OK);
        ReleaseDC(hWnd,dc);
        return false;
    }

    CreateRGBPalette(layer);

    ReleaseDC(hWnd,dc);
    return true;
   
}

#ifdef WGL_PROC_ADDRESS
#undef WGL_PROC_ADDRESS
#endif

#define WGL_PROC_ADDRESS(name)\
  layer->p##name = (void *)GetProcAddress(layer->hRenderer,#name);\
  if(!layer->p##name) goto Failed;

#ifdef GDI_PROC_ADDRESS
#undef GDI_PROC_ADDRESS
#endif

#define GDI_PROC_ADDRESS(name)\
  layer->p##name = (void *)GetProcAddress(layer->hGDI,#name);\
  if(!layer->p##name) goto Failed;

PGLIMLayer UEXPORT LoadGenericRenderer(char *dllName,HWND hWnd)
{
	GLIMLayer *layer;
	static char def_dll[32]="opengl32.dll";
	if(!dllName) dllName=def_dll;
	layer=PreloadLayer();
	if(!PreloadRenderer(layer,dllName)) goto Failed;
	layer->hGDI=LoadLibrary("GDI32.DLL");
	if(!layer->hGDI) goto Failed;
	WGL_PROC_ADDRESS(wglCreateContext);
	WGL_PROC_ADDRESS(wglDeleteContext);
	WGL_PROC_ADDRESS(wglGetCurrentContext);
	WGL_PROC_ADDRESS(wglGetCurrentDC);
	WGL_PROC_ADDRESS(wglGetProcAddress);
	WGL_PROC_ADDRESS(wglMakeCurrent);
	GDI_PROC_ADDRESS(SwapBuffers);
	GDI_PROC_ADDRESS(ChoosePixelFormat);
	GDI_PROC_ADDRESS(SetPixelFormat);
	GDI_PROC_ADDRESS(DescribePixelFormat);
	GDI_PROC_ADDRESS(GetPixelFormat);
  if(!bSetupPixelFormat(layer,hWnd)) goto Failed;
	return layer;
	Failed:;
	DestroyLayer(layer);
	return NULL;
}

#undef WGL_PROC_ADDRESS
#undef GDI_PROC_ADDRESS

void UEXPORT DestroyLayer(GLIMLayer *Layer)
{
  switch(Layer->LayerType)
  {
    case GenericLayer:
    {
      if(Layer->pwglMakeCurrent) 
        Layer->pwglMakeCurrent(NULL,NULL);
  	  if(Layer->pwglGetCurrentContext)
  	    if(Layer->pwglGetCurrentContext())
  	      if(Layer->pwglDeleteContext)
  	        Layer->pwglDeleteContext(Layer->pwglGetCurrentContext());
  	  /*if(Layer->pwglGetCurrentDC)
  	    if(Layer->pwglGetCurrentDC())
  	      if(Layer->pwglDeleteDC)
  	        Layer->pwglDeleteDC(Layer->pwglGetCurrentDC());*/
    }
  }
  if(Layer->hRenderer) 
    FreeLibrary(Layer->hRenderer);
  if(Layer->hGDI) 
    FreeLibrary(Layer->hGDI);
  free(Layer);
  Layer=NULL;
}

