#include <windows.h>
#include <gl\gl.h>
#pragma hdrstop

#define	V_LOCAL		static
#define	V_GLOBAL

#define HWACCEL_FOUND       1
#define PALETTE_NEEDED      2

vboolean	V_GLOBAL
vglIsAccelerated(glw_t glw)
{
	PIXELFORMATDESCRIPTOR  pfd;
	HDC  hdc;
	hdc=...
	DescribePixelFormat(hdc,GetPixelFormat(hdc),sizeof(PIXELFORMATDESCRIPTOR),&pfd);
	return (pfd.dwFlags & PFD_GENERIC_FORMAT) == (pfd.dwFlags & PFD_GENERIC_ACCELERATED);
}

#ifdef	__cplusplus
extern "C" {
#endif	/* __cplusplus */
    /*
     *	ntCreateGLContext() - creates OpenGL RC for drawing (but don't makes
     *	it current).
     *	Input:
     *		win_out: handle of output window;
     *		palh: pointer to store palette handle (8-bpp modes only).
     *	Outpit:
     *		Returns OpenGL RC or NULL on error.
     *	Notes.
     *		1) ntCreateGLContext() superclasses win_out;
     *		2) palh can be NULL in 16- or 32-bpp mode.
     */

HGLRC	V_GLOBAL
ntCreateGLContext(HWND win_out,HPALETTE *palh);

//     24- .
DECLSPEC  CreateDIBContext(HDC DibDC, HGLRC *lpRC);

//////////////////////////////////////////////////////////////////////////

static unsigned char threeto8[8] = {0, 0111>>1, 0222>>1, 0333>>1, 0444>>1,
  0555>>1, 0666>>1, 0377 };
static unsigned char twoto8[4] = {  0, 0x55, 0xaa, 0xff };
static unsigned char oneto8[2] = {  0, 255 };
static int defaultOverride[13] = { 0, 3, 24, 27, 64, 67, 88, 173, 181,
  236, 247, 164, 91 };

static PALETTEENTRY defaultPalEntry[20] = {
    { 0,   0,   0,    0 },
    { 0x80,0,   0,    0 },
    { 0,   0x80,0,    0 },
    { 0x80,0x80,0,    0 },
    { 0,   0,   0x80, 0 },
    { 0x80,0,   0x80, 0 },
    { 0,   0x80,0x80, 0 },
    { 0xC0,0xC0,0xC0, 0 },

    { 192, 220, 192,  0 },
    { 166, 202, 240,  0 },
    { 255, 251, 240,  0 },
    { 160, 160, 164,  0 },

    { 0x80,0x80,0x80, 0 },
    { 0xFF,0,   0,    0 },
    { 0,   0xFF,0,    0 },
    { 0xFF,0xFF,0,    0 },
    { 0,   0,   0xFF, 0 },
    { 0xFF,0,   0xFF, 0 },
    { 0,   0xFF,0xFF, 0 },
    { 0xFF,0xFF,0xFF, 0 }
};

static unsigned char
ComponentFromIndex(int i, UINT nbits, UINT shift)
{
  unsigned char val;
  val=(unsigned char)(i>>shift);
  switch(nbits)
  {
    case 1:
      val &= 0x1;
      return oneto8[val];
    case 2:
      val &= 0x3;
      return twoto8[val];
    case 3:
      val &= 0x7;
      return threeto8[val];
    default:
      return 0;
  }
}

DECLSPEC  CreateWindowContext(HWND hWnd, HGLRC *lpRC, HPALETTE *lphPal)
{
  int n, i;
  HDC dc;
  PIXELFORMATDESCRIPTOR pfd;
  LOGPALETTE *pPal;
  int iPixelFormat;
  DWORD Result=0;
  *lpRC=NULL;
  *lphPal=NULL;
  dc = GetDC(hWnd);
  memset(&pfd,0,sizeof(PIXELFORMATDESCRIPTOR));
  pfd.nSize=sizeof(PIXELFORMATDESCRIPTOR);
  pfd.nVersion=1;
  pfd.dwFlags=PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
  pfd.iPixelType=PFD_TYPE_RGBA;
  pfd.cDepthBits=16;
  pfd.iLayerType=PFD_MAIN_PLANE;
  //      
  iPixelFormat=ChoosePixelFormat(dc, &pfd);
  if(!iPixelFormat)
  {
    ReleaseDC(dc,hWnd);
    return Result;
  }
  //   
  if(!(SetPixelFormat(dc, iPixelFormat, &pfd)))
  {
    ReleaseDC(dc,hWnd);
    return Result;
  }
  n=GetPixelFormat(dc);
  //     
  DescribePixelFormat(dc,n,sizeof(PIXELFORMATDESCRIPTOR),&pfd);
  //  -    
  if((pfd.dwFlags & PFD_GENERIC_FORMAT)==0) // bugbug - this is wrong (see MGL example) !
  {
    Result|=HWACCEL_FOUND;
  }
  if(pfd.dwFlags & PFD_NEED_PALETTE)
  {
    Result|=PALETTE_NEEDED;
    n = 1 << pfd.cColorBits;
    pPal=(PLOGPALETTE)malloc(sizeof(LOGPALETTE)+n*sizeof(PALETTEENTRY));
    pPal->palVersion=0x300;
    pPal->palNumEntries=n;
    for(i=0;i<n;i++)
    {
      pPal->palPalEntry[i].peRed =
        ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
      pPal->palPalEntry[i].peGreen =
        ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
      pPal->palPalEntry[i].peBlue =
        ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
      pPal->palPalEntry[i].peFlags = 0;
    }
    if((pfd.cColorBits==8)&&(pfd.cRedBits==3)&&(pfd.cRedShift==0)&&
       (pfd.cGreenBits==3)&&(pfd.cGreenShift==3)&&
       (pfd.cBlueBits == 2)&&(pfd.cBlueShift==6))
    {
      for(i=1;i<=12;i++)
        pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
    }
    *lphPal = CreatePalette(pPal);
    free(pPal);
  }
  else
    *lphPal=NULL;
  *lpRC = wglCreateContext(dc);
  return Result;
}

DECLSPEC  CreateDIBContext(HDC DibDC, HGLRC *lpRC)
{
  int n, i;
  static PIXELFORMATDESCRIPTOR pfd;
  int iPixelFormat;
  *lpRC=NULL;
  memset(&pfd,0,sizeof(PIXELFORMATDESCRIPTOR));
  pfd.nSize=sizeof(PIXELFORMATDESCRIPTOR);
  pfd.nVersion=1;
  pfd.dwFlags=PFD_DRAW_TO_BITMAP|PFD_SUPPORT_OPENGL|PFD_SUPPORT_GDI;
  pfd.iPixelType=PFD_TYPE_RGBA;
  pfd.cColorBits=24;
  pfd.cDepthBits=32;
  pfd.iLayerType=PFD_MAIN_PLANE;
  //      
  iPixelFormat=ChoosePixelFormat(DibDC, &pfd);
  if(!iPixelFormat) return 0;
  //   
  if(!(SetPixelFormat(DibDC, iPixelFormat, &pfd))) return 0;
  *lpRC = wglCreateContext(DibDC);
  return 1;
}

BOOL WINAPI 
DllEntryPoint(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  switch(fdwReason)
  {
    case DLL_PROCESS_ATTACH:
    case DLL_THREAD_ATTACH:
    case DLL_PROCESS_DETACH:
    case DLL_THREAD_DETACH:
      break;
  }
  return TRUE;
}




/*
==================================
        
==================================
*/

unsigned char threeto8[8] = {
    0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377
};

unsigned char twoto8[4] = {
    0, 0x55, 0xaa, 0xff
};

unsigned char oneto8[2] = {
    0, 255
};

static int defaultOverride[13] = {
    0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91
};

static PALETTEENTRY defaultPalEntry[20] = {
    { 0,   0,   0,    0 },
    { 0x80,0,   0,    0 },
    { 0,   0x80,0,    0 },
    { 0x80,0x80,0,    0 },
    { 0,   0,   0x80, 0 },
    { 0x80,0,   0x80, 0 },
    { 0,   0x80,0x80, 0 },
    { 0xC0,0xC0,0xC0, 0 },

    { 192, 220, 192,  0 },
    { 166, 202, 240,  0 },
    { 255, 251, 240,  0 },
    { 160, 160, 164,  0 },

    { 0x80,0x80,0x80, 0 },
    { 0xFF,0,   0,    0 },
    { 0,   0xFF,0,    0 },
    { 0xFF,0xFF,0,    0 },
    { 0,   0,   0xFF, 0 },
    { 0xFF,0,   0xFF, 0 },
    { 0,   0xFF,0xFF, 0 },
    { 0xFF,0xFF,0xFF, 0 }
};

static unsigned char ComponentFromIndex(int i, UINT nbits, UINT shift)
{
    unsigned char val;

    val = (unsigned char) (i >> shift);
    switch (nbits) {

    case 1:
        val &= 0x1;
        return oneto8[val];

    case 2:
        val &= 0x3;
        return twoto8[val];

    case 3:
        val &= 0x7;
        return threeto8[val];

    default:
        return 0;
    }
}

static void CreateRGBPalette(GLIMLayer *layer)
{
    PIXELFORMATDESCRIPTOR pfd;
    LOGPALETTE *pPal;
    int n, i;

    n = layer->pGetPixelFormat(layer->pwglGetCurrentDC());
    layer->pDescribePixelFormat(layer->pwglGetCurrentDC(), n, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
    
    {
      int genericFormat = (pfd.dwFlags & PFD_GENERIC_FORMAT) != 0;
      int genericAccel = (pfd.dwFlags & PFD_GENERIC_ACCELERATED) != 0;
      /* Determine hardware driver status by the combination of these
       * two flags.  Both FALSE means ICD, both TRUE means MCD.
       */
      if(genericFormat && genericAccel)
        layer->RenderType = MCD;
      else
        if((!genericFormat) && (!genericAccel))
          layer->RenderType = ICD;
        else
          layer->RenderType = Software;
    }
    
    if (pfd.dwFlags & PFD_NEED_PALETTE) {
        n = 1 << pfd.cColorBits;
        pPal = (PLOGPALETTE)malloc(sizeof(LOGPALETTE) +
                n * sizeof(PALETTEENTRY));
        pPal->palVersion = 0x300;
        pPal->palNumEntries = n;
        for (i=0; i<n; i++) {
            pPal->palPalEntry[i].peRed =
                    ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
            pPal->palPalEntry[i].peGreen =
                    ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
            pPal->palPalEntry[i].peBlue =
                    ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
            pPal->palPalEntry[i].peFlags = 0;
        }

        /* fix up the palette to include the default GDI palette */
        if ((pfd.cColorBits == 8)                           &&
            (pfd.cRedBits   == 3) && (pfd.cRedShift   == 0) &&
            (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) &&
            (pfd.cBlueBits  == 2) && (pfd.cBlueShift  == 6)
           ) {
            for (i = 1 ; i <= 12 ; i++)
                pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
        }

        layer->ghPalette = CreatePalette(pPal);
        free(pPal);

        layer->ghpalOld = SelectPalette(layer->pwglGetCurrentDC(), layer->ghPalette, FALSE);
        n = RealizePalette(layer->pwglGetCurrentDC());
    }
}

static uboolean bSetupPixelFormat(GLIMLayer *layer,HWND hWnd)
{
  HDC dc;
  static PIXELFORMATDESCRIPTOR 
  pfd = {
  sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
  1,        // version number
  PFD_DRAW_TO_WINDOW |    // support window
    PFD_SUPPORT_OPENGL |    // support OpenGL
    PFD_DOUBLEBUFFER,   // double buffered
  PFD_TYPE_RGBA,      // RGBA type
  32,       // 32-bit color depth
  0, 0, 0, 0, 0, 0,   // color bits ignored
  0,        // no alpha buffer
  0,        // shift bit ignored
  0,        // no accumulation buffer
  0, 0, 0, 0,       // accum bits ignored
  32,       // 32-bit z-buffer  
  0,        // no stencil buffer
  0,        // no auxiliary buffer
  PFD_MAIN_PLANE,     // main layer
  0,        // reserved
  0, 0, 0       // layer masks ignored
  };
  int pixelformat;
  dc=GetDC(hWnd);
    if ( (pixelformat = layer->pChoosePixelFormat(dc, &pfd)) == 0 )
    {
        MessageBox(hWnd, "ChoosePixelFormat failed", "Error", MB_OK);
        ReleaseDC(hWnd,dc);
        return false;
        
    }

    if (layer->pSetPixelFormat(dc, pixelformat, &pfd) == FALSE)
    {
        MessageBox(hWnd, "SetPixelFormat failed", "Error", MB_OK);
        ReleaseDC(hWnd,dc);
        return false;
    }

    CreateRGBPalette(layer);

    ReleaseDC(hWnd,dc);
    return true;
   
}
