#include "refapi.hpp"
#include <windows.h>
#include <gl\gl.h>

#define MAX_TEXTURES    1024

class opengl_ref_c : public refapi_i
{
  public:
  virtual __stdcall uboolean initialize(int window_handle);
  virtual __stdcall uboolean finalize(void);
  //       3D-API.
  virtual __stdcall void set_render_state(int pname,int param);
  virtual __stdcall void load_modelview_matrix(matrix4_t matrix);
  virtual __stdcall void load_projection_matrix(matrix4_t matrix);
  virtual __stdcall uboolean load_texture(char *alias,int w,int h,
    void *texels,renderstate_t tiling,renderstate_t combine,
    renderstate_t magfilter,renderstate_t minfilter,unsigned *ptex=NULL);
  virtual __stdcall uboolean load_mipmap(char *alias,miphead_p mip,
    renderstate_t tiling,renderstate_t combine,unsigned *ptex=NULL);
  virtual __stdcall void free_texture(unsigned int tex);
  virtual __stdcall void free_texture(char *alias);
  virtual __stdcall void bind_texture(unsigned int tex);
  virtual __stdcall void bind_texture(char *alias);
  virtual __stdcall void *capture_frame_buffer(int &width,int &height);
  // transforms
  virtual __stdcall void pushmatrix();
  virtual __stdcall void popmatrix();
  virtual __stdcall void loadidentity();
  virtual __stdcall void loadmatrix4x4(matrix4x4_t src);
  virtual __stdcall void loadtransform(matrix3x4_t src);
  virtual __stdcall void scale(vec_t x,vec_t y,vec_t z);
  virtual __stdcall void scalev(vec3_t factor);
  virtual __stdcall void rotate_x(vec_t radians);
  virtual __stdcall void rotate_y(vec_t radians);
  virtual __stdcall void rotate_z(vec_t radians);
  virtual __stdcall void rotate(vec_t radians,vec_t x,vec_t y,vec_t z);
  virtual __stdcall void rotatev(vec_t radians,vec3_t direction);
  virtual __stdcall void translate(vec_t x,vec_t y,vec_t z);
  virtual __stdcall void translatev(vec3_t factor);
  virtual __stdcall void orto(vec_t left,vec_t right,vec_t bottom,vec_t top,vec_t near,vec_t far);
  virtual __stdcall void frustum(vec_t left,vec_t right,vec_t bottom,vec_t top,vec_t near,vec_t far);
  virtual __stdcall void viewport(int left,int bottom,int width,int height);
  // draw
  virtual __stdcall void clear(vec_t r,vec_t g,vec_t b,uboolean clear_z);
  
  virtual __stdcall void begin(primitivetype_t pname);
  virtual __stdcall void end();
  
  virtual __stdcall void color(vec_t r,vec_t g,vec_t b);
  virtual __stdcall void colorv(vec3_t rgb);
  virtual __stdcall void colorv(byte *rgba);
  
  virtual __stdcall void vertex(vec_t x,vec_t y,vec_t z);
  virtual __stdcall void vertexv(vec3_t xyz);
  
  virtual __stdcall void texcoord(vec_t u,vec_t v);
  virtual __stdcall void texcoordv(vec2_t uv);
  // fog
  virtual __stdcall void usefog(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec_t r,vec_t g,vec_t b,vec_t a);
  virtual __stdcall void usefogv(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec4_t rgba);
  // swap
  virtual __stdcall void swapbuffers();
  
  //private:
  HWND          hWnd;
  int           left,bottom;
  int           width,height;
  WNDPROC       prevWindowProc;
  char          *aliases[MAX_TEXTURES];
  unsigned int  tex[MAX_TEXTURES];
  
  public:
                opengl_ref_c(void) {prevWindowProc=NULL;}
                ~opengl_ref_c()
                {
                  if(prevWindowProc) 
                    SetClassLong(hWnd,GCL_WNDPROC,prevWindowProc);
                }
};


// implementation

static opengl_ref_c cur_ref = NULL;

static LRESULT CALLBACK 
opengl_callback(HWND hWnd,UINT Message,WPARAM wParam,LPARAM lParam)
{
  if(cur_ref)
  {
    switch(Message)
    {
      case WM_RESIZE:
        cur_ref->width=LOWORD(lParam);
        cur_ref->height=HIWORD(lParam);
        cur_ref->viewport(cur_ref->left,cur_ref->bottom,cur_ref->width,cur_ref->height);
        break;
      case WM_PAINT:
        cur_ref->swapbuffres();
        break;
      case WM_ERASEBKGND:
        return 1;
    }
    if(cur_ref->prevWindowProc)
      cur_ref->prevWindowProc(hWnd,Message,wParam,lParam);
    else
      DefWindowProc(hWnd,Message,wParam,lParam);
  }
  else
    DefWindowProc(hWnd,Message,wParam,lParam);
}

// standard OGL code =====================================================

// end of standard OGL code ==============================================

uboolean __stdcall opengl_ref_c::initialize(int window_handle)
{
  #error not implemented yet
}

uboolean __stdcall opengl_ref_c::finalize(void)
{
  HGLRC rc;
  HDC dc;
  if(prevWindowProc)
  {
    SetClassLong(hWnd,GCL_WNDPROC,prevWindowProc);
    prevWindowProc=NULL;
  }
  rc=wglGetCurrentContext();
  dc=wglGetCurrentDC();
  wglMakeCurrent(NULL,NULL);
  if(rc) wglDeleteContext(hRc);
  if(dc) ReleaseDC(NULL,hDc);
}

//       3D-API.
void __stdcall opengl_ref_c::set_render_state(int pname,int param)
{
  switch(pname)
  {
    case shademodel:
      switch(param)
      {
        case flat: glShadeModel(GL_FLAT); break;
        case gouraud: glShadeModel(GL_SMOOTH); break;
      }
      break;
    case ztest:
      if(param==zenable) 
        glEnable(GL_DEPTH_TEST); 
      else 
        glDisable(GL_DEPTH_TEST); 
      break;
    case alphatest:
      switch(param)
      {
        case alphadisable: 
          glDisable(GL_ALPHA_TEST); 
          break;
        case alphagreaterthen0:
          glEnable(GL_ALPHA_TEST); 
          glAlphaFunc(GL_GREATER,0.0);
          break;
        case alphagreaterthen128
          glEnable(GL_ALPHA_TEST); 
          glAlphaFunc(GL_GREATER,128.0);
          break;
      }
      break;
    case blendfunc
      switch(param)
      {
        case blendonezero: 
          glDisable(GL_BLEND);
          break;
        case blendoneone:
          glEnable(GL_BLEND);
          glBlendFunc(GL_ONE,GL_ONE);
          break;
        case blendonesrccolor:
          glEnable(GL_BLEND);
          glBlendFunc(GL_ONE,GL_SRC_COLOR);
          break;  #error
      }
      break;
    case:
    texmode
      break;
    case:
    texenable
      break;
    case:
    texdisable
      break;
    case:
    textiling
      break;
    case:
    texclamp
      break;
    case:
    texrepeat
      break;
    case:
    texmagfilter
      break;
    case:
    texnearest
      break;
    case:
    texlinear
      break;
    case:
    texminfilter
      break;
    case:
    texmipmap
      break;
    case:
    textrilinear
      break;
    case:
    texcombine
      break;
    case:
    texdecal
      break;
    case:
    texmodulate
      break;
    case:
    matrixmode
      break;
    case:
    matmodelview
      break;
    case:
    matprojection
      break;
    case:
    mattexture
      break;
  }
}

void __stdcall opengl_ref_c::loadmatrix(matrix4_t matrix)
{
  matrix4x4 glm;
  mat4Transpose(glm,matrix);
  #ifdef DOUBLEVEC
  glLoadMatricsd((PGLdouble)glm);
  #else
  glLoadMatricsf((PGLfloat)glm);
  #endif
}
  
uboolean __stdcall opengl_ref_c::load_texture(char *alias,int w,int h,
  void *texels,renderstate_t tiling,renderstate_t combine,
  renderstate_t magfilter,renderstate_t minfilter,unsigned *ptex=NULL)
{
  int i;
  for(i=0;i<MAX_TEXTURES && tex[i];i++);
  if(i==MAX_TEXTURES)
  {
    puts("i==MAX_TEXTURES: increase MAX_TEXTURES.");
    return false;
  }
  if(alias)
  {
    if(aliases[i]) free(aliases[i]);
    aliases[i]=strdup(alias);
  }
  glGenTextures(1,&tex[i]);
  glBindTexture(GL_TEXTURE_2D,tex[i]);
  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,w,h,0,GL_RGBA,GL_UNSIGNED_BYTE,
    texels);
  if(tiling==texclamp)
  {
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  }
  else
  {
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  }
  if(magfilter==texlinear) 
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  else
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  if(minfilter==texlinear)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  else
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  if(combine==texmodulate)
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  else
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
  if(ptex) *ptex=tex[i];
  glBindTexture(GL_TEXTURE_2D,0);
  return glGetError() == GL_NO_ERROR;
}  
  
uboolean __stdcall opengl_ref_c::load_mipmap(char *alias,miphead_p mip,
  renderstate_t tiling,renderstate_t combine,unsigned *ptex=NULL)
{
  int i;
  for(i=0;i<MAX_TEXTURES && tex[i];i++);
  if(i==MAX_TEXTURES)
  {
    puts("i==MAX_TEXTURES: increase MAX_TEXTURES.");
    return false;
  }
  if(alias)
  {
    if(aliases[i]) free(aliases[i]);
    aliases[i]=strdup(alias);
  }
  glGenTextures(1,&tex[i]);
  glBindTexture(GL_TEXTURE_2D,tex[i]);
  if(ptex) *ptex=tex[i];
  for(i=0;i<mip->levels;i++)
    glTexImage2D(GL_TEXTURE_2D,i,GL_RGBA,w,h,0,GL_RGBA,GL_UNSIGNED_BYTE,
      mip->levels[i]);
  if(tiling==texclamp)
  {
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  }
  else
  {
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  }
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
  if(combine==texmodulate)
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  else
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
  glBindTexture(GL_TEXTURE_2D,0);
  return glGetError() == GL_NO_ERROR;
}

void __stdcall opengl_ref_c::free_texture(unsigned int tex)

void __stdcall opengl_ref_c::free_texture(char *alias)

void __stdcall opengl_ref_c::bind_texture(unsigned int tex)

void __stdcall opengl_ref_c::bind_texture(char *alias)

void __stdcall opengl_ref_c::*capture_frame_buffer(int *width,int *height)
{
  byte *pix;
  *width=this->width;
  *height=this->height;
  pix=(byte*)malloc(width*height*4);
  glReadPixels(left,bottom,width,height,GL_RGBA,GL_UNSIGNED_BYTE);
}

// transforms
void __stdcall opengl_ref_c::pushmatrix();
void __stdcall opengl_ref_c::popmatrix();
void __stdcall opengl_ref_c::loadidentity();
void __stdcall opengl_ref_c::loadmatrix4x4(matrix4x4_t src);
void __stdcall opengl_ref_c::loadtransform(matrix3x4_t src);
void __stdcall opengl_ref_c::scale(vec_t x,vec_t y,vec_t z);
void __stdcall opengl_ref_c::scalev(vec3_t factor);
void __stdcall opengl_ref_c::rotate_x(vec_t radians);
void __stdcall opengl_ref_c::rotate_y(vec_t radians);
void __stdcall opengl_ref_c::rotate_z(vec_t radians);
void __stdcall opengl_ref_c::rotate(vec_t radians,vec_t x,vec_t y,vec_t z);
void __stdcall opengl_ref_c::rotatev(vec_t radians,vec3_t direction);
void __stdcall opengl_ref_c::translate(vec_t x,vec_t y,vec_t z);
void __stdcall opengl_ref_c::translatev(vec3_t factor);
void __stdcall opengl_ref_c::orto(vec_t left,vec_t right,vec_t bottom,vec_t top,vec_t near,vec_t far);
void __stdcall opengl_ref_c::frustum(vec_t left,vec_t right,vec_t bottom,vec_t top,vec_t near,vec_t far);
void __stdcall opengl_ref_c::viewport(int left,int bottom,int width,int height);
// draw
void __stdcall opengl_ref_c::clear(vec_t r,vec_t g,vec_t b,uboolean clear_z);

void __stdcall opengl_ref_c::begin(primitivetype_t pname);
void __stdcall opengl_ref_c::end();

void __stdcall opengl_ref_c::color(vec_t r,vec_t g,vec_t b);
void __stdcall opengl_ref_c::colorv(vec3_t rgb);
void __stdcall opengl_ref_c::colorv(byte *rgba);

void __stdcall opengl_ref_c::vertex(vec_t x,vec_t y,vec_t z);
void __stdcall opengl_ref_c::vertexv(vec3_t xyz);

void __stdcall opengl_ref_c::texcoord(vec_t u,vec_t v);
void __stdcall opengl_ref_c::texcoordv(vec2_t uv);
// fog
void __stdcall opengl_ref_c::usefog(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec_t r,vec_t g,vec_t b,vec_t a);
void __stdcall opengl_ref_c::usefogv(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec4_t rgba);
// swap
void __stdcall opengl_ref_c::swapbuffers();















  
static GLvoid CALLBACK  drawScene(GLvoid)
{
  if(single_instance) single_instance->drawscene();
  auxSwapBuffers();
}

static GLvoid CALLBACK Resize(GLsizei width,GLsizei height)
{
  GLfloat aspect;
  glViewport( 0, 0, width, height );
  aspect = (GLfloat) width / height;
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-aspect,aspect,1,-1,1,100);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}    

renderer_c::renderer_c(int w,int h,char *title,int rrate)
{
  if(rrate>-1)
  {
    if(!dispSetMode(w,h,16,rrate))
    {
      puts("auxrenderer_c::auxrenderer_c: ChangeDisplaySettings() failed.");
      ExitProcess(1);
    }
  }
  u_zero(aliases,sizeof(char*)*MAX_TEXTURES);
  u_zero(tex,sizeof(int)*MAX_TEXTURES);
  auxInitPosition(0,0,w,h);
  auxInitDisplayMode(AUX_RGB | AUX_DEPTH | AUX_DOUBLE);
  auxInitWindow(title);
  SetClassLong(GetActiveWindow(),GCL_HBRBACKGROUND,0);
  if(rrate>-1)
    SetWindowLong(GetActiveWindow(),GWL_STYLE,WS_POPUP);
  auxIdleFunc(drawScene);
  auxReshapeFunc(Resize);
  glClearColor(0.0,0.0,0.0,0.0);
  glClearDepth(10);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  Resize(w,h);
  if(initialize()) 
  {
    single_instance=this;
    auxMainLoop(drawScene);
    single_instance=NULL;
  }
  if(rrate>-1)
  {
    DEVMODE dm;
    u_zero(&dm,sizeof(dm));
    dm.dmSize=sizeof(dm);
    dm.dmPelsWidth=ow;
    dm.dmPelsHeight=oh;
    dm.dmFields=DM_PELSWIDTH|DM_PELSHEIGHT;
    ChangeDisplaySettings(&dm,0);
  }
}

renderer_c::~renderer_c()
{
  int i;
  finalize();
  for(i=0;i<MAX_TEXTURES;i++)
  {
    if(aliases[i]) free(aliases[i]);
    if(tex[i]) glDeleteTextures(1,tex+i);
  }
}

void renderer_c::clear(vec_t r,vec_t g,vec_t b,uboolean z_clear)
{
  glClearColor(r,g,b,0.5);
  if(z_clear)
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  else
    glClear(GL_COLOR_BUFFER_BIT);
}

void renderer_c::begin(primitivetype_t prim)
{
  switch(prim)
  {
    case points: glBegin(GL_POINTS); return;
    case lines: glBegin(GL_LINES); return;
    case linestrip: glBegin(GL_LINE_STRIP); return;
    case triangles: glBegin(GL_TRIANGLES); return;
    case tristrip: glBegin(GL_TRIANGLE_FAN); return;
    case trifan: glBegin(GL_TRIANGLE_STRIP); return;
    case quads: glBegin(GL_QUADS); return;
    case quadstrip: glBegin(GL_QUAD_STRIP); return;
    case polygon: glBegin(GL_POLYGON); return;
  }
  glBegin(GL_LINE_STRIP);
}

void renderer_c::end()
{
  glEnd();
}

void renderer_c::color(vec_t r,vec_t g,vec_t b)
{
  glColor3f(r,g,b);
}

void renderer_c::colorv(vec3_t rgb)
{
  glColor3fv(rgb);
}

void renderer_c::colorv(void *rgba)
{
  glColor3ubv((byte*)rgba);
}

void renderer_c::vertex(vec_t x,vec_t y,vec_t z)
{
  glVertex3f(x,y,z);
}

void renderer_c::vertexv(vec3_t xyz)
{
  glVertex3fv(xyz);
}

void renderer_c::texcoord(vec_t u,vec_t v)
{
  glTexCoord2f(u,v);
}

void renderer_c::texcoordv(vec2_t uv)
{
  glTexCoord2fv(uv);
}

void renderer_c::normal(vec_t x,vec_t y,vec_t z)
{
  glNormal3f(x,y,z);
}

void renderer_c::normalv(vec3_t xyz)
{
  glNormal3fv(xyz);
}

void renderer_c::translate(vec_t x,vec_t y,vec_t z)
{
  glTranslatef(x,y,z);
}

void renderer_c::translatev(vec3_t xyz)
{
  glTranslatef(xyz[0],xyz[1],xyz[2]);
}

void renderer_c::rotate(vec_t degrees,vec_t dir_x,vec_t dir_y,vec_t dir_z)
{
  glRotatef(degrees,dir_x,dir_y,dir_z);
}

void renderer_c::rotatev(vec_t degrees,vec3_t dir)
{
  glRotatef(degrees,dir[0],dir[1],dir[2]);
}

void renderer_c::scale(vec_t x,vec_t y,vec_t z)
{
  glScalef(x,y,z);
}

void renderer_c::scalev(vec3_t xyz)
{
  glScalef(xyz[0],xyz[1],xyz[2]);
}

void renderer_c::matrixmode(matrixmode_t mode)
{
  switch(mode)
  {
    case matprojection: glMatrixMode(GL_PROJECTION);return;
    case mattexture: glMatrixMode(GL_TEXTURE);return;
  }
  glMatrixMode(GL_MODELVIEW);
}

void renderer_c::pushmatrix()
{
  glPushMatrix();
}

void renderer_c::popmatrix()
{
  glPopMatrix();
}

void renderer_c::loadidentity()
{
  glLoadIdentity();
}

void renderer_c::loadmatrix(matrix3x4_t m)
{
  int i,j;
  matrix4_t t;
  for(i=0;i<3;i++)
    for(j=0;j<3;j++)
      t[i][j]=m[j][i];
  t[3][0]=m[0][3];
  t[3][1]=m[1][3];
  t[3][2]=m[2][3];
  t[0][3]=t[1][3]=t[2][3]=0;
  t[3][3]=1;
  glLoadMatrixf((GLfloat*)t);
}

// caps & fx
void renderer_c::enable(rendercaps_t pname)
{
  switch(pname)
  {
    case alpha: glEnable(GL_ALPHA_TEST); return;
    case blend: glEnable(GL_BLEND); return;
    case cullface: glEnable(GL_CULL_FACE); return;
    case depthtest: glEnable(GL_DEPTH_TEST); return;
    case fog: glEnable(GL_FOG); return;
    case pointsmooth: glEnable(GL_POINT_SMOOTH); return;
    case texture: glEnable(GL_TEXTURE_2D); return;
  }
}

void renderer_c::disable(rendercaps_t pname)
{
  switch(pname)
  {
    case alpha: glDisable(GL_ALPHA_TEST); return;
    case blend: glDisable(GL_BLEND); return;
    case cullface: glDisable(GL_CULL_FACE); return;
    case depthtest: glDisable(GL_DEPTH_TEST); return;
    case fog: glDisable(GL_FOG); return;
    case pointsmooth: glDisable(GL_POINT_SMOOTH); return;
    case texture: glDisable(GL_TEXTURE_2D); return;
  }
}

void renderer_c::fxcaps(fxcaps_t pname)
{
  switch(pname)
  {
    case alphaless: glAlphaFunc(GL_LESS,.5); return;
    case alphagreater: glAlphaFunc(GL_GREATER,.5); return;
    case blendoneone: glBlendFunc(GL_ONE,GL_ONE); return;
    case blendonesrccolor: glBlendFunc(GL_ONE,GL_SRC_COLOR); return;
    case foglinear: glFogi(GL_FOG_MODE,GL_LINEAR);
    case fogexp: glFogi(GL_FOG_MODE,GL_EXP);
  }
}

uboolean renderer_c::loadtexture(char *alias,int w,int h,void *texels,
  fxcaps_t tiling,fxcaps_t combine,fxcaps_t magfilter,fxcaps_t minfilter,
  unsigned int *ptex)
{
}
  
void renderer_c::texfree(unsigned int tex)
{
  int i;
  for(i=0;i<MAX_TEXTURES && this->tex[i]!=tex;i++);
  if(i==MAX_TEXTURES)
  {
    printf("renderer_c::freetex: texture %d not found !\n",tex);
  }
  else
  {
    glDeleteTextures(1,&tex);
    this->tex[i]=0;
    if(aliases[i]) 
    {
      free(aliases[i]);
      aliases[i]=NULL;
    }
  }
}

void renderer_c::texfree(char *alias)
{
  int i;
  for(i=0;i<MAX_TEXTURES;i++)
    if(aliases[i])
      if(!strcmp(aliases[i],alias)) break;
  if(i==MAX_TEXTURES)
  {
    printf("renderer_c::freetex: texture %s not found !\n",alias);
  }
  else
  {
    glDeleteTextures(1,&tex[i]);
    tex[i]=0;
    free(aliases[i]);
    aliases[i]=NULL;
  }
}

void renderer_c::texbind(unsigned int tex)
{
  glBindTexture(GL_TEXTURE_2D,tex);
}

void renderer_c::texbind(char *alias)
{
  int i;
  for(i=0;i<MAX_TEXTURES;i++)
    if(aliases[i])
      if(!strcmp(aliases[i],alias)) break;
  if(i==MAX_TEXTURES)
  {
    printf("renderer_c::bindtex: texture %s not found !\n",alias);
  }
  else
    glBindTexture(GL_TEXTURE_2D,tex[i]);
}

void renderer_c::usefog(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec_t r,vec_t g,vec_t b,vec_t a)
{
  vec4_t rgba;
  if(mode==fogexp)
    glFogi(GL_FOG_MODE,GL_EXP);
  else
    glFogi(GL_FOG_MODE,GL_LINEAR);
  glFogf(GL_FOG_START,start);
  glFogf(GL_FOG_END,end);
  glFogf(GL_FOG_DENSITY,dens);
  glFogfv(GL_FOG_COLOR,vec4Set(rgba,r,g,b,a));
}

void renderer_c::usefog(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec4_t rgba)
{
  if(mode==fogexp)
    glFogi(GL_FOG_MODE,GL_EXP);
  else
    glFogi(GL_FOG_MODE,GL_LINEAR);
  glFogf(GL_FOG_START,start);
  glFogf(GL_FOG_END,end);
  glFogf(GL_FOG_DENSITY,dens);
  glFogfv(GL_FOG_COLOR,rgba);
}
