#ifndef REFAPI_HPP
#define REFAPI_HPP

#include "..\..\..\include\u_lib.hpp"

typedef enum
{
  points,lines,linestrip,triangles,tristrip,trifan,quads,quadstrip,polygon
} primitivetype_t;

typedef enum
{
  // shading
  shademodel, flat, gouraud,
  // z-buffering
  ztest, zenable, zdisable,
  // alpha-chromakeying
  alphatest, alphadisable, alphagreaterthen0, alphagreaterthen128,
  // color blending
  blendfunc, blendonezero, blendoneone, blendonesrccolor,
  // texturing
  texmode, texenable, texdisable,
  textiling, texclamp, texrepeat,
  texmagfilter, texnearest, texlinear,
  texminfilter, texmipmap, textrilinear,
  texcombine, texdecal, texmodulate,
  // matrix mode
  matrixmode, matmodelview, matprojection, mattexture,
  // transforms
} renderstate_t;

class refapi_i
{
  public:
  virtual __stdcall uboolean initialize(int window_handle) = 0;
  virtual __stdcall uboolean finalize(void) = 0;
  //       3D-API.
  virtual __stdcall void set_render_state(int pname,int param) = 0;
  virtual __stdcall uboolean load_texture(char *alias,int w,int h,
    void *texels,renderstate_t tiling,renderstate_t combine,
    renderstate_t magfilter,renderstate_t minfilter,unsigned *ptex=NULL);
  virtual __stdcall uboolean load_mipmap(char *alias,miphead_p mip,
    renderstate_t tiling,renderstate_t combine,unsigned *ptex=NULL);
  virtual __stdcall void free_texture(unsigned int tex) = 0;
  virtual __stdcall void free_texture(char *alias) = 0;
  virtual __stdcall void bind_texture(unsigned int tex) = 0;
  virtual __stdcall void bind_texture(char *alias) = 0;
  virtual __stdcall void *capture_frame_buffer(int *width,int *height) = 0;
  // transforms
  virtual __stdcall void pushmatrix() = 0;
  virtual __stdcall void popmatrix() = 0;
  virtual __stdcall void loadidentity() = 0;
  virtual __stdcall void loadmatrix4x4(matrix4x4_t src) = 0;
  virtual __stdcall void loadtransform(matrix3x4_t src) = 0;
  virtual __stdcall void scale(vec_t x,vec_t y,vec_t z) = 0;
  virtual __stdcall void scalev(vec3_t factor) = 0;
  virtual __stdcall void rotate_x(vec_t radians) = 0;
  virtual __stdcall void rotate_y(vec_t radians) = 0;
  virtual __stdcall void rotate_z(vec_t radians) = 0;
  virtual __stdcall void rotate(vec_t radians,vec_t x,vec_t y,vec_t z) = 0;
  virtual __stdcall void rotatev(vec_t radians,vec3_t direction) = 0;
  virtual __stdcall void translate(vec_t x,vec_t y,vec_t z) = 0;
  virtual __stdcall void translatev(vec3_t factor) = 0;
  virtual __stdcall void orto(vec_t left,vec_t right,vec_t bottom,vec_t top,vec_t near,vec_t far) = 0;
  virtual __stdcall void frustum(vec_t left,vec_t right,vec_t bottom,vec_t top,vec_t near,vec_t far) = 0;
  virtual __stdcall void viewport(int left,int bottom,int width,int height) = 0;
  // draw
  virtual __stdcall void clear(vec_t r,vec_t g,vec_t b,uboolean clear_z) = 0;
  
  virtual __stdcall void begin(primitivetype_t pname) = 0;
  virtual __stdcall void end() = 0;
  
  virtual __stdcall void color(vec_t r,vec_t g,vec_t b) = 0;
  virtual __stdcall void colorv(vec3_t rgb) = 0;
  virtual __stdcall void colorv(byte *rgba) = 0;
  
  virtual __stdcall void vertex(vec_t x,vec_t y,vec_t z) = 0;
  virtual __stdcall void vertexv(vec3_t xyz) = 0;
  
  virtual __stdcall void texcoord(vec_t u,vec_t v) = 0;
  virtual __stdcall void texcoordv(vec2_t uv) = 0;
  // fog
  virtual __stdcall void usefog(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec_t r,vec_t g,vec_t b,vec_t a) = 0;
  virtual __stdcall void usefogv(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec4_t rgba) = 0;
  // swap
  virtual __stdcall void swapbuffers() = 0;
};

UEXPORT refapi_i *GetRefApi(void);
UEXPORT void ReleaseRefApi(refapi_i *refapi);

#endif

