#include <windows.h>
#include <dsound.h>
#include <stdio.h>
#include "../../include/v_lib.h"
#include "../../source/v_mt.h"

void_p	V_CALL
libvIsHandle(void *handle,char *caller,char *object,u32_t magic)
{
	if(!handle)
	{
		conlPrintf(LOG_ERROR,"%s(): NULL descriptor.\n",caller);
		return NULL;
	}
	if(*((u32_t*)handle) != magic)
	{
		conlPrintf(LOG_ERROR,"%s(): %s damaged: bad magic.\n",caller,
			object ? object : "object");
		return NULL;
	}
	return handle;
}

typedef void	*sound_engine_t;

#define	INVALID_INDEX			0xFFFFFFFF
#define	INVALID_BUFFER_VALUE	INVALID_INDEX

/*
 *	Implemenation
 */

typedef struct	ds_sound_info_s
{
	IDirectSoundBuffer	*idsb;
	char				wave_name[_MAX_PATH];	/* variable sized */
}	ds_sound_info_t, *ds_sound_info_p;

typedef struct	ds_sound_engine_s
{
#ifdef	CHECKED_BUILD
	u32_t				magic;
#endif	/* CHECKED_BUILD */
	/*
	HANDLE				dsound;
	*/
	IDirectSound		*ids;
	vp_stack_t			sounds;
	HRESULT				err;
}	ds_sound_engine_t, *ds_sound_engine_p;

#ifdef	THIS
#undef	THIS
#endif
#define	THIS(handle)	((ds_sound_engine_t*)(handle))

#ifdef	CHECKED_BUILD
V_LOCAL	char		*se_object="DSound engine";
V_LOCAL	u32_t		se_magic=0x44534E44;	/* DSND */
#endif	/* CHECKED_BUILD */

void	V_GLOBAL
seDestroyEngine(sound_engine_t engine)
{
	int i,n;
	ds_sound_info_t **sounds;
	if(!libvIsHandle(engine,"seDestroyEngine",se_object,se_magic))
		return;
	/* Release all buffers */
	sounds=(ds_sound_info_t**)vpsGetStorage(&THIS(engine)->sounds);
	n=vpsCount(&THIS(engine)->sounds);
	for(i=0;i<n;i++)
	{
		if(sounds[i])
		{
			IDirectSoundBuffer_Release(sounds[i]->idsb);
			libvFree(sounds[i]);
		}
	}
	vpsClear(&THIS(engine)->sounds);
	/* Release DirectSound */
	if(THIS(engine)->ids)
		THIS(engine)->ids->lpVtbl->Release(THIS(engine)->ids);
	/* Unload library */
	/* FreeLibrary(THIS(engine)->dsound); */
	libvFree(engine);
}

/*
 *	'app_window' is a handle (of type HWND) of main window.
 */
sound_engine_t	V_GLOBAL
seCreateEngine(void *app_window)
{
	ds_sound_engine_t *se;
	ALLOC_OBJECT(se,ds_sound_engine_t,0,se_magic);
	vpsInit(&se->sounds);
	/*
	 *	TODO: load dsound.dll dynamically, determine it's version,
	 *	and bail out if running under NT < 4.x
	 */
	se->err=DirectSoundCreate(NULL,&se->ids,NULL);
	if(se->err!=DS_OK)
	{
		libvFree(se);
		return NULL;
	}
	se->err=IDirectSound_SetCooperativeLevel(se->ids,app_window,DSSCL_PRIORITY);
	if(se->err!=DS_OK)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_WARNING,"seCreateEngine(): "
			"IDirectSound()->SetCooperativeLevel(DSSCL_PRIORITY) failed.\n");
#endif	/* CHECKED_BUILD */
		se->err=IDirectSound_SetCooperativeLevel(se->ids,app_window,DSSCL_NORMAL);
		if(se->err!=DS_OK)
		{
#ifdef	CHECKED_BUILD
			conlPuts(LOG_ERROR,"seCreateEngine(): "
				"IDirectSound()->SetCooperativeLevel(DSSCL_NORMAL) failed.\n");
#endif	/* CHECKED_BUILD */
			se->ids->lpVtbl->Release(se->ids);
			libvFree(se);
			return NULL;
		}
	}
	return (sound_engine_t)se;
}

/*
**	Returns sound ID or INVALID_BUFFER_VALUE on error
*/
u32_t	V_GLOBAL
seLoadSound(sound_engine_t engine,char *wave_name)
{
	wave_handle_t wave;
	WAVEFORMATEX wf;
	DSBUFFERDESC bd;
	int name_len;
	void *ptr1, *ptr2;
	IDirectSoundBuffer *idsb;
	DWORD size1, size2;
	ds_sound_info_t *dsi;
	u32_t buffer_num;
	if(!libvIsHandle(engine,"seLoadSound",se_object,se_magic))
		return INVALID_BUFFER_VALUE;
	wave=wavLoad(wave_name);
	if(!wave)
		return INVALID_BUFFER_VALUE;
#if 0
	if(wavNumChannels(wave)!=1)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"seLoadSound('%s'): only mono sound supported.\n",
			wave_name);
#endif	/* CHECKED_BUILD */
		libvFree(data);
		return INVALID_BUFFER_VALUE;
	}
#endif /* 0 */
	memZero(&wf,sizeof(WAVEFORMATEX));
	wf.wFormatTag=WAVE_FORMAT_PCM;
	wf.nChannels=(SHORT)wavNumChannels(wave);
	wf.nSamplesPerSec=wavSamplingFrequency(wave);
	wf.wBitsPerSample=(SHORT)wavBitsPerSample(wave);
	wf.nBlockAlign=(SHORT)((wf.wBitsPerSample>>3) * wavNumChannels(wave));
	wf.nAvgBytesPerSec=wf.nSamplesPerSec * wf.nBlockAlign;
	memset(&bd,0,sizeof(DSBUFFERDESC));
	bd.dwSize=sizeof(DSBUFFERDESC);
	bd.dwFlags=DSBCAPS_CTRLFREQUENCY|DSBCAPS_CTRLPAN|DSBCAPS_CTRLVOLUME;
	bd.dwBufferBytes=wavNumSamples(wave) * wf.nBlockAlign;
	bd.lpwfxFormat=&wf;
	THIS(engine)->err=IDirectSound_CreateSoundBuffer(THIS(engine)->ids,&bd,&idsb,NULL);
	if(THIS(engine)->err!=DS_OK)
	{
		/* Reduce number of controls and try again */
		bd.dwFlags=DSBCAPS_CTRLVOLUME;
		THIS(engine)->err=IDirectSound_CreateSoundBuffer(THIS(engine)->ids,&bd,&idsb,NULL);
		if(THIS(engine)->err!=DS_OK)
		{
			wavClose(wave);
			return INVALID_BUFFER_VALUE;
		}
	}
	/* lock buffer and copy data */
	THIS(engine)->err=IDirectSoundBuffer_Lock(idsb,0,bd.dwSize,&ptr1,&size1,
		&ptr2,&size2,DSBLOCK_ENTIREBUFFER);
	if(THIS(engine)->err!=DS_OK || size1!=bd.dwSize)
	{
		idsb->lpVtbl->Release(idsb);
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"seLoadSound('%s'): IDirectSoundBuffer->Lock() failed.\n",
			wave_name);
#endif	/* CHECKED_BUILD */
		THIS(engine)->err=DSERR_GENERIC;
		wavClose(wave);
		return INVALID_BUFFER_VALUE;
	}
	memCopy(ptr1,wavData(wave),bd.dwSize);
	IDirectSoundBuffer_Unlock(idsb,ptr1,size1,ptr2,size2);
	wavClose(wave);
	name_len=strlen(wave_name);
	dsi=(ds_sound_info_t*)libvMalloc(sizeof(void*)+name_len+4);
	dsi->idsb=idsb;
	memcpy(dsi->wave_name,wave_name,name_len+1);
	buffer_num=vpsStore(&THIS(engine)->sounds,dsi);
	return buffer_num;
}

/*
 *	Returns sound info with IDirectSoundBuffer or NULL.
 */
ds_sound_info_p	V_LOCAL
#ifdef	CHECKED_BUILD
get_sound_info(char *caller,sound_engine_t *engine,u32_t buffer_num)
#else	/* !CHECKED_BUILD */
#define	get_sound_info(caller,engine,buffer)	\
		get_sound_info_quiet(engine,buffer)
get_sound_info_quiet(sound_engine_t *engine,u32_t buffer_num)
#endif	/* !CHECKED_BUILD
 */
{
	u32_t len;
	ds_sound_info_t *info=NULL;
	if(!libvIsHandle(engine,caller,se_object,se_magic))
		return NULL;
	len=vpsCount(&THIS(engine)->sounds);
	if(buffer_num<len)
		info=(ds_sound_info_t*)vpsGet(&THIS(engine)->sounds,buffer_num);
	if(!info)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"%s(%d): %s buffer.\n",
			caller,buffer_num,(buffer_num<len) ? "empty" : "invalid");
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	return info;
}

vboolean	V_GLOBAL
sePlaySound(sound_engine_t *engine,u32_t buffer_num)
{
	ds_sound_info_t *info;
	info=get_sound_info("sePlaySound",engine,buffer_num);
	if(!info)
		return false;
	IDirectSoundBuffer_SetCurrentPosition(info->idsb,0);
	THIS(engine)->err=IDirectSoundBuffer_Play(info->idsb,0,0,0);
	return THIS(engine)->err == DS_OK;
}

vboolean	V_GLOBAL
seFreeSound(sound_engine_t *engine,u32_t buffer_num)
{
	ds_sound_info_t *info;
	info=get_sound_info("seFreeSound",engine,buffer_num);
	if(!info)
		return false;
	info->idsb->lpVtbl->Release(info->idsb);
	THIS(engine)->err=DS_OK;
	vpsSet(&THIS(engine)->sounds,buffer_num,NULL);
	return true;
}

/*
 *	Float attenuation in decibels; 0 mean full folume.
 */
vboolean	V_GLOBAL
seSetVolume(sound_engine_t *engine,u32_t buffer_num,u32_t attenuation)
{
	ds_sound_info_t *info;
	LONG la;
	info=get_sound_info("seSetVolume",engine,buffer_num);
	if(!info)
		return false;
	/*
	 *	Attenuation in hundredths of decibels, as documented for
	 *	IDirectSoundBuffer->SetVolume
	 */
	la=(LONG)(100 * attenuation);
	THIS(engine)->err=
		IDirectSoundBuffer_SetVolume(info->idsb,la);
	return THIS(engine)->err==DS_OK;
}

/*
 *	Float pan in decibels; 0 mean full folume for both channels,
 *	negative mean right attenuation, positive - left attenuation.
 */
vboolean	V_GLOBAL
seSetPan(sound_engine_t *engine,u32_t buffer_num,vec_t pan)
{
	LONG lp;
	ds_sound_info_t *info;
	info=get_sound_info("seSetPan",engine,buffer_num);
	if(!info)
		return false;
	lp=(LONG)(100 * pan);
	THIS(engine)->err=IDirectSoundBuffer_SetPan(info->idsb,lp);
	return THIS(engine)->err==DS_OK;
}

vboolean	V_GLOBAL
seSetRate(sound_engine_t *engine,u32_t buffer_num,u32_t rate)
{
	ds_sound_info_t *info;
	info=get_sound_info("seSetRate",engine,buffer_num);
	if(!info)
		return false;
	THIS(engine)->err=IDirectSoundBuffer_SetFrequency(info->idsb,rate);
	return THIS(engine)->err==DS_OK;
}

/*
 *	TODO: write seLoopBuffer().
 */


int main()
{
	glw_t win;
	sound_engine_t se;
	u32_t sound;
	
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);

	win=vglCreateFullscreenWindow("test",256,256,0,0);

	se=seCreateEngine(GetForegroundWindow());
	if(!se)
		return -1;
	sound=seLoadSound(se,"../../samples/media/guide.wav");
	if(sound!=INVALID_BUFFER_VALUE)
	{
		sePlaySound(se,sound);
		fgetc(stdin);
	}
	
	seDestroyEngine(se);
	vglDestroyWindow(win);
	
	return 0;
}
