#include <stdlib.h>
#include <math.h>

#include "fft.h"

vec_t Parzen(vec_t n,vec_t j)
{
   vec_t _fret;
   _fret = 1.0 - fabs((j - 0.5 * (n - 1.0)) / (0.5 * (n + 1.0)));
   return(_fret);
}


vec_t Hanning(vec_t n,vec_t j)
{
   vec_t _fret;
   _fret = 0.5 * (1.0 - cos(2.0 * M_PI * j / n));
   return(_fret);
}

vec_t Hamming(vec_t n,vec_t j)
{
   vec_t _fret;
   _fret = 0.54 - 0.46 * cos(2.0 * M_PI * j / n);
   return(_fret);
}

vec_t ExactBlackman(vec_t n,vec_t j)
{
   vec_t _fret;
   _fret = 0.42 - 0.50 * cos(2.0 * M_PI * j / n) + 0.08 * cos(4.0 * M_PI * j / n);
   return(_fret);
}

vec_t Welch(vec_t n,vec_t j)
{
   vec_t _fret;
    _fret = (j - 0.5 * (n - 1.0)) / (0.5 * (n + 1.0));
   _fret = 1 - _fret*_fret;
   return(_fret);
}

void UEXPORT WindowFFTData(vec_t xreal[],vec_t yimag[],int numdat,int window)
{
   int i;
   vec_t multiplier;
   for ( i = 0; i <= numdat - 1; ++i ) {
      switch ( window ) {
         case 0:
            multiplier = 1.0;
            break;
         case 1:
            multiplier = Parzen(numdat,i);
            break;
         case 2:
            multiplier = Hanning(numdat,i);
            break;
         case 3:
            multiplier = Welch(numdat,i);
            break;
         case 4:
            multiplier = Hamming(numdat,i);
            break;
         case 5:
            multiplier = ExactBlackman(numdat,i);
            break;
         default:
            multiplier = 1.0;
            break;
      }
      xreal[i] = multiplier * xreal[i];
      yimag[i] = multiplier * yimag[i];
   }
}

void FFT_swap(vec_t *s1, vec_t *s2)
{
   vec_t temp;
   temp = (*s1);
   (*s1) = (*s2);
   (*s2) = temp;
}

void UEXPORT fft(vec_t xreal[], vec_t yimag[], int numdat, int flag)
{
   int maxpower;
   int arg;
   int cntr;
   int pnt0;
   int pnt1;
   int i;
   int j;
   int a;
   int b;
   int k;
   vec_t sign;
   vec_t prodreal;
   vec_t prodimag;
   vec_t harm;
   vec_t *cosary;
   vec_t *sinary;
   cosary = (vec_t*)malloc(numdat*sizeof(vec_t));
   sinary = (vec_t*)malloc(numdat*sizeof(vec_t));
   j = 0;
   if ( flag != 0 ) {
      sign = 1.0;
      for ( i = 0; i <= numdat - 1; ++i ) {
         xreal[i] = xreal[i] / numdat;
         yimag[i] = yimag[i] / numdat;
      }
   }
   else {
      sign = -1.0;
   }
   for ( i = 0; i <= numdat - 2; ++i ) {
      if ( i < j ) {
         FFT_swap(&xreal[i],&xreal[j]);
         FFT_swap(&yimag[i],&yimag[j]);
      }
      k = numdat / 2;
      while ( k <= j ) {
         j = j - k;
         k = k / 2;
      }
      j = j + k;
   }
   maxpower = 0;
   i = numdat;
   while ( i != 1 ) {
      maxpower = maxpower + 1;
      i = i / 2;
   }
   harm = 2 * M_PI / numdat;
   for ( i = 0; i <= numdat - 1; ++i ) {
      sinary[i] = sign * sin(harm * i);
      cosary[i] = cos(harm * i);
   }
   a = 2;
   b = 1;
   for ( cntr = 1; cntr <= maxpower; ++cntr ) {
      pnt0 = numdat / a;
      pnt1 = 0;
      for ( k = 0; k <= b - 1; ++k ) {
         i = k;
         while ( i < numdat ) {
            arg = i + b;
            if ( k == 0 ) {
               prodreal = xreal[arg];
               prodimag = yimag[arg];
            }
            else {
               prodreal = xreal[arg] * cosary[pnt1] - yimag[arg] * sinary[pnt1];
               prodimag = xreal[arg] * sinary[pnt1] + yimag[arg] * cosary[pnt1];
            }
            xreal[arg] = xreal[i] - prodreal;
            yimag[arg] = yimag[i] - prodimag;
            xreal[i] = xreal[i] + prodreal;
            yimag[i] = yimag[i] + prodimag;
            i = i + a;
         }
         pnt1 = pnt1 + pnt0;
      }
      a = 2 * a;
      b = b * 2;
   }
   free(cosary);
   free(sinary);
}

void UEXPORT FFTCalc(vec_t xreal[],vec_t yimag[],int numdat)
{
   fft(xreal,yimag,numdat,0);
}

void UEXPORT FFTInvCalc(vec_t xreal[],vec_t yimag[], int numdat)
{
   fft(xreal,yimag,numdat,1);
}

vec_t SquareAndSum(vec_t a,vec_t b)
{
   vec_t _fret;
   _fret = a*a+b*b;
   return(_fret);
}

void UEXPORT PowerSpectrumCalc(vec_t xreal[], vec_t yimag[],int numdat,vec_t delta)
{
   vec_t timespan;
   vec_t normal;
   int i;
   vec_t n;
   n = numdat;
   normal = 1.0 / (n*n);
   fft(xreal,yimag,numdat,0);
   xreal[0] = SquareAndSum(xreal[0],yimag[0]) * normal;
   for ( i = 1; i <= (numdat / 2) - 1; ++i ) {
      xreal[i] = normal * (SquareAndSum(xreal[i],yimag[i]) +
        SquareAndSum(xreal[numdat - i],yimag[numdat - i]));
   }
   i = numdat / 2;
   xreal[i] = SquareAndSum(xreal[i],yimag[i]) * normal;
   timespan = (1.0*numdat) * delta;
   for ( i = 0; i <= (numdat/2); ++i ) {
      yimag[i] = (1.0*i) / timespan;
   }
}

void UEXPORT FFT2DCalc(vec_t xreal[],vec_t yimag[],
               int c,int r,int flag)
{
   int a;
   int i;
   int j;
   vec_t *yVector;
   vec_t *xVector;
   xVector = (vec_t*)malloc(r*sizeof(vec_t));
   yVector = (vec_t*)malloc(r*sizeof(vec_t));
   for ( j = 0; j <= r - 1; ++j ) {
      for ( a = 0; a <= c - 1; ++a ) {
         xVector[a] = xreal[j*c+a];
         yVector[a] = yimag[j*c+a];
      }
      if ( flag == 0 ) {
         FFTCalc(xVector,yVector,c);
      }
      else {
         FFTInvCalc(xVector,yVector,c);
      }
      for ( a = 0; a <= c - 1; ++a ) {
         xreal[j*c+a] = xVector[a];
         yimag[j*c+a] = yVector[a];
      }
   }
   for ( i = 0; i <= c - 1; ++i ) {
      for ( a = 0; a <= r - 1; ++a ) {
         xVector[a] = xreal[a*c+i];
         yVector[a] = yimag[a*c+i];
      }
      if ( flag == 0 ) {
         FFTCalc(xVector,yVector,r);
      }
      else {
         FFTInvCalc(xVector,yVector,r);
      }
      for ( a = 0; a <= r - 1; ++a ) {
         xreal[a*c+i] = xVector[a];
         yimag[a*c+i] = yVector[a];
      }
   }
   free(xVector);
   free(yVector);
}
