#include	"../../include/v_lib.h"
#include	"../../source/v_mt.h"
#include	"v_fsutil.h"
#include	"v_scanfs.h"

#ifndef		RIGHT_SLASHED
#define		PASS_PATH(cb,path)	\
	{\
		char	ospath[_MAX_PATH];\
		strCopy(ospath,(path),_MAX_PATH);\
		fsNative2Unix(ospath);\
		cb(ospath);\
	}
#else		/* !RIGHT_SLASHED */
#define		PASS_PATH(cb,path)	cb(path)
#endif		/* !RIGHT_SLASHED */

/*
 *	Enumeration of files in given directory and all subdirs.
 *	(implementation of vfs_t->enum_files for disk filesystem).
 */

#ifdef	HAVE_DIRENT

/*
 *	Ok, we have POSIX functions for that.
 */

void	V_CALL
fs_scan_dir_r(
	char *root_dir,
	void (V_CALL *dir_callback)(char *unix_path),
	void (V_CALL *file_callback)(char *unix_file_name),
	vboolean recursive)
{
	DIR		*dir;
	struct	dirent *ent;
	struct	stat st;
    char 	fname[_MAX_PATH];
	int		root_len;
#ifdef	__UNIX__
	/* Disallow recursive seach in /proc.. and /dev... */
	if(recursive && (!strncmp(root_dir,"/proc",5) || 
		!strncmp(root_dir,"/dev",4)))
		return;
#endif	/* __UNIX__ */
	if(dir_callback)
		PASS_PATH(dir_callback,root_dir);
	dir = opendir(root_dir);
	if(!dir)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"fsScan(): unable to opendir('%s').\n",root_dir);
#endif	/* CHECKED_BUILD */
		return;
	}
	strCopy(fname,root_dir,_MAX_PATH);
	root_len=strlen(root_dir);
	for(ent=readdir(dir);ent;ent=readdir(dir))
	{
		if(ent->d_name[0]=='.')	/* Skip hidden files */
			continue;
		fname[root_len]=0;
		strCat(fname,ent->d_name,_MAX_PATH-strlen(fname));
		if(stat(fname,&st)==-1)
		{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"fsScan(): stat('%s') failed.\n",fname);
#endif	/* CHECKED_BUILD */
			continue;
		}
		st.st_mode &= S_IFMT;
		if(st.st_mode == S_IFDIR)
		{
			strCat(fname,"/",_MAX_PATH-strlen(fname));
			if(recursive)
				fs_scan_dir_r(fname,dir_callback,file_callback,recursive);
#if 0		/* Must be enabled for correct recursuion in '/dev..' */
			{
				off_t off;
				off=telldir(dir);
				/*
				 *	Avoid recursion in virtual subdirs of /dev/fd and 
				 *	/proc/self/fd directories.
				 *	       
				 *	   /dev/fd/3;  -  
				 *	  !!!
				 */
				closedir(dir);
				fs_scan_dir_r(fname,dir_callback,file_callback,recursive);
				dir=opendir(root_dir);
				if((int)off>0)
					seekdir(dir,off);
				else
					break;
			}
#endif
			else
				if(dir_callback)
					PASS_PATH(dir_callback,fname);
		}
		else if(st.st_mode == S_IFREG)
		{
			if(file_callback)
			{
				strCopy(fname,root_dir,_MAX_PATH);
				strCat(fname,ent->d_name,_MAX_PATH-strlen(fname));
				PASS_PATH(file_callback,fname);
			}
		}
	}
	closedir(dir);
}

#else	/* HAVE_DIRENT */

/*
 *	We have DOS-style api for that.
 */

void	V_LOCAL
fs_scan_dir_r(
	char *root_dir,
	void (V_CALL *dir_callback)(char *unix_path),
	void (V_CALL *file_callback)(char *unix_file_name),
	vboolean recursive)
{
	struct _finddata_t fd;
	long handle, rc;
	char curdir[_MAX_DIR], name[_MAX_PATH];
	char path_delimiter[2]="/";
	if(!dir_callback && !file_callback)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"fsScan(): both callbacks is NULL.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	_getcwd(curdir,_MAX_DIR);
	fsNative2Unix(curdir);
	/* root */
	if(curdir[strlen(curdir)-1]=='/')
		path_delimiter[0]=0;
	if(dir_callback)
		dir_callback(curdir);
	handle=_findfirst("*.*",&fd);
	rc=handle;
	while(rc!=-1) 
	{
		if(fd.name[0]=='.')
		{
			rc=_findnext(handle,&fd);
			continue;
		}
		if(fd.attrib & _A_SUBDIR)
		{
			if(recursive)
			{
				if(_chdir(fd.name)==0)
				{
					fs_scan_dir(dir_callback,file_callback,recursive);
					_chdir("..");
				}
#ifdef			CHECKED_BUILD
				else
				{
					conlPrintf(LOG_ERROR,"fsScan(): can't _chdir(%s).\n",
						fd.name);
				}
#endif			/* CHECKED_BUILD */
			}
			else /* !recursive - call dir callback for every dir */
			{
				if(dir_callback)
				{
					strCopy(name,curdir,_MAX_PATH);
					strCat(name,path_delimiter,_MAX_PATH-strlen(name));
					fsNative2Unix(fd.name);
					strCat(name,fd.name,_MAX_PATH-strlen(name));
					dir_callback(name);
				}
			}
		}
		else
		{
			if(file_callback)
			{
				strCopy(name,curdir,_MAX_PATH);
				strCat(name,path_delimiter,_MAX_PATH-strlen(name));
				strCat(name,fd.name,_MAX_PATH-strlen(name));
				file_callback(name);
			}
		}
		rc=_findnext(handle,&fd);
	}
	_findclose(handle);
}

#endif	/* HAVE_DIRENT */

void	V_CALL
fs_scan_dir(void (V_CALL *dir_callback)(char *unix_path),
			void (V_CALL *file_callback)(char *unix_file_name),
			vboolean recursive)
{
	int len;
	char root_dir[_MAX_PATH];
	if(!dir_callback && !file_callback)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"fsScan(): both callbacks is NULL.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	_getcwd(root_dir,_MAX_DIR);
	len=strlen(root_dir);
	if(root_dir[len-1]!='/')
	{
		root_dir[len]='/';
		root_dir[len+1]=0;
	}
	fs_scan_dir_r(root_dir,dir_callback,file_callback,recursive);
}
