#include <v_lib.h>
#include <../source/v_mt.h>
#include "alloc.h"

/*
 *	File object
 */
typedef struct vfs_f_s
{
#ifdef	CHECKED_BUILD
	u32_t	magic;
#endif	/* CHECKED_BUILD */
	/*	Handle of real object */
	void	*instance;
	/*	Destructor */
	vboolean (V_CALL *close)(void *object);
	/*	Character-dev and block-dev IO */
	u32_t	(V_CALL *read)(void *object,void *buf,u32_t len);
	u32_t	(V_CALL *write)(void *object,void *buf,u32_t len);
	/*	Block-dev IO only */
	u32_t	(V_CALL *seek)(void *object,u32_t offset,int whence);
	u32_t	(V_CALL *tell)(void *object);
	u32_t	(V_CALL *size)(void *object);
	/*	IOctl */
	int		(V_CALL	*ioctl)(void *object,int code,void *parameters);
}	vfs_f_t, *vfs_f_p;

/*
 *	Filesystem object
 */
typedef struct vfs_s
{
#ifdef	CHECKED_BUILD
	u32_t	magic;
#endif	/* CHECKED_BUILD */
	void	*instance;
	void	(*destroy)(struct vfs_s *self);
	void	(*enum_files)(struct vfs_s *self,
		void (V_CALL *dir_callback)(char *full_unix_style_dir_name),
		void (V_CALL *file_callback)(char *full_unix_style_file_name),
		vboolean recursive);
	vfs_f_p	(*create_file)(struct vfs_s *self,char *file_name,char mode_char);
	/*	There is no 'close_file' because pointer to destructor stored
	 *	file object, not filesystem object */
}	vfs_t, *vfs_p;

/*
 *	Implementation of base file object
 */
V_LOCAL u32_t	file_magic=0x46494C45;			/* {'F','I','L','E'} */
V_LOCAL char	*file_object_name="file";

#define COMPLAIN_STUB(caller)\
	conlPrintf(LOG_ERROR,"%s(): in given object this is stub only.\n",\
		(caller));

/*	Member functions */
#include "argsused.h"
V_LOCAL	u32_t V_CALL read_stub(void *object,void *buf,u32_t len)
{
	COMPLAIN_STUB("fileRead");
	return 0;
}

#include "argsused.h"
V_LOCAL	u32_t V_CALL write_stub(void *object,void *buf,u32_t len)
{
	COMPLAIN_STUB("fileWrite");
	return 0;
}

#include "argsused.h"
V_LOCAL	u32_t V_CALL seek_stub(void *object,u32_t offset,int whence)
{
	COMPLAIN_STUB("fileSeek");
	return 0;
}

#include "argsused.h"
V_LOCAL	u32_t V_CALL tell_stub(void *object)
{
	COMPLAIN_STUB("fileTell");
	return 0;
}

#include "argsused.h"
V_LOCAL	u32_t V_CALL size_stub(void *object)
{
	COMPLAIN_STUB("fileSize");
	return 0;
}

#include "argsused.h"
V_LOCAL	vboolean V_CALL close_stub(void *object)
{
	COMPLAIN_STUB("fileClose");
	return true;	/* Allow caller to free() memory */
}

#include "argsused.h"
V_LOCAL	int V_CALL ioctl_stub(void *object,int code,void *parameters)
{
	COMPLAIN_STUB("fileIOControl");
	return 0;
}

/*	Default constructor */
vfs_f_p	V_LOCAL
vfsCreateStubObject(void)
{
	vfs_f_t *f;
	f=(vfs_f_t*)memAlloc(sizeof(vfs_f_t));
	memZero(f,sizeof(vfs_f_t));
	f->magic=file_magic;
	f->read=read_stub;
	f->write=write_stub;
	f->seek=seek_stub;
	f->tell=tell_stub;
	f->size=size_stub;
	f->close=close_stub;
	f->ioctl=ioctl_stub;
	return f;
}

/*
 *	Global functions to operate with files
 */
u32_t	V_GLOBAL
fRead(file_t handle,void *buf,u32_t len)
{
	if(!libvIsHandle(handle,"fileRead",file_object_name,file_magic))
		return 0;
	return ((vfs_f_t*)handle)->read((void*)handle,buf,len);
}

u32_t	V_GLOBAL
fWrite(file_t handle,void *buf,u32_t len)
{
	if(!libvIsHandle(handle,"fileWrite",file_object_name,file_magic))
		return 0;
	return ((vfs_f_t*)handle)->write((void*)handle,buf,len);
}

u32_t	V_GLOBAL
fSeek(file_t handle,u32_t offset,int whence)
{
	if(!libvIsHandle(handle,"fileSeek",file_object_name,file_magic))
		return 0;
	return ((vfs_f_t*)handle)->seek((void*)handle,offset,whence);
}

u32_t	V_GLOBAL
fTell(file_t handle)
{
	if(!libvIsHandle(handle,"fileTell",file_object_name,file_magic))
		return 0;
	return ((vfs_f_t*)handle)->tell((void*)handle);
}

u32_t	V_GLOBAL
fSize(file_t handle)
{
	if(!libvIsHandle(handle,"fileSize",file_object_name,file_magic))
		return 0;
	return ((vfs_f_t*)handle)->size((void*)handle);
}

vboolean	V_GLOBAL
fClose(file_t handle)
{
	if(!libvIsHandle(handle,"fileClose",file_object_name,file_magic))
		return false;
	((vfs_f_t*)handle)->close((void*)handle);
	memFree(handle);
	return true;
}

int	V_GLOBAL
fIOControl(file_t handle,u32_t code,void *parameters)
{
	if(!libvIsHandle(handle,"fileIOctl",file_object_name,file_magic))
		return 0;
	return ((vfs_f_t*)handle)->ioctl((void*)handle,code,parameters);
}

/*
 *	Implementation of base filesystem object.
 *	Note. Filesystem function NOT exposed because all known filesystem
 *	will be mounted automatically by libvPerThreadInit().
 */

V_LOCAL u32_t		vfs_magic=0x5646494C;		/* {'V','F','I','L'} */
V_LOCAL char		*vfs_object_name="filesystem";
V_LOCAL vp_stack_t	mtab;						/* Mounted VFS' */

int main()
{
	return 0;
}
