#include <v_lib.h>

#if 0

typedef struct vfs_file_object_s
{
#ifdef	CHECKED_BUILD
	u32_t	magic;
#endif	/* CHECKED_BUILD */
	/*	Handle of real object */
	void	*this_instance;
	/*	Destructor */
	vboolean (V_CALL *close_proc)(void *object);
	/*	Character - dev IO */
	u32_t	(V_CALL *read_proc)(void *object,void *buf,u32_t len);
	u32_t	(V_CALL *write_proc)(void *object,void *buf,u32_t len);
	/*	Block-dev IO */
	u32_t	(V_CALL *seek_proc)(void *object,u32_t offset,int whence);
	u32_t	(V_CALL *tell_proc)(void *object);
	u32_t	(V_CALL *size_proc)(void *object);
	/*	IOctl */
	int		(V_CALL	*ioctl_proc)(void *object,int code, ...);
}	vfs_file_object_t, *file_object_p;

V_LOCAL u32_t	file_magic=0x46494C45;			/* FILE */
V_LOCAL char	*file_object_name="file";

#define COMPLAIN_STUB(caller)\
	conlPrintf(LOG_ERROR,"%s(): in given object this is stub only.\n",\
		(caller));

#include "argsused.h"
V_LOCAL	V_CALL u32_t read_stub(void *object,void *buf,u32_t len)
{
	COMPLAIN_STUB("fileRead");
	return 0;
}

#include "argsused.h"
V_LOCAL	V_CALL u32_t write_stub(void *object,void *buf,u32_t len)
{
	COMPLAIN_STUB("fileWrite");
	return 0;
}

#include "argsused.h"
V_LOCAL	V_CALL u32_t seek_stub(void *object,u32_t offset,int whence)
{
	COMPLAIN_STUB("fileSeek");
	return 0;
}

#include "argsused.h"
V_LOCAL	V_CALL u32_t tell_stub(void *object)
{
	COMPLAIN_STUB("fileTell");
	return 0;
}

#include "argsused.h"
V_LOCAL	V_CALL u32_t size_stub(void *object)
{
	COMPLAIN_STUB("fileSize");
	return 0;
}

#include "argsused.h"
V_LOCAL	V_CALL vboolean close_stub(void *object)
{
	COMPLAIN_STUB("fileClose");
	return true;	/* Allow caller to free() memory */
}

#include "argsused.h"
V_LOCAL	V_CALL int ioctl_stub(void *object,int code)
{
	COMPLAIN_STUB("fileIOControl");
	return 0;
}

vfs_file_p	V_LOCAL
vfsCreateStubObject(void)
{
	vfs_file_t *f;
	f=(vfs_file_t*)calloc(sizeof(vfs_file_t));
	f->mafgic=file_magic;
	f->read_proc=read_stub;
	f->write_proc=write_stub;
	f->seek_proc=seek_stub;
	f->tell_proc=tell_stub;
	f->size_proc=size_stub;
	f->close_proc=close_stub;
	f->ioctl_proc=ioctl_stub
	return f;
}

u32_t	V_GLOBAL
fileRead(file_t handle,void *buf,u32_t len)
{
	if(!libvIsHandle(handle,"fileRead",file_object_name,file_magic))
		return 0;
	return ((vfs_file_t*)handle)->read_proc((void*)handle,buf,len);
}

u32_t	V_GLOBAL
fileWrite(file_t handle,void *buf,u32_t len)
{
	if(!libvIsHandle(handle,"fileWrite",file_object_name,file_magic))
		return 0;
	return ((vfs_file_t*)handle)->write_proc((void*)handle,buf,len);
}

u32_t	V_GLOBAL
fileSeek(file_t handle,u32_t offset,int whence)
{
	if(!libvIsHandle(handle,"fileSeek",file_object_name,file_magic))
		return 0;
	return ((vfs_file_t*)handle)->seek_proc((void*)handle,offset,whence);
}

u32_t	V_GLOBAL
fileTell(file_t handle)
{
	if(!libvIsHandle(handle,"fileTell",file_object_name,file_magic))
		return 0;
	return ((vfs_file_t*)handle)->tell_proc((void*)handle);
}

u32_t	V_GLOBAL
fileSize(file_t handle)
{
	if(!libvIsHandle(handle,"fileSize",file_object_name,file_magic))
		return 0;
	return ((vfs_file_t*)handle)->size_proc((void*)handle);
}

vboolean	V_GLOBAL
fileClose(file_t file)
{
	if(!libvIsHandle(handle,"fileClose",file_object_name,file_magic))
		return false;
	((vfs_file_t*)handle)->close_proc((void*)handle);
	libvFree(handle);
	return true;
}

#endif /* 0 */

/*
 *	Duplex channel
 */
typedef struct	ipc_pipe_s
{
	u32_t		magic;
	u8_fifo_t	fifo;
	mutex_t		lock;
	cond_t		cond;
	vboolean	block;
}	ipc_pipe_t, *ipc_pipe_p;

ipc_pipe_p	V_LOCAL
ipcPipeCreate(void)
{
	ipc_pipe_t *p;
	p=(ipc_pipe_t*)malloc(sizeof(ipc_pipe_t));
	p->magic=0;
	u8fInit(&p->fifo);
	p->lock=mutexCreate();
	p->cond=condCreate(true);
	p->block=false;
	return p;
}

void	V_LOCAL
ipcPipeDestroy(ipc_pipe_t *p)
{
	u8fClear(&p->fifo);
	mutexDestroy(p->lock);
	condDestroy(p->cond);
}

u32_t	V_LOCAL
ipcPipeWrite(ipc_pipe_t *p,void *buf,u32_t len)
{
	mutexLock(p->lock);
	u8fPushBuffer(&p->fifo,buf,len);
	condSet(p->cond);
	mutexUnlock(p->lock);
	return len;
}

u32_t	V_LOCAL
ipcPipeLen(ipc_pipe_t *p)
{
	u32_t len;
	mutexLock(p->lock);
	len=u8fCount(&p->fifo);
	mutexUnlock(p->lock);
	return len;
}

u32_t	V_LOCAL
ipcPipeRead(ipc_pipe_t *p,void *buf,u32_t len)
{
	u32_t avail;
	mutexLock(p->lock);
	if(!p->block)
	{
		/* Non-blocking mode */
		avail=u8fCount(&p->fifo);
		if(avail<len)
			len=avail;
		u8fPopBuffer(&p->fifo,buf,len);
		mutexUnlock(p->lock);
		return len;
	}
	for(;;)
	{
		/* Blocking mode */
		mutexLock(&p->lock);
		avail=u8fCount(&p->fifo);
		if(avail>=len)
		{
			printf("Enought data: req %d, avail %d\n",len,avail);
			u8fPopBuffer(&p->fifo,buf,len);
			break;
		}
		printf("Not enought data: req %d, avail %d\n",len,avail);
		mutexUnlock(p->lock);
		condWait(p->cond);
	}
	mutexUnlock(p->lock);
	return len;
}

void	V_LOCAL
ipcPipeWait(ipc_pipe_t *p)
{
	condWait(p->cond);
}

ipc_pipe_p	p;
thread_t	thread;
vboolean	done;

void V_CALL	t2_entry(void *arg)
{
	char buf[128];
	p->block=true;
	for(;;)
	{
//		ipcPipeWait(p);
		ipcPipeRead(p,buf,4);
		buf[4]=0;
		printf("Cmd: '%s'\n",buf);
		if(!memcmp(buf,"exit",4))
			break;
	}
	done=true;
	threadExit(0);
}

int main()
{
	char buf[128];
	p=ipcPipeCreate();
	thread=threadCreate(t2_entry,NULL);
	for(done=false;!done;)
	{
		memset(buf,'=',128);
		fgets(buf,128,stdin);
		buf[5]=0;
		ipcPipeWrite(p,buf,4);
	}
	threadJoin(thread,NULL);
	ipcPipeDestroy(p);
	return 0;
}
