#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* DOS -> UNIX */
int
translate_forward(FILE *out,FILE *in)
{
	unsigned char ch;
	for(;;)
	{
		ch=(char)fgetc(in);
    	if(feof(in)) break;
		/* strip 0x0D */
		if(ch!=0x0D)
			fputc(ch,out);
	}
}

/* UNIX -> DOS */
int
translate_backward(FILE *out,FILE *in)
{
	unsigned char ch;
	for(;;)
	{
		ch=(char)fgetc(in);
    	if(feof(in)) break;
    	/* for each 0x0A, append leading 0x0D */
		if(ch==0x0A)
		{
			fputc(0x0D,out);
			fputc(0x0A,out);
		}
		else
			fputc(ch,out);
	}
}

int
translate_file(char *outname,char *inname,int reverse)
{
	FILE *out;
	FILE *in;
	out=fopen(outname,"wb");
	if(!out)
	{
		printf("Can't create '%s'.\n",outname);
		return -1;
	}
	in=fopen(inname,"rb");
	if(!in)
	{
		printf("Can't open '%s'.\n",inname);
		return -1;
	}
	if(reverse)
		translate_backward(out,in);
	else
		translate_forward(out,in);
	fclose(out);
	fclose(in);
	return 0;
}

char*
tempname(char *srcname)
{
	static char retval[512];
#ifdef	__linux
	sprintf(retval,"/tmp/%s.tmp",srcname);
	return retval;
#else
	char *temp;
	temp=getenv("TEMP");
	if(!temp)
		temp=getenv("TMP");
	if(!temp)
		sprintf(retval,".\\%s.tmp",srcname);
	else
		sprintf(retval,"%s\\%s.tmp",temp,srcname);
	return retval;
#endif
}

int
translate_file_inplace(char *name,int reverse)
{
	FILE *out;
	FILE *in;
	char tempout[512];
	strcpy(tempout,tempname(name));
	out=fopen(tempout,"wb");
	if(!out)
	{
		printf("Can't create temporary file '%s'.\n",tempout);
		return -1;
	}
	in=fopen(name,"rb");
	if(!in)
	{
		printf("Can't open '%s'.\n",name);
		return -1;
	}
	if(reverse)
		translate_backward(out,in);
	else
		translate_forward(out,in);
	fclose(out);
	fclose(in);
	unlink(name);
	return rename(tempout,name);
}

void
banner(void)
{
	puts("DOS->Unix & vise versa translator ver 1.1. Smirnov V.A., 2000.");
	puts("This software copyrighted under GPL license.");
}

void
usage(void)
{
	puts("Usage: dos2unix <input text file> [output text file] [-r(everse)].");
}

int main(int argc,char *argv[])
{
	FILE *in,*out;
	int reverse;
	banner();
	if(argc<2 || argc>4)
	{
		usage();
		return -1;
	}
	reverse = !strcmp(argv[argc-1],"-r") || !strcmp(argv[argc-1],"-reverse");
	if(reverse)
		argc--;
	if(argc==2)
		return translate_file_inplace(argv[1],reverse);
	else
		return translate_file(argv[2],argv[1],reverse);
}
