#include <v_lib.h>

typerdef	void	*registry_t;

#ifdef	__cplusplus
extern "C" {
#endif	/* !__cplusplus */
	registry_t	V_GLOBAL	regLoadFile(char *file_name);
	vboolean	V_GLOBAL	regSaveFile(registry_t reg,char *file_name);
#ifdef	__WIN32__
	registry_t	V_GLOBAL	regOpenSystemRegistry(void);
#endif	/* !__WIN32__ */
	void		V_GLOBAL	regClose(registry_t reg);

	vboolean	V_GLOBAL	regOpenKey(registry_t reg,char *key_name);
	vboolean	V_GLOBAL	regCreateKey(registry_t reg,char *key_name);
	vboolean	V_GLOBAL	regCloseKey(registry_t reg,char *key_name);

	vboolean	V_GLOBAL	regGetDword(registry_t reg,u32_t *dest);
	vboolean	V_GLOBAL	regGetString(registry_t reg,char *dest,int dest_size);
	vboolean	V_GLOBAL	regGetBinary(registry_t reg,void *dest,int dest_size);

	vboolean	V_GLOBAL	regSetDword(registry_t reg,u32_t *dest);
	vboolean	V_GLOBAL	regSetString(registry_t reg,char *dest,int dest_size);
	vboolean	V_GLOBAL	regSetBinary(registry_t reg,void *dest,int dest_size);
#ifdef	__cplusplus
}
#endif	/* !__cplusplus */

/*
**	Implementation
*/

#define	DWORD_ALIGN(val)	((~((~(val)) & (~(unsigned)3)))+1)

/*
**	TODO: convert simple list to tree.
*/

/*
**	Registry types
*/

typedef enum
{
	reg_type_dword,
	reg_type_string,
	reg_type_binary,
}	reg_type_t;

/*
**	Registry value
*/

typedef struct	reg_value_s
{
	reg_type_t	type;
	u32_t		data_size;
		/* for binaries */
	char		*name;
	void		*data;
}	reg_value_t, *reg_value_p;

V_LOCAL reg_value_p
regAllocValue(reg_type_t type,char *name,void *data,u32_t data_size)
{
	reg_value_p v;
	u32_t n_len,d_len, na_len, da_len;
	n_len=strlen(name)+1;
	switch(type)
	{
		case reg_type_string:
			d_len = strlen((char*)data)+1;
			break;
		case reg_type_dword:
			d_len = sizeof(u32_t);
			break;
		default: /* type == reg_type_binary */
			d_len = data_size;
			break;
	}
	/* Align lens on dword boundary, alocate storage */
	na_len=DWORD_ALIGN(n_len);
	da_len=DWORD_ALIGN(d_len);
	v=(reg_value_p)malloc(sizeof(reg_value_t)+na_len+da_len;
	v->type=type;
	v->data_size=d_len;
	v->name=(char*)(v+1);
	v->data=(void*)(d->name+na_len);
	memcpy(v->name,name,n_len);
	memcpy(v->data,data,v->data_size);
	return v;
}

#define	regAllocDword(name,value)
 \
		regAllocValue(reg_type_dword,name,&(value),0)

#define	regAllocString(name,str)
 \
		regAllocValue(reg_type_string,name,str,0)

#define	regAllocBinary(name,buffer,len)
 \
		regAllocValue(reg_type_binary,name,buffer,len)

/*
**	Registry section
*/

typedef struct reg_section_s
{
	char		*name;
	vp_stack_t	values;
}	reg_section_t, *reg_section_p;

/*
**	Registry
*/

typedef struct	reg_sections_s
{
	u32_t		magic;
	vboolean	modified;
	vp_stack_t	sections;
}	reg_sections_t, *reg_sections_p;
	
#error	TODO

load plain list of sections
