#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glaux.h>
#pragma hdrstop
#include "../../include/v_lib.hpp"
#include <stdio.h>

#ifdef  DOUBLEVEC
#error     !
#endif

typedef enum
{
  alpha,blend,cullface,depthtest,fog,pointsmooth,texture
} rendercaps_t;

typedef enum
{
  alphaless,alphagreater,blendoneone,blendonesrccolor,
  decal,modulate,nearest,linear,mipmap,trilinear,clamp,repeat,
  foglinear,fogexp
} fxcaps_t;

typedef enum
{
  points,lines,linestrip,triangles,tristrip,trifan,quads,quadstrip,polygon
} primitivetype_t;

typedef enum
{
  matmodelview,matprojection,mattexture
} matrixmode_t;

#define MAX_TEXTURES    1024

class renderer_c
{
  int ow,oh;
  char *aliases[MAX_TEXTURES];
  unsigned int tex[MAX_TEXTURES];
  public:
  //  rrate > -1 -      
  renderer_c(int w,int h,char *title,int rrate=-1); 
  virtual ~renderer_c();
  
  void    clear(vec_t r,vec_t g,vec_t b,vboolean clear_z);
  void    begin(primitivetype_t prim);
  void    end();
  
  void    color(vec_t r,vec_t g,vec_t b);
  void    colorv(vec3_t rgb);
  void    colorv(void *rgba);
  
  void    vertex(vec_t x,vec_t y,vec_t z);
  void    vertexv(vec3_t xyz);
  
  void    texcoord(vec_t u,vec_t v);
  void    texcoordv(vec2_t uv);
  
  void    normal(vec_t x,vec_t y,vec_t z);
  void    normalv(vec3_t xyz);

  void    translate(vec_t x,vec_t y,vec_t z);
  void    translatev(vec3_t xyz);

  void    rotate(vec_t degrees,vec_t dir_x,vec_t dir_y,vec_t dir_z);
  void    rotatev(vec_t degrees,vec3_t dir);

  void    scale(vec_t x,vec_t y,vec_t z);
  void    scalev(vec3_t xyz);

  void    matrixmode(matrixmode_t mode);
  void    pushmatrix();
  void    popmatrix();
  void    loadidentity();
//  void    loadmatrix(matrix4_t m);
  void    loadmatrix(matrix3x4_t m);

  // caps & fx
  void    enable(rendercaps_t pname);
  void    disable(rendercaps_t pname);
  void    fxcaps(fxcaps_t pname);
  
  // texturing
  vboolean loadtexture(char *alias,int w,int h,void *texels,fxcaps_t tiling,
    fxcaps_t combine,fxcaps_t magfilter,fxcaps_t minfilter,unsigned int *ptex=NULL);
  void    texfree(unsigned int tex);
  void    texfree(char *alias);
  void    texbind(unsigned int tex);
  void    texbind(char *alias);
  
  // fog
  void    usefog(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec_t r,vec_t g,vec_t b,vec_t a);
  void    usefog(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec4_t rgba);
      
  virtual vboolean initialize();
  virtual void drawscene();
  virtual void finalize();
};

// cut here

static renderer_c *single_instance=NULL;

static GLvoid CALLBACK  drawScene(GLvoid)
{
  if(single_instance) single_instance->drawscene();
  auxSwapBuffers();
}

static GLvoid CALLBACK Resize(GLsizei width,GLsizei height)
{
  GLfloat aspect;
  glViewport( 0, 0, width, height );
  aspect = (GLfloat) width / height;
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-aspect,aspect,1,-1,1,100);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}    

renderer_c::renderer_c(int w,int h,char *title,int rrate)
{
  if(rrate>-1)
  {
   //
  }
  v_zero(aliases,sizeof(char*)*MAX_TEXTURES);
  v_zero(tex,sizeof(int)*MAX_TEXTURES);
  auxInitPosition(0,0,w,h);
  auxInitDisplayMode(AUX_RGB | AUX_DEPTH | AUX_DOUBLE);
  auxInitWindow(title);
  SetClassLong(GetActiveWindow(),GCL_HBRBACKGROUND,0);
  if(rrate>-1)
    SetWindowLong(GetActiveWindow(),GWL_STYLE,WS_POPUP);
  auxIdleFunc(drawScene);
  auxReshapeFunc(Resize);
  glClearColor(0.0,0.0,0.0,0.0);
  glClearDepth(10);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  Resize(w,h);
  if(initialize()) 
  {
    single_instance=this;
    auxMainLoop(drawScene);
    single_instance=NULL;
  }
  if(rrate>-1)
  {
    DEVMODE dm;
    v_zero(&dm,sizeof(dm));
    dm.dmSize=sizeof(dm);
    dm.dmPelsWidth=ow;
    dm.dmPelsHeight=oh;
    dm.dmFields=DM_PELSWIDTH|DM_PELSHEIGHT;
    ChangeDisplaySettings(&dm,0);
  }
}

renderer_c::~renderer_c()
{
  int i;
  finalize();
  for(i=0;i<MAX_TEXTURES;i++)
  {
    if(aliases[i]) free(aliases[i]);
    if(tex[i]) glDeleteTextures(1,tex+i);
  }
}

void renderer_c::clear(vec_t r,vec_t g,vec_t b,vboolean z_clear)
{
  glClearColor(r,g,b,0.5);
  if(z_clear)
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  else
    glClear(GL_COLOR_BUFFER_BIT);
}

void renderer_c::begin(primitivetype_t prim)
{
  switch(prim)
  {
    case points: glBegin(GL_POINTS); return;
    case lines: glBegin(GL_LINES); return;
    case linestrip: glBegin(GL_LINE_STRIP); return;
    case triangles: glBegin(GL_TRIANGLES); return;
    case tristrip: glBegin(GL_TRIANGLE_FAN); return;
    case trifan: glBegin(GL_TRIANGLE_STRIP); return;
    case quads: glBegin(GL_QUADS); return;
    case quadstrip: glBegin(GL_QUAD_STRIP); return;
    case polygon: glBegin(GL_POLYGON); return;
  }
  glBegin(GL_LINE_STRIP);
}

void renderer_c::end()
{
  glEnd();
}

void renderer_c::color(vec_t r,vec_t g,vec_t b)
{
  glColor3f(r,g,b);
}

void renderer_c::colorv(vec3_t rgb)
{
  glColor3fv(rgb);
}

void renderer_c::colorv(void *rgba)
{
  glColor3ubv((byte*)rgba);
}

void renderer_c::vertex(vec_t x,vec_t y,vec_t z)
{
  glVertex3f(x,y,z);
}

void renderer_c::vertexv(vec3_t xyz)
{
  glVertex3fv(xyz);
}

void renderer_c::texcoord(vec_t u,vec_t v)
{
  glTexCoord2f(u,v);
}

void renderer_c::texcoordv(vec2_t uv)
{
  glTexCoord2fv(uv);
}

void renderer_c::normal(vec_t x,vec_t y,vec_t z)
{
  glNormal3f(x,y,z);
}

void renderer_c::normalv(vec3_t xyz)
{
  glNormal3fv(xyz);
}

void renderer_c::translate(vec_t x,vec_t y,vec_t z)
{
  glTranslatef(x,y,z);
}

void renderer_c::translatev(vec3_t xyz)
{
  glTranslatef(xyz[0],xyz[1],xyz[2]);
}

void renderer_c::rotate(vec_t degrees,vec_t dir_x,vec_t dir_y,vec_t dir_z)
{
  glRotatef(degrees,dir_x,dir_y,dir_z);
}

void renderer_c::rotatev(vec_t degrees,vec3_t dir)
{
  glRotatef(degrees,dir[0],dir[1],dir[2]);
}

void renderer_c::scale(vec_t x,vec_t y,vec_t z)
{
  glScalef(x,y,z);
}

void renderer_c::scalev(vec3_t xyz)
{
  glScalef(xyz[0],xyz[1],xyz[2]);
}

void renderer_c::matrixmode(matrixmode_t mode)
{
  switch(mode)
  {
    case matprojection: glMatrixMode(GL_PROJECTION);return;
    case mattexture: glMatrixMode(GL_TEXTURE);return;
  }
  glMatrixMode(GL_MODELVIEW);
}

void renderer_c::pushmatrix()
{
  glPushMatrix();
}

void renderer_c::popmatrix()
{
  glPopMatrix();
}

void renderer_c::loadidentity()
{
  glLoadIdentity();
}

void renderer_c::loadmatrix(matrix3x4_t m)
{
  int i,j;
  matrix4_t t;
  for(i=0;i<3;i++)
    for(j=0;j<3;j++)
      t[i][j]=m[j][i];
  t[3][0]=m[0][3];
  t[3][1]=m[1][3];
  t[3][2]=m[2][3];
  t[0][3]=t[1][3]=t[2][3]=0;
  t[3][3]=1;
  glLoadMatrixf((GLfloat*)t);
}

// caps & fx
void renderer_c::enable(rendercaps_t pname)
{
  switch(pname)
  {
    case alpha: glEnable(GL_ALPHA_TEST); return;
    case blend: glEnable(GL_BLEND); return;
    case cullface: glEnable(GL_CULL_FACE); return;
    case depthtest: glEnable(GL_DEPTH_TEST); return;
    case fog: glEnable(GL_FOG); return;
    case pointsmooth: glEnable(GL_POINT_SMOOTH); return;
    case texture: glEnable(GL_TEXTURE_2D); return;
  }
}

void renderer_c::disable(rendercaps_t pname)
{
  switch(pname)
  {
    case alpha: glDisable(GL_ALPHA_TEST); return;
    case blend: glDisable(GL_BLEND); return;
    case cullface: glDisable(GL_CULL_FACE); return;
    case depthtest: glDisable(GL_DEPTH_TEST); return;
    case fog: glDisable(GL_FOG); return;
    case pointsmooth: glDisable(GL_POINT_SMOOTH); return;
    case texture: glDisable(GL_TEXTURE_2D); return;
  }
}

void renderer_c::fxcaps(fxcaps_t pname)
{
  switch(pname)
  {
    case alphaless: glAlphaFunc(GL_LESS,.5); return;
    case alphagreater: glAlphaFunc(GL_GREATER,.5); return;
    case blendoneone: glBlendFunc(GL_ONE,GL_ONE); return;
    case blendonesrccolor: glBlendFunc(GL_ONE,GL_SRC_COLOR); return;
    case foglinear: glFogi(GL_FOG_MODE,GL_LINEAR);
    case fogexp: glFogi(GL_FOG_MODE,GL_EXP);
  }
}

vboolean renderer_c::loadtexture(char *alias,int w,int h,void *texels,
  fxcaps_t tiling,fxcaps_t combine,fxcaps_t magfilter,fxcaps_t minfilter,
  unsigned int *ptex)
{
  int i;
  for(i=0;i<MAX_TEXTURES && tex[i];i++);
  if(i==MAX_TEXTURES)
  {
    puts("i==MAX_TEXTURES: increase MAX_TEXTURES.");
    return false;
  }
  if(alias)
  {
    if(aliases[i]) free(aliases[i]);
    aliases[i]=strdup(alias);
  }
  glGenTextures(1,&tex[i]);
  glBindTexture(GL_TEXTURE_2D,tex[i]);
  if(minfilter==mipmap||minfilter==trilinear)
    gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGBA,w,h,GL_RGBA,GL_UNSIGNED_BYTE,
      texels);
  else
    glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,w,h,0,GL_RGBA,GL_UNSIGNED_BYTE,
      texels);
  if(tiling==clamp)
  {
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  }
  else
  {
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  }
  if(magfilter==linear) 
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  else
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  switch(minfilter)
  {
    case nearest:
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      break;
    case linear:
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      break;
    case mipmap:
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
      break;
    case trilinear:
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
      break;
  }
  if(combine==decal)
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
  else
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  if(ptex) *ptex=tex[i];
  glBindTexture(GL_TEXTURE_2D,0);
  return glGetError() == GL_NO_ERROR;
}
  
void renderer_c::texfree(unsigned int tex)
{
  int i;
  for(i=0;i<MAX_TEXTURES && this->tex[i]!=tex;i++);
  if(i==MAX_TEXTURES)
  {
    printf("renderer_c::freetex: texture %d not found !\n",tex);
  }
  else
  {
    glDeleteTextures(1,&tex);
    this->tex[i]=0;
    if(aliases[i]) 
    {
      free(aliases[i]);
      aliases[i]=NULL;
    }
  }
}

void renderer_c::texfree(char *alias)
{
  int i;
  for(i=0;i<MAX_TEXTURES;i++)
    if(aliases[i])
      if(!strcmp(aliases[i],alias)) break;
  if(i==MAX_TEXTURES)
  {
    printf("renderer_c::freetex: texture %s not found !\n",alias);
  }
  else
  {
    glDeleteTextures(1,&tex[i]);
    tex[i]=0;
    free(aliases[i]);
    aliases[i]=NULL;
  }
}

void renderer_c::texbind(unsigned int tex)
{
  glBindTexture(GL_TEXTURE_2D,tex);
}

void renderer_c::texbind(char *alias)
{
  int i;
  for(i=0;i<MAX_TEXTURES;i++)
    if(aliases[i])
      if(!strcmp(aliases[i],alias)) break;
  if(i==MAX_TEXTURES)
  {
    printf("renderer_c::bindtex: texture %s not found !\n",alias);
  }
  else
    glBindTexture(GL_TEXTURE_2D,tex[i]);
}

void renderer_c::usefog(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec_t r,vec_t g,vec_t b,vec_t a)
{
  vec4_t rgba;
  if(mode==fogexp)
    glFogi(GL_FOG_MODE,GL_EXP);
  else
    glFogi(GL_FOG_MODE,GL_LINEAR);
  glFogf(GL_FOG_START,start);
  glFogf(GL_FOG_END,end);
  glFogf(GL_FOG_DENSITY,dens);
  glFogfv(GL_FOG_COLOR,vec4Set(rgba,r,g,b,a));
}

void renderer_c::usefog(fxcaps_t mode,vec_t start,vec_t end,vec_t dens,vec4_t rgba)
{
  if(mode==fogexp)
    glFogi(GL_FOG_MODE,GL_EXP);
  else
    glFogi(GL_FOG_MODE,GL_LINEAR);
  glFogf(GL_FOG_START,start);
  glFogf(GL_FOG_END,end);
  glFogf(GL_FOG_DENSITY,dens);
  glFogfv(GL_FOG_COLOR,rgba);
}








int w,h;
byte *pix;

vboolean renderer_c::initialize()
{
  pix=targaLoad("..\\..\\cache\\loaders\\texture.tga",&w,&h);
  loadtexture("test",w,h,pix,clamp,decal,nearest,nearest);
  targaFree(pix);
  return true;
}

void test_quad(renderer_c *r)
{
  r->begin(quads);
    r->color(1,0.5,0.5);
    r->texcoord(0,0);
    r->vertex(-50,-50,0);
    r->color(0.5,1,0.5);
    r->texcoord(1,0);
    r->vertex(-50,50,0);
    r->color(1,0.5,1);
    r->texcoord(1,1);
    r->vertex(50,50,0);
    r->color(0.5,1,1);
    r->texcoord(0,1);
    r->vertex(50,-50,0);
  r->end();
}

void renderer_c::drawscene()
{
  clear(0,0,0,true);
  
  matrixmode(matmodelview);
  loadidentity();
  
  translate(0,0,-3);
  scale((vec_t)0.02,(vec_t)0.02,1);
  disable(cullface);
  enable(texture);

  texbind("test");

  scale(1,-1,1);
  rotate(-90,0,0,1);
  
  pushmatrix();
  test_quad(this);
  popmatrix();
  
  Sleep(10);
  return;
}

void renderer_c::finalize()
{
  texfree("test");
  return;
}



/////////////////////////////////////////////////////////////////////////

void main()
{
  renderer_c r(512,512,"test",-1);
}
