#include <libv.h>

typedef struct line_strip_s
{
	int			num_segments;	/* == num_vertices - 1 */
	vec3_t		*vlist;
}	line_strip_t, *line_strip_p;

typedef struct	line_strips_s
{
	int				num_strips;
	void			*labels;
	line_strip_t	*strips;
}	line_strips_t, *line_strips_p;

typedef struct stack_t line_segs_stack_t, *line_segs_stack_p;

void
lssInit(line_segs_stack_t *lss)
{
	stackInit(&lss->a,sizeof(vec3_t));
}

void
lssPushSeg(line_segs_stack_t *lss,vec3_t a,vec3_t b)
{
	stackPush((stack_t*)lss,a);
	stackPush((stack_t*)lss,b);
}

vboolean
lssPopSeg(line_segs_stack_t *lss,vec3_t a,vec3_t b)
{
	if(stackPop((stack_t*)lss,b))
		return stackPop((stack_t*)lss,a);
	else
		return false;
}

void
lssGetSeg(line_segs_stack_t *lss,vec3_t a,vec3_t b,unsigned pos)
{
	stackGet((stack_t*)lss,a);
	stackGet((stack_t*)lss,b);
}

	
/*
 *	Returns malloc()'ed block
 */
line_strip_p lstMakeFromStack(stack_t *st)
{
	int i,size;
	line_strip_t *lst;
	V_ASSERT(st && st->block_size==sizeof(vec3_t));
	size=sizeof(line_strip_t)+stackCount(st)*sizeof(vec3_t);
	lst=(line_strip_t*)malloc(size);
	V_ASSERT(lst);
	lst->vlist=(vec3_t*)(lst+1);
	for(i=0;stackPop(st,lst->vlist[i]);i++);
}

vboolean
IsSegDegenerateEps(vec3_t a,vec3_t b,vec_t eps)
{
	return	vec3IsEqualEpsilon(a,b,eps);
}

vboolean
IsSegsOverlappedEps(vec3_t l1a,vec3_t l1b,vec3_t l2a,vec3_t l2b,vec_t eps)
{
	vec3_t da,db;
	vec3Sub(da,l1a,l2a);
	vec3Sub(db,l1b,l2b);
	if(vec3Abs(da)<eps && vec3Abs(da)<eps)
		return true;
	vec3Sub(da,l1a,l2b);
	vec3Sub(db,l1b,l2a);
	return vec3Abs(da)<eps && vec3Abs(da)<eps;
}

/*
 *	GetConnectionSegmentEps() - chooses segment which can be connected to
 *	given. Returns:
 *		- src_size: to which point of src must be connected dst;
 *			0 - for left, 1 - for right.
 *		- dst_size: to which point of dst must be connected src;
 *			0 - for left, 1 - for right.
 *		- dst_ind:	index of dst segment in segs.
 */

vboolean	GetConnectionSegmentEps();

/*
	There multiple segs to which src can be connected; but
	resulting polyline must be as straight as possible.
	vec_t cosine, cosine_min;
	cosine_min=1;
	for(i=0;i<segs;i++)
	{
		cosine=vecDot(seg,src)/(vec3Abs(seg)*abs_src);
		if(cosine<cosine_min)
			index=i;
	}
*/

/*
 *	ConnectSegmentEps
 *	Returns:
 *		1. If src was 'sleep' it becomes 'live'.
 *		2. If src was 'live' it becomes 'died'.
 */




/*
void v3sInit(stack_t *st)
{
	stackInit(st,sizeof(vec3_t));
}

void v3sPush(stack_t *st,vec3_t v)
{
	stackPush(st,v);
}

vboolean v3sPop(stack_t *st,vec3_t v)
{
	return stackPop(st,v);
}

unsigned v3sCount(stack_t *st)
{
	return stackCount(st);
}

void v3sClear(stack_t *st)
{
	stackClear(st);
}
*/

