#include  "../include/v_lib.h"
#include  "../source/v_mt.h"
#include  <windows.h>
#if		defined(_MSC_VER) || defined(__BORLANDC__)
#pragma	hdrstop
#endif	/* defined(_MSC_VER) || defined(__BORLANDC__) */

/* NEW: */

void	V_GLOBAL
vidGetCurrentMode(u32_t *width,u32_t *height,u32_t *bpp,u32_t *refresh)
{
	HDC	dc;
    dc=GetDC(NULL);
    if(width)
    	*width=GetDeviceCaps(dc,DESKTOPHORZRES);
    if(height)
    	*height=GetDeviceCaps(dc,DESKTOPVERTRES);
    if(bpp)
    	*bpp=GetDeviceCaps(dc,BITSPIXEL);
    if(refresh)
    	*refresh=GetDeviceCaps(dc,VREFRESH);
	ReleaseDC(NULL,dc);
}

/*  Win32 - specific ================================================== */

typedef struct gdidata_s
{
  char      cls[256];           /* window class name  */
  char      title[256];         /* window title */
  ATOM      wndclass;           /* return value of RegisterClass  */
  WNDCLASS  wc;                 /* window class */
  HINSTANCE inst;               /* executable file handle */
  HWND      wnd;                /* window handle as returned by CreateWindow  */
  BOOL      wndactive;          /* activity flag  */
  HBITMAP   bmp;                /* drawing surface  */
  HDC       dc;                 /* bitmap device context  */
  LOGBRUSH  lbr;                /* parameters of current brush  */
  LOGPEN    lpn;                /* parameters of current pen  */
  LOGFONT   lfn;                /* parameters of current font */
  HBRUSH    br;                 /* handle of current brush  */
  HPEN      pn;                 /* handle of current pen  */
  HPEN      npn;                /* handle of NULL (hollow) pen  */
  HBRUSH    nbr;                /* handle of NULL (hollow) brush  */
  HFONT     fn;                 /* handle of current font */
  POINT     pt[6];              /* array to store args for 'PolyXXX' GDI calls  */
  COLORREF  rgb[256];           /* RGB colors of current palette  */
  HPEN      apn[256];           /* pens with colors of palette  */
  HBRUSH    abr[256];           /* brushes with colors of palette */
  /*        TODO: include data for DirectDraw connectivity  */
  /*
  IDirectDraw *DD;
  IDirectDrawSurface *DDPrimary;
  IDirectDrawSurface *DDBackBuffer;
  */
  /*        TODO: include data for OpenGL connectivity  */
  /*
  void      (V_CALL *pwglGetPixelFormat)( ...
  */
} gdidata_t, *gdidata_p;

/*      Win32 window procedure ======================================== */

static int win32VkToKey(int vk_code)
{
  return vk_code;
}

/*
static int win32KeyToVk(int key)
{
  return key;
}
*/

static int win32MBFromWPARAM(WPARAM wparam)
{
  int res=0;
  if(wparam & MK_LBUTTON) res|=1;
  if(wparam & MK_MBUTTON) res|=2;
  if(wparam & MK_RBUTTON) res|=4;
  return res;
}
  
LRESULT CALLBACK 
graph_window_proc(HWND wnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
  gc_t        gc;
  gdidata_p   gdic;
  PAINTSTRUCT ps;
  thread_data_t *td;
  td=GetThreadData();
  gc=(gc_t)GetWindowLong(wnd,GWL_USERDATA);
  if(gc)
    gdic=(gdidata_p)gc->ctx_data;
  else
    gdic=NULL; // suppress warning message: in fact, this is VERY cheap ...
  if(!gc || !gdic) {
    conlPrintf(LOG_TRACE,"graph_window_proc(): cant handle event %d of "
      "window '%X'.\n",msg,wnd);
    return DefWindowProc(wnd,msg,wparam,lparam);
  }
  /* if system event handler present on the client side, then call them first */
  if(gc->sysevent) {
    if(!gc->sysevent((u32_t)wnd,(u32_t)msg,(u32_t)wparam,(u32_t)lparam)) {
      conlPrintf(LOG_TRACE,"graph_window_proc(): msg %d not passed to"
        " DefWindowProc().\n",msg);
      return DefWindowProc(wnd,msg,wparam,lparam);
    }
  }
  switch(msg)
  {
    case WM_ACTIVATE:
      gdic->wndactive=LOWORD(wparam) != WA_INACTIVE;
      if(gdic->wndactive) {
        if(gc->on_activate) gc->on_activate();
      } else {
        if(gc->on_deactivate) gc->on_deactivate();
      }
      break;
    case WM_DESTROY:
      if(gc->on_destroy)
        gc->on_destroy();
      graphDeleteContext(gc);
      // PostQuitMessage(0);
      return 0;
    case WM_KEYDOWN:
      if(gc->on_keydown)
        if(gc->on_keydown(win32VkToKey((int)wparam))) break;
      return 0;
    case WM_KEYUP:
      if(gc->on_keyup)
        if(gc->on_keyup(win32VkToKey((int)wparam))) break;
      return 0;
    case WM_LBUTTONDOWN:
      if(gc->on_mousedown)
        if(gc->on_mousedown(LOWORD(lparam),HIWORD(lparam),win32MBFromWPARAM(wparam)))
          break;
      return 0;
    case WM_RBUTTONDOWN:
      if(gc->on_mousedown)
        if(gc->on_mousedown(LOWORD(lparam),HIWORD(lparam),win32MBFromWPARAM(wparam)))
          break;
      return 0;
    case WM_LBUTTONUP:
      if(gc->on_mouseup)
        if(gc->on_mouseup(LOWORD(lparam),HIWORD(lparam),win32MBFromWPARAM(wparam)))
          break;
      return 0;
    case WM_RBUTTONUP:
      if(gc->on_mouseup)
        if(gc->on_mouseup(LOWORD(lparam),HIWORD(lparam),win32MBFromWPARAM(wparam)))
          break;
      return 0;
    case WM_MOUSEMOVE:
      if(gc->on_mousemove)
        if(gc->on_mousemove(LOWORD(lparam),HIWORD(lparam),win32MBFromWPARAM(wparam)))
          break;
      return 0;
    case WM_PAINT:
      BeginPaint(wnd,&ps);
      BitBlt(ps.hdc,0,0,gc->sizex,gc->sizey,gdic->dc,0,0,SRCCOPY);
      if(!gdic->wndactive)
        Sleep(10);
      EndPaint(wnd,&ps);
      return 0;
  }
  return DefWindowProc(wnd,msg,wparam,lparam);
}

int     V_GLOBAL graphSysMessageloop()
{
  MSG msg;
  thread_data_t *td;
  td=GetThreadData();
  conlPuts(LOG_INFO,"graphSysMessageloop(): entering Win32 message loop.\n");
  for(;;)
  {
    while((PeekMessage(&msg,NULL,0,0,PM_REMOVE)))
    {
      if(!td->num_ctx) {
        conlPuts(LOG_INFO,"graphSysMessageloop(): no contexts;"
          " exiting from Win32 message loop.\n");
        break;
      }
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
    if(!td->num_ctx) {
      conlPuts(LOG_INFO,"graphSysMessageloop(): no contexts;"
        " exiting from Win32 message loop.\n");
      break;
    }
    if(td->_gc_)
      if(td->_gc_->on_idle)
        td->_gc_->on_idle();
    else
      WaitMessage();
  }
  return msg.wParam;
}

vboolean V_GLOBAL graphProcessSysMessages()
{
  MSG msg;
  while((PeekMessage(&msg,NULL,0,0,PM_REMOVE)))
  {
    if(msg.message==WM_QUIT) 
    {
      return false;
    }
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
  return true;
}

//                

/*
  NOTE. This data shared by all threads.
*/

static int  cls_count = 0;

static int  choose_refresh(gc_t gc)
{
  int rtab[]=
  {
    75,   // < 640 
    120,  // = 640
    100,  // = 800
    85,   // = 960
    85,   // = 1024
    75,   // = 1152
    60,   // = 1280
    60,   // >= 1600
  };
  if(gc->sizex < 640) return rtab[0];
  if(gc->sizex == 640) return rtab[1];
  if(gc->sizex == 800) return rtab[2];
  if(gc->sizex == 960) return rtab[3];
  if(gc->sizex == 1024) return rtab[4];
  if(gc->sizex == 1152) return rtab[5];
  if(gc->sizex == 1280) return rtab[6];
  return rtab[7];
}

/* 06.02.00 - moved here from dmodes.cpp */

#define   MAX_DISPLAY_MODES       128
#define   MAX_REFRESH             64

#define   EDM_BPP_256COLOR        1
#define   EDM_BPP_HICOLOR         2
#define   EDM_BPP_TRUECOLOR       4
#define   EDM_BPP_ALL             7

typedef struct modedesc_s
{
  int   xres,yres,bpp;
  int   refcount;
  int   refresh[MAX_REFRESH];
} modedesc_t;

typedef struct displaymodes_s
{
  int         count;
  modedesc_t  modes[MAX_DISPLAY_MODES];
} displaymodes_t;

static void dispEnumModes(displaymodes_t *modes,int enumflags)
{
  DEVMODE dm;
  int i,x,y,bpp,rr;
  int index;
  if(!modes) 
  {
    conlPrintf(LOG_ERROR,"dispEnumModes(): modes==NULL\n");
    return;
  }
  memset(modes,0,sizeof(displaymodes_t));
  i=0;
  while(EnumDisplaySettings(NULL,i,&dm) != FALSE)
  {
    x=dm.dmPelsWidth;
    y=dm.dmPelsHeight;
    bpp=dm.dmBitsPerPel;
    rr=dm.dmDisplayFrequency;
    i++;
    if((bpp == 8) && ((enumflags & EDM_BPP_256COLOR)==0))
      continue;
    if(((bpp==15)||(bpp==16)) && ((enumflags & EDM_BPP_HICOLOR)==0))
      continue;
    if(((bpp==24)||(bpp==32)) && ((enumflags & EDM_BPP_TRUECOLOR)==0))
      continue;
    //          
    for(index=0;index<modes->count;index++)
    {
      if((modes->modes[index].xres == x) && (modes->modes[index].yres == y)
        && (modes->modes[index].bpp == bpp))
        break;
    }
    if(index == modes->count)
    {
      if(index == MAX_DISPLAY_MODES)
      {
        conlPrintf(LOG_WARNING,"dispEnumModes(): index==MAX_DISPLAY_MODES\n");
        return;
      }
      modes->modes[index].xres=x;
      modes->modes[index].yres=y;
      modes->modes[index].bpp=bpp;
      modes->count++;
    }
    if(modes->modes[index].refcount < MAX_REFRESH)
      modes->modes[index].refresh[modes->modes[index].refcount++]=rr;
  }
  conlPrintf(LOG_INFO,"dispEnumModes(): %d modes.\n",modes->count);
}

/*
static void dispGetMode(int *x,int *y,int *bpp)
{
  if(x) *x=GetSystemMetrics(SM_CXSCREEN);
  if(y) *y=GetSystemMetrics(SM_CYSCREEN);
  if(bpp) {
    HDC dc=GetDC(NULL);
    *bpp=GetDeviceCaps(dc,BITSPIXEL);
    ReleaseDC(NULL,dc);
  }
}
*/

static vboolean dispSetMode(int x,int y,int bpp,int refresh)
{
  DEVMODE dm;
  int retval;
  ZeroMemory(&dm,sizeof(dm));
  dm.dmSize=sizeof(dm);
  dm.dmPelsWidth=x;
  dm.dmPelsHeight=y;
  dm.dmBitsPerPel=bpp;
  dm.dmDisplayFrequency=0;
  dm.dmFields=DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
  if(refresh>0)
  {
    dm.dmDisplayFrequency=refresh;
    dm.dmFields|=DM_DISPLAYFREQUENCY;
  }
  conlPrintf(LOG_WARNING,"dispSetMode(): sx=%d, sy=%d, bpp=%d, refresh=%d Hz. "
    "Calling CDS... ",x,y,bpp,refresh);
  retval=ChangeDisplaySettings(&dm,0) == DISP_CHANGE_SUCCESSFUL;
  if(retval)
    conlPuts(LOG_WARNING,"Ok.\n");
  else
    conlPrintf(LOG_WARNING,"error %d.\n",GetLastError());
  return retval;
}

static void dispRestoreMode()
{
  ChangeDisplaySettings(NULL,0);
}

/* end of dmodes.cpp */

static int win32GetBpp(void)
{
  HDC dc;
  int bpp;
  dc=GetDC(NULL);
  bpp=GetDeviceCaps(dc,BITSPIXEL);
  ReleaseDC(NULL,dc);
  return bpp;
}
  
static vboolean change_display_settings(gc_t c)
{
  displaymodes_t m;
  int cx,cy,i,bpp;
  bpp=win32GetBpp();
  cx=GetSystemMetrics(SM_CXSCREEN);
  cy=GetSystemMetrics(SM_CYSCREEN);
  if(c->bpp==bpp && c->sizex==cx && c->sizey==cy) {
    conlPrintf(LOG_INFO,"change_display_settings(): using current display settings.\n");
    return true;
  }
  dispEnumModes(&m,EDM_BPP_ALL);
  for(i=0;i<m.count;i++)
    if(m.modes[i].xres==c->sizex && m.modes[i].yres==c->sizey && m.modes[i].bpp==c->bpp)
      break;
  if(i==m.count) {
    conlPrintf(LOG_ERROR,"change_display_settings(): mode %dX%dX%d not found.\n",
      c->sizex,c->sizey,c->bpp);
    return false;
  }
  conlPrintf(LOG_INFO,"Attempting to set %d Hz refresh rate...\n",
    choose_refresh(c));
  if(!dispSetMode(c->sizex,c->sizey,c->bpp,choose_refresh(c)))
  {
    conlPuts(LOG_INFO,"Falling back to 60 Hz...\n");
    if(!dispSetMode(c->sizex,c->sizey,c->bpp,60))
      return false;
  }
  return true;
}

void V_GLOBAL graphDeleteContext(gc_t c)
{
  int i;
  gdidata_p gdic;
  if(!c) {
    conlPuts(LOG_WARNING,"graphDeleteContext(NULL): can't delete.\n");
    return;
  }
  gdic=(gdidata_p)c->ctx_data;
  GetThreadData()->num_ctx--;
  if(c->caps & CTX_FULLSCREEN)
    dispRestoreMode();
  if(graphGetCurrentContext() == c) {
    conlPrintf(LOG_INFO,
      "graphDeleteContext(): %8X==graphGetCurrentContext(). Active"
      " context is NULL now.\n",c);
    graphMakeCurrent(NULL);
  }
  if(gdic->dc) 
    DeleteDC(gdic->dc);
  if(gdic->bmp) 
    DeleteObject(gdic->bmp);
  if(gdic->br) 
    DeleteObject(gdic->br);
  if(gdic->pn) 
    DeleteObject(gdic->pn);
  if(gdic->fn) 
    DeleteObject(gdic->fn);
  if(gdic->wnd) 
    DestroyWindow(gdic->wnd);
  if(gdic->wndclass) 
    UnregisterClass(gdic->cls,gdic->inst);
  for(i=0;i<256;i++)
  {
    if(gdic->apn[i]) DeleteObject(gdic->apn[i]);
    if(gdic->abr[i]) DeleteObject(gdic->abr[i]);
  }
  free(gdic);
  free(c);
}

void    V_GLOBAL graphFlush(gc_t c)
{
  if(!c) return;
  InvalidateRect(((gdidata_p)c->ctx_data)->wnd,NULL,0);
}

// here

static void V_CALL gdi_setcolor(gc_t _gc_,int r,int g,int b)
{
  HPEN pn;
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  gdic->lpn.lopnColor=RGB(r,g,b);
  pn=CreatePenIndirect(&gdic->lpn);
  DeleteObject(gdic->pn);
  gdic->pn=pn;
  SelectObject(gdic->dc,gdic->pn);
}

static void V_CALL gdi_setindex(gc_t _gc_,int index)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  SelectObject(gdic->dc,gdic->apn[index]);
}

static void V_CALL gdi_setbkcolor(gc_t _gc_,int r,int g,int b)
{
  HBRUSH br;
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  gdic->lbr.lbColor=RGB(r,g,b);
  br=CreateBrushIndirect(&gdic->lbr);
  DeleteObject(gdic->br);
  gdic->br=br;
  SelectObject(gdic->dc,gdic->br);
}

static void V_CALL gdi_setbkindex(gc_t _gc_,int index)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  SelectObject(gdic->dc,gdic->abr[index]);
}

static void V_CALL gdi_settextcolor(gc_t _gc_,int r,int g,int b)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  SetTextColor(gdic->dc,RGB(r,g,b));
}

static void V_CALL gdi_settextindex(gc_t _gc_,int index)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  SetTextColor(gdic->dc,gdic->rgb[index]);
}

static void V_CALL gdi_moveto(gc_t _gc_,int x,int y)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  gdic->pt[0].x=x;
  gdic->pt[0].y=y;
}

static void V_CALL gdi_lineto(gc_t _gc_,int x,int y)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  gdic->pt[1].x=x;
  gdic->pt[1].y=y;
  Polyline(gdic->dc,gdic->pt,2);
  gdic->pt[0].x=x;
  gdic->pt[0].y=y;
}
  
static void V_CALL gdi_line(gc_t _gc_,int x1,int y1,int x2,int y2)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  gdic->pt[1].x=x1;
  gdic->pt[1].y=y1;
  gdic->pt[2].x=x2;
  gdic->pt[2].y=y2;
  Polyline(gdic->dc,gdic->pt+1,2);
}

static void V_CALL gdi_rectangle(gc_t _gc_,int x1,int y1,int x2,int y2)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  gdic->pt[1].x=x1;
  gdic->pt[1].y=y1;
  gdic->pt[2].x=x2;
  gdic->pt[2].y=y1;
  gdic->pt[3].x=x2;
  gdic->pt[3].y=y2;
  gdic->pt[4].x=x1;
  gdic->pt[4].y=y2;
  gdic->pt[5].x=x1;
  gdic->pt[5].y=y1;
  Polyline(gdic->dc,gdic->pt+1,5);
}

static void V_CALL gdi_bar(gc_t _gc_,int x1,int y1,int x2,int y2)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  PatBlt(gdic->dc,x1,y1,x2-x1+1,y2-y1+1,PATCOPY);
}
  
static void V_CALL gdi_clearviewport(gc_t _gc_)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  PatBlt(gdic->dc,0,0,_gc_->sizex,_gc_->sizey,PATCOPY);
}

static void V_CALL gdi_circle(gc_t _gc_,int x,int y,int radius)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  SelectObject(gdic->dc,gdic->nbr);
  Ellipse(gdic->dc,x-radius,y-radius,x+radius,y+radius);
  SelectObject(gdic->dc,gdic->br);
}

static void V_CALL gdi_ellipse(gc_t _gc_,int x1,int y1,int x2,int y2)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  SelectObject(gdic->dc,gdic->nbr);
  Ellipse(gdic->dc,x1,y1,x2,y2);
  SelectObject(gdic->dc,gdic->br);
}

static void V_CALL gdi_triangle(gc_t _gc_,int x1,int y1,int x2,int y2,int x3,int y3)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  SelectObject(gdic->dc,gdic->npn);
  gdic->pt[1].x=x1;
  gdic->pt[1].y=y1;
  gdic->pt[2].x=x2;
  gdic->pt[2].y=y2;
  gdic->pt[3].x=x3;
  gdic->pt[3].y=y3;
  gdic->pt[4].x=x1;
  gdic->pt[4].y=y1;
  Polygon(gdic->dc,gdic->pt+1,4);
  SelectObject(gdic->dc,gdic->pn);
}

static void V_CALL gdi_getextent(gc_t _gc_,char *str,int *sx,int *sy)
{
  gdidata_p gdic;
  SIZE sz;
  gdic=(gdidata_p)_gc_->ctx_data;
  GetTextExtentPoint32(gdic->dc,str,strlen(str),&sz);
  *sx=sz.cx;
  *sy=sz.cy;
}

static void V_CALL gdi_outtext(gc_t _gc_,char *str)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  TextOut(gdic->dc,gdic->pt[0].x,gdic->pt[0].y,str,strlen(str));
}

static void V_CALL gdi_outtextxy(gc_t _gc_,int x,int y,char *str)
{
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  TextOut(gdic->dc,x,y,str,strlen(str));
}

static vboolean V_CALL gdi_loadpalette(gc_t _gc_,char *unix_resname)
{
  int r,g,b;
  int i,rh;
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  rh=rfind(unix_resname);
  if(rh!=-1)
  {
    SelectObject(gdic->dc,gdic->npn);
    SelectObject(gdic->dc,gdic->nbr);
    for(i=0;i<256;i++)
    {
      r=rread8(rh);
      g=rread8(rh);
      b=rread8(rh);
      gdic->rgb[i]=RGB(r,g,b);
      if(gdic->apn[i]) DeleteObject(gdic->apn[i]);
      gdic->apn[i]=CreatePen(PS_SOLID,1,gdic->rgb[i]);
      if(gdic->abr[i]) DeleteObject(gdic->abr[i]);
      gdic->abr[i]=CreateSolidBrush(gdic->rgb[i]);
    }
    rclose(rh);
  }
  else {
    u8_t *ptr;
    HRSRC rw;
    rw=FindResource(gdic->inst,unix_resname,RT_RCDATA);
    ptr=(u8_t*)LockResource(LoadResource(gdic->inst,rw));
    if(!ptr)
    {
      conlPrintf(LOG_WARNING,"loadpalette('%s'): not found.\n",unix_resname);
      return false;
    }
    SelectObject(gdic->dc,gdic->npn);
    SelectObject(gdic->dc,gdic->nbr);
    for(i=0;i<256;i++)
    {
      r=*ptr++;
      g=*ptr++;
      b=*ptr++;
      gdic->rgb[i]=RGB(r,g,b);
      if(gdic->apn[i]) DeleteObject(gdic->apn[i]);
      gdic->apn[i]=CreatePen(PS_SOLID,1,gdic->rgb[i]);
      if(gdic->abr[i]) DeleteObject(gdic->abr[i]);
      gdic->abr[i]=CreateSolidBrush(gdic->rgb[i]);
    }
  }
  setindex(0);
  setbkindex(0);
  settextindex(0);
  return true;
}

static void V_CALL gdi_usefont(gc_t _gc_,int sizex,int sizey,
                                  vec_t direction,u32_t effects,
                                  char *name)
{
  HFONT newf;
  gdidata_p gdic;
  gdic=(gdidata_p)_gc_->ctx_data;
  gdic->lfn.lfHeight=sizey;
  gdic->lfn.lfWidth=sizex;
  gdic->lfn.lfEscapement=(int)(ToDegrees(direction) * 10);
  gdic->lfn.lfOrientation=0;
  gdic->lfn.lfWeight=FW_DONTCARE;
  gdic->lfn.lfItalic=(unsigned char)((effects & EF_ITALIC) ? TRUE : FALSE);
  gdic->lfn.lfUnderline=(unsigned char)((effects & EF_UNDERLINE) ? TRUE : FALSE);
  gdic->lfn.lfStrikeOut=(unsigned char)((effects & EF_STRIKEOUT) ? TRUE : FALSE);
  gdic->lfn.lfCharSet=RUSSIAN_CHARSET;
  gdic->lfn.lfOutPrecision=OUT_DEFAULT_PRECIS;
  gdic->lfn.lfClipPrecision=CLIP_DEFAULT_PRECIS;
  gdic->lfn.lfQuality=DEFAULT_QUALITY;
  gdic->lfn.lfPitchAndFamily=DEFAULT_PITCH;
#ifdef UNICODE  
  wcscpy(gdic->lfn.lfFaceName,(wchar_t*)name);
#else
  strcpy(gdic->lfn.lfFaceName,name);
#endif  
  newf=CreateFontIndirect(&gdic->lfn);
  SelectObject(gdic->dc,newf);
  DeleteObject(gdic->fn);
  gdic->fn=newf;
}

gc_t V_GLOBAL graphCreateContext(graph_settings_t *start)
{
  int sx,sy;
  HDC wdc;
  gc_t gc;
  gdidata_p gdic;
  int style;
  thread_data_t *td;
  gc=(gc_t)malloc(sizeof(rendering_context_t));
  gdic=(gdidata_p)malloc(sizeof(gdidata_t));
  V_ASSERT(gc);
  V_ASSERT(gdic);
  memset(gc,0,sizeof(rendering_context_t));
  memset(gdic,0,sizeof(rendering_context_t));
  gc->ctx_data=gdic;
  /*
  TODO: class name is platform-dependent; implement another way to pass it.
  if(start && start->cls)
    strncpy(c->cls,start->cls,255);
  else
  */
    sprintf(gdic->cls,"graphClass%000000X",cls_count++);
  /*
  TODO: window title is platform-dependent; implement another way to pass it.
  if(start && start->title)
    strncpy(c->title,start->title,255);
  else
  */
    strcpy(gdic->title,"Graph");
  gdic->inst=GetModuleHandle(NULL);
  if(start && start->sizex)
    gc->sizex=start->sizex;
  else
    gc->sizex=640;
  if(start && start->sizey)
    gc->sizey=start->sizey;
  else
    gc->sizey=480;
  if(start && start->bpp)
    gc->bpp=start->bpp;
  else
  {
    gc->bpp=win32GetBpp();
    if(start && (start->caps & CTX_FULLSCREEN))
      conlPrintf(LOG_INFO,"Using default bit depth: %d.\n",gc->bpp);
  }
  sx=GetSystemMetrics(SM_CXSCREEN);
  sy=GetSystemMetrics(SM_CYSCREEN);
  gdic->wc.lpszClassName=gdic->cls;
  gdic->wc.hInstance=gdic->inst;
  gdic->wc.style = CS_HREDRAW | CS_VREDRAW;
  gdic->wc.lpfnWndProc=graph_window_proc;
  gdic->wc.cbClsExtra = 0;
  gdic->wc.cbWndExtra = 0;
  /* 'icon' must NOT be created with MAKEINTRESOURCE */
  /*
  TODO: icon resname is platform-dependent; implement another way to pass it.
  if(start && start->icon)
    gdic->wc.hIcon = LoadIcon(gdic->inst,start->icon);
  else
  */
    gdic->wc.hIcon = LoadIcon(NULL,IDI_WINLOGO);
  gdic->wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  gdic->wc.hbrBackground = GetStockObject(HOLLOW_BRUSH);
  gdic->wc.lpszMenuName=NULL;
  gdic->wndclass=::RegisterClass(&gdic->wc);
  if(!gdic->wndclass) 
  {
    conlPrintf(LOG_WARNING,"graphCreateContext(): "
      "can't RegisterClass() '%s'\n",gdic->wc.lpszClassName);
    free(gdic);
    free(gc);
    return NULL;
  }
  conlPrintf(LOG_INFO,"graphCreateContext(): registered class '%s'.\n",
    gdic->wc.lpszClassName);
  style=((sx==gc->sizex && sy==gc->sizey) || (start && start->caps & CTX_FULLSCREEN)) 
    ? WS_POPUP : WS_POPUP | WS_CAPTION;

  gdic->wnd=CreateWindow(
    gdic->cls,
    gdic->title,
    style,
    0,0,
    gc->sizex,
    gc->sizey,
    NULL,
    NULL,
    gdic->inst,
    NULL);
  if(!gdic->wnd)
  {
    conlPuts(LOG_WARNING,"graphCreateContext(): can't CreateWindow().\n");
    UnregisterClass(gdic->wc.lpszClassName,gdic->inst);
    free(gdic);
    free(gc);
    return NULL;
  }
  conlPrintf(LOG_INFO,
    "graphCreateContext(): window created (class '%s').\n",
    gdic->wc.lpszClassName);
  SetWindowLong(gdic->wnd,GWL_USERDATA,(LONG)gc);
  ShowWindow(gdic->wnd,SW_SHOWNORMAL);
  if(style!=WS_POPUP)
  {
    RECT rc;
    GetClientRect(gdic->wnd,&rc);
    int width=2*gc->sizex-rc.right;
    int height=2*gc->sizey-rc.bottom;
    SetWindowPos(gdic->wnd,HWND_NOTOPMOST,0,0,width,height,SWP_NOZORDER|SWP_NOMOVE);
  }
  if(start && (start->caps & CTX_FULLSCREEN)) {
    if(!change_display_settings(gc)) {
      DestroyWindow(gdic->wnd);
      UnregisterClass(gdic->wc.lpszClassName,gdic->inst);
      free(gdic);
      free(gc);
      return NULL;
    }
  }
  if(style==WS_POPUP)
    gc->caps|=CTX_FULLSCREEN;
  if(start && (start->caps & CTX_3D))
  {
    /* TODO: implement OpenGL loading */
    /*
    if(!vglLoadOpenGL("3dfxgl.dll")) {
      conlPrintf(...
    */
    
    conlPuts(LOG_ERROR,"graphCreateContext(): OpenGL code not implemented yet.\n");
    DestroyWindow(gdic->wnd);
    UnregisterClass(gdic->wc.lpszClassName,gdic->inst);
    free(gdic);
    free(gc);
    gc=NULL;
        
    return gc;
  }
  
  if(start && (start->caps & CTX_ACCELERATED))
  {
    /* TODO: implement DirectDraw 2D accelerated backbuffering */
    /* 
    CoInitialize();
    if(!SUCCEDED(CoCreateInstance(IID_IDirectDraw,...)) {
      conlPrintf(LOG_ERROR,"CoCreateInstance(): can't create DDraw object.\n");
      ...
    */
    conlPuts(LOG_ERROR,"graphCreateContext(): DirectDraw 2D code not implemented yet.\n");
    DestroyWindow(gdic->wnd);
    UnregisterClass(gdic->wc.lpszClassName,gdic->inst);
    free(gdic);
    free(gc);
    gc=NULL;

    return gc;
  }
  
  /* creating bitmap for backbuffer */
  wdc=GetDC(gdic->wnd);
  gdic->dc=CreateCompatibleDC(wdc);
  gdic->bmp=CreateCompatibleBitmap(wdc,gc->sizex,gc->sizey);
  SelectObject(gdic->dc,gdic->bmp);

  /* default apperance of window */
  gdic->lbr.lbStyle=BS_SOLID;
  gdic->lbr.lbColor=RGB(0,0,0);
  
  gdic->lpn.lopnStyle=PS_SOLID;
  gdic->lpn.lopnWidth.x=1;
  gdic->lpn.lopnWidth.y=1;
  gdic->lpn.lopnColor=RGB(255,255,255);
  
  gdic->lfn.lfHeight=(gc->sizex <= 640) ? 10 : ((gc->sizex == 800) ? 12 : 15 );
  gdic->lfn.lfWidth=0;
  gdic->lfn.lfEscapement=0;
  gdic->lfn.lfOrientation=0;
  gdic->lfn.lfWeight=FW_DONTCARE;
  gdic->lfn.lfItalic=0;
  gdic->lfn.lfUnderline=0;
  gdic->lfn.lfStrikeOut=0;
  gdic->lfn.lfCharSet=RUSSIAN_CHARSET;
  gdic->lfn.lfOutPrecision=OUT_DEFAULT_PRECIS;
  gdic->lfn.lfClipPrecision=CLIP_DEFAULT_PRECIS;
  gdic->lfn.lfQuality=DEFAULT_QUALITY;
  gdic->lfn.lfPitchAndFamily=DEFAULT_PITCH;
#ifdef UNICODE  
  wcscpy(gdic->lfn.lfFaceName,(wchar_t*)"Courier");
#else
  strcpy(gdic->lfn.lfFaceName,"Courier");
#endif  
  /* ========================== */
  gdic->br=CreateBrushIndirect(&gdic->lbr);
  gdic->pn=CreatePenIndirect(&gdic->lpn);
  gdic->fn=CreateFontIndirect(&gdic->lfn);
  gdic->npn=GetStockObject(NULL_PEN);
  gdic->nbr=GetStockObject(HOLLOW_BRUSH);
  SelectObject(gdic->dc,gdic->br);
  SelectObject(gdic->dc,gdic->pn);
  SelectObject(gdic->dc,gdic->fn);
  SetBkMode(gdic->dc,TRANSPARENT);
  SetTextColor(gdic->dc,gdic->lpn.lopnColor);
  SetBkColor(gdic->dc,gdic->lbr.lbColor);
  PatBlt(gdic->dc,0,0,gc->sizex,gc->sizey,BLACKNESS);
  gdic->wndactive=1;
  BitBlt(wdc,0,0,gc->sizex,gc->sizey,gdic->dc,0,0,SRCCOPY);
  ReleaseDC(gdic->wnd,wdc);

#ifdef  __cplusplus
#define R_FORCE_ASSIGNMENT(d32,s32) *((u32_t*)&(d32))=(u32_t)(s32)
#else
#define R_FORCE_ASSIGNMENT(d32,s32) (d32)=(s32)
#endif

  R_FORCE_ASSIGNMENT(gc->r2d.clearviewport,gdi_clearviewport);
  R_FORCE_ASSIGNMENT(gc->r2d.moveto,gdi_moveto);
  R_FORCE_ASSIGNMENT(gc->r2d.lineto,gdi_lineto);
  R_FORCE_ASSIGNMENT(gc->r2d.line,gdi_line);
  R_FORCE_ASSIGNMENT(gc->r2d.rectangle,gdi_rectangle);
  R_FORCE_ASSIGNMENT(gc->r2d.bar,gdi_bar);
  R_FORCE_ASSIGNMENT(gc->r2d.circle,gdi_circle);
  R_FORCE_ASSIGNMENT(gc->r2d.ellipse,gdi_ellipse);
  R_FORCE_ASSIGNMENT(gc->r2d.triangle,gdi_triangle);
  R_FORCE_ASSIGNMENT(gc->r2d.usefont,gdi_usefont);
  R_FORCE_ASSIGNMENT(gc->r2d.getextent,gdi_getextent);
  R_FORCE_ASSIGNMENT(gc->r2d.outtext,gdi_outtext);
  R_FORCE_ASSIGNMENT(gc->r2d.outtextxy,gdi_outtextxy);
  R_FORCE_ASSIGNMENT(gc->r2d.setcolor,gdi_setcolor);
  R_FORCE_ASSIGNMENT(gc->r2d.setbkcolor,gdi_setbkcolor);
  R_FORCE_ASSIGNMENT(gc->r2d.settextcolor,gdi_settextcolor);
  R_FORCE_ASSIGNMENT(gc->r2d.loadpalette,gdi_loadpalette);
  R_FORCE_ASSIGNMENT(gc->r2d.setindex,gdi_setindex);
  R_FORCE_ASSIGNMENT(gc->r2d.setbkindex,gdi_setbkindex);
  R_FORCE_ASSIGNMENT(gc->r2d.settextindex,gdi_settextindex);

#undef R_FORCE_ASSIGNMENT
  
  gc->caps|=CTX_PALETTED|CTX_COLORRGB|CTX_ENHFONTS;
  td=GetThreadData();
  td->num_ctx++;
  return gc;
}

