#include "mmplus.h"
#pragma  hdrstop

// =======================================================================
void CWavePlayer::ResetAll()
{
  iNumDevs=0;
  hSoundDevice=0;
  memset(&wh,0,sizeof(WAVEHDR));
  lpWave=NULL;
  dwSamplesTotal=0;
  bLastError=TRUE;
  iPlayerState=WPL_CLOSED;
}

// =======================================================================
void CWavePlayer::ReleaseAll()
{
  if(iPlayerState>WPL_CLOSED) Close();
}

CWavePlayer::CWavePlayer(HWND hWnd)
{
  ResetAll();
  this->hWnd=hWnd;
}

CWavePlayer::~CWavePlayer()
{
  ReleaseAll();
}

void CWavePlayer::Open(CWave *lpWave)
{
  MMRESULT mmr;
  ReleaseAll();
  ResetAll();
  iNumDevs=waveOutGetNumDevs();
  if(iNumDevs==0) return;
  if(!lpWave) return;
  this->lpWave=lpWave;
  dwSamplesTotal=lpWave->GetWaveSize()/lpWave->GetWaveChannels()/
    (lpWave->GetWaveBitsPerSample()>>3);
  mmr=::waveOutOpen(&hSoundDevice,WAVE_MAPPER,lpWave->GetWaveFormatEx(),
    (DWORD)hWnd,0,(hWnd) ? CALLBACK_WINDOW : CALLBACK_NULL);
  if(mmr!=MMSYSERR_NOERROR) return;
  bLastError=FALSE;
  iPlayerState=WPL_OPENED;
  return;
}

void CWavePlayer::Prepare(CWave *lpWave, BOOL bLooped)
{
  MMRESULT mmr;
  bLastError=TRUE;
  if(iPlayerState>WPL_OPENED)
  {
    if(iPlayerState>WPL_PREPARED) ::waveOutReset(hSoundDevice);
    ::waveOutUnprepareHeader(hSoundDevice,&wh,sizeof(WAVEHDR));
  }
  else if(iPlayerState<WPL_OPENED) return;
  wh.lpData=lpWave->GetWaveBits();
  wh.dwBufferLength=lpWave->GetWaveSize();
  if(bLooped)
  {
    wh.dwFlags=WHDR_BEGINLOOP;
  }
  else
  {
    wh.dwFlags=0;
  }
  wh.dwLoops=1;
  mmr=::waveOutPrepareHeader(hSoundDevice,&wh,sizeof(WAVEHDR));
  if(mmr!=MMSYSERR_NOERROR) return;
  bLastError=FALSE;
  iPlayerState=WPL_PREPARED;
  return;
}

void CWavePlayer::Play()
{
  if(iPlayerState>=WPL_PREPARED)
  {
    ::waveOutWrite(hSoundDevice,&wh,sizeof(WAVEHDR));
    iPlayerState=WPL_PLAYING;
  }
}

void CWavePlayer::Stop()
{
  if(iPlayerState>=WPL_PLAYING)
  {
    ::waveOutReset(hSoundDevice);
    iPlayerState=WPL_PREPARED;
  }
}

void CWavePlayer::Pause()
{
  if(iPlayerState==WPL_PLAYING)
  {
    ::waveOutPause(hSoundDevice);
    iPlayerState=WPL_PAUSED;
  }
}

void CWavePlayer::Resume()
{
  if(iPlayerState==WPL_PAUSED)
  {
    ::waveOutRestart(hSoundDevice);
  }
  iPlayerState=WPL_PLAYING;
}

void CWavePlayer::Close()
{
  if(iPlayerState>=WPL_PLAYING)
  {
    ::waveOutReset(hSoundDevice);
    iPlayerState=WPL_PREPARED;
  }
  if(iPlayerState>=WPL_PREPARED)
  {
    ::waveOutUnprepareHeader(hSoundDevice,&wh,sizeof(WAVEHDR));
    iPlayerState=WPL_OPENED;
  }
  if(iPlayerState>=WPL_OPENED)
  {
    ::waveOutClose(hSoundDevice);
    iPlayerState=WPL_CLOSED;
    hSoundDevice=NULL;
  }
  return;
}

BOOL CWavePlayer::IsPlaying()
{
  return iPlayerState==WPL_PLAYING;
}

DWORD CWavePlayer::GetProgress1000()
{
  MMTIME mmt;
  if(iPlayerState>=WPL_PLAYING)
  {
    mmt.wType=TIME_SAMPLES;
    ::waveOutGetPosition(hSoundDevice,&mmt,sizeof(MMTIME));
    return (mmt.u.sample*1000)/dwSamplesTotal;
  }
  return 0;
}

void CWavePlayer::SetVolume(DWORD dwLeft, DWORD dwRight)
{
  if(iPlayerState>WPL_PREPARED)
  {
    dwLeft&=0xFFFF;
    dwRight&=0xFFFF;
    ::waveOutSetVolume(hSoundDevice,dwLeft|(dwRight<<16));
  }
}

