#include  <windows.h>
#include  <mmsystem.h>
#pragma   hdrstop
#include  "w_rec.h"

#define         BUFFER_LENGTH     (8 * 1024)

HWAVEIN         hDev;
WAVEFORMATEX    wfe;
WAVEHDR         whd[2];
int             cur;
char            buf1[BUFFER_LENGTH], buf2[BUFFER_LENGTH];
int             total;
HANDLE          hEvData, hEvCont, hRecorder;

TCopySoundCallback  CopySound;
TNotifyProc     BeforeStart;
TNotifyProc     AfterStop;

#define         IsEventSignaled(Ev) (WaitForSingleObject(Ev,0)==WAIT_OBJECT_0)

/*
========================
  HandleWimData
========================
*/

static int HandleWimData()
{
  int done;
  if(!cur) cur=1; else cur=0;
  if(IsEventSignaled(hEvCont))
  {
    if(MMSYSERR_NOERROR!=waveInAddBuffer(hDev,&whd[cur],sizeof(WAVEHDR)))
    {
      ResetEvent(hEvCont);
      waveInClose(hDev);
      return FALSE;
    }
    if(MMSYSERR_NOERROR!=waveInStart(hDev))
    {
      ResetEvent(hEvCont);
      waveInClose(hDev);
      return FALSE;
    }
  }
  if(!cur) done=1; else done=0;
  if(MMSYSERR_NOERROR!=waveInUnprepareHeader(hDev,&whd[done],sizeof(WAVEHDR)))
  {
    ResetEvent(hEvCont);
    waveInClose(hDev);
    return FALSE;
  }
  total=total+whd[done].dwBytesRecorded;
  if(CopySound)
    CopySound(whd[done].lpData,whd[done].dwBytesRecorded);
  if(IsEventSignaled(hEvCont))
  {
    if(MMSYSERR_NOERROR!=waveInPrepareHeader(hDev,&whd[done],sizeof(WAVEHDR)))
    {
      ResetEvent(hEvCont);
      waveInClose(hDev);
      return FALSE;
    }
  }
  else
  {
    waveInClose(hDev);
    if(AfterStop) AfterStop();
  }
  return TRUE;
}

/*
========================
RecorderThread
========================
*/

DWORD WINAPI RecorderThread(void *data)
{
  WaitForSingleObject(hEvCont,INFINITE);
  while(IsEventSignaled(hEvCont))
  {
    ResetEvent(hEvData);
    WaitForSingleObject(hEvData,INFINITE);
    HandleWimData();
  }
  return 0;
}

/*
========================
  wrecInitialize
========================
*/

BOOL __declspec(dllexport) __stdcall wrecInitialize(TCopySoundCallback CopyCallback,TNotifyProc Before,TNotifyProc After)
{
  hDev = NULL;
  cur = 0;
  total = 0;
  if(waveInGetNumDevs()<1) return FALSE;
  CopySound = CopyCallback;
  BeforeStart = Before;
  AfterStop = After;
  //   
  hEvData=CreateEvent(NULL,TRUE,FALSE,NULL);
  hEvCont=CreateEvent(NULL,TRUE,FALSE,NULL);
  hRecorder=CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)RecorderThread,NULL,0,NULL);
  if(!hEvData || !hEvCont || !hRecorder) return FALSE;
  return TRUE;
}

/*
========================
  wrecFinalize
========================
*/

void __declspec(dllexport) __stdcall wrecFinalize()
{
  if(IsEventSignaled(hEvCont)) wrecStop();
  WaitForSingleObject(hRecorder,INFINITE);
  CloseHandle(hEvData);
  CloseHandle(hEvCont);
  CloseHandle(hRecorder);
}

//========================
//  Prepare
//========================

BOOL __declspec(dllexport) __stdcall wrecPrepare(int Frequency,int Bits,int Channels)
{
  int exitcode;
  wfe.wFormatTag=WAVE_FORMAT_PCM;
  wfe.nSamplesPerSec=Frequency;
  wfe.wBitsPerSample=Bits;
  wfe.nChannels=Channels;
  wfe.nBlockAlign=(wfe.nChannels*wfe.wBitsPerSample) >> 3;
  wfe.nAvgBytesPerSec=wfe.nSamplesPerSec*wfe.nBlockAlign;
  wfe.cbSize=0;
  if(MMSYSERR_NOERROR!=waveInOpen(&hDev,WAVE_MAPPER,&wfe,(int)hEvData,0,CALLBACK_EVENT))
    return FALSE;
  memset(whd,0,sizeof(whd));
  whd[0].lpData=buf1;
  whd[0].dwBufferLength=BUFFER_LENGTH;
  whd[1].lpData=buf2;
  whd[1].dwBufferLength=BUFFER_LENGTH;
  cur=0;
  if(MMSYSERR_NOERROR!=waveInPrepareHeader(hDev,&whd[0],sizeof(WAVEHDR)))
    return FALSE;
  if(MMSYSERR_NOERROR!=waveInPrepareHeader(hDev,&whd[1],sizeof(WAVEHDR)))
    return FALSE;
  return TRUE;
}

//========================
//  Start
//========================


int __declspec(dllexport) __stdcall wrecStart()
{
  total=0;
  if(BeforeStart) BeforeStart();
  if(MMSYSERR_NOERROR!=waveInAddBuffer(hDev,&whd[cur],sizeof(WAVEHDR)))
    return FALSE;
  if(MMSYSERR_NOERROR!=waveInStart(hDev))
    return FALSE;
  SetEvent(hEvCont);
  return TRUE;
}

//========================
//  Stop
//========================

void __declspec(dllexport) __stdcall wrecStop()
{
  ResetEvent(hEvCont);
  waveInReset(hDev);
}

//========================
//  Misc
//========================

int __declspec(dllexport) __stdcall wrecGetBytesRecorded()
{
  return total;
}

// ============================ THE END ==================================

// #define DEBUG_TEST

#ifdef DEBUG_TEST

#include <stdio.h>

int pos = 0;
#define maxlen 8 * 10 * 11025
char lb[maxlen];
int done=0;

void __stdcall copyfunc(char *buf,int len)
{
  if(pos+len>=maxlen)
  {
    printf("Copying %d bytes, exiting\n",maxlen-pos);
    memcpy(lb+pos,buf,maxlen-pos);
    wrecStop();
    done=1;
  }
  else
  {
    printf("Copying %d bytes...\n",len);
    memcpy(lb+pos,buf,len);
    pos+=len;
  }
}
  
void main()
{
  FILE *out;
  wrecInitialize(copyfunc,NULL,NULL);
  if(!wrecPrepare(44100,16,1))
  {
    puts("wrecPrep failed");
    return;
  }
  if(!wrecStart())
  {
    puts("wrecStart failed");
    return;
  }
  while(!done) Sleep(10);
  out=fopen("out.raw","wb");
  fwrite(lb,1,maxlen,out);
  fclose(out);
}

#endif
