/*<html><body bgcolor=#B4C4B0><font color=#000000><strong><pre>*/
#ifndef   V_BTREE_H
#define   V_BTREE_H

/*
**	Binary tree
*/
typedef struct btree_s
{
	void			*data;
	struct btree_s	*parent;
	struct btree_s	*children[2];
}	btree_t, *btree_p;

#ifdef	__cplusplus
extern "C" {
#endif	/* __cplusplus */
	/*	Allocates new tree node and optionally connects it with
	 *	parent. Input:
	 *		data	- pointer to store in ->data field;
	 *		parent	- parent node for new node (can be NULL);
	 *		child_index - if 0, connect new node at left of parent.
	 */
	btree_p	V_GLOBAL	btAllocNode(void *data,btree_t *parent,
		int child_index);
	/*	Same as above, but allocates additional space and copies
	 *	'len' bytes from 'data' to node->data.
	 */
	btree_p	V_GLOBAL	btStore(void *data,size_t len,btree_t *parent,
		int child_index);
	/*
	 *	Returns root of tree with given node.
	 */
	btree_p	V_GLOBAL	btRootNode(btree_t *node);
	/*
	**	Traverses binary tree in order:
	**		root
	**			root->left
	**				root->left->left
	**				... etc
	**			root->right
	**				root->right->left
	**					root->right->left->left
	**					... etc
	*/
	void	V_GLOBAL	btTraverseLeft(btree_t *root,
		void (V_CALL *handler)(void*));
	/*
	**	Traverses binary tree in order:
	**		root
	**			root->left
	**			root->right
	**				root->left->left
	**				root->left->right
	**				root->right->left
	**				root->right->right
	**					... etc
	*/
	void	V_GLOBAL	btTraverseLeftRF(btree_t *root,
		void (V_CALL *handler)(void*));
	/*
	**	Traverses binary tree in order:
	**		root
	**			root->right
	**				root->right->right
	**				... etc
	**			root->left
	**				root->left->right
	**				... etc
	*/
	void	V_GLOBAL	btTraverseRight(btree_t *root,
		void (V_CALL *handler)(void*));
	/*
	**	Traverses binary tree in order:
	**		root
	**			root->right
	**			root->left
	**				root->right->right
	**				root->right->left
	**				root->left->right
	**				root->left->left
	**					... etc
	*/
	void	V_GLOBAL	btTraverseRightRF(btree_t *root,
		void (V_CALL *handler)(void*));
	/*
	**	Removes all nodes from leftmost to rightmost, optionally
	**	calling 'destructor' for data in each node.
	*/
	void	V_GLOBAL	btDeleteAllNodes(btree_t *root,
		void (V_CALL *destructor)(void*));
#ifdef	__cplusplus
}
#endif	/* __cplusplus */

#endif	/* V_BTREE_H */
/*</pre></body></html>*/
/*EOF*/
