/*<html><body bgcolor=#000000><font color=#B0B0B0><pre>*/
#ifndef		V_FS_H
#define		V_FS_H

typedef		unsigned		file_t;	/* "inode" :) of resource handle */

#ifdef	__cplusplus
extern "C" {
#endif	/* __cplusplus */
	/*	Converts DOS path to UNIX (CYGWIN) path
	 *	Input:
	 *		- path: string of len _MAX_PATH to be converted.
	 *	Output:
	 *		'path' converted; returns pointer to 'path'.
	 */
	char_p	V_GLOBAL	fsDos2Unix(char *path);
	/* Converts UNIX (CYGWIN) path to DOS path */
	char_p	V_GLOBAL	fsUnix2Dos(char *path);
	/*	========== FS service routines ================================ */
	/*	Enumerates contents of current directory (and, optionally, all
	 *	subdirectories). Contents of PAK directories NOT enumerated.
	 *	Input:
	 *		- dir_callback (can be NULL): caller - supplied callback
	 *			for directories;
	 *		- file_callback (can be NULL): caller - supplied callback
	 *			for files;
	 *		- recursive: if nonzero, fsScan enumerates all subdirs.
	 *	Notes.
	 *		1. dir_callback always called BEFORE directory processing;
	 *	for example, if recursive == 0 than fsScan called 'dir_callback',
	 *	and than, for all files in directory, 'file_callback'.
	 *		2. Under MS OSes, names of files and directories passed in
	 *	CYGWIN - style: "c:\winnt" becomes "//c/winnt", etc.
	 */
	void		V_GLOBAL	fsScan(
		void (V_CALL *dir_callback)(char *full_unix_style_dir_name),
		void (V_CALL *file_callback)(char *full_unix_style_file_name),
		vboolean recursive);
	/*	Enumerates contents of all mounted PAK files.
	 *	Input:
	 *		pak_callback - will be called before enumeration of
	 *			PAK file 'pak_name';
	 *		file_callback - will be called for every entry in 'pak_name'.
	 *	Note.
	 */
	void		V_GLOBAL	fsScanPaks(
		void (V_CALL *pak_callback)(char *pak_name),
		void (V_CALL *file_callback)(char *pak_entry));
	/* Splits full UNIX (of CYGWIN32-style) path */
	void		V_GLOBAL	fsSplitPath(char *src,char *drive,char *dir,
		char *name,char *ext);
	/* Makes full UNIX (or CYGWIN32-style) path */
	void		V_GLOBAL	fsMakePath(char *dest_path,char *drive,
		char *dir,char *name,char *ext);
	/* Replaces extension in filename */
	void		V_GLOBAL	fsReplaceExt(char *dst,char *name,char *ext);
	/*	========== PAK service routines =============================== */
	/*	Mounts *.pak files in given directory and all subdirectories */
	vboolean	V_GLOBAL	fsInit(char *path);
	/*	Cleanup: releases all objects */
	void		V_GLOBAL	fsShutdown();
	/*	========== file service routines ============================== */
	/*	All names in UNIX form. For MS OSes, see note #2 for 'fsScan()'.
	 */
	/*	Invalid handle value is 0, NOT -1 ! */
	file_t		V_GLOBAL	fileOpenRead(char *name);
	file_t		V_GLOBAL	fileOpenWrite(char *name);
	file_t		V_GLOBAL	fileOpenAppend(char *name);
	void		V_GLOBAL	fileClose(file_t file);

	vboolean	V_GLOBAL	fileEof(file_t file);
	int			V_GLOBAL	fileSeek(file_t file,int offset,int whence);
	int			V_GLOBAL	fileTell(file_t file);
	int			V_GLOBAL	fileSize(file_t file);

	int			V_GLOBAL	fileRead(file_t file,void *buf,int len);
	char_t		V_GLOBAL	fileGetc(file_t file);
	int			V_GLOBAL	fileGets(file_t file,char *buffer,int maxlen);
	/*	Loads entire file into malloc()'ed buffer. */
	void_p		V_GLOBAL	fileLoad(char *name,int *len);
	/*	Reads little-endian word */
	u16_t		V_GLOBAL	fileGetLittleShort(file_t f);
	/*	Reads little-endian dword */
	u32_t		V_GLOBAL	fileGetLittleLong(file_t f);

	int			V_GLOBAL	fileWrite(file_t file,void *buf,int len);
	vboolean	V_GLOBAL	filePutc(file_t file,int c);
	/* NOTE: Writes system-dependent EOL after the string. */
	vboolean	V_GLOBAL	filePuts(file_t file,char *str);
	/* End-of-line sequence NOT APPENDED */
	vboolean	V_GLOBAL_VA	filePrintf(file_t file,const char *fmt,...);
	/* Saves entire buffer. Buffer remains unchanged. */
	vboolean	V_GLOBAL	fileSave(char *name,void *buf,int len);
	/*	Writes little-endian word */
	void		V_GLOBAL	filePutLittleShort(file_t f,u16_t w);
	/*	Writes little-endian dword */
	void		V_GLOBAL	filePutLittleLong(file_t f,u32_t dw);
#ifdef	__cplusplus
}

#endif	/* __cplusplus */

#endif	/* V_FS_H */
/*</pre></body></html>*/
/*EOF*/
