/*<html><body bgcolor=#B4C4B0><font color=#000000><strong><pre>*/
#ifndef V_TYPES_H
#define V_TYPES_H

/*
 *	Limits.
 */
#if     !defined(MAXCHAR)
#define	MAXCHAR							0x7F
#define	MINCHAR							0x80
#endif  /* MAXCHAR */
#if     !defined(MAXSHORT)
#define	MAXSHORT						0x7FFF
#define	MINSHORT						0x8000
#endif  /* MAXSHORT */
#if     !defined(MAXLONG)
#define MAXLONG							0x7FFFFFFF
#define MINLONG							0x80000000
#endif  /* MAXLONG */
#if     !defined(MAXLONGLONG) && !defined(INT64BROKEN)
#define MAXLONGLONG						0x7FFFFFFFFFFFFFFF
#define MINLONGLONG						0x8000000000000000
#endif  /* MAXLONGLONG */
#if     !defined(MAXFLOAT)
#define MAXFLOAT						3.4e+38
#endif  /* MAXFLOAT */
#if     !defined(MAXDOUBLE)
#define MAXDOUBLE						1.7e+308
#endif  /* MAXDOUBLE */

#ifndef	DOUBLEVEC
#define	MAXVEC							((vec_t)MAXFLOAT)
#else
#define	MAXVEC							((vec_t)MAXDOUBLE)
#endif	/* DOUBLEVEC */

/*
 *	Upper limit for estimated fp round-off error (and for square of error)
 */
#ifndef	FLOAT_ZERO_EPSILON
#define	FLOAT_ZERO_EPSILON				1e-6
#endif	/* FLOAT_ZERO_EPSILON */
#ifndef	FLOAT_ZERO_EPSILON_SQUARE
#define	FLOAT_ZERO_EPSILON_SQUARE		1e-12
#endif	/* FLOAT_ZERO_EPSILON_SQUARE */

/*	Base types. Byte lenght of such types is guaranted.
 */
typedef unsigned char			u8_t,	*u8_p;
typedef signed char				i8_t,	*i8_p;
typedef unsigned short			u16_t,	*u16_p;
typedef signed short			i16_t,	*i16_p;
typedef unsigned long			u32_t,	*u32_p;
typedef signed long				i32_t,	*i32_p;
#ifndef	INT64BROKEN
#if		!defined(__UNIX__) && !defined(__CYGWIN32__)
typedef unsigned __int64		u64_t,	*u64_p;
typedef signed __int64			i64_t,	*i64_p;
#else
typedef unsigned long long int	u64_t,	*u64_p;
typedef long long int			i64_t,	*i64_p;
#endif	/* __UNIX__ */
#endif	/* INT64BROKEN */

/*	Pointers: void and string.
 */
typedef		void				void_t,	*void_p, **void_pp;
typedef		char				char_t, *char_p, **char_pp;

/*	Another one boolean type :)
 */
#if			!defined(__cplusplus) && defined(__WATCOMC__)
typedef		enum				{ false, true }	vboolean;
#else		/* !defined(__cplusplus) && defined(__WATCOMC__) */
typedef 	int					vboolean;
#define 	false				0
#define 	true				1
#endif		/* !defined(__cplusplus) && defined(__WATCOMC__) */

/*	Base floating-point scalar type.
 *	Names of type and macro stolen from Quake.
 */ 
#ifdef		DOUBLEVEC
typedef		double				vec_t, *vec_p;
#else		/* !DOUBLEVEC */
typedef		float				vec_t, *vec_p;
#endif		/* DOUBLEVEC  */

/*
 *	NOTE: config macro MATRIX_DOUBLE controls precision of floatin-point
 *		any-order vector and matrix calculations. It is undependent
 * 		from DOUBLEVEC setting because such matrices are often used in 
 *		scientific calculations, which requires extra precision.
 */

#ifdef	MATRIX_DOUBLE
typedef	double					mvec_t;
#else
typedef	float					mvec_t;
#endif

typedef	mvec_t					*mvec_p;

/*
 *	Vector and matrix types (low-order).
 *	Values in matrices stored in row-major order (unlike in OpenGL)
 */
typedef vec_t					vec2_t[2];
typedef vec_t					vec3_t[3];
typedef vec_t					vec4_t[4];
typedef vec_t					matrix2_t[2][2];
typedef vec_t					matrix3_t[3][3];
typedef vec_t					matrix4_t[4][4];
typedef vec_t					matrix3x4_t[3][4];

/*
 *	RGB(A) colors.
 */
typedef u8_t					color3b_t[3];
typedef u8_t					color4b_t[4];

/*
 *	Handy macros for making powers of 2.
 *	TODO: move them to other file.
 */

/*		Returns 1 if value is power of 2; otherwise returns 0 */
#define IS_POWER_TWO(x)			(((x) & ((x)-1)) ? 0 : 1)

/*		Sets retval to power of 2 which is not less than value */
/*		TODO: are there better ways to do it ? */
#define	NOT_LESS_POWER_TWO(retval,value,type)\
		{\
			type mask;\
			if((value)==(type)0) {\
				retval=0;\
			} else {\
				mask=(type)1 << ((sizeof(type)<<3)-1);\
				while(!(mask & (value)))\
					mask>>=1;\
				retval = (mask < (value)) ? mask << 1 : mask;\
			}\
		}

/*		Sets retval to power of 2 which is greater than value */
/*		TODO: are there better ways to do it ? */
#define	GREATER_POWER_TWO(retval,value,type)\
		{\
			type mask;\
			if((value)==(type)0) {\
				retval=1;\
			} else {\
				mask=(type)1 << ((sizeof(type)<<3)-1);\
				while(!(mask & (value)))\
					mask>>=1;\
				retval = (mask <= (value)) ? mask << 1 : mask;\
			}\
		}

#endif	/* V_TYPES_H */
/*</pre></strong></body></html>*/
/*EOF*/
