#include <v_lib.h>

void
dump_tl(FILE *out,vec3_t v,color4b_t c,vec2_t uv)
{
	fprintf(out,
		"[%3.2f,%3.2f,%3.2f]-(%d,%d,%d,%d)-{%3.2f,%3.2f}",
		v[0],v[1],v[2],
		c[0],c[1],c[2],c[4],
		uv[0],uv[1]);
}

void	V_CALL 
point(void *dev,vec3_t v,color4b_t c,vec2_t uv)
{
	fprintf((FILE*)dev,"POINT\t");
	dump_tl((FILE*)dev,v,c,uv);
	fprintf((FILE*)dev,"\n");
}

void	V_CALL
line(void *dev,vec3_t v1,color4b_t c1,vec2_t uv1,
	vec3_t v2,color4b_t c2,vec2_t uv2)
{
	fprintf((FILE*)dev,"LINE\t");
	dump_tl((FILE*)dev,v1,c1,uv1);
	fprintf((FILE*)dev," | ");
	dump_tl((FILE*)dev,v2,c2,uv2);
	fprintf((FILE*)dev,"\n");
}

void	V_CALL
begin_polygon(void *dev,int n)
{
	fprintf((FILE*)dev,"POLYGON: %d vertices:\n",n);
}

void	V_CALL
polygon_vertex(void *dev,vec3_t v,color4b_t c,vec2_t uv)
{
	fprintf((FILE*)dev,"VERTEX: ");
	dump_tl((FILE*)dev,v,c,uv);
	fprintf((FILE*)dev,"\n");
}

void	V_CALL
pass_through(void *dev,vec_t v)
{
	fprintf((FILE*)dev,"Pass through: '%3.2f'\n",v);
}

gl_tl_callbacks_t	debug_cbs=
{
	point,
	line,
	begin_polygon,
	polygon_vertex,
	pass_through
};

#define	WIDTH		10
#define	HEIGHT		10

void draw(int mode)
{
	u8_t image[16]=
	{
		0x80,0x90,0xff,0xff,0xff,0xe0,0xd0,0xc0,
		0xc0,0xd0,0xe0,0xff,0xff,0xff,0x90,0x80
	};

	glClearColor(.5f,.5f,.5f,0);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0,1,1,0,-1,1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glClear(GL_COLOR_BUFFER_BIT);
	
	glTexImage1D(GL_TEXTURE_1D,0,1,16,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,
		image);
	glTexParameterf(GL_TEXTURE_1D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameterf(GL_TEXTURE_1D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexEnvf(GL_TEXTURE_1D,GL_TEXTURE_ENV_MODE,GL_DECAL);
	glTexEnvf(GL_TEXTURE_1D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	glTexEnvf(GL_TEXTURE_1D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	/*
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE,GL_ONE);
	*/
	glEnable(GL_TEXTURE_1D);
	
	glBegin(mode);
		glColor3f(1,1,0);
		glTexCoord2f(0,0);
		glVertex3f(-0.2f,	.5f,	1.0f);
		
		glColor3f(0,1,1);
		glTexCoord2f(2,0);
		glVertex3f(.9f, .1f, 0);
		
		glColor3f(1,0,1);
		glTexCoord2f(2,0);
		glVertex3f(.9f,.9f, 0);
	glEnd();

	vglSwapBuffers();
}

void main()
{
	GLfloat *fb;
	u32_t num;
	u8_t *ptr;
	FILE *out;
	
	out=fopen("out.txt","wb");
	
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	
	vglCreateOffscreenWindow(WIDTH,HEIGHT);

	draw(GL_TRIANGLES);
	ptr=vglReadPixels(NULL,NULL);
	tgaSave("out.tga",WIDTH,HEIGHT,ptr);
	free(ptr);
	
	fb=vglBeginFeedback(1000);

	glPassThrough(1);
	
	draw(GL_TRIANGLES);
	vglSwapBuffers();
	vglEndFeedback(&num);
	vglDecodeFeedback(&debug_cbs,out,fb,num);
	free(fb);

	fb=vglBeginFeedback(1000);

	glPassThrough(2);

	draw(GL_LINE_LOOP);
	vglSwapBuffers();
	vglEndFeedback(&num);
	vglDecodeFeedback(&debug_cbs,out,fb,num);
	free(fb);

	vglDestroyWindow(vglGetCurrentWindow());
	fclose(out);
}
