#include <windows.h>
#include <gl/gl.h>
#include <gl/glaux.h>
#define SGI_GL_H
#include <v_lib.h>

vec_t al,bet;
vec_t dist=-3;
vec_t scale=1;
tripatch_t *pat=NULL;
patchprofile_t *pp=NULL;

int   pts=1;
int   drawpatch = 1, drawtri = 1, drawlines = 1;

void CALLBACK up_proc() {  al-=5.0; }
void CALLBACK down_proc() {  al+=5.0; }
void CALLBACK left_proc() {  bet-=5.0; }
void CALLBACK right_proc() {  bet+=5.0; }
void CALLBACK plus_proc() {  dist+=(vec_t)0.1; }
void CALLBACK minus_proc() {  dist-=(vec_t)0.1; }
void CALLBACK a_proc() {  scale*=(vec_t)1.1; }
void CALLBACK z_proc() {  scale*=(vec_t)(1/1.1); }
void CALLBACK proc1() {  drawpatch = !drawpatch; }
void CALLBACK proc2() {  drawtri=!drawtri; }
void CALLBACK proc3() {  drawlines=!drawlines; }
void CALLBACK proc4() {  pts=!pts; }

int fc=0;
int cbase;

void uuDrawWireChains(wirechains_p mesh,vec_t val)
{
  int i,j;
  vec3_t  *vlist;
  char label[64];
  if(!mesh) return;
  vlist=mesh->vlist;
  for(i=0;i<mesh->numchains;i++)
  {
    glBegin(GL_LINE_STRIP);
    glColor3f(1,1,1);
    for(j=0;j<mesh->numvert[i];j++) {
      glVertex3fv(vlist[mesh->clist[i][j]]);
    }
    glEnd();
    if(pts && cbase) {
      glRasterPos3fv(mesh->llist[i].center);
      sprintf(label,"%1.3g",val);
      glListBase(cbase);
      glCallLists(strlen(label),GL_UNSIGNED_BYTE,label);
    }
  }
}

static void uuDrawPatchprofile(patchprofile_t *iso)
{
  int i;
  if(!iso) return;
  for(i=0;i<iso->splits;i++)
  {
    if(iso->wires[i])
      uuDrawWireChains(iso->wires[i],iso->chainvals[i]);
  }
}

static GLvoid CALLBACK  drawScene(GLvoid)
{
  int i;
  
  if(!pat) return;
  
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  
  glPushMatrix();
  glTranslatef(0,0,dist);
  glScalef(scale,scale,scale);
  glRotatef(al,1,0,0);
  glRotatef(bet,0,1,0);
  
  glDisable(GL_BLEND);
  glBlendFunc(GL_ONE,GL_SRC_COLOR);

  glEnable(GL_DEPTH_TEST);

  if(drawpatch) {
  
  glBegin(GL_TRIANGLES);
    for(i=0;i<pat->numtri;i++)
    {
      glColor3f((float).7,(float).3,(float).2);
      glVertex3fv(pat->vlist[pat->tlist[i][0]]);
      glColor3f((float).2,(float).8,(float).2);
      glVertex3fv(pat->vlist[pat->tlist[i][1]]);
      glColor3f((float).2,(float).3,(float).7);
      glVertex3fv(pat->vlist[pat->tlist[i][2]]);
    }
  glEnd();
  
  }
  
  glDisable(GL_DEPTH_TEST);

  if(drawlines) uuDrawPatchprofile(pp);
    
    /*
  if(drawtri) {
  
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_BLEND);
  
  glBegin(GL_TRIANGLES);
    glColor3f(1,0.2,0.2);
    glVertex3fv(v[0]);
    glColor3f(0.2,0.2,1);
    glVertex3fv(v[1]);
    glColor3f(0.2,1,0.2);
    glVertex3fv(v[2]);
  glEnd();
  glDisable(GL_BLEND);
  }
    */
    
  glPopMatrix();
  auxSwapBuffers();
}

static GLvoid CALLBACK Resize(GLsizei width,GLsizei height)
{
  GLfloat aspect;
  glViewport(0,0,width,height );
  aspect = (GLfloat)width/height;
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-aspect,aspect,1,-1,1,100);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}    



mvec_t V_CALL rot_para(mvec_t x,mvec_t y)
{
  // return .5*x*x+y*y - x*y/2 + .5*(x*x+y*y)*sin(sqrt(x*x+y*y)*5);
  // return x*x+y*y+2*x*x*x-x*y*y;
  return - x*x - y*y - 2*x*x*x + x*y*y + 1;
}

mvec_t	V_CALL	sheffe(mvec_t x,mvec_t y,mvec_t z)
{
  return 20 * x*y*z;
}


static int vtoi(vec_t arg)
{
  return (int)(1000 * arg);
}

int main(int argc, char *argv[])
{
  vec3_t center={0,0,0};
  logStartConsoleLog();
  /*
  logUpperLevel(LOG_TRACE);
  */
  
  if(argc>1 && !strcmp(argv[1],"-gibbs"))
  {
    pat=ppGridSheffePolynom(center,1,0,20,NULL,NULL,sheffe,0);
    pp=ppCreateTriangular(center,1,0,NULL,sheffe,5,0);
  }
  else
  {
    pat=ppGridFunction(-1,-1,1,1,20,20,NULL,rot_para,1);
    pp=ppCreateRectangular(-1,-1,1,1,rot_para,10,1);
  }
  
  auxInitPosition(0,0,640,480);
  auxInitDisplayMode(AUX_RGB|AUX_DEPTH|AUX_DOUBLE);
  auxInitWindow("test");
  SetClassLong(GetActiveWindow(),GCL_HBRBACKGROUND,0);
  auxReshapeFunc(Resize);
  auxKeyFunc(AUX_UP,(AUXKEYPROC)up_proc);
  auxKeyFunc(AUX_DOWN,(AUXKEYPROC)down_proc);
  auxKeyFunc(AUX_LEFT,(AUXKEYPROC)left_proc);
  auxKeyFunc(AUX_RIGHT,(AUXKEYPROC)right_proc);
  auxKeyFunc(AUX_a,(AUXKEYPROC)a_proc);
  auxKeyFunc(AUX_z,(AUXKEYPROC)z_proc);
  auxKeyFunc(AUX_s,(AUXKEYPROC)plus_proc);
  auxKeyFunc(AUX_x,(AUXKEYPROC)minus_proc);
  auxKeyFunc(AUX_1,(AUXKEYPROC)proc1);
  auxKeyFunc(AUX_2,(AUXKEYPROC)proc2);
  auxKeyFunc(AUX_3,(AUXKEYPROC)proc3);
  auxKeyFunc(AUX_4,(AUXKEYPROC)proc4);
  glClearColor(0.0,0.2,0.2,0.0);
  glClearDepth(1);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  Resize(640,480);
  cbase=glGenLists(256);
  wglUseFontBitmaps(wglGetCurrentDC(),0,256,cbase);
  puts("Use arrows and a,z,s,x to control. 1,3 toggles elements.");
  auxMainLoop(drawScene);
  glDeleteLists(cbase,256);
  patFree(pat);
  ppFree(pp);
  return 0;
}
