#include <v_lib.h>

#define		NUM_SUBDIVISIONS		10000

int main()
{
	u32_t i;
	vector_t *x,*y,*smoothed;
	cs_interpolator_t sp;
	timer_t t;
	mvec_t mult;
	input_record_t in_rec;
	
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	
	/* create input data */
	x=vecAlloc(10);
	y=vecAlloc(x->size);
	smoothed=vecAlloc(NUM_SUBDIVISIONS);
	
	for(i=0;i<x->size;i++)
	{
		x->elements[i]=(mvec_t)i;
		y->elements[i]=(9.0f * (rand() % 1024)) / 1024;
	}
	sp=csiCreate(x,y);
	
	/* initialize OpenGL */
	vglCreateWindow("Spline test (press ESC to exit)",
		0,100,512,384);
	glClearColor(0,0.3f,0.3f,0.5f);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0,9,0,9,-1,1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClear(GL_COLOR_BUFFER_BIT);
	
	/* draw input data */
	glColor4f(1,1,1,0.5f);
	glPointSize(9);
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glBegin(GL_POINTS);
		for(i=0;i<x->size;i++)
			glVertex2f(x->elements[i],y->elements[i]);
	glEnd(); /* GL_POINTS */
	glDisable(GL_POINT_SMOOTH);

	/* compute curve and smooth input data */
	mult=10.0f / smoothed->size;
	timerStart(&t);
	for(i=0;i<smoothed->size;i++)
		smoothed->elements[i]=csiInterpolate(sp,(mvec_t)i * mult);
	printf("%d points, %d milliseconds.\n",smoothed->size,timerMilliseconds(&t));
	timerStop(&t);

	/* draw smoothed data */
	glColor4f(1,1,1,0.5f);
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(2.5f);
	glBegin(GL_LINE_STRIP);
		for(i=0;i<smoothed->size;i++)
			glVertex2f((mvec_t)i * mult,smoothed->elements[i]);
	glEnd(); /* GL_LINE_STRIP */
	glDisable(GL_LINE_SMOOTH);
	glDisable(GL_BLEND);
	vglSwapBuffers();

	for(;;)
	{
		vglNextEvent(&in_rec);
		if(in_rec.first_code==key_escape)
			break;
	}

	vglDestroyWindow(vglGetCurrentWindow());
	free(x);
	free(y);
	free(smoothed);
	free(sp);
	return 0;
}
