#include <v_lib.h>

#define	NUM_THREADS		5

u32_t	key;

void	print_specific(void)
{
	printf("Thread specific is %d\n",
		(unsigned)threadGetSpecific(key));
}

void V_CALL
thread_proc(void *arg)
{
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	printf("Arg: %X\n",(unsigned)arg);
	threadSetSpecific(key,(void*)arg);
	print_specific();
	timeDelay((u32_t)1000);
	printf("Exit: %X\n",(unsigned)arg);
	threadExit((u32_t)arg);
}

mutex_t *lock;

void V_CALL
mutex_thread_proc(void *arg)
{
	vboolean already_locked;
	already_locked=!mutexTryLock(lock);
	printf("Thread %d: mutex %s, locking for 1 second\n",
		(unsigned)arg,already_locked ? "already locked" : "free");
	if(already_locked) /* seems to be not intuitive ... */
		mutexLock(lock);
	printf("Thread %d: mutex aquired\n",(unsigned)arg);
	timeDelay(1000);
	printf("Thread %d: unlocking mutex\n",(unsigned)arg);
	mutexUnlock(lock);
}

int main()
{
	int i;
	thread_t threads[NUM_THREADS];
	u32_t ret_vals[NUM_THREADS];
	/* libvPerThreadInit(); */
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
    key=threadCreateKey();
	if(!key)
	{
		printf("threadCreateKey() failed !\n");
		return -1;
	}
	puts("Setting specific to 100");
	threadSetSpecific(key,(void*)100);
	print_specific();
	lock=mutexCreate();
	printf("\nThread demo: spawning %d threads...\n\n",NUM_THREADS);
	for(i=0;i<NUM_THREADS;i++)
	{
		threads[i]=threadCreate(thread_proc,(void*)i);
		timeDelay(100);
	}
	print_specific();
	puts("Setting specific to 101");
	threadSetSpecific(key,(void*)101);
	print_specific();
	for(i=0;i<NUM_THREADS;i++)
		threadJoin(threads[i],&ret_vals[i]);
	for(i=0;i<NUM_THREADS;i++)
	{
		printf("Thread %x exited with code %lu\n",
			(unsigned)threads[i],ret_vals[i]);
	}
	printf("\nMutex demo: spawning %d threads...\n\n",NUM_THREADS);
	for(i=0;i<NUM_THREADS;i++)
	{
		threads[i]=threadCreate(mutex_thread_proc,(void*)i);
		timeDelay(10);
	}
	for(i=0;i<NUM_THREADS;i++)
		threadJoin(threads[i],NULL);
	mutexDestroy(lock);
	print_specific();
	threadDestroyKey(key);
	return 0;
}
