/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../include/v_lib.h"
#include	"../source/v_mt.h"

/*	=======================================================================
**	Helper API.
*/

btree_p	V_LOCAL
alloc_btree_node(btree_t *parent,int child_index,size_t extra)
{
	btree_t *n;
	if(parent && parent->children[child_index])
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"alloc_btree_node(): %s child already present.\n",
			child_index ? "right" : "left");
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	n=(btree_t*)memAlloc(sizeof(btree_t)+extra);
	n->parent=parent;
	n->children[0]=n->children[1]=NULL;
	if(parent)
		parent->children[child_index]=n;
	return n;
}

void	V_LOCAL
btTraverseFromLeft_r(btree_t *root,void (V_CALL *proc)(void*))
{
	if(!root)
		return;
	proc(root->data);
	btTraverseFromLeft_r(root->children[0],proc);
	btTraverseFromLeft_r(root->children[1],proc);
}

void	V_LOCAL
btTraverseFromRight_r(btree_t *root,void (V_CALL *proc)(void*))
{
	if(!root)
		return;
	proc(root->data);
	btTraverseFromRight_r(root->children[1],proc);
	btTraverseFromRight_r(root->children[0],proc);
}

void	V_LOCAL
btTraverseFromLeftRF_r(btree_t *root,void (V_CALL *proc)(void*))
{
	if(!root)
		return;
	if(root->children[0])
		proc(root->children[0]->data);
	if(root->children[1])
		proc(root->children[1]->data);
	btTraverseFromLeftRF_r(root->children[0],proc);
	btTraverseFromLeftRF_r(root->children[1],proc);
}

void	V_LOCAL
btTraverseFromRightRF_r(btree_t *root,void (V_CALL *proc)(void*))
{
	if(!root)
		return;
	if(root->children[1])
		proc(root->children[1]->data);
	if(root->children[0])
		proc(root->children[0]->data);
	btTraverseFromRightRF_r(root->children[1],proc);
	btTraverseFromRightRF_r(root->children[0],proc);
}

void	V_LOCAL
btDeleteAllNodesFast_r(btree_t *root)
{
	if(!root)
		return;
	btDeleteAllNodesFast_r(root->children[0]);
	btDeleteAllNodesFast_r(root->children[1]);
	memFree(root);
}

void	V_LOCAL
btDeleteAllNodes_r(btree_t *root,void (V_CALL *destructor)(void*))
{
	if(!root)
		return;
	btDeleteAllNodes_r(root->children[0],destructor);
	btDeleteAllNodes_r(root->children[1],destructor);
	destructor(root->data);
	memFree(root);
}

int	V_LOCAL
#ifdef	CHECKED_BUILD
check_traverse_args(char *caller,btree_t *node,void (V_CALL *proc)(void*))
{
	if(!node)
	{
		conlPrintf(LOG_ERROR,"%s(): nothing to traverse.\n",
			caller);
		return 0;
	}
	if(!proc)
	{
		conlPrintf(LOG_ERROR,"%s(): nothing to do with nodes.\n",
			caller);
		return 0;
	}
	return 1;
}
#else	/* !CHECKED_BUILD */
check_traverse_args_q(btree_t *node,void (V_CALL *proc)(void*))
{
	return (!node || !proc) ? 0 : 1;
}
#define	check_traverse_args(caller,node,proc)	\
		check_traverse_args_q(node,proc)
#endif

/*	=======================================================================
**	Global API.
*/

btree_p	V_GLOBAL
btAllocNode(void *data,btree_t *parent,int child_index)
{
	btree_t *n=alloc_btree_node(parent,child_index,0);
	if(n)
		n->data=data;
	return n;
}

btree_p	V_GLOBAL
btStore(void *data,size_t len,btree_t *parent,int child_index)
{
	btree_t *n;
	if(!data || !len)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"btStore(): nothing to store.\n");
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	n=alloc_btree_node(parent,child_index,len);
	if(n)
	{
		n->data=(void*)(n+1);
		memCopy(n->data,data,len);
	}
	return n;
}

btree_p	V_GLOBAL
btRootNode(btree_t *node)
{
	if(!node)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"btRootNode(): couldn't ascend from NULL.\n");
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	while(node->parent)
		node=node->parent;
	return node;
}

/*
**	Traverses binary tree in order:
**		root
**			root->left
**				root->left->left
**				... etc
**			root->right
**				root->right->left
**					root->right->left->left
**					... etc
*/
void	V_GLOBAL
btTraverseLeft(btree_t *root,void (V_CALL *handler)(void*))
{
	if(!check_traverse_args("btTraverseLeft",root,handler))
		return;
	btTraverseFromLeft_r(root,handler);
}

/*
**	Traverses binary tree in order:
**		root
**			root->left
**			root->right
**				root->left->left
**				root->left->right
**				root->right->left
**				root->right->right
**					... etc
*/
void	V_GLOBAL
btTraverseLeftRF(btree_t *root,void (V_CALL *handler)(void*))
{
	if(!check_traverse_args("btTraverseLeftRF",root,handler))
		return;
	handler(root->data);
	btTraverseFromLeftRF_r(root,handler);
}

/*
**	Traverses binary tree in order:
**		root
**			root->right
**				root->right->right
**				... etc
**			root->left
**				root->left->right
**				... etc
*/
void	V_GLOBAL
btTraverseRight(btree_t *root,void (V_CALL *handler)(void*))
{
	if(!check_traverse_args("btTraverseRight",root,handler))
		return;
	btTraverseFromRight_r(root,handler);
}

/*
**	Traverses binary tree in order:
**		root
**			root->right
**			root->left
**				root->right->right
**				root->right->left
**				root->left->right
**				root->left->left
**					... etc
*/
void	V_GLOBAL
btTraverseRightRF(btree_t *root,void (V_CALL *handler)(void*))
{
	if(!check_traverse_args("btTraverseRightRF",root,handler))
		return;
	handler(root->data);
	btTraverseFromRightRF_r(root,handler);
}

/*
**	Removes all nodes from leftmost to rightmost
*/
void	V_GLOBAL
btDeleteAllNodes(btree_t *root,void (V_CALL *destructor)(void*))
{
	if(!root)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"btDeleteAllNodes(): nothing to delete.");
#endif	/* CHECKED_BUILD */
		return;
	}
	if(destructor)
	{
		btDeleteAllNodes_r(root->children[0],destructor);
		btDeleteAllNodes_r(root->children[1],destructor);
		destructor(root->data);
	}
	else
	{
		btDeleteAllNodesFast_r(root->children[0]);
		btDeleteAllNodesFast_r(root->children[1]);
	}
	memFree(root);
}

/*</pre></strong></body></html>*/
/*EOF*/
