/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../include/v_lib.h"
#include	"../source/v_mt.h"

/*	Implementation of u8_fifo_t
 */
void	V_GLOBAL
u8fInit(u8_fifo_t *fifo)
{
	V_ASSERT(fifo);
	v_zero(fifo,sizeof(u8_fifo_t));
}

void	V_GLOBAL
u8fClear(u8_fifo_t *fifo)
{
	V_ASSERT(fifo);
	if(fifo->buf)
		free(fifo->buf);
	v_zero(fifo,sizeof(u8_fifo_t));
}

/*
 *	not_todo: don't convert to use realloc().
 */
 
void	V_LOCAL
u8fGrowBuffer(u8_fifo_t *fifo,unsigned next_minsize)
{
	u8_t *nextbuf;
	unsigned nextlen,len,headlen,taillen;
	/* How many elements in fifo */
	len = fifo->size-fifo->avail;
	/* How many elements between tail end end of buffer */
	headlen = fifo->size-fifo->tail;
	/* How many elements must be copied from tail (first block) */
	if(headlen>len)
		headlen=len;
	/* How many elements must be copied from start (second block) */
	taillen=len-headlen;
	NOT_LESS_POWER_TWO(nextlen,next_minsize,unsigned);
	nextbuf=(u8_t*)malloc(nextlen*sizeof(u8_t));
	/* Copy head bytes */
	if(headlen)
		v_copy(nextbuf,fifo->buf+fifo->tail,headlen*sizeof(u8_t));
	/* Copy tail bytes */
	if(taillen)
		v_copy(nextbuf+headlen,fifo->buf,taillen*sizeof(u8_t));
	fifo->head=headlen+taillen;
	fifo->tail=0;
	if(fifo->buf)
		free(fifo->buf);
	fifo->buf=nextbuf;
	fifo->size=nextlen;
	fifo->avail=fifo->size-fifo->head;
}

void	V_GLOBAL
u8fPurge(u8_fifo_t *fifo)
{
	V_ASSERT(fifo);
	fifo->head=fifo->tail=0;
	fifo->avail=fifo->size;
}

void	V_GLOBAL
u8fPush(u8_fifo_t *fifo,u8_t element)
{
	V_ASSERT(fifo);
	if(fifo->avail == 0)
		/* Allocate new storage */
		u8fGrowBuffer(fifo,fifo->size+1);
	*(fifo->buf + fifo->head) = element;
	fifo->head++;
	if(fifo->head == fifo->size)
		fifo->head = 0;
	fifo->avail--;
}

void	V_GLOBAL
u8fPushBuffer(u8_fifo_t *fifo,u8_t *src,unsigned num_elements)
{
	V_ASSERT(fifo && src && num_elements);
	if(fifo->avail<num_elements)
	{
		/* Allocate new storage */
		u8fGrowBuffer(fifo,fifo->size+num_elements);
		/* fifo->tail granted to be 0, so we can issue only one copy */
		v_copy(fifo->buf+fifo->head,src,num_elements*sizeof(u8_t));
		fifo->head+=num_elements;
	}
	else
	{
		/* In common case we must perform two copy operations */
		unsigned headlen, taillen;
		/* How many elements feet between head and end of buffer */
		headlen = fifo->size-fifo->head;
		/* How many elements must be copied at head (first block) */
		if(headlen>num_elements)
			headlen=num_elements;
		/* How many elements must be copied at start (second block) */
		taillen=num_elements-headlen;
		if(headlen)
			v_copy(fifo->buf+fifo->head,src,headlen);
		if(taillen)
		{
			v_copy(fifo->buf,src+headlen,taillen);
			fifo->head=taillen;
		}
		else
		{
			fifo->head+=num_elements;
			if(fifo->head==fifo->size)
				fifo->head=0;
		}
	}
	fifo->avail-=num_elements;
}

int		V_GLOBAL
u8fPop(u8_fifo_t *fifo,u8_t *element)
{
	V_ASSERT(fifo && element);
	if(fifo->avail == fifo->size)
		return 0;
	*element = *(fifo->buf + fifo->tail);
	fifo->tail++;
	if(fifo->tail == fifo->size)
		fifo->tail = 0;
	fifo->avail++;
	return 1;
}

/*	Returns number of copied bytes */
#if 0
int		V_GLOBAL
u8fPopBuffer_(u8_fifo_t *fifo,u8_t *dst,unsigned num_elements)
{
	unsigned len,headlen,taillen;
	V_ASSERT(dst && num_elements);
	len=fifo->size-fifo->avail;
	if(num_elements>len)
		num_elements=len;
	/* How many elements between tail end end of buffer */
	headlen = fifo->size-fifo->tail;
	/* How many elements must be copied from tail (first block) */
	if(headlen>num_elements)
		headlen=num_elements;
	/* How many elements must be copied from start (second block) */
	taillen=num_elements-headlen;
	/* Copy head bytes */
	if(headlen)
		v_copy(dst,fifo->buf+fifo->tail,headlen/* *sizeof(u8_t)*/);
	/* Copy tail bytes */
	if(taillen)
	{
		v_copy(dst+headlen,fifo->buf,taillen/* *sizeof(u8_t)*/);
		fifo->tail=taillen;
	}
	else
	{
		fifo->tail+=num_elements;
		if(fifo->tail==fifo->size)
			fifo->tail=0;
	}
	fifo->avail+=num_elements;
	return num_elements;
}
#endif

/*	Returns number of copied bytes */
int		V_GLOBAL
u8fPopBuffer(u8_fifo_t *fifo,u8_t *dst,unsigned num_elements)
{
	unsigned len,headlen,taillen;
	V_ASSERT(dst && num_elements);
	len=fifo->size-fifo->avail;
	if(num_elements>len)
		num_elements=len;
	/* How many elements between tail end end of buffer */
	headlen = fifo->size-fifo->tail;
	/* How many elements must be copied from tail (first block) */
	if(headlen>=num_elements)
	{
		headlen=num_elements;	/* taillen = 0 */
		v_copy(dst,fifo->buf+fifo->tail,headlen/* *sizeof(u8_t)*/);
		fifo->tail+=num_elements;
		if(fifo->tail==fifo->size)
			fifo->tail=0;
	}
	else
	{
		taillen=num_elements-headlen;
		v_copy(dst,fifo->buf+fifo->tail,headlen*sizeof(u8_t));
		v_copy(dst+headlen,fifo->buf,taillen*sizeof(u8_t));
		fifo->tail=taillen;
	}
	fifo->avail+=num_elements;
	return num_elements;
}

#ifdef	FIFO_NO_MACRO

unsigned	V_GLOBAL
u8fCount(u8_fifo_t *fifo)
{
	V_ASSERT(fifo);
	return fifo->size-fifo->avail;
}

#endif	/* FIFO_NO_MACRO */

/*	Implementation of fifo_t */

void	V_GLOBAL
fifoInit(fifo_t *fifo,unsigned element_size)
{
	V_ASSERT(fifo && element_size);
	v_zero(fifo,sizeof(fifo_t));
	fifo->element_size=element_size;
}

void	V_GLOBAL
fifoClear(fifo_t *fifo)
{
	V_ASSERT(fifo);
	if(fifo->buf)
		free(fifo->buf);
	fifoInit(fifo,fifo->element_size);
}

void	V_GLOBAL
fifoPush(fifo_t *fifo,void *element)
{
	V_ASSERT(fifo && element);
	if(fifo->avail == 0)
	{
		/* Allocate new storage */
		u8_t *nextbuf;
		unsigned nextlen,len,headlen,taillen;
		len = fifo->size-fifo->avail;
		headlen = fifo->size - (fifo->tail-fifo->buf) / fifo->element_size;
		if(headlen>len)
			headlen=len;
		taillen=len-headlen;
		GREATER_POWER_TWO(nextlen,fifo->size,unsigned);
		nextbuf=(u8_t*)malloc(nextlen*fifo->element_size);
		/* Copy head bytes */
		if(headlen)
			v_copy(nextbuf,fifo->tail,headlen*fifo->element_size);
		/* Copy tail bytes */
		if(taillen)
			v_copy(nextbuf+headlen*fifo->element_size,fifo->buf,
				taillen*fifo->element_size);
		if(fifo->buf)
			free(fifo->buf);
		fifo->buf=nextbuf;
		fifo->limit=fifo->buf + nextlen * fifo->element_size;
		fifo->head=fifo->buf + len * fifo->element_size;
		fifo->tail=fifo->buf;
		fifo->size=nextlen;
		fifo->avail=fifo->size-len;
	}
	v_copy(fifo->head,element,fifo->element_size);
	fifo->head+=fifo->element_size;
	if(fifo->head == fifo->limit)
		fifo->head = fifo->buf;
	fifo->avail--;
}

void	V_GLOBAL
fifoPurge(fifo_t *fifo)
{
	V_ASSERT(fifo);
	fifo->head=fifo->tail=0;
	fifo->avail=fifo->size;
}

int		V_GLOBAL
fifoPop(fifo_t *fifo,void *element)
{
	if(fifo->avail == fifo->size)
		return 0;
	v_copy(element,fifo->tail,fifo->element_size);
	fifo->tail+=fifo->element_size;
	if(fifo->tail == fifo->limit)
		fifo->tail = fifo->buf;
	fifo->avail++;
	return 1;
}

#ifdef	FIFO_NO_MACRO

unsigned	V_GLOBAL
fifoCount(fifo_t *fifo)
{
	return fifo->size-fifo->avail;
}


#endif /* FIFO_NO_MACRO */
/*</pre></strong></body></html>*/
/*EOF*/
