/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../include/v_lib.h"
#include	"../source/v_mt.h"

/*
**	OpenGL feedback stuff
*/

#define FEEDBACK_MODE	GL_3D_COLOR_TEXTURE

/*
 *	Puts OpenGL in feedback mode and returns pointer to
 *	libvMalloc()'ed feedback buffer.
 */
void_p	V_GLOBAL
vglBeginFeedback(int num_elements)
{
	GLfloat *buffer;
	buffer=(GLfloat*)libvMalloc(num_elements*sizeof(GLfloat));
	v_zero(buffer,num_elements+sizeof(GLfloat));
	glFeedbackBuffer(num_elements,FEEDBACK_MODE,buffer);
	glRenderMode(GL_FEEDBACK);
	return (void_p)buffer;
}

/*
 *	Takes OpenGL out of feedback mode. Returns true if
 *	revious mode was feedback mode.
 */
vboolean	V_GLOBAL
vglEndFeedback(u32_t *stored_elements)
{
	GLint m,stored;
	if(!stored_elements)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglEndFeedback(NULL): incorrect parameter.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	glGetIntegerv(GL_RENDER_MODE,&m);
	if(m != GL_FEEDBACK)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglEndFeedback(): current render mode != GL_FEEDBACK.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	stored=glRenderMode(GL_RENDER);
	if(!stored)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglEndFeedback(): no elements in feedback buffer.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	if(stored<0)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglEndFeedback(): feedback buffer overflowed.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	*stored_elements=(u32_t)stored;
	return true;
}

/*
 *	Decodes RGBA vertex and stores data in 'v', 'c' and 'uv'. Advances
 *	pointer '*start' to the next vertex / token. Returns 'false' if not
 *	enought source data to place into vertex parameters.
 *	Destination parameters ('v', 'c' and 'uv') can be NULL.
 */
vboolean	V_LOCAL
glf_decode_rgba_vertex(GLfloat **start,GLfloat *end,vec3_t v,color4b_t c,
	vec2_t uv)
{
	GLfloat *ptr=*start;
#if		FEEDBACK_MODE==GL_3D_COLOR_TEXTURE
	vec4_t cf;
	if((end-ptr)<11)
	{
		*start=end;
		return false;
	}
	/* Vertex */
	if(v)
	{
		v[0]=(vec_t)ptr[0];
		v[1]=(vec_t)ptr[1];
		v[2]=(vec_t)ptr[2];
	}
	/* Color */
	if(c)
	{
		cf[0]=(vec_t)ptr[3];
		cf[1]=(vec_t)ptr[4];
		cf[2]=(vec_t)ptr[5];
		cf[3]=(vec_t)ptr[6];
		rgbaByteFromFloat(c,cf);
	}
	/* Texture */
	if(uv)
	{
		uv[0]=(vec_t)ptr[7];
		uv[1]=(vec_t)ptr[8];
	}
	*start=ptr+11;
	return true;
#elif	/* FEEDBACK_MODE */
#error	Unknown feedback mode !
#endif	/* FEEDBACK_MODE */
}

/*
 *	Decodes OpenGL feedback buffer.
 *	Input:
 *		back_api	- list of drawing callbacks;
 *		dev			- device handle to pass to back_api's;
 *		buffer		- filled OpenGL feedback buffer;
 *		stored_elemens - number of GLfloat's in feedback buffer.
 *	Output:
 *		None.
 */
void	V_GLOBAL
vglDecodeFeedback(gl_tl_callbacks_t *back_api,void *dev,void_p buffer,
	int stored_elements)
{
	u32_t n;
	vec3_t v, v1;
	vec2_t uv, uv1;
	color4b_t c, c1;
	GLfloat *ptr, *end;
	if(!back_api || !buffer)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglDecodeBuffer(%X,%X,%X,%d): incorrect parameters.\n",
			back_api,dev,buffer,stored_elements);
#endif	/*CHECKED_BUILD */
		return;
	}
#ifdef	CHECKED_BUILD
	if(!stored_elements)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_WARNING,
			"vglDecodeBuffer(%X,%X,%X,%d): empty feedback buffer.\n",
			back_api,dev,buffer,stored_elements);
#endif	/*CHECKED_BUILD */
		return;
	}
	if(!dev)
	{
		conlPrintf(LOG_WARNING,
			"vglDecodeBuffer(%X,%X,%X,%d): NULL device handle.\n",
			back_api,dev,buffer,stored_elements);
	}
#endif	/*CHECKED_BUILD */
	ptr=(GLfloat*)buffer;
	end=ptr+stored_elements;
	while(ptr<end)
	{
		if(IsZero(*ptr-GL_POINT_TOKEN))
		{
			/* Point */
			ptr++;
			if(glf_decode_rgba_vertex(&ptr,end,v,c,uv))
				back_api->point(dev,v,c,uv);
		}
		else if(IsZero(*ptr-GL_LINE_TOKEN) || 
			IsZero(*ptr-GL_LINE_RESET_TOKEN))
		{
			ptr++;
			if(glf_decode_rgba_vertex(&ptr,end,v,c,uv) &
				glf_decode_rgba_vertex(&ptr,end,v1,c1,uv1))
					back_api->line(dev,v,c,uv,v1,c1,uv1);
		}
		else if(IsZero(*ptr-GL_POLYGON_TOKEN))
		{
			/* Polygon */
			if((end-ptr)<2)
			{
				ptr=end;
				break;
			}
			n=(u32_t)ptr[1];
			ptr+=2;
			if(!n)
				break;
			back_api->begin_polygon(dev,n);
			while(n--)
				if(glf_decode_rgba_vertex(&ptr,end,v,c,uv))
					back_api->polygon_vertex(dev,v,c,uv);
		}
		else if(IsZero(*ptr-GL_BITMAP_TOKEN))
		{
			/* Ignore */
			ptr++;
			glf_decode_rgba_vertex(&ptr,end,NULL,NULL,NULL);
		}
		else if(IsZero(*ptr-GL_DRAW_PIXEL_TOKEN))
		{
			/* Ignore */
			ptr++;
			glf_decode_rgba_vertex(&ptr,end,NULL,NULL,NULL);
		}
		else if(IsZero(*ptr-GL_COPY_PIXEL_TOKEN))
		{
			/* Ignore */
			ptr++;
			glf_decode_rgba_vertex(&ptr,end,NULL,NULL,NULL);
		}
		else if(IsZero(*ptr-GL_PASS_THROUGH_TOKEN))
		{
			if((end-ptr)<2)
			{
				ptr=end;
				break;
			}
			back_api->pass_though(dev,ptr[1]);
			ptr+=2;
		}
		else
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_WARNING,
				"vglDecodeBuffer(%X,%X,%X,%d): unknown token '%f'.\n",
				back_api,dev,buffer,stored_elements,*ptr);
#endif		/*CHECKED_BUILD */
			ptr++;
		}
	}
}

/*
 *	Creates metafile context, sets it's modelview and projection matrices
 *	idential to ones in current context, calls DrawProc(param) and closes
 *	metafile.
 *	Feedback buffer size automatically increases if needed.
 */
vboolean	V_GLOBAL
vglDrawToMetafile(char *file_name,u32_t mmWidth,u32_t mmHeight,
	void (V_CALL *DrawProc)(void *param),void *param)
{
	u32_t i,try_limit;
	glw_t cur;
	void *mf;
	vboolean success=0;
	GLfloat projection_matrix[16];
	GLfloat modelview_matrix[16];
	u32_t hint=256 * 1024;
	try_limit = 31 - (8 + 10);		/* have your 2 GB of RAM ? */
	if(!file_name || !file_name[0] || !DrawProc)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"vglDrawToMetafile(%s,%u,%u,%X,%X): "
			"incorrect parameter(s).\n",
			dbs(file_name),mmWidth,mmHeight,DrawProc,param);
#endif	/* CHECKED_BUILD */
		return false;
	}
	cur=vglGetCurrentWindow();
	if(!cur)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglDrawToMetafile(%s,%X,%X): current context is NULL.\n",
			dbs(file_name),DrawProc,param);
#endif	/* CHECKED_BUILD */
		return false;
	}
	glGetFloatv(GL_MODELVIEW_MATRIX,modelview_matrix);
	glGetFloatv(GL_PROJECTION_MATRIX,projection_matrix);
	for(i=0;i<try_limit && !success;i++)
	{
		mf=wmfCreateOffscreenWindow(file_name,mmWidth,mmHeight,hint);
		if(!mf)
		{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_ERROR,"vglDrawToMetafile(%s,%X,%X): "
				"wmfCreateOffscreenWindow() failed.\n",dbs(file_name),
				DrawProc,param);
#endif	/* CHECKED_BUILD */
			if(vglGetCurrentWindow() != cur)
				vglMakeCurrent(cur);
			return false;
		}
		glMatrixMode(GL_PROJECTION);
		glLoadMatrixf(projection_matrix);
		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf(modelview_matrix);
		DrawProc(param);
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_INFO,
			"vglDrawToMetafile(): attempting with FB size %u; \n",hint);
#endif	/* CHECKED_BUILD */
		success=wmfDestroyWindow(mf);
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_INFO,"vglDrawToMetafile(): %s.\n",
			success ? "success" : "FB too small !");
#endif	/* CHECKED_BUILD */
		hint<<=1;
	}
	if(vglGetCurrentWindow() != cur)
		vglMakeCurrent(cur);
	if(i==try_limit)
	{
		/* scene not fit in 2 GB - COOL ! */
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"vglDrawToMetafile(): try_limit reached, "
			"decrease scene complexity.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	return true;
}

/*
 *	Utility to produce screenshots.
 */
vboolean	V_GLOBAL
vglSaveFramebuffer(char *file_name,u32_t mmWidth,u32_t mmHeight)
{
	u8_t *pixels;
	u32_t rw, rh;
	char ext[_MAX_EXT];
	vboolean res;
	if(!file_name || !file_name[0])
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"vglSaveFramebuffer(%s,%u,%u): "
			"incorrect parameter.\n",
			dbs(file_name),mmWidth,mmHeight);
#endif	/* CHECKED_BUILD */
		return false;
	}
	pixels=vglReadPixels(&rw,&rh);
	if(!pixels)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"vglSaveFramebuffer(%s,%u,%u): "
			"vglReadPixels() failed.\n",
			dbs(file_name),mmWidth,mmHeight);
#endif	/* CHECKED_BUILD */
		return false;
	}
	/* determine file format */
	if(!strcasecmp(file_name,"//./clipboard"))
		res=bmpSaveEx(file_name,rw,rh,pixels,mmWidth,mmHeight);
	else
	{
		fsSplitPath(file_name,NULL,NULL,NULL,ext);
		if(!strcasecmp(ext,"bmp"))
			res=bmpSaveEx(file_name,rw,rh,pixels,mmWidth,mmHeight);
		else if(!strcasecmp(ext,"tga"))
		{
#ifdef	CHECKED_BUILD
			if(mmWidth || mmHeight)
				conlPrintf(LOG_WARNING,"vglSaveFramebuffer(%s,%u,%u): "
					"Targa format specified, ignoring image size "
					"%u by %u millimeters.\n",
					dbs(file_name),mmWidth,mmHeight,mmWidth,mmHeight);
#endif	/* CHECKED_BUILD */
			res=tgaSave(file_name,rw,rh,pixels);
		}
		else
		{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_ERROR,"vglSaveFramebuffer(%s,%u,%u): "
				"unknown file format '%s'.\n",
				dbs(file_name),mmWidth,mmHeight,ext);
#endif	/* CHECKED_BUILD */
			res=false;
		}
	}
	return res;
}

/*
 *	Texture stuff
 */

/*
**	Loads image and binds it to new texture object with given parameters.
*/
GLuint	V_GLOBAL
vglLoadTexture(char *file_name,
	u8_p (V_CALL *LoadProc)(char *name,u32_t *width,u32_t *height),
	GLenum filter,GLenum clamp,GLenum env)
{
	u32_t width,height;
	GLuint id;
	u8_t *texels;
	GLenum magf=filter;
	int level;
	if(!file_name || !file_name[0] || !LoadProc)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglLoadTexture('%s',%X): incorrect parameter(s).\n",
			dbs(file_name),LoadProc);
#endif	/* CHECKED_BUILD */
		return 0;
	}
	texels=LoadProc(file_name,&width,&height);
	if(!texels)
		return 0;
	if(!IS_POWER_TWO(width) || !IS_POWER_TWO(height))
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglLoadTexture(): wrong texture dimensions %dX%d.\n",
			width,height);
#endif	/* CHECKED_BUILD */
		libvFree(texels);
		return 0;
	}
	switch(filter)
	{
		case GL_LINEAR_MIPMAP_NEAREST:
		case GL_LINEAR_MIPMAP_LINEAR:
		case GL_LINEAR:
			magf=GL_LINEAR;
			break;
		case GL_NEAREST_MIPMAP_LINEAR:
		case GL_NEAREST_MIPMAP_NEAREST:
		case GL_NEAREST:
			magf=GL_NEAREST;
			break;
		default:
			magf=filter;
			break;
	}
	/* Generate texture and bind it to target */
	glGenTextures(1,&id);
	glBindTexture(GL_TEXTURE_2D,id);
	/* Download image data to renderer */
	level=0;
	if(magf!=filter)	/* generate all except last mipmap levels */
		for(;width>1 && height>1;level++)
		{
			glTexImage2D(GL_TEXTURE_2D,level,GL_RGBA,width,height,0,GL_RGBA,
				GL_UNSIGNED_BYTE,texels);
			texels=(u8_p)imgDownsample(texels,width,height,false);
			width>>=1;
			height>>=1;
		}
	glTexImage2D(GL_TEXTURE_2D,level,GL_RGBA,width,height,0,GL_RGBA,
		GL_UNSIGNED_BYTE,texels);
	libvFree(texels);
	/* Set texture parameters */
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,filter);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,magf);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,clamp);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,clamp);
	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,env);
	/* Clear current texture object */
	glBindTexture(GL_TEXTURE_2D,0);
	return id;
}

vboolean	V_GLOBAL
vglLoadSky(quake_sky_t dest,char *path,char *sky_name,char *file_ext)
{
	int i;
	char suffixes[6][3]={"up","ft","lf","bk","rt","dn"};
	char name[_MAX_FNAME];
	char file_name[_MAX_PATH];
	u8_p (V_CALL *LoadProc)(char *name,u32_t *width,u32_t *height);
	if(!dest || !sky_name || !sky_name[0] || !file_ext || !file_ext[0])
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglLoadSky(%X,'%s','%s','%s'): incorrect parameter(s).\n",
			dest,dbs(path),dbs(sky_name),dbs(file_ext));
#endif	/* CHECKED_BUILD */
		if(dest)
			v_zero(dest,sizeof(quake_sky_t));
		return false;	
	}
	/* Choose appropriate loading function */
	LoadProc=NULL;
	if(!strcmp(file_ext,"tga"))
		LoadProc=tgaLoad;
	else if(!strcmp(file_ext,"jpg"))
		LoadProc=jpgLoad;
	else if(!strcmp(file_ext,"bmp"))
		LoadProc=bmpLoad;
	if(!LoadProc)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglLoadSky(%X,'%s','%s','%s'): unknown image format '%s'\n",
			dest,path,sky_name,file_ext,file_ext);
#endif	/* CHECKED_BUILD */
		v_zero(dest,sizeof(quake_sky_t));
		return false;	
	}
	/* Load images */
	/* TODO: determine presence of HW renderer by ctx->accelerated */
	for(i=0;i<6;i++)
	{
		sprintf(name,"%s%s",sky_name,suffixes[i]);
		fsMakePath(file_name,NULL,path,name,file_ext);
#if	0
		dest[i]=vglLoadTexture(file_name,LoadProc,GL_NEAREST,GL_CLAMP,GL_DECAL);
#else
		dest[i]=vglLoadTexture(file_name,LoadProc,GL_LINEAR,GL_CLAMP,GL_MODULATE);
#endif		/* NO_ACCEL */
	}
	return true;
}

void	V_GLOBAL
vglDrawSky(quake_sky_t sky_textures,vec_t sky_intensity,vec_t sky_opacity)
{
	int i;
	GLfloat	mat[16];
	GLfloat	cube_vertices[8][3]={
		{	1,	1,	1},
		{	-1,	1,	1},
		{	-1,	-1,	1},
		{	1,	-1,	1},
		{	1,	1,	-1},
		{	-1,	1,	-1},
		{	-1,	-1,	-1},
		{	1,	-1,	-1}};
	int		cube_faces[6][4]={
		{	0,	1,	2,	3	},		/* up */
		{	0,	3,	7,	4	},		/* forward */
		{	1,	0,	4,	5	},		/* left */
		{	2,	1,	5,	6	},		/* back */
		{	3,	2,	6,	7	},		/* right */
		{	4,	7,	6,	5	}};		/* down */
	GLfloat	cube_bottom[4][2]=
	{
		{1,0},
		{1,1},
		{0,1},
		{0,0}
	};
	GLfloat	cube_top[4][2]=
	{
		{1,1},
		{0,1},
		{0,0},
		{1,0}
	};
	GLfloat	cube_wall[4][2]=
	{
		{0,0},
		{1,0},
		{1,1},
		{0,1}
	};
	/* setup state */
	glDisable(GL_CULL_FACE);
	glEnable(GL_TEXTURE_2D);
	/*	move into origin of the world: just reset translation part */
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glGetFloatv(GL_MODELVIEW_MATRIX,mat);
	mat[3]=mat[7]=mat[11]=mat[12]=mat[13]=mat[14]=0;
	mat[15]=1;
	glLoadMatrixf(mat);
	glScalef(10,10,10);
	/* draw box */
	glColor4f(sky_intensity,sky_intensity,sky_intensity,sky_opacity);
	if(IsZero(sky_opacity-1))
	{
		glDisable(GL_BLEND);
	}
	else
	{
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	}
	/* top */
	glBindTexture(GL_TEXTURE_2D,sky_textures[0]);
	glBegin(GL_QUADS);
	for(i=0;i<4;i++)
	{
		glTexCoord2fv(cube_top[i]);
		glVertex3fv(cube_vertices[cube_faces[0][i]]);
	}
	glEnd();
	/* forward */
	glBindTexture(GL_TEXTURE_2D,sky_textures[1]);
	glBegin(GL_QUADS);
	for(i=0;i<4;i++)
	{
		glTexCoord2fv(cube_wall[i]);
		glVertex3fv(cube_vertices[cube_faces[1][i]]);
	}
	glEnd();
	/* left */
	glBindTexture(GL_TEXTURE_2D,sky_textures[4]);
	glBegin(GL_QUADS);
	for(i=0;i<4;i++)
	{
		glTexCoord2fv(cube_wall[i]);
		glVertex3fv(cube_vertices[cube_faces[2][i]]);
	}
	glEnd();
	/* back */
	glBindTexture(GL_TEXTURE_2D,sky_textures[3]);
	glBegin(GL_QUADS);
	for(i=0;i<4;i++)
	{
		glTexCoord2fv(cube_wall[i]);;
		glVertex3fv(cube_vertices[cube_faces[3][i]]);
	}
	glEnd();
	/* right */
	glBindTexture(GL_TEXTURE_2D,sky_textures[2]);
	glBegin(GL_QUADS);
	for(i=0;i<4;i++)
	{
		glTexCoord2fv(cube_wall[i]);;
		glVertex3fv(cube_vertices[cube_faces[4][i]]);
	}
	glEnd();
	/* bottom */
	glBindTexture(GL_TEXTURE_2D,sky_textures[5]);
	glBegin(GL_QUADS);
	for(i=0;i<4;i++)
	{
		glTexCoord2fv(cube_bottom[i]);
		glVertex3fv(cube_vertices[cube_faces[5][i]]);
	}
	glEnd();
	glPopMatrix();
	glDisable(GL_TEXTURE_2D);
}

/*
 *	Vertex drawing functions.
 */
#include "argsused.h"
void	V_LOCAL
vDrawV(vec_t *v,vec_t *c,vec_t *n,vec_t *t)
{
#ifdef	DOUBLEVEC
	glVertex3dv((GLdouble*)v);
#else	/* DOUBLEVEC */
	glVertex3fv((GLfloat*)v);
#endif	/* DOUBLEVEC */
}

#include "argsused.h"
void	V_LOCAL
vDrawVN(vec_t *v,vec_t *c,vec_t *n,vec_t *t)
{
#ifdef	DOUBLEVEC
	glNormal3dv((GLdouble*)n);
	glVertex3dv((GLdouble*)v);
#else	/* DOUBLEVEC */
	glNormal3fv((GLfloat*)n);
	glVertex3fv((GLfloat*)v);
#endif	/* DOUBLEVEC */
}

#include "argsused.h"
void	V_LOCAL
vDrawVT(vec_t *v,vec_t *c,vec_t *n,vec_t *t)
{
#ifdef	DOUBLEVEC
	glTexCoord2dv((GLdouble*)t);
	glVertex3dv((GLdouble*)v);
#else	/* DOUBLEVEC */
	glTexCoord2fv((GLfloat*)t);
	glVertex3fv((GLfloat*)v);
#endif	/* DOUBLEVEC */
}

#include "argsused.h"
void	V_LOCAL
vDrawVNT(vec_t *v,vec_t *c,vec_t *n,vec_t *t)
{
#ifdef	DOUBLEVEC
	glTexCoord2dv((GLdouble*)t);
	glNormal3dv((GLdouble*)n);
	glVertex3dv((GLdouble*)v);
#else	/* DOUBLEVEC */
	glTexCoord2fv((GLfloat*)t);
	glNormal3fv((GLfloat*)n);
	glVertex3fv((GLfloat*)v);
#endif	/* DOUBLEVEC */
}

#include "argsused.h"
void	V_LOCAL
vDrawVC(vec_t *v,vec_t *c,vec_t *n,vec_t *t)
{
#ifdef	DOUBLEVEC
	glColor4dv((GLdouble*)c);
	glVertex3dv((GLdouble*)v);
#else	/* DOUBLEVEC */
	glColor4fv((GLfloat*)c);
	glVertex3fv((GLfloat*)v);
#endif	/* DOUBLEVEC */
}

#include "argsused.h"
void	V_LOCAL
vDrawVCN(vec_t *v,vec_t *c,vec_t *n,vec_t *t)
{
#ifdef	DOUBLEVEC
	glColor4dv((GLdouble*)c);
	glNormal3dv((GLdouble*)n);
	glVertex3dv((GLdouble*)v);
#else	/* DOUBLEVEC */
	glColor4fv((GLfloat*)c);
	glNormal3fv((GLfloat*)n);
	glVertex3fv((GLfloat*)v);
#endif	/* DOUBLEVEC */
}

#include "argsused.h"
void	V_LOCAL
vDrawVCT(vec_t *v,vec_t *c,vec_t *n,vec_t *t)
{
#ifdef	DOUBLEVEC
	glColor4dv((GLdouble*)c);
	glTexCoord2dv((GLdouble*)t);
	glVertex3dv((GLdouble*)v);
#else	/* DOUBLEVEC */
	glColor4fv((GLfloat*)c);
	glTexCoord2fv((GLfloat*)t);
	glVertex3fv((GLfloat*)v);
#endif	/* DOUBLEVEC */
}

void	V_LOCAL
vDrawVCNT(vec_t *v,vec_t *c,vec_t *n,vec_t *t)
{
#ifdef	DOUBLEVEC
	glColor4dv((GLdouble*)c);
	glTexCoord2dv((GLdouble*)t);
	glNormal3dv((GLdouble*)n);
	glVertex3dv((GLdouble*)v);
#else	/* DOUBLEVEC */
	glColor4fv((GLfloat*)c);
	glTexCoord2fv((GLfloat*)t);
	glNormal3fv((GLfloat*)n);
	glVertex3fv((GLfloat*)v);
#endif	/* DOUBLEVEC */
}

/*
 *	Triangle drawing functions.
 */

void	V_LOCAL
triDrawV(tlist_face_t *tri)
{
	int i;
	for(i=0;i<3;i++)
#ifdef	DOUBLEVEC
		glVertex3dv((GLdouble*)tri->verts[i].vert);
#else	/* DOUBLEVEC */
		glVertex3fv((GLfloat*)tri->verts[i].vert);
#endif	/* DOUBLEVEC */
}

void	V_LOCAL
triDrawVN(tlist_face_t *tri)
{
	int i;
	for(i=0;i<3;i++)
	{
#ifdef	DOUBLEVEC
		glNormal3dv((GLdouble*)tri->verts[i].normal);
		glVertex3dv((GLdouble*)tri->verts[i].vert);
#else	/* DOUBLEVEC */
		glNormal3fv((GLfloat*)tri->verts[i].normal);
		glVertex3fv((GLfloat*)tri->verts[i].vert);
#endif	/* DOUBLEVEC */
	}
}

void	V_LOCAL
triDrawVT(tlist_face_t *tri)
{
	int i;
	for(i=0;i<3;i++)
	{
#ifdef	DOUBLEVEC
		glTexCoord2dv((GLdouble*)tri->verts[i].texcoord);
		glVertex3dv((GLdouble*)tri->verts[i].vert);
#else	/* DOUBLEVEC */
		glTexCoord2fv((GLfloat*)tri->verts[i].texcoord);
		glVertex3fv((GLfloat*)tri->verts[i].vert);
#endif	/* DOUBLEVEC */
	}
}

void	V_LOCAL
triDrawVNT(tlist_face_t *tri)
{
	int i;
	for(i=0;i<3;i++)
	{
#ifdef	DOUBLEVEC
		glNormal3dv((GLdouble*)tri->verts[i].normal);
		glTexCoord2dv((GLdouble*)tri->verts[i].texcoord);
		glVertex3dv((GLdouble*)tri->verts[i].vert);
#else	/* DOUBLEVEC */
		glNormal3fv((GLfloat*)tri->verts[i].normal);
		glTexCoord2fv((GLfloat*)tri->verts[i].texcoord);
		glVertex3fv((GLfloat*)tri->verts[i].vert);
#endif	/* DOUBLEVEC */
	}
}

void	V_LOCAL
triDrawVC(tlist_face_t *tri)
{
	int i;
	for(i=0;i<3;i++)
	{
#ifdef	DOUBLEVEC
		glColor4dv((GLdouble*)tri->verts[i].color);
		glVertex3dv((GLdouble*)tri->verts[i].vert);
#else	/* DOUBLEVEC */
		glColor4fv((GLfloat*)tri->verts[i].color);
		glVertex3fv((GLfloat*)tri->verts[i].vert);
#endif	/* DOUBLEVEC */
	}
}

void	V_LOCAL
triDrawVCN(tlist_face_t *tri)
{
	int i;
	for(i=0;i<3;i++)
	{
#ifdef	DOUBLEVEC
		glColor4dv((GLdouble*)tri->verts[i].color);
		glNormal3dv((GLdouble*)tri->verts[i].normal);
		glVertex3dv((GLdouble*)tri->verts[i].vert);
#else	/* DOUBLEVEC */
		glColor4fv((GLfloat*)tri->verts[i].color);
		glNormal3fv((GLfloat*)tri->verts[i].normal);
		glVertex3fv((GLfloat*)tri->verts[i].vert);
#endif	/* DOUBLEVEC */
	}
}

void	V_LOCAL
triDrawVCT(tlist_face_t *tri)
{
	int i;
	for(i=0;i<3;i++)
	{
#ifdef	DOUBLEVEC
		glColor4dv((GLdouble*)tri->verts[i].color);
		glTexCoord2dv((GLdouble*)tri->verts[i].texcoord);
		glVertex3dv((GLdouble*)tri->verts[i].vert);
#else	/* DOUBLEVEC */
		glColor4fv((GLfloat*)tri->verts[i].color);
		glTexCoord2fv((GLfloat*)tri->verts[i].texcoord);
		glVertex3fv((GLfloat*)tri->verts[i].vert);
#endif	/* DOUBLEVEC */
	}
}

void	V_LOCAL
triDrawVCNT(tlist_face_t *tri)
{
	int i;
	for(i=0;i<3;i++)
	{
#ifdef	DOUBLEVEC
		glColor4fv((GLfloat*)tri->verts[i].color);
		glNormal3fv((GLfloat*)tri->verts[i].normal);
		glTexCoord2fv((GLfloat*)tri->verts[i].texcoord);
		glVertex3fv((GLfloat*)tri->verts[i].vert);
#else	/* DOUBLEVEC */
		glColor4fv((GLfloat*)tri->verts[i].color);
		glNormal3fv((GLfloat*)tri->verts[i].normal);
		glTexCoord2fv((GLfloat*)tri->verts[i].texcoord);
		glVertex3fv((GLfloat*)tri->verts[i].vert);
#endif	/* DOUBLEVEC */
	}
}

/*
 *	bit 0 - normal;
 *	bit 1 - texture;
 *	bit 2 - color.
 */
V_LOCAL	vertex_draw_proc_t	vert_draw_procs[8]=
{
	vDrawV,
	vDrawVN,
	vDrawVT,
	vDrawVNT,
	vDrawVC,
	vDrawVCN,
	vDrawVCT,
	vDrawVCNT
};

V_LOCAL	tri_draw_proc_t	tri_draw_procs[8]=
{
	triDrawV,
	triDrawVN,
	triDrawVT,
	triDrawVNT,
	triDrawVC,
	triDrawVCN,
	triDrawVCT,
	triDrawVCNT
};

u32_t	V_CALL
vglGetDrawBits(void)
{
	u32_t draw_bits=0;
	/* glNormalXXX calls useless if lighting disabled */
	if(glIsEnabled(GL_LIGHTING))
		draw_bits|=NORMAL_DRAW_BIT;
	/* glTextureXXX calls useless if texturing disabled */
	if(glIsEnabled(GL_TEXTURE_2D) || glIsEnabled(GL_TEXTURE_2D))
		draw_bits|=TEXTURE_DRAW_BIT;
	/*	what we do to decide wheither or not glColorXXX calls 
	 *	are necessary ? */
	draw_bits|=COLOR_DRAW_BIT;
	return draw_bits;
}

/*
 *	if draw_bits==-1 than appropriate function determined automatically
 */
vertex_draw_proc_t	V_CALL
vglGetVertexDrawProc(int draw_bits)
{
	if(draw_bits<0)
		return vert_draw_procs[vglGetDrawBits()];
	else
		return vert_draw_procs[draw_bits & 7];
}

tri_draw_proc_t	V_CALL
vglGetTriangleDrawProc(int draw_bits)
{
	if(draw_bits<0)
		return tri_draw_procs[vglGetDrawBits()];
	else
		return tri_draw_procs[draw_bits & 7];
}

/*</pre></strong></body></html>*/
/*EOF*/
