/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../include/v_lib.h"
#include	"../source/v_mt.h"
#include	"v_os.h"

#ifdef		__WIN32__

/*
 *	Console input - Win32 specific code.
 */

/*
 *	TODO: IPC commands (begins from ESC char):
 *		"color <16-bit code>"
 *		"clear"
 *		"gotoxy <0-based x, 0-based y>"
 *		etc.
 */
 
static i32_t
keycodeFromConsoleCode(i32_t scan,i32_t vk)
{
	switch(scan)
	{
		case 87:	return key_f11;
		case 88:	return key_f12;
		case 55:	
			if(vk==106)
				return key_grey_asterisk;
			else
				break;
		case 69:
			if(vk==19)
				return key_pause;
			else
				break;
		case 53:
			if(vk==111)
				return key_grey_slash;
			else
				break;
	}
	/*
	 *	Other scan codes idential to DOS codes.
	 */
	return keycodeDosCode(scan);
}

V_LOCAL	void
process_key_event(ipc_handle_t dst,KEY_EVENT_RECORD *ev)
{
	if(ev->wRepeatCount>1)
		return;
	inPostEvent(dst,ev->bKeyDown ? ev_key_down : ev_key_up,
		keycodeFromConsoleCode(ev->wVirtualScanCode,ev->wVirtualKeyCode),
		ev->wVirtualKeyCode);
}

V_LOCAL	u32_t
con_mouse_flags(DWORD bs)
{
	return bs;
}

V_LOCAL vboolean
btnPressed(u32_t prev,u32_t cur,keycode_t *btn)
{
	u32_t mask;
	vboolean pressed;
	pressed=false;
	*btn=key_unknown;
	for(mask=1;mask<8;mask<<=1)
	{
		if((prev & mask) != (cur & mask))
		{
			pressed=((prev & mask) < (cur & mask)) ? true : false;
			switch(mask)
			{
				case 1: *btn=key_left_mouse; break;
				case 2: *btn=key_right_mouse; break;
				case 4: *btn=key_middle_mouse; break; /* ??? BUG */
			}
			break;
		}
	}
	return pressed;
}

V_LOCAL	void
process_mouse_event(ipc_handle_t dst,MOUSE_EVENT_RECORD *ev)
{
	u32_t mf;
	thread_data_t *td;
	keycode_t btn;
	input_event_t up_down;
	switch(ev->dwEventFlags)
	{
		case 0:	/* single click */
			td=GetThreadData();
			mf=con_mouse_flags(ev->dwButtonState);
			if(btnPressed(td->con_prev_mouse_flags,mf,&btn))
				up_down=ev_key_down;
			else
				up_down=ev_key_up;
			if(btn)
				inPostEvent(dst,up_down,btn,0);
			td->con_prev_mouse_flags=mf;
			break;
		case MOUSE_MOVED:
			inPostEvent(dst,ev_mouse_move,ev->dwMousePosition.X,ev->dwMousePosition.Y);
			break;
		default:
			return;
	}
}

LONG				ref_count=0;
HANDLE				console_term_event;
HANDLE				console_input_thread;
HANDLE				std_in=NULL;
HANDLE				std_out=NULL;
u32_t				screen_width, screen_height;
vp_stack_t			exit_handlers;
CRITICAL_SECTION	er_access;
u16_t				last_attr;

BOOL __stdcall	console_handler_routine(DWORD dwCtrlType);

V_LOCAL	DWORD	CALLBACK
console_input_thread_entry(LPVOID param)
{
	HANDLE wait_objects[3]; /* TODO: wait for IPC command */
	INPUT_RECORD in_buf;
	vboolean done;
	DWORD num_events;
	ipc_handle_t srv;
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	ipcInit();
	srv=ipcServerEndpoint(param ? (char*)param : "ConsoleInput");
	if(!srv)
		ExitThread(0);
	wait_objects[0]=std_in;
	wait_objects[1]=console_term_event;
	FlushConsoleInputBuffer(std_in);
	for(done=false;done==false;)
	{
		switch(WaitForMultipleObjects(2,wait_objects,FALSE,1000))
		{
			case WAIT_TIMEOUT:
				// puts("no events withing one second ...");
				continue;
			case WAIT_OBJECT_0+1:
				FlushConsoleInputBuffer(std_in);
				done=true;
				continue;
		}
		if(!ReadConsoleInput(std_in,&in_buf,1,&num_events))
		{
			conlPrintf(LOG_ERROR,"ReadConsoleInput() failed, code %d\n",
				GetLastError());
			Sleep(10);
			continue;
		}
		switch(in_buf.EventType)
		{
			case KEY_EVENT:
				process_key_event(srv,&in_buf.Event.KeyEvent);
				break;
			case MOUSE_EVENT:
				process_mouse_event(srv,&in_buf.Event.MouseEvent);
				break;
		}
	}
	ipcClose(srv);
	ipcShutdown();
	logStopConsoleLog();
	ExitThread(1);
	return 1;	/* quiet compiler warning */
}

vboolean	V_GLOBAL
incStartServer(char *server_name)
{
	DWORD id;
	if(ref_count++>0)
		return true; /* already started */
	std_in=GetStdHandle(STD_INPUT_HANDLE);
	std_out=GetStdHandle(STD_OUTPUT_HANDLE);
	console_term_event=CreateEvent(NULL,TRUE,FALSE,NULL);
	InitializeCriticalSection(&er_access);
    last_attr=0x07;
    SetConsoleCtrlHandler(console_handler_routine,TRUE);
	screen_width=GetLargestConsoleWindowSize(std_out).X;
	screen_height=GetLargestConsoleWindowSize(std_out).Y;
	console_input_thread=CreateThread(NULL,0,console_input_thread_entry,
		server_name,0,&id);
	if(!console_input_thread || 
		WaitForSingleObject(console_input_thread,10)!=WAIT_TIMEOUT)
	{
		CloseHandle(console_term_event);
		console_input_thread=NULL;
		console_term_event=NULL;
		ref_count=0;
		return false;
	}
	return true;
}

void	V_GLOBAL
incStopServer(void)
{
	con_notify_event_t exit_handler;
	if(ref_count<=0)
		return;
	ref_count--;
	if(!ref_count)
	{
		SetEvent(console_term_event);
		WaitForSingleObject(console_input_thread,INFINITE);
		while(vpsPop(&exit_handlers,(void**)&exit_handler))
			if(exit_handler)
				exit_handler();
		CloseHandle(console_term_event);
		DeleteCriticalSection(&er_access);
		console_term_event=NULL;
		console_input_thread=NULL;
	}
}

/*	Control event handler */	
BOOL	__stdcall
console_handler_routine(DWORD dwCtrlType)
{
	switch(dwCtrlType)
	{
		case	CTRL_C_EVENT:
		{
			conTextAttr(0x0D);
			conPutsEx("CTRL_C_EVENT: exiting ...\n");
			incStopServer();
			ExitProcess(0);
		}
		case CTRL_BREAK_EVENT:
		{
			conTextAttr(0x0D);
			conPutsEx("CTRL_BREAK_EVENT: exiting ...\n");
			incStopServer();
			ExitProcess(0);
		}
		case CTRL_CLOSE_EVENT:
		{
			conTextAttr(0x0D);
			conPutsEx("CTRL_CLOSE_EVENT: exiting ...\n");
			incStopServer();
			ExitProcess(0);
		}
		case CTRL_LOGOFF_EVENT:
		{
			conTextAttr(0x0D);
			conPutsEx("CTRL_LOGOFF_EVENT: exiting ...\n");
			incStopServer();
			ExitProcess(0);
		}
		case CTRL_SHUTDOWN_EVENT:
		{
			conTextAttr(0x0D);
			conPutsEx("CTRL_SHUTDOWN_EVENT: exiting ...\n");
			incStopServer();
			ExitProcess(0);
		}
	}
	return TRUE;
}

/*
**	Console control procedures.
*/

/* Changes console title. */
void	V_GLOBAL
conSetCaption(char *caption)
{
	SetConsoleTitle(caption);
}

/*	Returns x-coord of left corner of terminal. */
u32_t	V_GLOBAL
conGetMaxX(void)
{
	return screen_width-1;
}

/* Returns y-coord of bottom corner of terminal. */
u32_t	V_GLOBAL
conGetMaxY(void)
{
	return screen_height-1;
}

/*	Changes cursor position. */
void	V_GLOBAL
conGotoXY(u32_t x,u32_t y)
{
	COORD xy;
	xy.X=(u16_t)x;
	xy.Y=(u16_t)y;
	SetConsoleCursorPosition(std_out,xy);
}

/*	Changes text and background color. */
void	V_GLOBAL
conTextAttr(u16_t attr)
{
	SetConsoleTextAttribute(std_out,attr);
    last_attr=attr;
}

/*	Clears entire screen. */

void	V_GLOBAL
conClearScreen(void)
{
	DWORD	cw;
	int		len;
	COORD	start;
	start.X=start.Y=0;
	len=screen_width*screen_height;
	SetConsoleCursorPosition(std_out,start);
	FillConsoleOutputCharacter(std_out,' ',len,start,&cw);
	FillConsoleOutputAttribute(std_out,last_attr,len,start,&cw);
}

/*	Clears end of current line */
void	V_GLOBAL
conClearLine(void)
{
	DWORD	cw;
	COORD	start;
	CONSOLE_SCREEN_BUFFER_INFO	ci;
	GetConsoleScreenBufferInfo(std_out,&ci);
	start.X=ci.dwCursorPosition.X;
	start.X=ci.dwCursorPosition.Y;
	FillConsoleOutputCharacter(std_out,' ',screen_width-start.X,start,&cw);
	FillConsoleOutputAttribute(std_out,last_attr,screen_width-start.X,
		start,&cw);
}

/*	Writes string to console */
void	V_GLOBAL
conPutsEx(char *str)
{
	DWORD	len;
	char	*out=dbs(str);
	WriteConsole(std_out,out,strlen(out),&len,NULL);
}

/*
**		These registers / unregisters exit notification events.
*/
void	V_GLOBAL
conRegisterAtExit(con_notify_event_t handler)
{
	EnterCriticalSection(&er_access);
	vpsPush(&exit_handlers,*((void**)&handler));
	LeaveCriticalSection(&er_access);
}

void	V_GLOBAL
conUnregisterAtExit(con_notify_event_t handler)
{
	u32_t i;
	EnterCriticalSection(&er_access);
	for(i=0;i<vpsCount(&exit_handlers);i++)
		if(vpsGet(&exit_handlers,i)==handler)
			vpsSet(&exit_handlers,i,NULL);
	LeaveCriticalSection(&er_access);
}

#elif	defined(__UNIX__)	/* __WIN32__ */

#error	TODO: Implement UNIX console input via CURSES library.

#elif	defined(__DOS32__)	/* defined(__UNIX__) */

#error	TODO: Implement DOS console input via HW interrupts.

#endif	/* defined(__DOS32__) */

/*</pre></strong></body></html>*/
/*EOF*/
