/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../include/v_lib.h"
#include	"../source/v_mt.h"
#include	"v_os.h"

thread_data_t		__thread_data=
{
	LIBV_TSD_MAGIC,		/* magic */
	0xFF,				/* constant_alpha - opaque */
	1000,1000,			/* metafile extents */
	NULL,				/* log_file */
	LOG_ERROR			/* log_upper_level */
	/* All other members must be 0/NULL */
};

void	V_GLOBAL
libvFree(void *block)
{
	free(block);
}

void_p	V_GLOBAL
libvMalloc(u32_t size)
{
	return malloc(size);
}

#ifdef		__WIN32__
void		V_CALL
libvError(char *msg)
{
	MessageBox(NULL,msg,"LibV",MB_OK|MB_ICONHAND);
}

LONG		V_LOCAL	detected_must_die=-1;

vboolean	V_CALL
WeAreUnderMustDie(void)
{
	if(detected_must_die==-1)
	{
		LONG md;
		OSVERSIONINFO vi;
		v_zero(&vi,sizeof(OSVERSIONINFO));
		vi.dwOSVersionInfoSize=sizeof(OSVERSIONINFO);
		GetVersionEx(&vi);
		md=vi.dwPlatformId != VER_PLATFORM_WIN32_NT ? 1 : 0;
		InterlockedExchange(&detected_must_die,md);
	}
	return (vboolean)detected_must_die;
}

#elif		defined(__UNIX__) /* !__WIN32__ */

void		V_CALL
libvError(char *msg)
{
	fprintf(stderr,"LibV error: %s.\n",msg);
}

#endif		/* __WIN32__ */

#ifdef		MULTITHREADED

#define		STORAGE_SIZE			(sizeof(thread_data_t))

vboolean	V_CALL
libvPerThreadInit(void)
{
	thread_data_t *storage;
	char buf[256];
	storage=(thread_data_t *)libvGetSpecific();
	if(storage)
	{
		if(storage->magic==LIBV_TSD_MAGIC)
		{
#ifdef	CHECKED_BUILD
			conlPuts(LOG_WARNING,"libvPerThreadInit(): internal "
				"TSD for current thread already initialized.\n");
#endif	/* CHECKED_BUILD */
			return true;
		}
#ifdef	CHECKED_BUILD
        libvError(
        	"libvPerThreadInit(): garbage in internal TSD; discarded.");
#endif	/* CHECKED_BUILD */
	}
	storage=(thread_data_t*)libvMalloc(STORAGE_SIZE);
	if(!storage)
	{
#ifdef	CHECKED_BUILD
		sprintf(buf,"libvPerThreadInit(): failed on allocation %d "
			"bytes for TSD.\n",STORAGE_SIZE);
        libvError(buf);
#endif	/* CHECKED_BUILD */
		return false;
	}
	v_copy(storage,&__thread_data,STORAGE_SIZE);
    libvSetSpecific(storage);
	return true;
}

vboolean	V_CALL
libvPerThreadShutdown(void)
{
	void *storage;
	storage=libvGetSpecific();
	if(storage)
	{
		if(((thread_data_t*)storage)->magic!=LIBV_TSD_MAGIC)
		{
#ifdef	CHECKED_BUILD
			conPuts("libvPerThreadShutdown(): "
				"garbaged TSD; probably memory leak.\n");
#endif	/* CHECKED_BUILD */
			return false;
		}
		libvFree(storage);
		return true;
	}
	return false;
}

#endif	/* !MULTITHREADED */

#ifdef	BUILD_DLL

#include "win32/argsused.h"
BOOL WINAPI
DllMain(HINSTANCE hInstance,DWORD fdwReason,LPVOID lpvReserved)
{
	switch(fdwReason)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
			return libvPerThreadInit() ? TRUE : FALSE;
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			return libvPerThreadShutdown() ? TRUE : FALSE;
	}
	return TRUE;
}

#endif	/* !BUILD_DLL */

/*
**	Platform - independent utility functions.
*/
#ifdef	CHECKED_BUILD
void_p	V_CALL
libvIsHandle(void *handle,char *caller,char *object,u32_t magic)
{
	if(!handle)
	{
		conlPrintf(LOG_ERROR,"%s(): NULL descriptor.\n",caller);
		return NULL;
	}
	if(*((u32_t*)handle) != magic)
	{
		conlPrintf(LOG_ERROR,"%s(): %s damaged: bad magic.\n",caller,
			object ? object : "object");
		return NULL;
	}
	return handle;
}
#endif	/* CHECKED_BUILD */

/*</pre></strong></body></html>*/
/*EOF*/
