/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../include/v_lib.h"
#include	"../source/v_mt.h"
#include 	"../source/v_os.h"

/*
**	The core: multitasking environment wrapper for NT and UNIX.
**	Known bugs:
**	1. Under DOS all threads will execute sequentially
**	(Hmm... My guess, this is bug of DOS, not of LibV)
**	2. Code in DLL behaves BAD when some system APIs spawn threads
**	in user space: looks like system performs mucking with TLS slots too.
**	3. And... static linking for Win32 is currently BROKEN !!!
**	Last update: 14.11.2000
*/

#ifdef	CHECKED_BUILD
#define	THREAD_MAGIC	0x74687264
#define	MUTEX_MAGIC		0x4D555445
#define	COND_MAGIC		0x434F4E44
V_LOCAL	char			*thread_object_name="thread";
V_LOCAL	char			*mutex_object_name="mutex";
V_LOCAL	char			*cond_object_name="cond";
#endif	/* CHECKED_BUILD */

/*
TODO: implement guarded section with global data.
	when (call to threadCreate())
		if (section not protected)
			protect section;
	when (call to threadJoin())
		if (no more threads created by user)
			unprotect section;
In section must be stored:
	1. Table of mounted filesystems.
	2. Table of shared handles.
and:
	implement method (based on thread id's ?) to prevent
	changes of protected/unprotected state from threads
	other than main thread
*/

V_LOCAL	mutex_t				libv_section_guard;
V_LOCAL	libv_global_data_t	libv_global_data={0};

V_LOCAL	libv_global_data_p
lock_protected(void)
{
	mutexLock(libv_section_guard);
	return &(libv_global_data);
}

V_LOCAL	libv_global_data_p
lock_unprotected(void)
{
	return &(libv_global_data);
}

V_LOCAL	void
unlock_protected(void)
{
	mutexUnlock(libv_section_guard);
}

V_LOCAL	void
unlock_unprotected(void)
{
	return;
}

V_LOCAL	libv_global_data_p	(*lock_global_ptr)(void)=lock_unprotected;
V_LOCAL	void				(*unlock_global_ptr)(void)=unlock_unprotected;

V_LOCAL void
libv_protect_section(void)
{
	if(libv_section_guard)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"libv_protect_section(): global data already thread-safe.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	libv_section_guard=mutexCreate();
	mutexLock(libv_section_guard);
	lock_global_ptr=lock_protected;
	unlock_global_ptr=unlock_protected;
	/* NOTE: we leave mutex in locked state */
	puts("SECTION protected");
}

V_LOCAL void
libv_unprotect_section(void)
{
	mutex_t lcopy;
	if(!libv_section_guard)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"libv_protect_section(): global data was thread-unsafe.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	if(!mutexTryLock(libv_section_guard))
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_WARNING,
			"libv_protect_section(): global data probably in use.\n");
#endif	/* CHECKED_BUILD */
		mutexLock(libv_section_guard);
	}
	lock_global_ptr=lock_unprotected;
	unlock_global_ptr=unlock_unprotected;
	lcopy=libv_section_guard;
	libv_section_guard=NULL;
	mutexUnlock(lcopy);
	mutexDestroy(lcopy);
	
	puts("SECTION unprotected");
}

/*
**	Returns contents of guarded section. After use, section
**	MUST be unlocked by libvUnlockGlobal().
*/
libv_global_data_p	V_CALL
libvLockGlobal(void)
{
	return lock_global_ptr();
}

/*
**	Unlocks global data, makes it possible to use by another threads.
*/
void	V_CALL
libvUnlockGlobal(void)
{
	unlock_global_ptr();
}

/*
TODO
extern	thread_data_t	__thread_data;
V_LOCAL	void	*main_thread_tsd=&__thread_data;
*/
V_LOCAL	void	*main_thread_tsd=NULL;
V_LOCAL	u32_t	tsd_key=0;

#if		defined(__WIN32__)

#include "win32/argsused.h"
V_LOCAL	LPVOID WINAPI
TlsGetValueStub(DWORD Key)
{
	return main_thread_tsd;
}

#include "win32/argsused.h"
V_LOCAL	BOOL WINAPI
TlsSetValueStub(DWORD Key,LPVOID Value)
{
	main_thread_tsd=Value;
	return TRUE;
}

LPVOID	(WINAPI *TlsGetValuePtr)(DWORD)=TlsGetValueStub;
BOOL	(WINAPI *TlsSetValuePtr)(DWORD,LPVOID)=TlsSetValueStub;

void	tsd_switch_to_system(void)
{
	TlsSetValuePtr=TlsSetValue;
	TlsGetValuePtr=TlsGetValue;
}

void	tsd_switch_to_stubs(void)
{
	TlsSetValuePtr=TlsSetValueStub;
	TlsGetValuePtr=TlsGetValueStub;
}

/*
 *	Thread object
 */
typedef struct	os_thread_s
{
#ifdef	CHECKED_BUILD
	u32_t			magic;
#endif	/* CHECKED_BUILD */
	HANDLE			handle;
	DWORD			id;
	void			*arg;
	thread_proc_t	start_address;
}	os_thread_t, *os_thread_p;

/*
 *	Mutex object
 */
typedef struct	os_mutex_s
{
#ifdef	CHECKED_BUILD
	u32_t		magic;
#endif	/* CHECKED_BUILD */
	HANDLE		mutex;
}	os_mutex_t, *os_mutex_p;

/*
 *	Conditional variable object
 */
typedef struct	os_cond_s
{
#ifdef	CHECKED_BUILD
	u32_t		magic;
#endif	/* CHECKED_BUILD */
	HANDLE		cond;
}	os_cond_t, *os_cond_p;

/*
 *	Thread entry point
 */
DWORD	CALLBACK
Win32ThreadEntry(LPVOID param)
{
	((os_thread_t *)param)->start_address(((os_thread_t *)param)->arg);
	threadExit(0);
	return 0;
}

#elif	defined(__UNIX__)	/* !__WIN32__ */

int
pthread_setspecific_stub(pthread_key_t key,const void *pointer)
{
	main_thread_tsd=(void*)pointer;
	return 0;
}

void*
pthread_getspecific_stub(pthread_key_t key)
{
	return main_thread_tsd;
}

int	(*pthread_setspecific_ptr)(pthread_key_t,const void *)=
	pthread_setspecific_stub;

void*	(*pthread_getspecific_ptr)(pthread_key_t)=
	pthread_getspecific_stub;

V_LOCAL	void
tsd_switch_to_system(void)
{
	pthread_setspecific_ptr=pthread_setspecific;
	pthread_getspecific_ptr=pthread_getspecific;
}

V_LOCAL	void
tsd_switch_to_stubs(void)
{
	pthread_setspecific_ptr=pthread_setspecific_stub;
	pthread_getspecific_ptr=pthread_getspecific_stub;
}

/*
 *	Thread object
 */
typedef struct	os_thread_s
{
#ifdef	CHECKED_BUILD
	u32_t			magic;
#endif	/* CHECKED_BUILD */
	pthread_t		id;
	pthread_attr_t	attr;
	void			*arg;
	thread_proc_t	start_address;
}	os_thread_t, *os_thread_p;

/*
 *	Mutex object
 */
typedef struct	os_mutex_s
{
#ifdef	CHECKED_BUILD
	u32_t			magic;
#endif	/* CHECKED_BUILD */
	pthread_mutex_t		handle;
	pthread_mutexattr_t	attr;
}	os_mutex_t, *os_mutex_p;

/*
 *	Conditional variable object
 */
typedef	struct	os_cond_s
{
#ifdef	CHECKED_BUILD
	u32_t			magic;
#endif	/* CHECKED_BUILD */
	vboolean		autoreset;
	vboolean		state;
	vboolean		pulse;
	pthread_cond_t	cond;
	pthread_mutex_t	mutex;
}	os_cond_t, *os_cond_p;

/*
 *	Thread entry point
 */
void*
PosixThreadEntry(void *param)
{
	((os_thread_t *)param)->start_address(((os_thread_t *)param)->arg);
	threadExit(0);
	return NULL;
}

#else	/* !__UNIX__; unknown platform (probably, __DOS32__) */

#include "argsused.h"
V_LOCAL	int
thread_setspecific_stub(void *key,const void *pointer)
{
	main_thread_tsd=(void*)pointer;
	return 0;
}

#include "argsused.h"
V_LOCAL	void*
thread_getspecific_stub(void *key)
{
	return main_thread_tsd;
}

int	(*thread_setspecific_ptr)(void*,const void *)=thread_setspecific_stub;
void*	(*thread_getspecific_ptr)(void*)=thread_getspecific_stub;

#define	tsd_switch_to_system()
#define	tsd_switch_to_stubs()

/*
 *	Thread object
 */
typedef struct	os_thread_s
{
#ifdef	CHECKED_BUILD
	u32_t			magic;
#endif	/* CHECKED_BUILD */
	void			*arg;
	thread_proc_t	start_address;
}	os_thread_t, *os_thread_p;

/*
 *	Mutex object
 */
typedef struct	os_mutex_s
{
#ifdef	CHECKED_BUILD
	u32_t			magic;
#endif	/* CHECKED_BUILD */
	void			*stub;
}	os_mutex_t, *os_mutex_p;

/*
 *	Conditional variable object
 */
typedef	struct	os_cond_s
{
#ifdef	CHECKED_BUILD
	u32_t			magic;
#endif	/* CHECKED_BUILD */
	vboolean		state;
}	os_cond_t, *os_cond_p;

/*
 *	Thread entry point
 */
V_LOCAL	void*
StubThreadEntry(void *param)
{
#ifdef	CHECKED_BUILD
	conlPrintf(LOG_WARNING,"StubThreadEntry(): "
		"multithreading not supported, "
		"executing %X(%X) sequentially ...\n",
		(u32_t)((os_thread_t *)param)->start_address,
		(u32_t)((os_thread_t *)param)->arg);
#endif	/* CHECKED_BUILD */
	((os_thread_t *)param)->start_address(((os_thread_t *)param)->arg);
	threadExit(0);	/* In fact, this is not necessary */
	return NULL;
}

#endif	/* __WIN32__ */

/*	=======================================================================
**	TSD management.
*/

u32_t	V_GLOBAL
threadCreateKey(void)
{
	u32_t key;
#ifdef	__WIN32__
	key=(u32_t)TlsAlloc();
#elif	defined(__UNIX__) /* !__WIN32__ */
	/* Unfortunately, key #0 is VALID ! */
	if(pthread_key_create((pthread_key_t*)&key,NULL))
		return 0;
	key++;
#else	/* !__UNIX__ */
	return (u32_t)malloc(sizeof(void*));
#endif	/* __WIN32__ */
	return key;
}

void	V_GLOBAL
threadDestroyKey(u32_t key)
{
#ifdef	__WIN32__
	TlsFree((DWORD)key);
#elif	defined(__UNIX__)
	pthread_key_delete((pthread_key_t)(key-1));
#else	/* !__UNIX__ */
	if(key)
		free((void*)key);
#endif	/* __WIN32__ */
}

void	V_GLOBAL
threadSetSpecific(u32_t key,void_p specific)
{
#ifdef	__WIN32__
	TlsSetValue((DWORD)key,(LPVOID)specific);
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_setspecific((pthread_key_t)(key-1),specific);
#else	/* !__UNIX__ */
	*((void**)key)=specific;
#endif	/* !__UNIX__ */
}

void_p		V_GLOBAL
threadGetSpecific(u32_t key)
{
#ifdef	__WIN32__
	return TlsGetValue((DWORD)key);
#elif	defined(__UNIX__) /* !__WIN32__ */
	return pthread_getspecific((pthread_key_t)(key-1));
#else	/* !__UNIX__ */
	return *((void**)key);
#endif	/* !__UNIX__ */
}

/*
**	Private key management
*/

V_LOCAL	vboolean
tsd_create_private_key(void)
{
	vboolean ok;
#ifdef	__WIN32__
	tsd_key=(u32_t)TlsAlloc();
	ok=tsd_key!=0;
#elif	defined(__UNIX__) /* !__WIN32__ */
	ok=pthread_key_create((pthread_key_t*)&tsd_key,NULL)==0;
#endif	/* !__UNIX__ */
	if(!ok)
	{
		/* Can't use conlPrintf() ! */
#ifdef	__WIN32__
		if(MessageBox(NULL,"tsd_per_thread_init(): "
			"TlsAlloc() failed, prepare for DrWatson() !",
			"LibV main thread initialization error",
			MB_OKCANCEL|MB_ICONHAND)==IDOK)
			return false;
		else
			ExitProcess(1);
#elif	defined(__UNIX__)
		printf("LibV->tsd_create_private_key(): pthread_key_create() failed,"
			" prepare for core dump !\n");
		return false;
#endif	/* !__WIN32__ */
	}
	return true;
}

V_LOCAL void
tsd_destroy_private_key(void)
{
#ifdef	__WIN32__
	TlsFree(tsd_key);
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_key_delete((pthread_key_t)tsd_key);
#endif	/* !__UNIX__ */
}

void	V_CALL
libvSetSpecific(void_p specific)
{
#ifdef	__WIN32__
	TlsSetValuePtr(tsd_key,specific);
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_setspecific_ptr(tsd_key,specific);
#else	/* !__UNIX__ */
	thread_setspecific_ptr(tsd_key,specific);
#endif	/* !__UNIX__ */
}

void_p	V_CALL
libvGetSpecific(void)
{
#ifdef	__WIN32__
	return TlsGetValuePtr(tsd_key);
#elif	defined(__UNIX__) /* !__WIN32__ */
	return pthread_getspecific_ptr(tsd_key);
#else	/* !__UNIX__ */
	return thread_getspecific_ptr(tsd_key);
#endif	/* !__UNIX__ */
}

/*	=======================================================================
**	Switching between thread-unsafe and thread-safe behavior.
*/
V_LOCAL	void	before_start_thread()
{
	libv_global_data_p data;
	data=libvLockGlobal();
	data->thread_count++;
	if(data->thread_count==1)
	{
		void *main_specific;
		/*	This is first thread, initialize key, store local tsd
		 *	and switch TSD functions */
		libv_protect_section();
		main_specific=libvGetSpecific();
		tsd_create_private_key();
		tsd_switch_to_system();
		libvSetSpecific(main_specific);
	}
	libvUnlockGlobal();
}

V_LOCAL	void	after_exit_thread()
{
	libv_global_data_p data;
	data=libvLockGlobal();
	data->thread_count--;
	if(data->thread_count==0)
	{
		void *main_specific;
		/*	This is the last thread, delete key, restore local tsd
		 *	and switch TSD functions back to stubs */
		main_specific=libvGetSpecific();
		tsd_switch_to_stubs();
		tsd_destroy_private_key();
		libvSetSpecific(main_specific);
		/* Now we can use global data without locks */
		libvUnlockGlobal();
		libv_unprotect_section();
		return;
	}
	libvUnlockGlobal();
}

/*	=======================================================================
 *	Threads
 */
thread_t	V_GLOBAL
threadCreate(thread_proc_t start_address,void *arg)
{
#ifdef	__UNIX__
	struct sched_param sp;
#endif	/* __UNIX__ */
	os_thread_t *thread;
	if(!start_address)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"threadCreate(NULL,%X): no thread entry provided.\n",
			arg);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	thread=(os_thread_t*)malloc(sizeof(os_thread_t));
	thread->start_address=start_address;
	thread->arg=arg;
#ifdef	CHECKED_BUILD
	thread->magic=THREAD_MAGIC;
#endif	/* CHECKED_BUILD */
	before_start_thread();
#ifdef	__WIN32__
	thread->handle=CreateThread(NULL,0,Win32ThreadEntry,thread,0,
		&thread->id);
	if(!thread->handle)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"threadCreate(%X,%X): CreateThread() "
			"failed.\n",start_address,arg);
#endif	/* CHECKED_BUILD */
		free(thread);
		return NULL;
	}
#elif	defined(__UNIX__)	/* !__WIN32__ */
	pthread_attr_init(&thread->attr);
	pthread_attr_setscope(&thread->attr,PTHREAD_SCOPE_PROCESS);
	pthread_attr_setdetachstate(&thread->attr,PTHREAD_CREATE_JOINABLE);
	if(!pthread_attr_getschedparam(&thread->attr,&sp))
	{
		sp.sched_priority=5;
		pthread_attr_setschedparam(&thread->attr,&sp);
	}
	if(pthread_create(&thread->id,&thread->attr,
		PosixThreadEntry,(void*)thread)==-1)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"threadCreate(%X,%X): pthread_create() "
			"failed.\n",start_address,arg);
#endif	/* CHECKED_BUILD */
		free(thread);
		return NULL;
	}
#else	/* !__UNIX__ */
	/* No support for threads: chain control of this thread */
	StubThreadEntry((void*)thread);
#endif	/* !__UNIX__ */
	return (thread_t)thread;
}

void		V_GLOBAL
threadExit(u32_t exit_code)
{
#ifdef	__WIN32__
	ExitThread(exit_code);
#elif	defined(__UNIX__)
	pthread_exit((void*)exit_code);
#else	/* !__UNIX__ */
	return;
#endif	/* __WIN32__ */
}

vboolean	V_GLOBAL
threadJoin(thread_t thread,u32_t *exit_code)
{
	thread=(thread_t)libvIsHandle((void*)thread,"threadJoin",
		thread_object_name,THREAD_MAGIC);
	if(!thread)
		return false;
#ifdef	__WIN32__
	if(WaitForSingleObject(((os_thread_t*)thread)->handle,INFINITE)
		!=WAIT_OBJECT_0)
		return false;
	if(exit_code)
	{
		DWORD ec;
		GetExitCodeThread(((os_thread_t*)thread)->handle,&ec);
		*exit_code=(u32_t)ec;
	}
#elif	defined(__UNIX__)
	if(pthread_join(((os_thread_t*)thread)->id,
		exit_code ? (void**)exit_code : NULL))
		return false;
#else	/* !__UNIX__ */
	/* Nothing to do */
#endif	/* __WIN32__ */
	after_exit_thread();
	free(thread);
	return true;
}

/*	=======================================================================
 *	Mutexes
 */
mutex_t		V_GLOBAL
mutexCreate()
{
	os_mutex_t *mutex;
	mutex=(os_mutex_t*)malloc(sizeof(os_mutex_t));
#ifdef	CHECKED_BUILD
	mutex->magic=MUTEX_MAGIC;
#endif	/* !CHECKED_BUILD */
#ifdef	__WIN32__
	mutex->mutex=CreateMutex(NULL,FALSE,NULL);
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_mutexattr_init(&mutex->attr);
#ifdef	__USE_UNIX98
	pthread_mutexattr_settype(&mutex->attr,PTHREAD_MUTEX_RECURSIVE_NP);
#endif	/* !__USE_UNIX98 */
#else	/* !__UNIX__ */
	/* Nothing to do */
#endif	/* !__UNIX__ */
	return mutex;
}

void	V_GLOBAL
mutexLock(mutex_t mutex)
{
	mutex=(mutex_t)libvIsHandle((void*)mutex,"mutexLock",
		mutex_object_name,MUTEX_MAGIC);
	if(!mutex)
		return;
#ifdef	__WIN32__
	WaitForSingleObject(((os_mutex_t *)mutex)->mutex,INFINITE);
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_mutex_lock(&((os_mutex_t*)mutex)->handle);
#else	/* !__UNIX__ */
	/* Nothing to do: locks always succeded */
#endif	/* !__UNIX__ */
}

vboolean	V_GLOBAL
mutexTryLock(mutex_t mutex)
{
	/*	Calling mutexTryLock(NULL) is NOT considered as
	 *	an error. */
	if(!mutex)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_INFO,
			"mutexTryLock(NULL): NULL handle, impossible to lock.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
#ifdef	CHECKED_BUILD
	mutex=(mutex_t)libvIsHandle((void*)mutex,"mutexTryLock",
		mutex_object_name,MUTEX_MAGIC);
	if(!mutex)
		return false;
#endif	/* CHECKED_BUILD */
#ifdef	__WIN32__
	if(WaitForSingleObject(((os_mutex_t *)mutex)->mutex,0)
		==WAIT_OBJECT_0)
		return true;
	else
		return false;
#elif	defined(__UNIX__)
	if(pthread_mutex_trylock(&((os_mutex_t *)mutex)->handle)!=0)
		return false;
	else
		return true;
#else	/* !__UNIX__ */
	/* Nothing to do: locks always succeded */
	return true;
#endif	/* __WIN32__ */
}

void	V_GLOBAL
mutexUnlock(mutex_t mutex)
{
	mutex=(mutex_t)libvIsHandle((void*)mutex,"mutexUnlock",
		mutex_object_name,MUTEX_MAGIC);
	if(!mutex)
		return;
#ifdef	__WIN32__
	ReleaseMutex(((os_mutex_t *)mutex)->mutex);
#elif	defined(__UNIX__)
	pthread_mutex_unlock(&((os_mutex_t *)mutex)->handle);
#else	/* !__UNIX__ */
	/* Nothing to do */
#endif	/* __WIN32__ */
	return;
}

void		V_GLOBAL
mutexDestroy(mutex_t mutex)
{
	mutex=(mutex_t)libvIsHandle((void*)mutex,"mutexDestroy",
		mutex_object_name,MUTEX_MAGIC);
	if(!mutex)
		return;
#ifdef	__UNIX__
	pthread_mutexattr_destroy(&((os_mutex_t*)mutex)->attr);
#endif	/* __UNIX__ */
	free(mutex);
}

/*	=======================================================================
 *	Conditional variables
 */
cond_t		V_GLOBAL
condCreate(vboolean autoreset)
{
	os_cond_t *cond;
	cond=(os_cond_t*)malloc(sizeof(os_cond_t));
#ifdef	CHECKED_BUILD
	cond->magic=COND_MAGIC;
#endif	/* !CHECKED_BUILD */
#ifdef	__WIN32__
	cond->cond=CreateEvent(NULL,autoreset ? FALSE : TRUE,FALSE,NULL);
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_cond_init(&cond->cond,NULL);
	pthread_mutex_init(&cond->mutex,NULL);
	cond->state=false;
	cond->pulse=false;
	cond->autoreset=autoreset;
#else	/* !__UNIX__ */
	cond->state=false;
#endif	/* !__UNIX__ */
	return (cond_t)cond;
}

vboolean	V_GLOBAL
condSet(cond_t cond)
{
	cond=(cond_t)libvIsHandle((void*)cond,"condSet",
		cond_object_name,COND_MAGIC);
	if(!cond)
		return false;
#ifdef	__WIN32__
	SetEvent(((os_cond_t*)cond)->cond);
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_mutex_lock(&((os_cond_t *)cond)->mutex);
	((os_cond_t *)cond)->state=true;
	pthread_cond_broadcast(&((os_cond_t *)cond)->cond);
	pthread_mutex_unlock(&((os_cond_t *)cond)->mutex);
#else	/* !__UNIX__ */
	((os_cond_t *)cond)->state=true;
#endif	/* !__UNIX__ */
	return true;
}

vboolean	V_GLOBAL
condPulse(cond_t cond)
{
	cond=(cond_t)libvIsHandle((void*)cond,"condPulse",
		cond_object_name,COND_MAGIC);
	if(!cond)
		return false;
#ifdef	__WIN32__
	PulseEvent(((os_cond_t*)cond)->cond);
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_mutex_lock(&((os_cond_t *)cond)->mutex);
	((os_cond_t *)cond)->state=true;
	((os_cond_t *)cond)->pulse=true;
	pthread_cond_broadcast(&((os_cond_t *)cond)->cond);
	pthread_mutex_unlock(&((os_cond_t *)cond)->mutex);
#else	/* !__UNIX__ */
	/* Nothing to do */
#endif	/* !__UNIX__ */
	return true;
}

vboolean	V_GLOBAL
condReset(cond_t cond)
{
	cond=(cond_t)libvIsHandle((void*)cond,"condReset",
		cond_object_name,COND_MAGIC);
	if(!cond)
		return false;
#ifdef	__WIN32__
	ResetEvent(((os_cond_t*)cond)->cond);
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_mutex_lock(&((os_cond_t *)cond)->mutex);
	((os_cond_t *)cond)->state=false;
	((os_cond_t *)cond)->pulse=false;
	pthread_mutex_unlock(&((os_cond_t *)cond)->mutex);
#else	/* !__UNIX__ */
	((os_cond_t *)cond)->state=false;
#endif	/* !__UNIX__ */
	return true;
}

vboolean	V_GLOBAL
condWait(cond_t cond)
{
	cond=(cond_t)libvIsHandle((void*)cond,"condWait",
		cond_object_name,COND_MAGIC);
	if(!cond)
		return false;
#ifdef	__WIN32__
	if(WaitForSingleObject(((os_cond_t*)cond)->cond,INFINITE)==
		WAIT_OBJECT_0)
		return true;
	else
		return false;
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_mutex_lock(&((os_cond_t *)cond)->mutex);
	if(((os_cond_t *)cond)->autoreset)
		((os_cond_t *)cond)->state=false;
	if(((os_cond_t *)cond)->pulse)
	{
		((os_cond_t *)cond)->pulse=false;
		((os_cond_t *)cond)->state=false;
	}
	while(!((os_cond_t *)cond)->state)
		pthread_cond_wait(&((os_cond_t *)cond)->cond,
			&((os_cond_t *)cond)->mutex);
	pthread_mutex_unlock(&((os_cond_t *)cond)->mutex);
	return true;
#else	/* !__UNIX__ */
#ifdef	CHECKED_BUILD
	if(((os_cond_t *)cond)->state == false)
	{
		conlPuts(LOG_WARNING,"condWait(): current cond state will cause "
			"deadlock; ignored.\n");
	}
#endif	/* CHECKED_BUILD */
	return true;
#endif	/* !__UNIX__ */
}

vboolean	V_GLOBAL
condTimedWait(cond_t cond,u32_t milliseconds)
{
#ifdef	__UNIX__
	struct timeval now;
    struct timespec timeout;
	struct timezone tz={0,0};
	vboolean state=false;
#endif	/* !__UNIX__ */
	cond=(cond_t)libvIsHandle((void*)cond,"condTimedWait",
		cond_object_name,COND_MAGIC);
	if(!cond)
		return false;
#ifdef	__WIN32__
	if(WaitForSingleObject(((os_cond_t*)cond)->cond,milliseconds)==
		WAIT_OBJECT_0)
		return true;
	else
		return false;
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_mutex_lock(&((os_cond_t *)cond)->mutex);
	if(((os_cond_t *)cond)->autoreset)
		((os_cond_t *)cond)->state=false;
	if(((os_cond_t *)cond)->pulse)
	{
		((os_cond_t *)cond)->pulse=false;
		((os_cond_t *)cond)->state=false;
	}
	if(((os_cond_t *)cond)->state || !milliseconds)
	{
		state=((os_cond_t *)cond)->state;
		pthread_mutex_unlock(&((os_cond_t *)cond)->mutex);
		return state;
	}
	gettimeofday(&now,&tz);
	timeout.tv_sec = now.tv_sec + milliseconds/1000;
	timeout.tv_nsec = (now.tv_usec+(milliseconds % 1000)*1000)*1000;
	for(;;)
	{
		if(pthread_cond_timedwait(&((os_cond_t *)cond)->cond,
			&((os_cond_t *)cond)->mutex,&timeout))
			break;
		state=((os_cond_t *)cond)->state;
		if(state)
			break;
	}
	pthread_mutex_unlock(&((os_cond_t *)cond)->mutex);
	return state;
#else	/* !__UNIX__ */
	state=((os_cond_t *)cond)->state;
	if(!state && milliseconds)
		timeDelay(milliseconds);
	return state;
#endif	/* !__UNIX__ */
}

void		V_GLOBAL
condDestroy(cond_t cond)
{
	cond=(cond_t)libvIsHandle((void*)cond,"condDestroy",
		cond_object_name,COND_MAGIC);
	if(!cond)
		return;
#ifdef	__WIN32__
	CloseHandle(((os_cond_t*)cond)->cond);
#elif	defined(__UNIX__) /* !__WIN32__ */
	pthread_cond_broadcast(&((os_cond_t *)cond)->cond);
	pthread_cond_destroy(&((os_cond_t *)cond)->cond);
	pthread_mutex_destroy(&((os_cond_t *)cond)->mutex);
#else	/* !__UNIX__ */
	/* Nothing to do */
#endif	/* !__UNIX__ */
	free(cond);
}

/*</pre></strong></body></html>*/
/*EOF*/
