/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../include/v_lib.h"
#include	"../source/v_mt.h"

/*	Iternal macros for "typed stacks"
 */
#define	TYPED_STACK_INIT(stack)\
		(stack).n_elem=(stack).next_elem=0;\
		(stack).buf=NULL;

#define	TYPED_STACK_PUSH(stack,element,type_elem)\
		if((stack).next_elem>=(stack).n_elem) {\
			/* Allocate new storage */\
			unsigned nextsize;\
			GREATER_POWER_TWO(nextsize,(stack).next_elem,unsigned);\
			(stack).buf=(type_elem*)realloc((stack).buf,nextsize*sizeof(type_elem));\
			(stack).n_elem=nextsize;\
		}\
		(stack).buf[(stack).next_elem++]=element;

#define	TYPED_STACK_PUSH_BUFFER(stack,buffer,num_elements,type_elem)\
		{\
			unsigned total;\
			total=(stack).next_elem+(num_elements);\
			if(total > (stack).n_elem) {\
				NOT_LESS_POWER_TWO(total,total,unsigned);\
				(stack).buf=(type_elem*)realloc((stack).buf,total*sizeof(type_elem));\
				(stack).n_elem=total;\
			}\
			v_copy((stack).buf+(stack).next_elem,(buffer),\
				(num_elements)*sizeof(type_elem));\
			(stack).next_elem+=(num_elements);\
		}

#define	TYPED_STACK_RESIZE(stack,newsize,type_elem)\
		if((newsize) > (stack).n_elem) {\
			/* We must allocate new storage */\
			unsigned nextlen;\
			NOT_LESS_POWER_TWO(nextlen,(newsize),unsigned);\
			(stack).buf=(type_elem*)realloc((stack).buf,nextlen*sizeof(type_elem));\
			/* Fill rest with zeros */\
			v_zero((stack).buf+(stack).next_elem,\
				((newsize)-(stack).next_elem)*sizeof(type_elem));\
			(stack).n_elem=nextlen;\
		} else {\
			/* We fit in current storage */\
			if((newsize) > (stack).next_elem)\
				v_zero((stack).buf+(stack).next_elem,\
					((newsize)-(stack).next_elem)*sizeof(type_elem));\
		}\
		(stack).next_elem=(newsize);\

/*	Function's templates for "typed stacks"
 */
#define	TYPED_STACK_DEFINE_INIT(type_prefix)\
void	V_GLOBAL \
type_prefix##sInit(type_prefix##_stack_t *stack)\
{\
	V_ASSERT(stack);\
	TYPED_STACK_INIT(*stack);\
}

#define	TYPED_STACK_DEFINE_CLEAR(type_prefix)\
void	V_GLOBAL \
type_prefix##sClear(type_prefix##_stack_t *stack)\
{\
	V_ASSERT(stack);\
	if(stack->buf)\
		free(stack->buf);\
	TYPED_STACK_INIT(*stack);\
}

#define	TYPED_STACK_DEFINE_PUSH(type_prefix,type)\
void	V_GLOBAL \
type_prefix##sPush(type_prefix##_stack_t *stack,type element)\
{\
	V_ASSERT(stack);\
	TYPED_STACK_PUSH(*stack,element,type);\
}

#define	TYPED_STACK_DEFINE_POP(type_prefix,type)\
int		V_GLOBAL \
type_prefix##sPop(type_prefix##_stack_t *stack,type *element)\
{\
	V_ASSERT(stack && element);\
	if(stack->next_elem) {\
		*element=stack->buf[--stack->next_elem];\
		return 1;\
	} else return 0;\
}

#define	TYPED_STACK_DEFINE_PUSH_BUFFER(type_prefix,type)\
void	V_GLOBAL \
type_prefix##sPushBuffer(type_prefix##_stack_t *stack,\
						type *src,unsigned num_elements)\
{\
	V_ASSERT(stack && src && num_elements);\
	TYPED_STACK_PUSH_BUFFER(*stack,src,num_elements,type);\
}

#define	TYPED_STACK_DEFINE_POP_BUFFER(type_prefix,type)\
int		V_GLOBAL \
type_prefix##sPopBuffer(type_prefix##_stack_t *stack,\
						type *dst,unsigned num_elements)\
{\
	V_ASSERT(stack && dst && num_elements);\
	if(num_elements>stack->next_elem)\
		num_elements=stack->next_elem;\
	stack->next_elem-=num_elements;\
	v_copy(dst,stack->buf + stack->next_elem,num_elements*sizeof(type));\
	return num_elements;\
}

#define	TYPED_STACK_DEFINE_RESIZE(type_prefix,type)\
void	V_GLOBAL \
type_prefix##sResize(type_prefix##_stack_t *stack,unsigned newsize)\
{\
	V_ASSERT(stack && newsize>=0);\
	TYPED_STACK_RESIZE(*stack,newsize,type);\
}

#define	TYPED_STACK_DEFINE_COUNT(type_prefix)\
unsigned	V_GLOBAL \
type_prefix##sCount(type_prefix##_stack_t *stack)\
{\
	V_ASSERT(stack);\
	return stack->next_elem;\
}

#define	TYPED_STACK_DEFINE_GET(type_prefix,type)\
type	V_GLOBAL \
type_prefix##sGet(type_prefix##_stack_t *stack,unsigned pos)\
{\
	V_ASSERT(stack && stack->n_elem > pos);\
	return stack->buf[pos];\
}

#define	TYPED_STACK_DEFINE_GET_STORAGE(type_prefix,type_ptr)\
type_ptr	V_GLOBAL \
type_prefix##sGetStorage(type_prefix##_stack_t *stack)\
{\
	V_ASSERT(stack);\
	return stack->buf;\
}

#define	TYPED_STACK_DEFINE_SET(type_prefix,type)\
void	V_GLOBAL \
type_prefix##sSet(type_prefix##_stack_t *stack,unsigned pos,type element)\
{\
	V_ASSERT(stack && pos>=0);\
	if(pos>=stack->n_elem)\
	{\
		TYPED_STACK_RESIZE(*stack,pos+1,type);\
	}\
	stack->buf[pos]=element;\
}

#define	TYPED_STACK_DEFINE_CONCAT(type_prefix,type)\
void	V_GLOBAL \
type_prefix##sConcat(type_prefix##_stack_t *dst,type_prefix##_stack_t *src)\
{\
	V_ASSERT(src && dst);\
	TYPED_STACK_PUSH_BUFFER(*dst,src->buf,src->next_elem,type);\
}

/*	Implementation of u8_stack_t and vp_stack_t
 *	( Looks as ugly as MFC's message tables :)
 */

	TYPED_STACK_DEFINE_INIT(u8)
	TYPED_STACK_DEFINE_CLEAR(u8)
	TYPED_STACK_DEFINE_PUSH(u8,u8_t)
	TYPED_STACK_DEFINE_PUSH_BUFFER(u8,u8_t)
	TYPED_STACK_DEFINE_POP(u8,u8_t)
	TYPED_STACK_DEFINE_POP_BUFFER(u8,u8_t)
	TYPED_STACK_DEFINE_RESIZE(u8,u8_t)
	TYPED_STACK_DEFINE_CONCAT(u8,u8_t)
	TYPED_STACK_DEFINE_SET(u8,u8_t)

	TYPED_STACK_DEFINE_INIT(vp)
	TYPED_STACK_DEFINE_CLEAR(vp)
	TYPED_STACK_DEFINE_PUSH(vp,void_p)
	TYPED_STACK_DEFINE_PUSH_BUFFER(vp,void_p)
	TYPED_STACK_DEFINE_POP(vp,void_p)
	TYPED_STACK_DEFINE_POP_BUFFER(vp,void_p)
	TYPED_STACK_DEFINE_RESIZE(vp,void_p)
	TYPED_STACK_DEFINE_CONCAT(vp,void_p)
	TYPED_STACK_DEFINE_SET(vp,void_p)

#ifdef	STACK_NO_MACRO

	TYPED_STACK_DEFINE_COUNT(u8)
	TYPED_STACK_DEFINE_GET(u8,u8_t)
	TYPED_STACK_DEFINE_GET_STORAGE(u8,u8_p)

	TYPED_STACK_DEFINE_COUNT(vp)
	TYPED_STACK_DEFINE_GET(vp,void_p)
	TYPED_STACK_DEFINE_GET_STORAGE(vp,void_pp)

#endif	/* STACK_NO_MACRO */

/*
 *	One useful method of 'stack of pointers'
 */
u32_t	V_GLOBAL
vpsStore(vp_stack_t *st,void *ptr)
{
	u32_t i,len;
	void **objects;
	V_ASSERT(st);
	len=vpsCount(st);
	objects=(void**)vpsGetStorage(st);
	V_ASSERT(objects);
	for(i=0;i<len;i++)
		if(!objects[i])
		{
			objects[i]=ptr;
			return i;
		}
	vpsPush(st,ptr);
	return len;
}

/*	Implementation of "Stack of blocks"
 */
void	V_GLOBAL
stackInit(bstack_t *stack,unsigned element_size)
{
	V_ASSERT(stack && element_size);
	stack->n_elem=stack->next_elem=0;
	stack->buf=NULL;
	stack->element_size=element_size;
}

void	V_GLOBAL
stackClear(bstack_t *stack)
{
	V_ASSERT(stack);
	stack->n_elem = stack->next_elem = 0;
	if(stack->buf)
	{
		free(stack->buf);
		stack->buf=NULL;
	}
	stack->ptr = NULL;
}

void	V_GLOBAL
stackPush(bstack_t *stack,void *element)
{
	V_ASSERT(stack && element);
	if(stack->next_elem>=stack->n_elem)
	{
		/* Allocate new storage */
		unsigned nextsize;
		GREATER_POWER_TWO(nextsize,stack->next_elem,unsigned);
		stack->buf=(unsigned char *)realloc(stack->buf,
			nextsize * stack->element_size);
		stack->n_elem=nextsize;
		stack->ptr=stack->buf + stack->next_elem * stack->element_size;
	}
	v_copy(stack->ptr, element, stack->element_size);
	stack->ptr+=stack->element_size;
	stack->next_elem++;
}

int		V_GLOBAL
stackPop(bstack_t *stack,void *element)
{
	V_ASSERT(stack && element);
	if(stack->next_elem)
	{
		stack->next_elem--;
		stack->ptr-=stack->element_size;
		v_copy(element,stack->ptr,stack->element_size);
		return 1;
	}
	return 0;
}

int		V_GLOBAL
stackPopBuffer(bstack_t *stack,void *dst,unsigned num_elements)
{
	V_ASSERT(stack && dst && num_elements);
	if(num_elements>stack->next_elem)
		num_elements=stack->next_elem;
	stack->next_elem-=num_elements;
	stack->ptr-=num_elements*stack->element_size;
	v_copy(dst,stack->ptr,num_elements*stack->element_size);
	return num_elements;
}

void	V_GLOBAL
stackPushBuffer(bstack_t *stack,void *buffer,unsigned num_elements)
{
	unsigned total;
	V_ASSERT(stack && buffer && num_elements);
	total=stack->next_elem+num_elements;
	if(total > stack->n_elem)
	{
		NOT_LESS_POWER_TWO(total,total,unsigned);
		stack->buf=(unsigned char *)realloc(stack->buf,
			total*stack->element_size);
		stack->n_elem=total;
		stack->ptr=stack->buf + stack->next_elem * stack->element_size;
	}
	v_copy(stack->ptr,buffer,num_elements*stack->element_size);
	stack->ptr+=num_elements*stack->element_size;
	stack->next_elem+=num_elements;
}

void	V_GLOBAL
stackResize(bstack_t *stack,unsigned newsize)
{
	V_ASSERT(stack);
	if(newsize > stack->n_elem)
	{
    	/* We must allocate new storage */
    	unsigned nextlen;
    	NOT_LESS_POWER_TWO(nextlen,newsize,unsigned);
    	stack->buf=(unsigned char*)realloc(stack->buf,
    		nextlen*stack->element_size);
    	stack->n_elem=nextlen;
		stack->ptr=stack->buf + stack->next_elem * stack->element_size;
    	/* Fill rest with zeros */
    	v_zero(stack->ptr,(newsize - stack->next_elem) * stack->element_size);
    }
    else
    {
		if(newsize > stack->next_elem)
    		v_zero(stack->ptr,(newsize-stack->next_elem)*stack->element_size);
    }
	stack->next_elem=newsize;
	stack->ptr=stack->buf + stack->next_elem * stack->element_size;
}

void	V_GLOBAL
stackSet(bstack_t *stack,unsigned pos,void *element)
{
	V_ASSERT(stack && element && pos>=0)
	if(pos>stack->n_elem)
		stackResize(stack,pos);
	stackPush(stack,element);
}

void	V_GLOBAL
stackConcat(bstack_t *dst,bstack_t *src)
{
	stackPushBuffer(dst,src->buf,src->next_elem);
}

#ifdef		STACK_NO_MACRO

unsigned	V_GLOBAL
stackCount(bstack_t *stack)
{
	return stack->next_elem;
}

void	V_GLOBAL
stackGet(bstack_t *stack,void *dst,unsigned pos)
{
	v_copy(dst,(unsigned char*)stack->buf+pos*stack->element_size,
		stack->element_size);
}

void_p	V_GLOBAL
stackGetStorage(bstack_t *stack)
{
	return (stack)->buf;
}

#endif		/* STACK_NO_MACRO */
/*</pre></strong></body></html>*/
/*EOF*/
