/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../include/v_lib.h"
#include	"../source/v_mt.h"

#ifndef	WAVE_FORMAT_PCM
#define	WAVE_FORMAT_PCM		1
#endif	/* WAVE_FORMAT_PCM */

#ifdef	_MSC_VER
#pragma warning(disable:4103)
#endif	/* _MSC_VER */

/*
**	Microsoft WAV format
*/
#include "pack0.h"
typedef struct	wave_format_s
{
	u16_t	format_tag;			/*	only WAVE_FORMAT_PCM supported */
	u16_t	num_channels;		/*	must be either 1 or 2 */
	u32_t	sampling_frequency;
	u32_t	bytes_per_second;	/*	== samples_per_second * block_align */
	u16_t	block_align;		/*	== num_channels*bits_per_sample / 8 */
	u16_t	bits_per_sample;
}	wave_format_t, *wave_format_p;
#include "ppack.h"

#ifdef	_MSC_VER
#pragma warning(default:4103)
#endif	/* _MSC_VER */

#ifdef	CHECKED_BUILD
#define	WAVE_MAGIC					0x57415645	/* WAVE */
V_LOCAL	char		*wave_object=	"wave";
#endif	/* CHECKED_BUILD */

/*
**	Actual description of the wave data.
*/
typedef struct	wave_data_s
{
#ifdef	CHECKED_BUILD
	u32_t	magic;
	char	wave_name[_MAX_PATH]; /* name of data file, used internally */
#endif	/* CHECKED_BUILD */
	u16_t	num_channels;		/*	must be either 1 or 2 */
	u32_t	sampling_frequency;
	u32_t	bytes_per_second;	/*	== samples_per_second * block_align */
	u16_t	block_align;		/*	== num_channels*bits_per_sample / 8 */
	u16_t	bits_per_sample;
	/*		!!!					These fields not to store in WAV file ! */
	u32_t	num_samples;		/* total number of samples (not bytes) */
	u32_t	num_bytes;			/* total data size in bytes */
	u8_t	*wave_data;			/* only valid after wavLoad() */
	file_t	source;				/* only valid after wavOpenRead() */
	file_t	dest;				/* only valid after wavOpenWrite() */
	u32_t	transferred;		/* number of transferred bytes */
}	wave_data_t, *wave_data_p;

typedef struct chunk_head_s
{
	char	id[4];		/* identifier: "RIFF" for file header */
	u32_t	len;		/* remaining chunk length after header */
}	chunk_head_t, *chunk_head_p;

/*
**	Reads header of the next chunk in RIFF file.
**	Returns false if unexpected EOF reached.
*/
V_LOCAL vboolean
riff_read_head(file_t in,chunk_head_t *head_ret)
{
	u32_t pos,rd;
	pos=fileTell(in);
	fileRead(in,head_ret,4);
	head_ret->len=fileGetLittleLong(in);
	rd=fileTell(in)-pos;
	if(rd!=8)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"riff_read_head(): unable to read chunk header.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	return true;
}

/*
**	Reads chunk and returns pointer to memAlloc()'ed block (or
**	NULL if unexpected EOF reached).
**	File pointer must be propely positioned (to first byte
**	after chunk header).
*/
V_LOCAL void_p
riff_read_chunk_data(file_t in,chunk_head_t *chead)
{
	void *cdata;
	cdata=memAlloc(chead->len);
	if((u32_t)fileRead(in,cdata,chead->len)!=chead->len)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"riff_read_chunk_data(): "
			"unexpected end of file in chunk '%c%c%c%c'.\n",
			chead->id[0],chead->id[1],chead->id[2],chead->id[3]);
#endif	/* CHECKED_BUILD */
		memFree(cdata);
		return NULL;
	}
	return cdata;
}

/*
**	Reads wave format chunk.
*/
V_LOCAL	vboolean
riff_read_fmt(file_t in,chunk_head_t *chead,wave_data_t *fmt)
{
	wave_format_t *cdata;
	vboolean ret=false;
	for(;;)
	{
		cdata=(wave_format_t*)riff_read_chunk_data(in,chead);
		if(!cdata)
			break;
		if(cdata->format_tag!=WAVE_FORMAT_PCM)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"riff_read_fmt(): format tag (%X)!=WAVE_FORMAT_PCM.\n",
				cdata->format_tag);
#endif		/* CHECKED_BUILD */
			break;
		}
#ifdef	BIG_ENDIAN
		fmt->num_channels=SWAP_TWO_BYTES(cdata->num_channels);
		fmt->bits_per_sample=SWAP_TWO_BYTES(cdata->bits_per_sample);
		fmt->sampling_frequency=SWAP_FOUR_BYTES(cdata->sampling_frequency);
#else	/* !BIG_ENDIAN */
		fmt->num_channels=cdata->num_channels;
		fmt->bits_per_sample=cdata->bits_per_sample;
		fmt->sampling_frequency=cdata->sampling_frequency;
#endif	/* !BIG_ENDIAN */
		fmt->block_align=
			(u16_t)(fmt->num_channels * (fmt->bits_per_sample >> 3));
		fmt->bytes_per_second=fmt->sampling_frequency * fmt->block_align;
		ret=true;
		break;
	}
	memFree(cdata);
	return ret;
}

/*
**	Opens wave file, reads wave format ant seeks to the start of data.
*/
wave_handle_t	V_GLOBAL
wavOpenRead(char *name)
{
	chunk_head_t head;
	char typeid[5]={0,0,0,0,0};
	int chunk_count;
	wave_data_t *fmt;
	u32_t source_pos;
	ALLOC_OBJECT(fmt,wave_data_t,0,WAVE_MAGIC);
	fmt->source=fileOpenRead(name);
	if(!fmt->source)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"waveOpenRead('%s'): unable to open file.\n",
			dbs(name));
#endif	/* CHECKED_BUILD */
		memFree(fmt);
		return NULL;
	}
	/* Read main header */
	if(!riff_read_head(fmt->source,&head))
	{
		fileClose(fmt->source);
		memFree(fmt);
		return NULL;
	}
	/* Is it a valid RIFF file ? */
	if(memcmp(head.id,"RIFF",4))
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"waveOpenRead('%s'): expected RIFF fourcc, found '%c%c%c%c'.\n",
			name,head.id[0],head.id[1],head.id[2],head.id[3]);
#endif	/* CHECKED_BUILD */
		fileClose(fmt->source);
		memFree(fmt);
		return NULL;
	}
	/* Read typeid */
	fileRead(fmt->source,typeid,4);
	/* Check for type of data */
	if(memcmp(typeid,"WAVE",4))
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"waveOpenRead('%s'): expected WAVE data, found '%s'.\n",
			name,typeid);
#endif	/* CHECKED_BUILD */
		fileClose(fmt->source);
		memFree(fmt);
		return NULL;
	}
	/* Find 'fmt ' and 'data' chunks */
	for(chunk_count=0;chunk_count<2;)
	{
		if(!riff_read_head(fmt->source,&head))
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,"waveOpenRead('%s'): unexpected EOF.\n",
				name);
#endif	/* CHECKED_BUILD */
			break;
		}
		if(!memcmp(head.id,"fmt ",4)) /* read wave format */
		{
			if(!riff_read_fmt(fmt->source,&head,fmt))
				break;
			chunk_count++;
			continue;
		} else if(!memcmp(head.id,"data",4))
		{
			/* save parameters of data block */
			source_pos=fileTell(fmt->source);
			fmt->num_bytes=head.len;
			/* Skip data block; data will be read later */
			fileSeek(fmt->source,head.len,SEEK_CUR);
			chunk_count++;
			continue;
		} else /* skip other chunks */
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_TRACE,
				"waveOpenRead('%s'): unknown chunk '%c%c%c%c'; ignored.\n",
				name,head.id[0],head.id[1],head.id[2],head.id[3]);
#endif		/* CHECKED_BUILD */
			fileSeek(fmt->source,head.len,SEEK_CUR);
		}
	}
	if(chunk_count<2)
	{
		/* we are failed */
		fileClose(fmt->source);
		memFree(fmt);
		return NULL;
	}
	/* adjust number of samples and data size */
	fmt->num_samples=fmt->num_bytes/fmt->block_align;
	/* Seek to start of data */
	fileSeek(fmt->source,source_pos,SEEK_SET);
#ifdef	CHECKED_BUILD
	/* Store file name in fmt */
	strCopy(fmt->wave_name,name,_MAX_PATH);
#endif	/* CHECKED_BUILD */
	return (wave_handle_t*)fmt;
}

/*
**	Reads next block of data from file
*/
u32_t	V_GLOBAL
wavRead(wave_handle_t handle,void *dest,u32_t len)
{
	u32_t rlen;
	wave_data_t *fmt;
	fmt=(wave_data_t*)libvIsHandle(handle,"wavRead",
		wave_object,WAVE_MAGIC);
	if(!fmt)
		return 0;
	/* Was wave file opened for input ? */
	if(!fmt->source)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_WARNING,"wavRead(): file '%s' "
			"was opened for output.\n",fmt->wave_name);
#endif	/* CHECKED_BUILD */
		return 0;
	}
	/* Remaining length of data */
	rlen=fmt->num_bytes-fmt->transferred;
	if(rlen<len)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_WARNING,
			"wavRead(): only %d bytes of data available.\n",rlen);
#endif	/* CHECKED_BUILD */
		len=rlen;
	}
	/* Read block */
	rlen=fileRead(fmt->source,dest,len);
	fmt->transferred+=rlen;
	return rlen;
}

/*
**	Closes wave file, frees all resources.
*/
vboolean	V_GLOBAL
wavClose(wave_handle_t handle)
{
	wave_data_t *fmt;
	fmt=(wave_data_t*)libvIsHandle(handle,"wavClose",
		wave_object,WAVE_MAGIC);
	if(!fmt)
		return false;
	if(fmt->dest)
	{
		/* File was opened for output: now we must write some bytes
		 * to header. */
		fileSeek(fmt->dest,0x04,SEEK_SET);
		/* Write size of main chunk */
		filePutLittleLong(fmt->dest,0x24+fmt->transferred);
		fileWrite(fmt->dest,"WAVEfmt ",8);
		/* Write size of 'data' chunk */
		fileSeek(fmt->dest,0x28,SEEK_SET);
		filePutLittleLong(fmt->dest,fmt->transferred);
		fileClose(fmt->dest);
		memFree(fmt);
		return true;
	}
	if(fmt->source)
		fileClose(fmt->source);
	if(fmt->wave_data)
		memFree(fmt->wave_data);
	memFree(fmt);
	return true;
}

/*
**	Starts writing of wav file. Returned handle must be later 
**	passed to wavClose().
*/
wave_handle_t	V_GLOBAL
wavOpenWrite(char *name,u32_t num_channels,u32_t sampling_frequency,
	u32_t bits_per_sample)
{
	wave_data_t *fmt;
	ALLOC_OBJECT(fmt,wave_data_t,0,WAVE_MAGIC);
	fmt->dest=fileOpenWrite(name);
	if(!fmt->dest)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"waveOpenWrite('%s'): can't create file.\n",
			dbs(name));
#endif	/* CHECKED_BUILD */
		memFree(fmt);
		return NULL;
	}
	fmt->num_channels=(u16_t)num_channels;
	fmt->sampling_frequency=(u32_t)sampling_frequency;
	fmt->bits_per_sample=(u16_t)bits_per_sample;
	fmt->block_align=
		(u16_t)(fmt->num_channels * (fmt->bits_per_sample >> 3));
	fmt->bytes_per_second=fmt->sampling_frequency * fmt->block_align;
	/* Write main header */
	fileWrite(fmt->dest,"RIFF",4);
	/* Now we cant determine all data in header, so it 
	 * must be fixed later */
	filePutLittleLong(fmt->dest,0x24);
	/* Write typeid and id of 'fmt ' chunk */
	fileWrite(fmt->dest,"WAVEfmt ",8);
	/* Write size of fmt chunk */
	filePutLittleLong(fmt->dest,0x10);
	/* Write wave parameters */
	filePutLittleShort(fmt->dest,WAVE_FORMAT_PCM);
	filePutLittleShort(fmt->dest,fmt->num_channels);
	filePutLittleLong(fmt->dest,fmt->sampling_frequency);
	filePutLittleLong(fmt->dest,fmt->bytes_per_second);
	filePutLittleShort(fmt->dest,fmt->block_align);
	filePutLittleShort(fmt->dest,fmt->bits_per_sample);
	/* Write id of 'data' chunk */
	fileWrite(fmt->dest,"data",4);
	/* Write size of 'data' chunk */
	filePutLittleLong(fmt->dest,0);
#ifdef	CHECKED_BUILD
	/* Store file name in fmt */
	strCopy(fmt->wave_name,name,_MAX_PATH);
#endif	/* CHECKED_BUILD */
	return (wave_handle_t*)fmt;
}

u32_t	V_GLOBAL
wavWrite(wave_handle_t handle,void *src,u32_t len)
{
	u32_t wlen;
	wave_data_t *fmt;
	fmt=(wave_data_t*)libvIsHandle(handle,"wavWrite",
		wave_object,WAVE_MAGIC);
	if(!fmt)
		return 0;
	/* Was wave file opened for output ? */
	if(!fmt->dest)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_WARNING,"wavRead(): file '%s' "
			"was opened for input.\n",fmt->wave_name);
#endif	/* CHECKED_BUILD */
		return 0;
	}
	wlen=fileWrite(fmt->dest,src,len);
	fmt->transferred+=wlen;
	return wlen;
}

/*
**	Reads all wave data.
**	Pointer to data can be obtained by wavGetData().
*/
wave_handle_t	V_GLOBAL
wavLoad(char *name)
{
	wave_data_t *src;
	src=(wave_data_t*)wavOpenRead(name);
	if(!src)
		return NULL;
	/* Now allocate memory and read wave data */
	src->wave_data=memAlloc(src->num_bytes);
	wavRead((wave_handle_t)src,src->wave_data,src->num_bytes);
	/* We woudn't call waveClose(); instead we just close input handle */
	fileClose(src->source);
	src->source=0;
	return (wave_handle_t)src;
}

u32_t	V_GLOBAL
wavNumChannels(wave_handle_t handle)
{
	wave_data_t *fmt;
	fmt=(wave_data_t*)libvIsHandle(handle,"wavNumChannels",
		wave_object,WAVE_MAGIC);
	if(!fmt)
		return 0;
	return fmt->num_channels;
}

u32_t	V_GLOBAL
wavSamplingFrequency(wave_handle_t handle)
{
	wave_data_t *fmt;
	fmt=(wave_data_t*)libvIsHandle(handle,"wavSamplingFrequency",
		wave_object,WAVE_MAGIC);
	if(!fmt)
		return 0;
	return fmt->sampling_frequency;
}

u32_t	V_GLOBAL
wavBitsPerSample(wave_handle_t handle)
{
	wave_data_t *fmt;
	fmt=(wave_data_t*)libvIsHandle(handle,"wavBitsPerSample",
		wave_object,WAVE_MAGIC);
	if(!fmt)
		return 0;
	return fmt->bits_per_sample;
}

u32_t	V_GLOBAL
wavNumSamples(wave_handle_t handle)
{
	wave_data_t *fmt;
	fmt=(wave_data_t*)libvIsHandle(handle,"wavNumSamples",
		wave_object,WAVE_MAGIC);
	if(!fmt)
		return 0;
	return fmt->num_samples;
}

/*	!!! Only for handles returned by wavLoad(). */
void_p	V_GLOBAL
wavData(wave_handle_t handle)
{
	wave_data_t *fmt;
	fmt=(wave_data_t*)libvIsHandle(handle,"wavData",
		wave_object,WAVE_MAGIC);
	if(!fmt)
		return NULL;
	return fmt->wave_data;
}

/*</pre></strong></body></html>*/
/*EOF*/
