/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../include/v_lib.h"
#include	"../source/v_mt.h"
#include	"v_os.h"

#ifdef	__WIN32__

/*
 *	These procs returns metafile extents.
 */
u32_t	V_GLOBAL
wmfGetMaxX(void)
{
	return GetThreadData()->meta_x_extent;
}

u32_t	V_GLOBAL
wmfGetMaxY(void)
{
	return GetThreadData()->meta_y_extent;
}

/*
 *	This changes metafile extents
 */
void	V_GLOBAL
wmfSetExtents(u32_t x_ext,u32_t y_ext)
{
	thread_data_t *td;
	td=GetThreadData();
	td->meta_x_extent=x_ext;
	td->meta_y_extent=y_ext;
}

/*
 *	wmfCreate() - creates metafile DC.
 *	Input:
 *		mmWidth, mmHeight - width and height in millimeters.
 *	Output:
 *		Allocated and initialized wmf_t struct or NULL on error.
 *	Notes.
 *		1. If at least one of mmWidth and mmHeight parameters is zero,
 *	extents of metafile set equals to size of A4 sheet.
 *		3. Logical sizes of metafile are always META_X_EXTENT by 
 *			META_Y_EXTENT units.
 *		2. Win32 only: if name is "//./clipboard", subsequent call
 *	to wmfClose() places data into system clipboard.
 */
wmf_p	V_GLOBAL
wmfCreate(char *name,u32_t mmWidth,u32_t mmHeight)
{
	HDC dc;
	wmf_t *mf;
	u32_t w, h, x_ext, y_ext;
	thread_data_t *td;
	if(!name || !name[0])
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"wmfCreate('%s',%d,%d): incorrect file name.\n",
			dbs(name),mmWidth,mmHeight);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	td=GetThreadData();
	if(!td->meta_x_extent)
		td->meta_x_extent=1000;
	if(!td->meta_y_extent)
		td->meta_y_extent=1000;
	x_ext=td->meta_x_extent;
	y_ext=td->meta_y_extent;
	dc=CreateMetaFile(NULL);
	if(!dc)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"wmfCreate('%s',%d,%d): CreateMetafile() failed (code %d).\n",
			dbs(name),mmWidth,mmHeight,GetLastError());
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	if(!mmWidth || !mmHeight)
	{
		/* Set width and height to fit picture in A4 sheet */
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_WARNING,
			"wmfCreate('%s',%d,%d): using default size (A4).\n",
			dbs(name),mmWidth,mmHeight);
#endif	/* CHECKED_BUILD */
		w=10000;
		h=6700;
	}
	else
	{
		/* Convert millimeters to inches * 1000 */
		w=(u32_t)(mmWidth * 1000 / 25.2);
		h=(u32_t)(mmHeight * 1000 / 25.2);
	}
	SetMapMode(dc,MM_ANISOTROPIC);
	SetWindowExtEx(dc,x_ext,y_ext,NULL);
	SetWindowOrgEx(dc,0,0,NULL);
	mf=(wmf_t*)malloc(sizeof(wmf_t));
	v_zero(mf,sizeof(wmf_t));
	mf->dc=dc;
	mf->width=(u32_t)(w * 25.2 / 1000);
	mf->height=(u32_t)(h * 25.2 / 1000);
	mf->in1000_width=w;
	mf->in1000_height=h;
	strCopy(mf->name,name,_MAX_PATH);
	return mf;
}

void_p		V_GLOBAL
wmfGetDevice(wmf_t *mf)
{
	V_ASSERT(mf);
	return mf->dc;
}

/*		Placeable metafile signature */
#define	PMPHKEY			0x9AC6CDD7

#ifdef	_MSC_VER
#pragma pack(push)
#endif	/* _MSC_VER */

#pragma pack(2)
/*	Placeable metafile header */
typedef struct pmfh_s
{
	u32_t	key;
	u16_t	emptywmf;
	u16_t	left, top, right, bottom;
	u16_t	inch;
	u32_t	reserved;
	u16_t	checksum;
}	pmfh_t;

#ifdef	_MSC_VER
#pragma pack(push)
#else
#pragma pack(4)
#endif	/* _MSC_VER */

/*	Computes placeable metafile checksum */
u16_t	V_LOCAL
pmfhCheckSum(pmfh_t *pmfh)
{
	int i;
	u16_t sum=0;
	u16_t *ptr=(u16_t*)pmfh;
	for(i=0;i<10;i++)
		sum^=ptr[i];
	return sum;
}

/*
 *	wmfClose() - saves metafile.
 *	Input:
 *		Address of wmf_t object.
 *	Output:
 *		true if success, false otherwise; wmf_t object cleaned and freed.
 */
vboolean	V_GLOBAL
wmfClose(wmf_t *mf)
{
	u32_t	size;
	u8_t	*bits;
	pmfh_t	aph;
	file_t	out;
	HMETAFILE hmf;
	if(!mf)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"wmfClose(%X): invalid handle of wmf_t object.\n",mf);
#endif	/* CHECKED_BUILD */
		return false;
	}
	hmf=CloseMetaFile(mf->dc);
	if(!hmf)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"wmfClose(%X): CloseMetaFile() failed (code %d).\n",mf,
			GetLastError());
#endif	/* CHECKED_BUILD */
		return false;
	}
	if(!strcmp(mf->name,"//./clipboard") || !strcmp(mf->name,"\\\\.\\clipboard"))
	{
		/*	Save data to clipboard. Tell me, why VCL provides no way
		 *	(at least, no SIMPLE way) to do it ?
		 */
		HGLOBAL hmfp;
		METAFILEPICT *mfp;
		if(!OpenClipboard(NULL))
		{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"wmfClose(%X): can't OpenClipboard() (code %d).\n",mf,
				GetLastError());
#endif	/* CHECKED_BUILD */
			DeleteMetaFile(hmf);
			free(mf);
			return false;
		}
		/* Allocate piece of GMEM_DDESHARE'd memory */
		hmfp=GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE,sizeof(METAFILEPICT));
		mfp=(METAFILEPICT*)GlobalLock(hmfp);
		V_ASSERT(mfp); 
		v_zero(mfp,sizeof(METAFILEPICT));
		EmptyClipboard();
		/* Dimensions in MM_HIMETRIC units */
		mfp->mm=MM_ANISOTROPIC;
		mfp->xExt=mf->width*100;
		mfp->yExt=mf->height*100;
		mfp->hMF=hmf;
		/* Release HGLOBAL */
		GlobalUnlock(hmfp);
		/* Pass it to system */
		SetClipboardData(CF_METAFILEPICT,hmfp);
		/* We are done ! */
		CloseClipboard();
		DeleteMetaFile(hmf);
		free(mf);
		return true;
	}
	/* Save contents to file */
	size=GetMetaFileBitsEx(hmf,0,NULL);
	if(!size)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"wmfClose(%X): GetMetaFileBitsEx() failed (code %d).\n",mf,
			GetLastError());
#endif	/* CHECKED_BUILD */
		DeleteMetaFile(hmf);
		return false;
	}
	bits=(u8_t*)malloc(size);
	GetMetaFileBitsEx(hmf,size,bits);
	aph.key=PMPHKEY;
	aph.emptywmf=0;
	aph.left=0;
	aph.top=0;
	aph.right=(u16_t)mf->in1000_width;
	aph.bottom=(u16_t)mf->in1000_height;
	aph.inch=1000;
	aph.reserved=0;
	aph.checksum=pmfhCheckSum(&aph);
	out=fileOpenWrite(mf->name);
	if(!out)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"wmfClose('%s'): can't create output file.\n",dbs(mf->name));
#endif	/* CHECKED_BUILD */
		DeleteMetaFile(hmf);
		free(bits);
		free(mf);
		return false;
	}
	filePutLittleLong(out,aph.key);
	filePutLittleShort(out,aph.emptywmf);
	filePutLittleShort(out,aph.left);
	filePutLittleShort(out,aph.top);
	filePutLittleShort(out,aph.right);
	filePutLittleShort(out,aph.bottom);
	filePutLittleShort(out,aph.inch);
	filePutLittleLong(out,aph.reserved);
	filePutLittleShort(out,aph.checksum);
	fileWrite(out,bits,size);
	fileClose(out);
	DeleteMetaFile(hmf);
	free(bits);
	free(mf);
	return true;
}

/*
 *	GDI op seems to be broken ...
 */
vboolean V_GLOBAL
wmfTextExtent(wmf_t *mf,char *str,u32_t *cx,u32_t *cy)
{
	HDC ref_dc;
	HFONT old_font, real_font;
	BOOL retval;
	SIZE sz;
	ref_dc=GetDC(NULL);
	/* begin hack */
	real_font=SelectObject(mf->dc,GetStockObject(SYSTEM_FONT));
	SelectObject(mf->dc,real_font);
	/* end hack: we obtained current dc's font handle ! */
	old_font=SelectObject(ref_dc,real_font);
	retval=GetTextExtentPoint32(ref_dc,str,strlen(str),&sz);
	SelectObject(ref_dc,old_font);
	ReleaseDC(NULL,ref_dc);
	if(cx)
		*cx=sz.cx;
	if(cy)
		*cy=sz.cy;
	return retval ? true : false;
}

/*
 *	Changes font for DC of metafile.
 *	Input:
 *		mf - wmf_t object;
 *		char_height - height of character (in metafile's logical
 *			units; 100 mean 1/10 of metafile's height etc.);
 *		deg_ori - orientation in degrees (counter-clockwise);
 *		weight - weight from 0 to 900;
 *		italic - if 1, italic font used;
 *		face - font face name.
 */
void	V_GLOBAL
wmfChangeFont(wmf_t *mf,int char_height,int deg_ori,int weight,int italic,
	char *face)
{
	HFONT hnew;
	if(weight<0)
		weight=0;
	else
		if(weight>900)
			weight=900;
	hnew=CreateFont(char_height,0,deg_ori*10,deg_ori*10,weight,italic,0,0,
#ifndef	ALWAYS_ENGLISH
		RUSSIAN_CHARSET,
#else	/* ALWAYS_ENGLISH */
		DEFAULT_CHARSET,
#endif	/* ALWAYS_ENGLISH */
		OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,DEFAULT_PITCH,face ? face : "Courier New");
	if(hnew)
	{
		SelectObject(mf->dc,hnew);
		if(mf->hfont)
			DeleteObject(mf->hfont);
		mf->hfont=hnew;
	}
}

/*
 *	OpenGL feedback connectivity
 */
#define			FDMF_MAGIC			0xFD062241

typedef	struct	fd_mf_s
{
	u32_t		magic;
	LOGBRUSH	lbrush;
	LOGPEN		lpen;
	HPEN		null_pen;
	HBRUSH		null_brush;
	color4b_t	colors[3];
	vec2_t		vertices[3];
	POINT		pts[3];
	u32_t		vertex_count;
	glw_t		offscreen_window;
	void		*opengl_feedback_buffer;
	u32_t		max_elements;
	u32_t		stored_elements;
	matrix2_t	R;
	vec2_t		T;
	wmf_t		*mf;
}	fd_mf_t, *fd_mf_p;

u32_t V_LOCAL
average_rgb_2(color4b_t c1,color4b_t c2)
{
	int r,g,b;
	r=c1[0]+c2[0];
	g=c1[1]+c2[1];
	b=c1[2]+c2[2];
	return RGB((u8_t)(r >> 1),(u8_t)(g >> 1),(u8_t)(b >> 1));
}

u32_t V_LOCAL
average_rgb_3(color4b_t c123[])
{
	int r=c123[0][0],g=c123[0][1],b=c123[0][2];
	int i;
	for(i=1;i<3;i++)
	{
		r+=c123[i][0];
		g+=c123[i][1];
		b+=c123[i][2];
	}
	return RGB((u8_t)(r / 3),(u8_t)(g / 3),(u8_t)(b / 3));
}

#include "argsused.h"
V_LOCAL	void	V_CALL 
fd_mf_point(void *dev,vec3_t v,color4b_t c,vec2_t uv)
{
	fd_mf_t *mf=(fd_mf_t *)dev;
	vec2_t vt;
	mat2VecMul(vt,mf->R,v);
	vec2Add(vt,vt,mf->T);
	SetPixel(mf->mf->dc,(int)vt[0],(int)vt[1],RGB(c[0],c[1],c[2]));
}

#include "argsused.h"
V_LOCAL	void	V_CALL
fd_mf_line(void *dev,vec3_t v1,color4b_t c1,vec2_t uv1,
	vec3_t v2,color4b_t c2,vec2_t uv2)
{
	fd_mf_t *mf=(fd_mf_t *)dev;
	HDC dc=mf->mf->dc;
	vec2_t vt1,vt2;
	mat2VecMul(vt1,mf->R,v1);
	vec2Add(vt1,vt1,mf->T);
	mat2VecMul(vt2,mf->R,v2);
	vec2Add(vt2,vt2,mf->T);
	mf->lpen.lopnColor=average_rgb_2(c1,c2);
	SelectObject(dc,CreatePenIndirect(&mf->lpen));
	MoveToEx(dc,(int)vt1[0],(int)vt1[1],NULL);
	LineTo(dc,(int)vt2[0],(int)vt2[1]);
	DeleteObject(SelectObject(dc,mf->null_pen));
}

#include "argsused.h"
V_LOCAL	void	V_CALL
fd_mf_begin_polygon(void *dev,int n)
{
	((fd_mf_t *)dev)->vertex_count=0;
}

#include "argsused.h"
V_LOCAL	void	V_CALL
fd_mf_polygon_vertex(void *dev,vec3_t v,color4b_t c,vec2_t uv)
{
	fd_mf_t *mf=(fd_mf_t *)dev;
	vec2_t vt;
	rgbaByteCopy(mf->colors[mf->vertex_count],c);
	mat2VecMul(vt,mf->R,v);
	vec2Add(vt,vt,mf->T);
	vec2Copy(mf->vertices[mf->vertex_count],vt);
	mf->vertex_count++;
	if(mf->vertex_count==3)
	{
		/* Draw next triangle of triangle fan */
		int i;
		HDC dc;
		dc=mf->mf->dc;
		mf->lbrush.lbColor=average_rgb_3(mf->colors);
		SelectObject(dc,CreateBrushIndirect(&mf->lbrush));
		for(i=0;i<3;i++)
		{
			mf->pts[i].x=(LONG)mf->vertices[i][0];
			mf->pts[i].y=(LONG)mf->vertices[i][1];
		}
		Polygon(dc,mf->pts,3);
		DeleteObject(SelectObject(dc,mf->null_brush));
		mf->vertex_count=1;
	}
}

#include "argsused.h"
V_LOCAL	void	V_CALL
fd_mf_pass_through(void *dev,vec_t v)
{
	/* stub */
	return;
}

V_LOCAL	gl_tl_callbacks_t	fd_mf_cbs=
{
	fd_mf_point,
	fd_mf_line,
	fd_mf_begin_polygon,
	fd_mf_polygon_vertex,
	fd_mf_pass_through
};

fd_mf_p		V_LOCAL
fd_mf_begin(char *name,u32_t mm_width,u32_t mm_height)
{
	wmf_p mf;
	fd_mf_p fmf;
	mf=wmfCreate(name,mm_width,mm_height);
	if(!mf)
		return NULL;
	fmf=(fd_mf_t*)malloc(sizeof(fd_mf_t));
	v_zero(fmf,sizeof(fd_mf_t));
	fmf->magic=FDMF_MAGIC;
	fmf->mf=mf;
	fmf->null_pen=GetStockObject(NULL_PEN);
	fmf->null_brush=GetStockObject(HOLLOW_BRUSH);
	fmf->lbrush.lbStyle=BS_SOLID;
	fmf->lpen.lopnStyle=PS_SOLID;
	fmf->lpen.lopnWidth.x=1;
	fmf->lpen.lopnWidth.y=1;
	SelectObject(mf->dc,fmf->null_pen);
	SelectObject(mf->dc,fmf->null_brush);
	return fmf;
}

/*
 *	wmfCreateOffscreenWindow() - creates 'Metafile' OpenGL context.
 *	Input:
 *		name - name of WMF file to save;
 *		mmWidth, mmHeight - dimensions of metafile in millimeters;
 *		feedback_buffer_hint - appr. size of OpenGL feedback buffer
 *			in BYTES.
 *	Output:
 *		Feedback metafile handle or NULL.
 */
void_p	V_GLOBAL
wmfCreateOffscreenWindow(char *name,u32_t mmWidth,u32_t mmHeight,
	u32_t feedback_buffer_hint)
{
	fd_mf_t *out;
	glw_t w;
	vec_t x_ext, y_ext;
	thread_data_t *td;
	int pw = 16, ph = 16;	/* dummy window - takes a little RAM */
	w=vglCreateMappedWindow(pw,ph);
	if(!w)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"wmfCreateOffscreenWindow('%s',%d,%d,%d): "
			"vglCreateMappedWindow() failed.\n",dbs(name),
			mmWidth,mmHeight,feedback_buffer_hint);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	/* For feedback: not less than 1 Meg of RAM ! */
	/*
	if(feedback_buffer_hint < 1024*1024)
		feedback_buffer_hint = 1024*1024;
	*/
	wmfSetExtents(10000,10000);
	out=fd_mf_begin(name,mmWidth,mmHeight);
	if(!out)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"wmfCreateOffscreenWindow('%s',%d,%d,%d): "
			"wmfCreate() failed.\n",dbs(name),
			mmWidth,mmHeight,feedback_buffer_hint);
#endif	/* CHECKED_BUILD */
	}
	out->max_elements=feedback_buffer_hint/sizeof(GLfloat);
	out->opengl_feedback_buffer=vglBeginFeedback(out->max_elements);
	if(!out->opengl_feedback_buffer)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"wmfCreateOffscreenWindow('%s',%d,%d,%d): "
			"vglBeginFeedback() failed.\n",dbs(name),
			mmWidth,mmHeight,feedback_buffer_hint);
#endif	/* CHECKED_BUILD */
		DeleteMetaFile(CloseMetaFile(out->mf->dc));
		free(out->mf);
		free(out);
		return NULL;
	}
	/* looks paranoidal ... */
	td=GetThreadData();
	x_ext=(vec_t)td->meta_x_extent;
	y_ext=(vec_t)td->meta_y_extent;
	out->R[0][0]=(vec_t)(x_ext / pw);
	out->R[1][1]=-(vec_t)(y_ext / ph);
	out->R[0][1]=out->R[1][0]=0;
	out->T[0]=0;
	out->T[1]=y_ext;
	out->offscreen_window=w;
	return out;
}

/*
 *	Saves metafile and frees all objects.
 */	
vboolean	V_GLOBAL
wmfDestroyWindow(void *fdmf)
{
	vboolean retval;
	fd_mf_t *fmf=(fd_mf_t*)fdmf;
	if(!fmf || fmf->magic!=FDMF_MAGIC)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"wmfDestroyWindow(%X): wrong handle.\n",fmf);
#endif	/* CHECKED_BUILD */
		return false;
	}
	retval=vglEndFeedback(&fmf->stored_elements);
	if(retval && fmf->stored_elements < fmf->max_elements)
	{
		vglDecodeFeedback(&fd_mf_cbs,(void*)fmf,
			fmf->opengl_feedback_buffer,fmf->stored_elements);
		wmfClose(fmf->mf);
	}
	else
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"wmfDestroyWindow(%X): vglEndFeedback() failed.\n",fmf);
#endif	/* CHECKED_BUILD */
		retval=false;
	}
	free(fmf->opengl_feedback_buffer);
	vglDestroyWindow(fmf->offscreen_window);
	free(fmf);
	return retval;
}

#elif	defined(__UNIX__)	/* __WIN32__ */
#error	TODO: Implement UNIX Metafile I/O.
#elif	defined(__DOS32__)	/* defined(__UNIX__) */
#error	TODO: Implement DOS Metafile I/O.
#endif	/* defined(__DOS32__) */

/*</pre></strong></body></html>*/
/*EOF*/
