/*	Mimmap'ed RGBA image header */
typedef struct miphead_s
{
	u32_t	width, height;		/* Dimensions of level # 0 */
	u32_t	num_levels;			/* Number of levels */
	u8_t	**level;			/* Array of pointers to levels */
}	miphead_t, *miphead_p;

#ifdef 	__cplusplus
extern	"C" {
#endif	/* __cplusplus */
	/*	Loading and saving summed-area filtered images 
		(TODO: add support for loading Quake *.w8 files) */
	miphead_p	V_GLOBAL	mipLoad(char *mipfile);
	vboolean	V_GLOBAL	mipSave(char *mipfile,miphead_t *mip);
	/*	This performs summed-area filtering of source image */
	miphead_p	V_GLOBAL	mipBuild(u8_t *pixels,
		u32_t width,u32_t height,u32_t levels);
#ifdef __cplusplus
}
#endif	/* __cplusplus */

/*
 *	mipAlloc()
 */
 
miphead_p	V_LOCAL
mipAlloc(u32_t width,u32_t height,u32_t levels)
{
	u32_t size,i,w,h;
	u32_t sizes[MAX_MIP_LEVELS];
	miphead_p head;
	if(levels>MAX_MIP_LEVELS || !width || !height || 
		!IS_POWER_TWO(width) || !IS_POWER_TWO(height))
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"mipAlloc(%d,%d,%d): incorrect parameter(s).\n",
			width,height,levels);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	w=width;
	h=height;
	size=sizeof(miphead_t)+levels*sizeof(u8_t*);
	for(i=0;i<levels;i++)
	{
		sizes[i]=4*sizeof(u8_t)*w*h;
		size+=sizes[i];
		w>>=1;
		h>>=1;
	}
	head=(miphead_p)malloc(size);
	head->width=width;
	head->height=height;
	head->num_levels=levels;
	head->level=(u8_t**)(head+1);
	head->level[0]=(u8_t*)(head->level)+levels*sizeof(u8_t*);
	for(i=1;i<levels;i++) 
		head->level[i]=head->level[i-1]+sizes[i-1];
	return head;
}
		
void	V_LOCAL
DownSampleImage(u8_t *dest,u8_t *src,u32_t w,u32_t h)
{
	u32_t x,y,ad_src,w_s=w;
	u32_t r[4],g[4],b[4],a[4];
	u32_t offset;
	offset=0;
	if(w<2 || h<2 || !IS_POWER_TWO(w) || !IS_POWER_TWO(h) || !dest || !src)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"DownSampleImage(%X,%X,%d,%d): incorrect parameter(s).\n",
			dest,src,w,h);
#endif	/* CHECKED_BUILD */
		return;
	}
	w>>=1;
	h>>=1;
	for(y=0;y<h;y++)
	{ 
		for(x=0;x<w;x++)
		{
			ad_src = (x<<1) + (y<<1) * w_s;
			r[0]=src[(ad_src<<2)];
			g[0]=src[(ad_src<<2)+1];
			b[0]=src[(ad_src<<2)+2];
			a[0]=src[(ad_src<<2)+3];
			
			r[1]=src[((ad_src+1)<<2)];
			g[1]=src[((ad_src+1)<<2)+1];
			b[1]=src[((ad_src+1)<<2)+2];
			a[1]=src[((ad_src+1)<<2)+3];
			
			r[2]=src[((ad_src+w_s)<<2)];
			g[2]=src[((ad_src+w_s)<<2)+1];
			b[2]=src[((ad_src+w_s)<<2)+2];
			a[2]=src[((ad_src+w_s)<<2)+3];
			
			r[3]=src[((ad_src+w_s+1)<<2)];
			g[3]=src[((ad_src+w_s+1)<<2)+1];
			b[3]=src[((ad_src+w_s+1)<<2)+2];
			a[3]=src[((ad_src+w_s+1)<<2)+3];
			
			dest[offset++]=(u8_t)((r[0]+r[1]+r[2]+r[3])>>2);
			dest[offset++]=(u8_t)((g[0]+g[1]+g[2]+g[3])>>2);
			dest[offset++]=(u8_t)((b[0]+b[1]+b[2]+b[3])>>2);
			dest[offset++]=(u8_t)((a[0]+a[1]+a[2]+a[3])>>2);
		}
	}
}

miphead_p	V_GLOBAL
mipBuild(u8_t *pixels,u32_t width,u32_t height,u32_t levels)
{
	u32_t i;
	miphead_p mip;
	if(!pixels || width<=0 || height<=0 || levels<=0)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"mipBuild(%X,%d,%d,%d): incorrect parameter(s).\n",
			pixels,width,height,levels);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	mip=mipAlloc(width,height,levels);
	if(!mip)
	{
		/* mipAlloc() already reported error; stay quiet */
		return NULL;
	}
	v_copydw(mip->level[0],pixels,width*height);
	for(i=1;i<levels;i++)
	{
		DownSampleImage(mip->level[i],mip->level[i-1],width,height);
		width>>=1;
		height>>=1;
	}
	return mip;
}

#define	MIP_MAGIC_STRING	"RGBM"

vboolean V_GLOBAL
mipSave(char *mipfile,miphead_t *mip)
{
	u32_t i,w,h,ml;
	file_t f;
	if(!mipfile || !mip)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"mipSave('%s',%X): incorrect parameter(s).\n",
			dbs(mipfile),mip);
#endif	/* CHECKED_BUILD */
		return false;
	}
	f=fileOpenWrite(mipfile);
	if(!f)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"mipSave('%s'): can't create file.\n",mipfile);
#endif	/* CHECKED_BUILD */
		return false;
	}
	ml=strlen(MIP_MAGIC_STRING);
	if(ml)
		fileWrite(f,MIP_MAGIC_STRING,ml);
	filePutLittleLong(f,mip->width);
	filePutLittleLong(f,mip->height);
	filePutLittleLong(f,mip->num_levels);
	w=mip->width;
	h=mip->height;
	for(i=0;i<mip->num_levels;i++)
	{
		fileWrite(f,mip->level[i],w*h<<2);
		w>>=1;
		h>>=1;
	}
	fileClose(f);
    return true;
}

miphead_p	V_GLOBAL
mipLoad(char *mipfile)
{
	u32_t w,h,levels,i,ml;
	miphead_t *mip;
	file_t f;
	char *buf;
	if(!mipfile)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"mipLoad('%s'):incorrect parameter.\n",
			dbs(mipfile));
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	f=fileOpenRead(mipfile);
	if(!f)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"mipLoad('%s'): file not found.\n",mipfile);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	ml=strlen(MIP_MAGIC_STRING);
	buf=(char*)malloc(ml);
	fileRead(f,buf,ml);
	if(strncmp(buf,MIP_MAGIC_STRING,ml))
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"mipLoad('%s'): magic wrong, file corrupted.\n",mipfile);
#endif	/* CHECKED_BUILD */
		fileClose(f);
		return NULL;
	}
	w=fileGetLittleLong(f);
	h=fileGetLittleLong(f);
	levels=fileGetLittleLong(f);
	mip=mipAlloc(w,h,levels);
	if(!mip)
	{
		/* mipAlloc() already reported error; stay quiet */
		return NULL;
	}
	for(i=0;i<levels;i++)
	{
		fileRead(f,mip->level[i],w*h*4);
		w>>=1;
		h>>=1;
	}
	fileClose(f);
	return mip;
}

