//	v_prof.c

#ifdef DELPHI_COMPAT

typedef void (V_CALL *delphi_planar_func)(mvec_t *x,mvec_t *y,mvec_t *res);
typedef void (V_CALL *delphi_spatial_func)(mvec_t *x,mvec_t *y,mvec_t *z,mvec_t *res);

/*
		NOTE. This data not moved to TLS due to possible removing v_pprof from
		V Library into standalone program.
*/

static delphi_planar_func   planar_func=NULL;
static delphi_spatial_func  spatial_func=NULL;

mvec_t	V_LOCAL	V_CALL
planar_thk(mvec_t x,mvec_t y)
{
	mvec_t res;
	planar_func(&x,&y,&res);
	return res;
}

mvec_t	V_LOCAL	V_CALL
spatial_thk(mvec_t x,mvec_t y,mvec_t z)
{
	mvec_t res;
	spatial_func(&x,&y,&z,&res);
	return res;
}

#ifdef __cplusplus
extern "C" {
#endif
	tripatch_p      V_GLOBAL ppGridFunctionTHK(vec_p l,vec_p t,vec_p r,vec_p b,
		int hsplit,int vsplit,patch_settings_p settings,
		delphi_planar_func planar,int basenum);
	tripatch_p      V_GLOBAL ppGridSheffePolynomTHK(vec3_t center,vec_p radius,
		vec_p angle0,int split,patch_settings_p settings,vec3_t *gibbs_abc,
		delphi_spatial_func sheffe,int basenum);
	patchprofile_p  V_GLOBAL ppCreateRectangularTHK(vec_p l,vec_p t,vec_p r,vec_p b,
		delphi_planar_func func,int zsplits,int baseplane);
	patchprofile_p  V_GLOBAL ppCreateTriangularTHK(vec3_t center,vec_p radius,
		vec_p angle0,vec3_t *gibbs_abc,delphi_spatial_func sheffe,int zsplits,
		int baseplane);
#ifdef __cplusplus
}
#endif
 
 
tripatch_p V_GLOBAL ppGridFunctionTHK(vec_p l,vec_p t,vec_p r,vec_p b,
	int hsplit,int vsplit,patch_settings_p settings,delphi_planar_func planar,
	int basenum)
{
	planar_func=planar;
	return ppGridFunction(*l,*t,*r,*b,hsplit,vsplit,settings,planar_thk,basenum);
}

tripatch_p V_GLOBAL ppGridSheffePolynomTHK(vec3_t center,vec_p radius,vec_p angle0,
	int split,patch_settings_p settings,vec3_t *gibbs_abc,delphi_spatial_func sheffe,
	int basenum)
{
	spatial_func=sheffe;
	return ppGridSheffePolynom(center,*radius,*angle0,split,settings,gibbs_abc,
		spatial_thk,basenum);
}

patchprofile_p V_GLOBAL ppCreateRectangularTHK(vec_p l,vec_p t,vec_p r,vec_p b,
	delphi_planar_func func,int zsplits,int baseplane)
{
	patchprofile_p pp;
	planar_func=func;
	pp=ppCreateRectangular(*l,*t,*r,*b,planar_thk,zsplits,baseplane);
	if(pp)
		pp->planar=(planar_func_t)func;
	return pp;
}

patchprofile_p V_GLOBAL ppCreateTriangularTHK(vec3_t center,vec_p radius,vec_p angle0,
	vec3_t *gibbs_abc,delphi_spatial_func sheffe,int zsplits,int baseplane)
{
	patchprofile_p pp;
	spatial_func=sheffe;
	pp=ppCreateTriangular(center,*radius,*angle0,gibbs_abc,spatial_thk,zsplits,baseplane);
	if(pp)
		pp->spatial=(spatial_func_t)sheffe;
	return pp;
}

#endif /* DELPHI_COMPAT */

//	v_matrix.c

#ifdef	DELPHI_COMPAT

/*
 *	matDetermTHK
 */
#ifdef	__cplusplus
extern "C"
#endif	/* __cplusplus */
void	V_GLOBAL
matDetermTHK(matrix_t *m,vboolean keepon,mvec_t *res)
{
	if(res)
		*res=matDeterm(m,keepon);
}

#endif	/* DELPHI_COMPAT */
