#ifndef		V_ARR_HPP
#define		V_ARR_HPP

/*
 *	array_tt - template	for	class "dynamic array".
 *	Methods:
 *		add(elem_t elem) - add element at	the	end	of array;
 *		set(int index,elem_t elem) - write element at	pos	'index'. If	
 *			there are no elements before this position, all previous
 *			positions fills with zeros.
 *		overloaded operation[i] -	use	it for READ	elements. No bounds
 *			checks performed, so you should be careful.
 *		count() - returns number of elements (of course, it may differs
 *			from size of array)
 *		clear() - clears array (frees all storages and sets number of
 *			elements to zero)
 *		concat() - concatenates current array with other array.
 *
 *	NOTE.
 *	  .  :      
 *	    ,     -
 *	  ! -   :  N , 
 *	10N,   N,  11N, ...  . : 
 *	   .  :  
 *	 ,      8-16 ,
 *	    .
*/

#ifdef		__cplusplus

class		string_c;

#ifdef _MSC_VER
extern "C++"
#endif
template	<class elem_t> class array_tt
{
private:
  elem_t	*buf;
  int		n_elem;
  int		next_elem;
  friend	class string_c;
public:
			array_tt()
			{
				n_elem=8;
				next_elem=0;
				buf=(elem_t*)malloc(n_elem*sizeof(elem_t));
			}
			array_tt(array_tt &ref)
			{
				n_elem=ref.n_elem;
				next_elem=ref.next_elem;
				buf=(elem_t*)malloc(n_elem*sizeof(elem_t));
				v_copy(buf,ref.buf,n_elem*sizeof(elem_t));
			}
			~array_tt()
			{
				free(buf);
			}
  void		add(elem_t elem)
			{
			  	if(next_elem>=n_elem)
				{
					elem_t *next_buf=(elem_t*)malloc((n_elem<<1)*sizeof(elem_t));
					v_copy(next_buf,buf,n_elem*sizeof(elem_t));
					free(buf);
					buf=next_buf;
					n_elem<<=1;
				}
				buf[next_elem++]=elem;
			}
  void		clear()
			{
				free(buf);
				n_elem=16;
				next_elem=0;
				buf=(elem_t*)malloc(n_elem*sizeof(elem_t));
			}
  void		concat(array_tt	&arr)
			{
			  // TODO: optimize	me !
				int i;
				for(i=0;i<arr.next_elem;i++)
					add(arr[i]);
			}
  elem_t	operator[](int index)
  			{
  				return buf[index];
  			}
  //		bit	strange, but this code is FAST ENOUGHT (removing check don't
  //		boost the speed	AT ALL on Klamath-233 while	collecting u_32;
  //		increase by	1.8	while collecting u_8, and by 2.1 while collecting u_16)
  void		set(int	index,elem_t elem) 
			{
				if(index<next_elem)
					buf[index]=elem;
				else
				{
					while(index>next_elem)
				  		add(0);
					add(elem);
				}
			}
  int		count()
  			{
  				return next_elem;
  			}
  void		growto(int newsize)	// 4.1.00
			{
				if(newsize<=n_elem)
					next_elem=newsize;
				else
					set(newsize,0);
			}
};	// array_tt: last implemented 7.12.99

#ifdef _MSC_VER
extern "C++"
#endif
template <class	class_t> class storage_tt
{
private:
  array_tt<class_t*>	objects;
public:
						~storage_tt()
						{
							clear();
						}
  void					add()
  						{
  							objects.add(new	class_t);
  						}
  void					add(class_t	&ref)
						{
							add(); 
							v_copy(objects[objects.count()-1],&ref,
								sizeof(class_t));
						}
  void					growto(int larger_size)	
						{
							while(larger_size>objects.count())
								add(); 
						}
  void					clear()
						{
							for(int i=0;i<objects.count();i++)
								delete objects[i];
						}
  int					count()
  						{
  							return objects.count();
  						}
  class_t				&operator[](int	index)
  						{
  							return *objects[index];
  						}
  class_t				&last(void)
  						{
  							return *objects[objects.count()-1];
  						}
};	// storage_tt: implemented 8.12.99

#endif	/* __cplusplus */

#endif	/* V_ARR_HPP */
