/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../include/v_lib.h"
#include	"../source/v_mt.h"
#include	"v_os.h"

#define	GL_W_MAGIC\
		(((u32_t)('V')<<16) + ((u32_t)('G')<<8) + (u32_t)('L'))

#if	defined(__WIN32__)

/*
 *	WIN32: wglXXX already declared in wingdi.h.
 *	All prototypes for functions we need to call are in sgi/gl.h;
 *	we DON'T need to include extra OpenGL header (assuming calling
 *	conventions for target platform properly set in sgi/gl.h).
 */

/*
 *	============ STRUCT GLS - analog to OpenGL context ===================
 */
typedef struct		gl_w_settings_s
{
	u32_t			magic;			/* GL_W_MAGIC */
	u32_t			orgx, orgy;		/* Position */
	u32_t			width,height;	/* Extents */
	u32_t			bpp;			/* bit depth (fullscreen only) */
	vboolean		fullscreen;		/* ==0 for overlapped windows */
	u32_t			refresh;		/* refresh rate (fullscreen only) */
	vboolean		bitmap_based;	/* ==1 for offscreen windows */
	vboolean		accelerated;	/* ==1 for HW renderer */
	char			*vendor;		/* extracted from OpenGL pipeline */
	char			*renderer;
	char			*version;
	u32_t			num_extensions;
	char			**extensions;
	HWND			parent;			/* used for child windows */
	HBITMAP			bmp;			/* used for offscreen windows */
	BITMAPINFOHEADER	bmh;
	void			*bits;
	HPALETTE		pal;			/* used in 8-bpp mode */
	char			title[128];		/* window title */
	HWND			wnd;			/* window handle */
	BOOL			wndactive;		/* activity flag */
	HDC				dc;				/* window of bitmap DC */
	HGLRC			rc;				/* OpenGL RC */
	DWORD			rc_owner;		/* RC owner thread ID (GL client) */
	ipc_handle_t	control_pipe;	/* ipc channel to post messages to */
	gl_outline_font_t *font;		/* selected font */
	vp_stack_t		fonts;			/* fonts created in 'rc' */
	HANDLE			access;			/* lock for context */
	/* rest is HACKS ... */
	UINT			last_msg;			/* used to store last msg */
	i32_t			last_mx, last_my;	/* used to store last mousemove */
}	gl_w_settings_t, *gl_w_settings_p;

/* 		How much memory on your 3D-accelerator card ? :) */
/*		Seriously, TODO: convert to use vp_stack_t */
#define	MAX_GL_WINDOWS		8
/*		This thread dispatches all messages and manages OpenGL
 *		windows. */
V_LOCAL	HANDLE				gl_host_thread=NULL;
V_LOCAL	DWORD				gl_host_thread_id;
/*		Global data access mutex, managed by caller thread */
V_LOCAL	HANDLE				gl_access=NULL;
/*		Number of currently created windows, managed by gl_host_thread */
V_LOCAL	int					gl_num_windows=0;
/*		All OpenGL windows belongs to this class. Class managed by 
 *		gl_host_thread */
V_LOCAL	char				*gl_class_name="GenericOpenGLWindowClass";
V_LOCAL	WNDCLASS			gl_wc;
/*		Per - window settings. Windowing-part of each entry managed by
 *		gl_host_thread, OpenGL-part managed by caller thread. Caller
 *		allocates (and can free) entries, but MUST NOT alter gl_win[]
 *		contents (it can only READ it). */
V_LOCAL	gl_w_settings_p		gl_win[MAX_GL_WINDOWS];
/*		Index of current OpenGL window. Managed by caller thread */
/*		Event set to signaled by gl_host_thread when it process message */
HANDLE	message_delivered	= NULL;

/*	======================================================================
 *	Communication stuff.
 *
 *		There is an interesting bit: we use multiple threads while
 *	OpenGL implemetation not permits us to do it. The solution is
 *	to perform RegisterClass() and CreateWindow() in gl_host_thread,
 *	and all OpenGL calls in caller thread. We needs some method to
 *	tell gl_host_thread what we want to perform; PostThreadMessage
 *	is suitable.
 *	TODO (too late for now !): convert to use 'control_pipe'.
 */

/*	We want to create window */
#define		gl_msg_create_window		(WM_USER+1)
/*	We want to destroy window */
#define		gl_msg_destroy_window		(WM_USER+2)
/*	We want to terminate gl_host_thread */
#define		gl_msg_terminate_thread		(WM_QUIT)
/*	gl_host_thread want to tell us about last operation */
#define		gl_msg_answer				(WM_USER+3)
/*	We want to redirect input to given ipc channel */
#define		gl_msg_redirect_input		(WM_USER+4)
/*	We want to resize window */
#define		gl_msg_resize				(WM_USER+5)
/*	We want to move window */
#define		gl_msg_reposition			(WM_USER+6)
/*	We want to remove focus from thread windows */
#define		gl_msg_set_focus			(WM_USER+7)
/*	From server: "Something have been wrote in IPC channel" */
#define		gl_msg_notification			(WM_USER+100)

/*
 *	issue_gl_message().
 *	Posts message to gl_host_thread, and waits for answer.
 *	Returns true if message successfully delivered, false if timeout.
 *	If defined(CHECKED_BUILD), prints errors to console.
 *	Must be called from OpenGL client thread (not gl_host_thread).
 */

vboolean	V_LOCAL
#ifdef	CHECKED_BUILD
issue_gl_message(UINT message,WPARAM wparam,LPARAM lparam,
	u32_t wait_timeout,
	WPARAM *result,char *caller_name)
#define	ISSUE_GL_MESSAGE					issue_gl_message
#else	/* CHECKED_BUILD */
issue_gl_message(UINT message,WAPARAM wparam,LPARAM lparam,
	u32_t wait_timeout,
	LPARAM *result)
#define	ISSUE_GL_MESSAGE(msg,wp,lp,caller)	issue_gl_message(msg,wp,lp)
#endif	/* CHECKED_BUILD */
{
	MSG answer;
	timer_t timer;
	u32_t quant;
#ifdef	CHECKED_BUILD
	char *msg_term=" in issue_gl_message()";
#endif	/* CHECKED_BUILD */
	ResetEvent(message_delivered);
	if(!PostThreadMessage(gl_host_thread_id,message,wparam,lparam))
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"%s(): PostThreadMessage(%d,%d,%d) failed"
			"%s (code %d).\n",
		caller_name,message,wparam,lparam,msg_term,GetLastError());
#endif	/* CHEDKED_BUILD */
		return false;
	}
	if(!wait_timeout)
		return true;
	/*
	 *	Line:
	 *	"WaitForSingleObject(message_delivered,wait_timeout);"
	 *	magically screws up some deep inside OS when creating child 
	 *	windows (with top-level windows - all Ok).
	 */
	quant=wait_timeout / 500;
	/* don't be evil */
	if(!quant)
		quant=1;
	timerStart(&timer);
	while(timerMilliseconds(&timer)<wait_timeout)
	{
		if(PeekMessage(&answer,NULL,gl_msg_answer,gl_msg_answer,
			PM_NOREMOVE))
			break;
		timeDelay(quant);
	}
	timerStop(&timer);
	/*
	 *	Event is still usefull.
	 */
	if(WaitForSingleObject(message_delivered,0)==WAIT_TIMEOUT)
	{
		/* timeout */
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"%s(): gl_host_thread not responding !\n",
			caller_name);
#endif	/* CHECKED_BUILD */
	}
	else
	{
		if(!PeekMessage(&answer,NULL,gl_msg_answer,gl_msg_answer,PM_REMOVE))
		{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_ERROR,"%s(): wait terminated%s.\n",
				caller_name,msg_term);
#endif	/* CHECKED_BUILD */
		}
		else
		{
			if(result)
				*result=answer.wParam;
			if(answer.lParam==0)
				return true;
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"%s(): %s\n",caller_name,(char*)answer.lParam);
#endif		/* CHECKED_BUILD */
			/* error msg was strdup()'ed, free() it */
			free((void*)answer.lParam);
		}
	}
	return false;
}
	
/*
 *	issue_answer().
 *	Posts answer message to client thread.
 *	Must be called from gl_host_thread, NOT client of OpenGL.
 */
void		V_LOCAL
issue_answer(DWORD dest_thread_id,WPARAM wparam,char *error_msg)
{
	if(!error_msg)
		PostThreadMessage(dest_thread_id,gl_msg_answer,wparam,0);
	else
		PostThreadMessage(dest_thread_id,gl_msg_answer,wparam,
			(LPARAM)strdup(error_msg));
	/* if(!message_was_posted) */
	SetEvent(message_delivered);
}

/*	======================================================================
 *	Specific comm functions.
 */
 
/*
 *	issue_create_window().
 *	Returns (index of created window)+1 or 0 on error.
 *	Must be called by future RC owner (client of OpenGL).
 *	Note. Upon exit gls->access owned by caller.
 */
glw_t	V_LOCAL
issue_create_window(char *title,u32_t orgx,u32_t orgy,u32_t extx,u32_t exty,
	HWND parent,vboolean fullscreen,u32_t bpp,u32_t refresh,
	vboolean bitmap_based)
{
	gl_w_settings_t *gls;
	glw_t res;
	/* Allocate memory for context. Memory must be freed either
	 * in gl_host_thread if it fails to create window, or later in
	 * caller thread if gl_host_thread successed */
	gls=(gl_w_settings_t *)malloc(sizeof(gl_w_settings_t));
	v_zero(gls,sizeof(gl_w_settings_t));
	gls->magic=GL_W_MAGIC;
	gls->orgx=orgx;
	gls->orgy=orgy;
	gls->width=extx;
	gls->height=exty;
	gls->fullscreen=fullscreen;
	gls->parent=parent;
	gls->rc_owner=GetCurrentThreadId();
	strCopy(gls->title,title,128);
	gls->bpp=bpp;
	gls->refresh=refresh;
	gls->bitmap_based=bitmap_based;
	gls->access=CreateMutex(NULL,FALSE,NULL);
	/*	Pass partially initialized struct gls to gl_host_thread.
	 *	Timeout must be quite large because switching screen mode
	 *	can take VERY long time.
	 */
	if(!ISSUE_GL_MESSAGE(gl_msg_create_window,(WPARAM)gls,0,
		10000,(WPARAM*)&res,"issue_create_window"))
		return 0;
	else
		return res;
}

/*
 *	issue_detsroy_window().
 *	Returns true on success.
 *	Must be called by RC owner (client of OpenGL).
 */
vboolean	V_LOCAL
issue_destroy_window(glw_t index_plus_1)
{
	return ISSUE_GL_MESSAGE(gl_msg_destroy_window,(WPARAM)index_plus_1,0,
		1000,NULL,"issue_destroy_window");
}

/*
 *	issue_redirect_input().
 *	Returns true on success.
 *	Must be called by RC owner (client of OpenGL).
 */
vboolean	V_LOCAL
issue_redirect_input(glw_t index_plus_1,char *new_server_name)
{
	return ISSUE_GL_MESSAGE(gl_msg_redirect_input,(WPARAM)index_plus_1,
		(LPARAM)new_server_name,1000,NULL,"issue_redirect_input");
}

/*
 *	Returns true on success.
 *	Must be called by RC owner (client of OpenGL).
 */
vboolean	V_LOCAL
issue_resize(glw_t index_plus_1,u32_t width,u32_t height)
{
	return ISSUE_GL_MESSAGE(gl_msg_resize,(WPARAM)(index_plus_1),
		MAKELONG(width,height),1000,NULL,"issue_resize");
}

/*
 *	Returns true on success.
 *	Must be called by RC owner (client of OpenGL).
 */
vboolean	V_LOCAL
issue_move(glw_t index_plus_1,u32_t left,u32_t top)
{
	return ISSUE_GL_MESSAGE(gl_msg_reposition,(WPARAM)(index_plus_1),
		MAKELONG(left,top),1000,NULL,"issue_move");
}

/*
 *	Returns true on success.
 *	Must be called by RC owner (client of OpenGL).
 */
vboolean	V_LOCAL
issue_set_focus(glw_t index_plus_1)
{
	return ISSUE_GL_MESSAGE(gl_msg_set_focus,(WPARAM)(index_plus_1),0,
		0,NULL,"issue_set_focus");
}

/*	======================================================================
 *	Common OpenGL stuff.
 */

V_LOCAL	u8_t	threeto8[8] = {0, 0111>>1, 0222>>1, 0333>>1, 0444>>1,
  0555>>1, 0666>>1, 0377 };
V_LOCAL	u8_t	twoto8[4] = {  0, 0x55, 0xaa, 0xff };
V_LOCAL	u8_t	oneto8[2] = {  0, 255 };
V_LOCAL	int 	defaultOverride[13] = { 0, 3, 24, 27, 64, 67, 88, 173,
	181, 236, 247, 164, 91 };

V_LOCAL	PALETTEENTRY defaultPalEntry[20] = {
	{ 0,   0,   0,    0 },
	{ 0x80,0,   0,    0 },
	{ 0,   0x80,0,    0 },
	{ 0x80,0x80,0,    0 },
	{ 0,   0,   0x80, 0 },
	{ 0x80,0,   0x80, 0 },
	{ 0,   0x80,0x80, 0 },
	{ 0xC0,0xC0,0xC0, 0 },

	{ 192, 220, 192,  0 },
	{ 166, 202, 240,  0 },
	{ 255, 251, 240,  0 },
	{ 160, 160, 164,  0 },

	{ 0x80,0x80,0x80, 0 },
	{ 0xFF,0,   0,    0 },
	{ 0,   0xFF,0,    0 },
	{ 0xFF,0xFF,0,    0 },
	{ 0,   0,   0xFF, 0 },
	{ 0xFF,0,   0xFF, 0 },
	{ 0,   0xFF,0xFF, 0 },
	{ 0xFF,0xFF,0xFF, 0 }
};

V_LOCAL	u8_t
ComponentFromIndex(int i,u32_t nbits,u32_t shift)
{
	u8_t val;
	val=(u8_t)(i>>shift);
	switch(nbits)
	{
		case 1:
			val &= 0x1;
			return oneto8[val];
		case 2:
			val &= 0x3;
			return twoto8[val];
		case 3:
			val &= 0x7;
			return threeto8[val];
		default:
			return 0;
	}
}

/*
 *	Uses 'dc' and 'bitmap_based' fields. Modifies 'rc' and 'pal' fields.
 *	Must be called from main thread, NOT gl_host_thread !
 */
vboolean	V_LOCAL
vglCreateContext(gl_w_settings_t *gls)
{
	int rep, pixfmt;
	PIXELFORMATDESCRIPTOR pfd;
	gls->pal=NULL;
	gls->rc=NULL;
	memset(&pfd,0,sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize=sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion=1;
	if(gls->bitmap_based)
	{
		pfd.cDepthBits=16;
		pfd.cColorBits=24;
		pfd.dwFlags=PFD_DRAW_TO_BITMAP|PFD_SUPPORT_OPENGL;
	}
	else
	{
		pfd.cDepthBits=(unsigned char)gls->bpp;
		pfd.dwFlags=PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
	}
	pfd.cAlphaBits=8;
	pfd.iPixelType=PFD_TYPE_RGBA;
	pfd.iLayerType=PFD_MAIN_PLANE;
	pixfmt=ChoosePixelFormat(gls->dc,&pfd);
	if(!pixfmt)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglCreateContext(): ChoosePixelFormat() failed (code %d).\n",
			GetLastError());
#endif	/* CHECKED_BUILD */
		return false;
	}
	for(rep=4;rep;rep--)
	{
		if(SetPixelFormat(gls->dc,pixfmt,&pfd))
			break;
		Sleep(1000);
	}
	if(!rep)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"vglCreateContext(): SetPixelFormat() failed"
			" (code %d, OpenGL %d).\n",GetLastError(),glGetError());
#endif	/* CHECKED_BUILD */
		return false;
	}
	DescribePixelFormat(gls->dc,pixfmt,sizeof(pfd),&pfd);
	if(pfd.dwFlags & PFD_NEED_PALETTE)
	{
		unsigned int i, n;
		LOGPALETTE *lp;
		n = 1 << pfd.cColorBits;
		lp=(PLOGPALETTE)malloc(sizeof(LOGPALETTE)+n*sizeof(PALETTEENTRY));
		lp->palVersion=0x300;
		lp->palNumEntries=(unsigned short)n;
		for(i=0;i<n;i++)
		{
			lp->palPalEntry[i].peRed =
				ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
			lp->palPalEntry[i].peGreen =
				ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
			lp->palPalEntry[i].peBlue =
				ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
			lp->palPalEntry[i].peFlags = 0;
		}
		if((pfd.cColorBits==8)&&(pfd.cRedBits==3)&&(pfd.cRedShift==0)&&
			(pfd.cGreenBits==3)&&(pfd.cGreenShift==3)&&
			(pfd.cBlueBits == 2)&&(pfd.cBlueShift==6))
		{
			for(i=1;i<=12;i++)
				lp->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
		}
		gls->pal = CreatePalette(lp);
		free(lp);
		SelectPalette(gls->dc,gls->pal,FALSE);
		RealizePalette(gls->dc);
	}
	gls->rc=wglCreateContext(gls->dc);
	if(!gls->rc)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglCreateContext(): wglCreateContext failed (code %d).\n",
			GetLastError());
#endif	/* CHECKED_BUILD */
		if(gls->pal)
		{
			SelectPalette(gls->dc,GetStockObject(DEFAULT_PALETTE),FALSE);
			DeleteObject(gls->pal);
			gls->pal=NULL;
		}
		return false;
	}
	/* Other non-critical operations */
#ifndef	PFD_GENERIC_ACCELERATED
#define PFD_GENERIC_ACCELERATED     0x00001000
#endif	/* PFD_GENERIC_ACCELERATED */
	gls->accelerated=(pfd.dwFlags & PFD_GENERIC_FORMAT) == 
		(pfd.dwFlags & PFD_GENERIC_ACCELERATED);
	/* Initialize font cache */
	vpsInit(&gls->fonts);
	gls->font=NULL;
	return true;
}

void	V_LOCAL	vglReleaseFont(gl_outline_font_t *fnt);

/*
 *	Uses and modifies 'rc' and 'pal' fields.
 *	Must be called from main thread, NOT gl_host_thread !
 */
void	V_LOCAL
vglDeleteContext(gl_w_settings_t *gls)
{
	u32_t i,num;
	gl_outline_font_t *f;
#ifdef	CHECKED_BUILD
	conlPrintf(LOG_INFO,"vglDeleteContext(): %s OpenGL on %s ver. %s\n",
		gls->vendor,gls->renderer,gls->version);
#endif	/* CHECKED_BUILD */
	/* Delete all fonts */
	num=vpsCount(&gls->fonts);
	for(i=0;i<num;i++)
	{
		f=(gl_outline_font_t*)vpsGet(&gls->fonts,i);
		if(f)
		{
#ifdef		CHECKED_BUILD
			/* There is nothing wrong, but we must notify caller */
			if(f->ref)
				/* font was produced implicitly and there was no 
				 * way for caller to delete it */
				conlPrintf(LOG_INFO,"vglDeleteContext(): duplicated "
					"font at slot %d have been released.\n",i);
			else
				/* caller must be more accurate ! */
				conlPrintf(LOG_WARNING,
					"vglDeleteContext(): non-empty font slot %d.\n",i);
#endif	/* CHECKED_BUILD */
			vglReleaseFont(f);
		}
	}
	/* Make no context current */
	if(gls->rc)
	{
		wglMakeCurrent(gls->dc,NULL);
		wglDeleteContext(gls->rc);
		gls->rc=NULL;
	}
	/* Delete palette */
	if(gls->pal)
	{
		SelectPalette(gls->dc,GetStockObject(DEFAULT_PALETTE),FALSE);
		DeleteObject(gls->pal);
		gls->pal=NULL;
	}
	/* free strdup()'ed memory */
	if(gls->vendor)
		free(gls->vendor);
	if(gls->renderer)
		free(gls->renderer);
	if(gls->version)
		free(gls->version);
	if(gls->extensions)
	{
		for(i=0;i<gls->num_extensions;i++)
			free(gls->extensions[i]);
		free(gls->extensions);
	}
}

/*
 *	Translates wParam and lParam into LibV key code.
 */
keycode_t	V_LOCAL
keycodeWin32Code(u32_t wp,u32_t lp)
{
	keycode_t key;
	unsigned char scan=(unsigned char)((lp>>16) & 0xFF);
	unsigned char ext=(unsigned char)((lp & (1<<24)) != 0);
	switch(wp)
	{
		case VK_ESCAPE:	key=key_escape; break;
		case '1':		key=key_1; break;
		case '2':		key=key_2; break;
		case '3':		key=key_3; break;
		case '4':		key=key_4; break;
		case '5':		key=key_5; break;
		case '6':		key=key_6; break;
		case '7':		key=key_7; break;
		case '8':		key=key_8; break;
		case '9':		key=key_9; break;
		case '0':		key=key_0; break;
		case 189:		key=key_minus; break;
		case 187:		key=key_equal; break;
		case VK_BACK:	key=key_backspace; break;
		case VK_TAB:	key=key_tab; break;
		case 'Q':		key=key_Q; break;
		case 'W':		key=key_W; break;
		case 'E':		key=key_E; break;
		case 'R':		key=key_R; break;
		case 'T':		key=key_T; break;
		case 'Y':		key=key_Y; break;
		case 'U':		key=key_U; break;
		case 'I':		key=key_I; break;
		case 'O':		key=key_O; break;
		case 'P':		key=key_P; break;
		case 219:		key=key_left_bracket; break;
		case 221:		key=key_right_bracket; break;
		case VK_RETURN:	key=ext ? key_grey_enter : key_enter; break;
		case 17:		key=
			ext ? key_right_control : key_left_control; break;
		case 18:		key=ext ? key_right_alt : key_left_alt; break;
		case 'A':		key=key_A; break;
		case 'S':		key=key_S; break;
		case 'D':		key=key_D; break;
		case 'F':		key=key_F; break;
		case 'G':		key=key_G; break;
		case 'H':		key=key_H; break;
		case 'J':		key=key_J; break;
		case 'K':		key=key_K; break;
		case 'L':		key=key_L; break;
		case 186:		key=key_colon; break;
		case 222:		key=key_apostrophe; break;
		case 192:		key=key_back_quote; break;
		case 16:		key=
			(scan==42) ? key_left_shift : key_right_shift; break;
		case 220:		key=key_back_slash; break;
		case 'Z':		key=key_Z; break;
		case 'X':		key=key_X; break;
		case 'C':		key=key_C; break;
		case 'V':		key=key_V; break;
		case 'B':		key=key_B; break;
		case 'N':		key=key_N; break;
		case 'M':		key=key_M; break;
		case 188:		key=key_comma; break;
		case 190:		key=key_dot; break;
		case 191:		key=key_slash; break;
		case VK_F1:		key=key_f1; break;
		case VK_F2:		key=key_f2; break;
		case VK_F3:		key=key_f3; break;
		case VK_F4:		key=key_f4; break;
		case VK_F5:		key=key_f5; break;
		case VK_F6:		key=key_f6; break;
		case VK_F7:		key=key_f7; break;
		case VK_F8:		key=key_f8; break;
		case VK_F9:		key=key_f9; break;
		case VK_F10:	key=key_f10; break;
		case VK_F11:	key=key_f11; break;
		case VK_F12:	key=key_f12; break;
		case VK_NUMLOCK:	key=key_numlock; break;
		case VK_SCROLL:	key=key_scrolllock; break;
		case 32:		key=key_space; break;
		case 20:		key=key_capslock; break;
		case 91:		key=key_left_win; break;
		case 92:		key=key_right_win; break;
		case 93:		key=key_context; break;
		case 111:		key=key_grey_slash; break;
		case 106:		key=key_grey_asterisk; break;
		case 109:		key=key_grey_minus; break;
		case 107:		key=key_grey_plus; break;
		default:
			switch(scan)
			{
				case 71:	key=(ext) ? key_home : key_numpad_7; break;
				case 72:	key=(ext) ? key_up : key_numpad_8; break;
				case 73:	key=(ext) ? key_pageup : key_numpad_9; break;
				case 75:	key=(ext) ? key_left : key_numpad_4; break;
				case 76:	key=key_numpad_5; break;
				case 77:	key=(ext) ? key_right : key_numpad_6; break;
				case 79:	key=(ext) ? key_end : key_numpad_1; break;
				case 80:	key=(ext) ? key_down : key_numpad_2; break;
				case 81:	key=(ext) ? key_pagedown : key_numpad_3; break;
				case 82:	key=(ext) ? key_insert : key_numpad_0; break;
				case 83:	key=(ext) ? key_delete : key_numpad_dot; break;
				case 69:	key=key_pause; break;
				case 55:	key=key_printscreen; break;
				default:	key=0; break;
			}
		break;
	}
	return key;
}

vboolean V_LOCAL
wm_process_message(gl_w_settings_t *gls,UINT msg,WPARAM wparam,
	LPARAM lparam,thread_data_t *td)
{
	input_event_t event;
	i32_t first_code, second_code;
	vboolean mouse_event=false;
	switch(msg)
	{
		case WM_SYSKEYDOWN:
		case WM_KEYDOWN:
#ifndef		ENABLE_AUTOREPEAT
			if(lparam & 0x40000000)
				return true;
#endif		/* ENABLE_AUTOREPEAT */
			event=ev_key_down;
			first_code=keycodeWin32Code(wparam,lparam);
			switch(first_code)
			{
				case key_left_shift:
				case key_right_shift:
					td->lower_case=false;
					break;
				case key_capslock:
					td->caps_lock=GetKeyState(20) & 1;
					break;
				case key_numlock:
					td->num_lock=GetKeyState(VK_NUMLOCK) & 1;
					break;
				case key_space:
					if(GetAsyncKeyState(VK_SPACE) & 0x8000)
						SetFocus(NULL);
			}
			second_code=keycodeGetChar(first_code,
				td->caps_lock ? !td->lower_case : td->lower_case,NULL);
			break;
		case WM_KEYUP:
		case WM_SYSKEYUP:
			event=ev_key_up;
			first_code=keycodeWin32Code(wparam,lparam);
			switch(first_code)
			{
				/* TODO: workaround Win32 bug with shifts 
				   (see Descent 3 readme for details) */
				case key_left_shift:
					td->lower_case=true;
					inPostEvent(gls->control_pipe,ev_key_up,key_left_shift,0);
					/* if(!(GetAsyncKeyState(VK_SHIFT) & 0x8000)) */
					/* Generate key_up + key_right_shift event (NOT WORKING !) */
					/* inPostEvent(dst,ev_key_up,key_right_shift,0); */
					return true;
				case key_right_shift:
					td->lower_case=true;
					inPostEvent(gls->control_pipe,ev_key_up,key_right_shift,0);
					/* if(!(GetAsyncKeyState(VK_SHIFT) & 0x8000)) */
					/* Generate key_up + key_left_shift (NOT WORKING !) */
					/* inPostEvent(dst,ev_key_up,key_left_shift,0); */
					return true;
			}
			second_code=0;
			break;
		case WM_MOUSEMOVE:
			event=ev_mouse_move;
			first_code=LOWORD(lparam);
			second_code=HIWORD(lparam);
			if(first_code & 0x8000)
				first_code|=0xFFFF0000;
			if(second_code & 0x8000)
				second_code|=0xFFFF0000;
#ifndef		DISABLE_MOUSE_FILTERING
			if( gls->last_mx==first_code &&
				gls->last_my==second_code)
				return true;
			gls->last_mx=first_code;
			gls->last_my=second_code;
#endif		/* DISABLE_MOUSE_FILTERING */
			break;
		case WM_LBUTTONDOWN:
			event=ev_key_down;
			first_code=key_left_mouse;
			second_code=0;
			mouse_event=true;
			break;
		case WM_MBUTTONDOWN:
			event=ev_key_down;
			first_code=key_middle_mouse;
			second_code=0;
			mouse_event=true;
			break;
		case WM_RBUTTONDOWN:
			event=ev_key_down;
			first_code=key_right_mouse;
			second_code=0;
			mouse_event=true;
			break;
		case WM_LBUTTONUP:
			event=ev_key_up;
			first_code=key_left_mouse;
			second_code=0;
			mouse_event=true;
			break;
		case WM_MBUTTONUP:
			event=ev_key_up;
			first_code=key_middle_mouse;
			second_code=0;
			mouse_event=true;
			break;
		case WM_RBUTTONUP:
			event=ev_key_up;
			first_code=key_right_mouse;
			second_code=0;
			mouse_event=true;
			break;
		default:
			return false;
	}
	inPostEvent(gls->control_pipe,event,first_code,second_code);
	/*
	 *	Post mouse_move event after mouse press/release events.
	 */
	if(mouse_event)
	{
		if(event==ev_key_down)
			SetCapture(gls->wnd);
		else
			ReleaseCapture();
		event=ev_mouse_move;
		first_code=LOWORD(lparam);
		second_code=HIWORD(lparam);
		if(first_code & 0x8000)
			first_code|=0xFFFF0000;
		if(second_code & 0x8000)
			second_code|=0xFFFF0000;
		inPostEvent(gls->control_pipe,event,first_code,second_code);
	}
	return true;
}

/*	======================================================================
 *	Callback for all OpenGL windows
 */
LRESULT CALLBACK
gl_window_proc(HWND wnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	int slot_plus_1;
	gl_w_settings_t *gls;
	thread_data_t *td;
	slot_plus_1=GetWindowLong(wnd,GWL_USERDATA);
	td=GetThreadData();
	if(!slot_plus_1 || slot_plus_1>MAX_GL_WINDOWS)
		gls=NULL;
	else
		gls=gl_win[slot_plus_1-1];
	if(!gls)
		return DefWindowProc(wnd,msg,wparam,lparam);
	WaitForSingleObject(gls->access,INFINITE);
	switch(msg)
	{
		case WM_ACTIVATE:
			gls->wndactive=LOWORD(wparam) != WA_INACTIVE;
			/* Extra stuff for fullscreen windows */
			if(gls->fullscreen)
			{
				/* TODO: implement video mode switching */
				if(gls->wndactive)
					SetWindowPos(gls->wnd,HWND_TOP,0,0,0,0,
						SWP_NOSIZE|SWP_NOMOVE);
				else
					SetWindowPos(gls->wnd,HWND_NOTOPMOST,0,0,0,0,
						SWP_NOSIZE|SWP_NOMOVE);
			}
			break;
		case WM_CLOSE:
			break;
		case WM_PAINT:
        	if(td->gl_current_slot_plus_1 == 0)
        	{
        		RECT rc;
        		PAINTSTRUCT ps;
        		HDC dc=BeginPaint(wnd,&ps);
        		GetClientRect(wnd,&rc);
        		PatBlt(dc,0,0,ps.rcPaint.right,ps.rcPaint.bottom,BLACKNESS);
        		EndPaint(wnd,&ps);
        	}
			/* We can't use OpenGL ops here - rc is invalid in context
			 * of gl_host_thread. Inform owner. */
			if(gls->control_pipe)
			{
				if(gls->last_msg!=WM_PAINT)
					inPostEvent(gls->control_pipe,ev_os_dependent,WM_PAINT,0);
			}
			break;
		case WM_MOVE:
		{
			gls->orgx=(u32_t)LOWORD(lparam);
			gls->orgy=(u32_t)HIWORD(lparam);
			if(gls->control_pipe)
				inPostEvent(gls->control_pipe,ev_os_dependent,WM_MOVE,lparam);
			break;
		}
		case WM_SIZE:
		{
			gls->width=(u32_t)LOWORD(lparam);
			gls->height=(u32_t)HIWORD(lparam);
			if(gls->control_pipe)
				inPostEvent(gls->control_pipe,ev_os_dependent,WM_MOVE,lparam);
			break;
		}
		case WM_PALETTECHANGED:
		{
			if(wnd!=(HWND)wparam)
			{
				SelectPalette(gls->dc,gls->pal,FALSE);
				RealizePalette(gls->dc);
			}
			break;
		}
		case WM_QUERYNEWPALETTE:
		{
			SelectPalette(gls->dc,gls->pal,FALSE);
			RealizePalette(gls->dc);
			ReleaseMutex(gls->access);
			return TRUE;
		}
		default:
		{
			if(gls->parent)
			{
				/* Send notification message to the parent window */
				PostMessage(gls->parent,gl_msg_notification,msg,wparam);
				/* we can't move this into wm_process_message() because
				 * it will not be called if no input channel */
				if(msg==WM_LBUTTONDOWN || msg==WM_RBUTTONDOWN || msg==WM_MBUTTONDOWN)
					SetFocus(gls->wnd);
			}
			if(gls->control_pipe)
				if(wm_process_message(gls,msg,wparam,lparam,td))
					break;
			ReleaseMutex(gls->access);
			return DefWindowProc(wnd,msg,wparam,lparam);
		}
	}
	gls->last_msg=msg;
	ReleaseMutex(gls->access);
	return 0;
}

/*
 *	Registers OpenGL window class.
 *	Must be called from gl_host_thread.
 */
ATOM	V_LOCAL
vgl_register_class(WNDCLASS *wc)
{
	wc->lpszClassName=gl_class_name;
	wc->hInstance=GetModuleHandle(NULL);
	wc->style = CS_OWNDC | CS_NOCLOSE /* CLIP ??? */;
	wc->lpfnWndProc=gl_window_proc;
	wc->cbClsExtra = 0;
	wc->cbWndExtra = 0;
	wc->hIcon = LoadIcon(NULL,IDI_WINLOGO);
	wc->hCursor = LoadCursor(NULL, IDC_ARROW);
	wc->hbrBackground = GetStockObject(HOLLOW_BRUSH);
	wc->lpszMenuName=NULL;
	return RegisterClass(wc);
}

/*
 *	Changes screen resolution, display depth and, under NT,
 *	display refresh rate.
 *	Must be called from gl_host_thread.
 */
vboolean	V_GLOBAL
vidSetMode(int width,int height,int bpp,int refresh_rate)
{
	DEVMODE dm;
	v_zero(&dm,sizeof(dm));
	dm.dmSize=sizeof(dm);
	dm.dmPelsWidth=width;
	dm.dmPelsHeight=height;
	dm.dmBitsPerPel=bpp;
	dm.dmFields=DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
	if(refresh_rate>0)
	{
		dm.dmDisplayFrequency=refresh_rate;
		dm.dmFields|=DM_DISPLAYFREQUENCY;
	}
	return ChangeDisplaySettings(&dm,0) == DISP_CHANGE_SUCCESSFUL;
}

/*
 *	Sets default display settings.
 *	Must be called from gl_host_thread.
 */
void	V_GLOBAL
vidDefaultMode(void)
{
	ChangeDisplaySettings(NULL,0);
}

/*
 *	Creates OpenGL window / bitmap.
 *	Uses fields: bitmap_based, width, height, fullscreen, bpp, refresh,
 *				 parent.
 *	Modifies fields: wnd, bmp, dc.
 *	Must be called from gl_host_thread.
 *	Returns (index of created window)+1 on success, 0 on error.
 *	After this call mutex gls->access MUST be unlocked !!!
 */
glw_t	V_LOCAL
vgl_create_window(gl_w_settings_t *gls)
{
	int slot;
	u32_t sx, sy;
	for(slot=0;slot<MAX_GL_WINDOWS;slot++)
		if(gl_win[slot]==NULL)
			break;
	if(slot==MAX_GL_WINDOWS)
	{
#ifdef	CHECKED_BUILD
		if(gl_num_windows==MAX_GL_WINDOWS)
			conlPrintf(LOG_ERROR,"vgl_create_window(): no more slots.\n");
		else
			conlPrintf(LOG_ERROR,
				"vgl_create_window(): 'gl_win' corrupted !.\n");
#endif	/* CHECKED_BUILD */
		return 0;
	}
	/*
	 *	For bitmap - based contexts, we must only create
	 *	output bitmap.
	 */
	WaitForSingleObject(gls->access,INFINITE);
	if(gls->bitmap_based)
	{
		gls->bmh.biSize=sizeof(BITMAPINFOHEADER);
		gls->bmh.biWidth=gls->width;
		gls->bmh.biHeight=gls->height;
		gls->bmh.biPlanes=1;
		gls->bmh.biBitCount=24;
		gls->bmh.biCompression=BI_RGB;
		gls->dc=CreateCompatibleDC(NULL);
		gls->bmp=CreateDIBSection(gls->dc,(LPBITMAPINFO)&gls->bmh,
			DIB_RGB_COLORS,&gls->bits,NULL,0);
		SelectObject(gls->dc,gls->bmp);
		gl_win[slot]=gls;
		gl_num_windows++;
		ReleaseMutex(gls->access);
		return slot+1;
	}
	if(gls->fullscreen)
	{
		HDC refdc;
		u32_t sbpp;
		u32_t refresh;
		refdc=GetDC(NULL);
		sbpp=GetDeviceCaps(refdc,BITSPIXEL);
		sx=GetDeviceCaps(refdc,DESKTOPHORZRES);
		sy=GetDeviceCaps(refdc,DESKTOPVERTRES);
		refresh=GetDeviceCaps(refdc,VREFRESH);
		ReleaseDC(NULL,refdc);
		/*	Create top-level window without any widgets */
		/*	We must do it here to get out of full-screen console
			process; otherwise ChangeDisplaySettings WILL FAIL. */
		gls->wnd=CreateWindow(
			gl_class_name,
			gls->title,
			WS_POPUP,
			0,0,
			sx,sy,
			NULL,
			NULL,
			GetModuleHandle(NULL),
			NULL);
		if(gls->wnd)
		{
			ShowWindow(gls->wnd,SW_SHOWNORMAL);
			SetActiveWindow(gls->wnd);
			Sleep(100);
		}
		else
		{
			/* Can't create window */
#ifdef		CHECKED_BUILD
				conlPrintf(LOG_ERROR,
					"CreateWindow() failed (error '%d').\n",
					GetLastError());
#endif		/* CHECKED_BUILD */
			ReleaseMutex(gls->access);
			return 0;
		}
		if(gls->width<=0 && gls->height<=0)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_INFO,
				"vgl_create_window(): using desktop dimensions "
				"%d by %d.\n",sx, sy);
#endif		/* CHECKED_BUILD */
			gls->width=sx;
			gls->height=sy;
		}
		/*	under Chicago we can't change bpp; under Nashvile and
		 *	Memphis we can, but can't set refresh rate ...
		 */
		if(gls->bpp==0 || WeAreUnderMustDie())
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_INFO,
				"vgl_create_window(): %susing desktop display "
				"depth %d.\n",
				WeAreUnderMustDie() ? "Windows 95/98 detected, " : "",
				sbpp);
#endif		/* CHECKED_BUILD */
			gls->bpp=sbpp;
		}
		if(WeAreUnderMustDie())
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_INFO,
				"vgl_create_window(): Windows 95/98 detected, IGNORING "
				"refresh rate '%d Hz'.\n",gls->refresh);
#endif		/* CHECKED_BUILD */
			gls->refresh=0;
		}
		if(sbpp != gls->bpp || sx != gls->width || sy != gls->height)
		{
			/* Change screen resolution */
#ifdef		CHECKED_BUILD
			if((gls->width>sx || gls->height>sy) && gls->refresh>=refresh)
				conlPrintf(LOG_WARNING,
					"vgl_create_window(): refresh rate %d Hz may be to "
					"high (current is %d Hz).",gls->refresh,refresh);
			conlPrintf(LOG_INFO,
				"vgl_create_window(): calling CDS(%d,%d,%d,%d Hz) ... ",
				gls->width,gls->height,gls->bpp,gls->refresh);
#endif		/* CHECKED_BUILD */
			if(!vidSetMode(gls->width,gls->height,gls->bpp,gls->refresh))
			{
#ifdef		CHECKED_BUILD
				conlPrintf(LOG_INFO,"failed (error '%d').\n",
				GetLastError());
#endif		/* CHECKED_BUILD */
				DestroyWindow(gls->wnd);
				gls->wnd=NULL;
				ReleaseMutex(gls->access);
				return 0;
			}
#ifdef		CHECKED_BUILD
			conlPuts(LOG_INFO,"success.\n");
#endif		/* CHECKED_BUILD */
		}
		/* resize window to full screen */
		SetWindowPos(gls->wnd,HWND_TOP,0,0,gls->width,gls->height,0);
	}
	else	/* !fullscreen */
	{
		if(gls->parent)
		{
			/* Create child window */
			gls->wnd=CreateWindow(
				gl_class_name,
				gls->title,
				WS_CHILD,
				gls->orgx,gls->orgy,
				gls->width,
				gls->height,
				gls->parent,
				NULL,
				GetModuleHandle(NULL),
				NULL);
		}
		else
		{
			/* Create top-level window without sizing border */
			/* @&%# !!! damned user32.dll !!! */
			/*
			if(gls->wnd && gls->width<sx && gls->height<sy)
			{
				RECT rc;
				u32_t rw, rh;
				GetClientRect(gls->wnd,&rc);
				
				rw=2*gls->width-rc.right;
				rh=2*gls->height-rc.bottom;
				
				works only if started from cmd.exe !!! DAMN
				
				SetWindowPos(gls->wnd,HWND_NOTOPMOST,0,0,rw,rh,SWP_NOMOVE);
			}
			*/
			/* Workaround: */
			gls->wnd=CreateWindow(
				gl_class_name,
				gls->title,
				WS_POPUP | WS_CAPTION,
				gls->orgx,gls->orgy,
				gls->width + 2 * GetSystemMetrics(SM_CXFIXEDFRAME),
				gls->height + 2 * GetSystemMetrics(SM_CYFIXEDFRAME) + 
					GetSystemMetrics(SM_CYCAPTION),
				NULL,
				NULL,
				GetModuleHandle(NULL),
				NULL);
		}
	}
	if(!gls->wnd)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_INFO,
			"vgl_create_window(): CreateWindow() returns error '%d'.\n",
			GetLastError());
#endif	/* CHECKED_BUILD */
		ReleaseMutex(gls->access);
		return 0;
	}
	gls->dc=GetDC(gls->wnd);
	SetWindowLong(gls->wnd,GWL_USERDATA,slot+1);
	ShowWindow(gls->wnd,SW_SHOWNORMAL);
	gls->wndactive=true;
	gl_win[slot]=gls;
	gl_num_windows++;
	ReleaseMutex(gls->access);
	return slot+1;
}

/*
 *	Destroys OpenGL window / bitmap.
 *	Uses fields: wnd, bitmap_based, bmp, fullscreen.
 *	Modifies fields: dc, wnd / bmp.
 *	Modifies gl_win[] contents.
 *	Must be called from gl_host_thread. Assumes gl_access is closed.
 *	Assumes slot_plus_1 are valid.
 *	Note. Waits for 1 second at gls->access and if other operations 
 *	with whis context are pending, fails.
 */
vboolean	V_LOCAL
vgl_destroy_window(glw_t slot_plus_1)
{
	gl_w_settings_t *gls;
	gls=gl_win[slot_plus_1-1];
	V_ASSERT(gls);
	if(WaitForSingleObject(gls->access,1000)!=WAIT_OBJECT_0)
		return false;
	gl_win[slot_plus_1-1]=NULL;
	gl_num_windows--;
	if(gls->bmp)
	{
		DeleteObject(gls->bmp);
		gls->bmp=NULL;
	}	
	if(gls->control_pipe)
	{
		ipcClose(gls->control_pipe);
		gls->control_pipe=0;
	}
	if(gls->fullscreen)
		vidDefaultMode();
	if(gls->dc)
	{
		ReleaseDC(gls->wnd,gls->dc);
		gls->dc=NULL;
	}
	if(gls->wnd)
	{
		DestroyWindow(gls->wnd);
		gls->wnd=NULL;
	}
	ReleaseMutex(gls->access);
	CloseHandle(gls->access);
	gls->access=NULL;
	return true;
}

/*	======================================================================
 *	Core functionality: host thread
 *	======================================================================
 *	TODO: reduce code size by moving common checks outside switch().
 */

#ifdef	_MSC_VER
#pragma	warning(disable:4100)	/* Unreferenced formal parameter */
#endif	/* _MSC_VER */

#ifdef	__BORLANDC__
#pragma argsused
#endif	/* __BORLANDC__ */

V_LOCAL	DWORD	WINAPI
gl_host_thread_start_address(LPVOID param)
{
	MSG msg;
	thread_data_t *td;
	gl_w_settings_t *gls;
	glw_t slot_plus_1;
	DWORD caller;
	char srv_name[_MAX_PATH];
	td=GetThreadData();
#ifdef	CHECKED_BUILD
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
#endif	/* CHECKED_BUILD */
	td->lower_case=true;
	td->caps_lock=GetKeyState(20) & 1;
	td->num_lock=GetKeyState(VK_NUMLOCK) & 1;
	/*	We begin to change global data */
	/*
	WaitForSingleObject(gl_access,INFINITE)
	gl_host_thread_id=GetCurrentThreadId();
	ReleaseMutex(gl_access);
	*/
	/* Create message queue */
	PeekMessage(&msg,NULL,WM_USER,WM_USER,PM_NOREMOVE);
	/* Register window class */
	if(!vgl_register_class(&gl_wc))
	{
		/* We can't RegisterClass() */
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"gl_host_thread(): RegisterClass() failed (code %d).\n",
			GetLastError());
#endif	/* CHECKED_BUILD */
		ExitThread(0);
	}
	/*
	 *	Message loop.
	 */
	gl_num_windows=0;
	for(;;)
	{
		if(!WaitMessage())
		{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_WARNING,
				"gl_host_thread(): WaitMessage() failed (%d).\n",
				GetLastError());
#endif		/* CHECKED_BUILD */
			continue;
		}
		while(PeekMessage(&msg,NULL,0,0,PM_REMOVE))
		{
			switch(msg.message)
			{
				case gl_msg_create_window:
					/* struct gl_w_settings allocated by caller */
					gls=(gl_w_settings_t*)msg.wParam;
					if(!gls || gls->magic!=GL_W_MAGIC)
					{
#ifdef					CHECKED_BUILD
						conlPrintf(LOG_ERROR,
							"gl_host_thread(): can't create window: "
							"incorrect contents of wParam.\n");
#endif					/* CHECKED_BUILD */
						/* no way to return answer; caller thread will
						 * dumb wait until timeout ... 
						 * Should be assert(0) here ???
						 */
						break;
					}
					WaitForSingleObject(gl_access,INFINITE);
					slot_plus_1=vgl_create_window(gls);
					ReleaseMutex(gl_access);
					caller=gls->rc_owner;
					/* If vgl_create_window() fails, free gls */
					if(!slot_plus_1)
					{
						CloseHandle(gls->access);
						free(gls);
					}
					/*	When caller receives this message it must continue
					 *	initialization (or report error) */
					issue_answer(caller,slot_plus_1,NULL);
					break;
				case gl_msg_destroy_window:
					slot_plus_1=msg.wParam;
					if(slot_plus_1==0 || slot_plus_1>MAX_GL_WINDOWS || 
						gl_win[slot_plus_1-1]==NULL)
					{
#ifdef					CHECKED_BUILD
						conlPrintf(LOG_ERROR,
							"gl_msg_destroy_window: wrong wParam '%d'.\n",
							slot_plus_1);
#endif					/* CHECKED_BUILD */
						break;
					}
					WaitForSingleObject(gl_access,INFINITE);
					/* we want to track caller thread id */
					gls=gl_win[slot_plus_1-1];
					caller = gls ? gls->rc_owner : 0;
					vgl_destroy_window(slot_plus_1);
					ReleaseMutex(gl_access);
					/* 'done'; caller can free struct gls */
					if(caller)
						issue_answer(caller,1,NULL);
					break;
				case gl_msg_redirect_input:
					slot_plus_1=msg.wParam;
					if(slot_plus_1==0 || slot_plus_1>MAX_GL_WINDOWS || 
						gl_win[slot_plus_1-1]==NULL)
					{
#ifdef					CHECKED_BUILD
						conlPrintf(LOG_ERROR,
							"gl_msg_redirect_input: wrong wParam '%d'.\n",
							slot_plus_1);
#endif					/* CHECKED_BUILD */
						break;
					}
					strCopy(srv_name,(char*)msg.lParam,_MAX_PATH);
					WaitForSingleObject(gl_access,INFINITE);
					gls=gl_win[slot_plus_1-1];
					if(gls->control_pipe)
						ipcClose(gls->control_pipe);
					gls->control_pipe=ipcServerEndpoint(srv_name);
					/* BUGBUG, but other ops will not be time-consuming */
					ReleaseMutex(gl_access);
					if(!gls->control_pipe)
					{
						if(gls->rc_owner)
						{
							char buf[128+_MAX_PATH];
							sprintf(buf,"gl_msg_redirect_input: "
								"can't start server '%s'.",srv_name);
							issue_answer(gls->rc_owner,0,strdup(buf));
						}
					}
					else
						if(gls->rc_owner)
							issue_answer(gls->rc_owner,1,NULL);
					break;
				case gl_msg_resize:
					slot_plus_1=msg.wParam;
					if(slot_plus_1==0 || slot_plus_1>MAX_GL_WINDOWS || 
						gl_win[slot_plus_1-1]==NULL)
					{
#ifdef					CHECKED_BUILD
						conlPrintf(LOG_ERROR,
							"gl_msg_resize: wrong wParam '%d'.\n",
							slot_plus_1);
#endif					/* CHECKED_BUILD */
						break;
					}
					WaitForSingleObject(gl_access,INFINITE);
					gls=gl_win[slot_plus_1-1];
					SetWindowPos(gls->wnd,HWND_NOTOPMOST,0,0,
						LOWORD(msg.lParam),HIWORD(msg.lParam),
						SWP_NOMOVE|SWP_NOZORDER);
					ReleaseMutex(gl_access);
					/* 'done' */
					if(gls->rc_owner)
						issue_answer(gls->rc_owner,1,NULL);
					break;
				case gl_msg_reposition:
					slot_plus_1=msg.wParam;
					if(slot_plus_1==0 || slot_plus_1>MAX_GL_WINDOWS || 
						gl_win[slot_plus_1-1]==NULL)
					{
#ifdef					CHECKED_BUILD
						conlPrintf(LOG_ERROR,
							"gl_msg_reposition: wrong wParam '%d'.\n",
							slot_plus_1);
#endif					/* CHECKED_BUILD */
						break;
					}
					WaitForSingleObject(gl_access,INFINITE);
					gls=gl_win[slot_plus_1-1];
					SetWindowPos(gls->wnd,HWND_NOTOPMOST,
						LOWORD(msg.lParam),HIWORD(msg.lParam),
						640,480,
						SWP_NOSIZE|SWP_NOZORDER);
					ReleaseMutex(gl_access);
					/* 'done' */
					if(gls->rc_owner)
						issue_answer(gls->rc_owner,1,NULL);
					break;
				case gl_msg_set_focus:
					slot_plus_1=msg.wParam;
					if(slot_plus_1==0 || slot_plus_1>MAX_GL_WINDOWS || 
						gl_win[slot_plus_1-1]==NULL)
					{
						/* All Ok, caller tell us to remove focus */
						SetFocus(NULL);
						issue_answer(gls->rc_owner,1,NULL);
						break;
					}
					gls=gl_win[slot_plus_1-1];
					SetFocus(gls->wnd);
					/* 'done' */
					if(gls->rc_owner)
						issue_answer(gls->rc_owner,1,NULL);
					break;
				case gl_msg_terminate_thread:
					/* terminate */
					WaitForSingleObject(gl_access,INFINITE);
					/* destroy all windows */
					for(slot_plus_1=1;slot_plus_1<=MAX_GL_WINDOWS;
						slot_plus_1++)
						if(gl_win[slot_plus_1-1])
							vgl_destroy_window(slot_plus_1);
					UnregisterClass(gl_class_name,GetModuleHandle(NULL));
					ReleaseMutex(gl_access);
					ExitThread(1);
				default:
					/* Post other messages to active window */
					TranslateMessage(&msg);
					DispatchMessage(&msg);
					break;
			}
		}
	}
#ifndef	__BORLANDC__
	return 0;
#endif	/* __BORLANDC__ */
}

/*
 *	thrIsAlive(): Waits for 'milliseconds' and returns true if thread
 *	is still running.
 */
vboolean	V_GLOBAL
thrIsAlive(HANDLE thread,u32_t milliseconds)
{
	if(thread && WaitForSingleObject(thread,milliseconds)==STATUS_TIMEOUT)
	{
		DWORD exit_code;
		GetExitCodeThread(thread,&exit_code);
		if(exit_code==STATUS_PENDING)
			return true;
	}
	return false;
}

/*
 *	vglStartHostThread
 */

vboolean	V_LOCAL
vglStartHostThread(void)
{
	gl_access=CreateMutex(NULL,TRUE,NULL);
	message_delivered=CreateEvent(NULL,TRUE,FALSE,NULL);
	v_zero(gl_win,MAX_GL_WINDOWS*sizeof(gl_w_settings_t*));
	gl_host_thread=CreateThread(NULL,0,gl_host_thread_start_address,NULL,0,
		&gl_host_thread_id);
	ReleaseMutex(gl_access);
	if(thrIsAlive(gl_host_thread,100))
		return true;
#ifdef	CHECKED_BUILD
	conlPrintf(LOG_ERROR,
		"vglStartHostThread(): windowing thread failed to initialize"
		" (code %d).\n",GetLastError());
#endif	/* CHECKED_BUILD */
	gl_host_thread=NULL;
	gl_host_thread=0;
	CloseHandle(message_delivered);
	message_delivered=NULL;
	CloseHandle(gl_access);
	gl_access=NULL;
	return false;
}

/*
 *	vglTerminateHostThread
 */
void	V_LOCAL
vglTerminateHostThread(void)
{
	if(!thrIsAlive(gl_host_thread,10))
		return;
	/* Break gl_host_thread's message loop */
	PostThreadMessage(gl_host_thread_id,gl_msg_terminate_thread,0,0);
	/* Wait for completion */
	WaitForSingleObject(gl_host_thread,INFINITE);
	gl_host_thread=NULL;
	gl_host_thread=0;
	CloseHandle(gl_access);
	gl_access=NULL;
}
  
/*
 *	vglCreateView
 */
glw_t	V_LOCAL
vglCreateView(char *title,u32_t orgx,u32_t orgy,u32_t extx,u32_t exty,
	HWND parent,vboolean fullscreen,u32_t bpp,u32_t refresh,
	vboolean bitmap_based)
{
	gl_w_settings_t *gls;
	glw_t slot_plus_1;
	int i;
	parser_t *ep;
	char *exts;
	vboolean retval;
	/* If host thread not started, start it */
	if(!gl_access && !vglStartHostThread())
		return 0;
	/* create window */
	slot_plus_1=issue_create_window(title,orgx,orgy,extx,exty,parent,
		fullscreen,bpp,refresh,bitmap_based);
	if(!slot_plus_1)
	{
		/* Stay quiet, issue_create() already reported error */
		/* gls already free()'d in issue_create() */
		return 0;
	}
	WaitForSingleObject(gl_access,INFINITE);
	/* gls allocated in issue_create_window() */
	gls=gl_win[slot_plus_1-1];
	/* initialize OpenGL context */
	WaitForSingleObject(gls->access,INFINITE);
	retval=vglCreateContext(gls);
	ReleaseMutex(gls->access);
	ReleaseMutex(gl_access);
	/* If we failed to init OpenGL, we must destroy window */
	if(!retval)
	{
		/* stay quiet; vglCreateContext() already reported error */
		issue_destroy_window(slot_plus_1);
		free(gls);
		return 0;
	}
	/* Make window current */
	if(!vglMakeCurrent(slot_plus_1))
	{
		vglDeleteContext(gls);
		issue_destroy_window(slot_plus_1);
		free(gls);
		return 0;
	}
	/* Misc */
	WaitForSingleObject(gls->access,INFINITE);
	gls->vendor=(char*)glGetString(GL_VENDOR);
	if(gls->vendor)
		gls->vendor=strdup((const char*)gls->vendor);
	gls->renderer=(char*)glGetString(GL_RENDERER);
	if(gls->renderer)
		gls->renderer=strdup((const char*)gls->renderer);
	gls->version=(char*)glGetString(GL_VERSION);
	if(gls->version)
		gls->version=strdup((const char*)gls->version);
	gls->extensions=NULL;
	gls->num_extensions=0;
	exts=(char*)glGetString(GL_EXTENSIONS);
	if(exts)
	{
		ep=tokStart(exts,NULL);
		while(tokNext(ep))
			gls->num_extensions++;
		if(gls->num_extensions)
		{
			gls->extensions=(char**)malloc(gls->num_extensions*sizeof(char*));
			tokRestart(ep);
			for(i=0;tokNext(ep);i++)
				gls->extensions[i]=strdup(tokLast(ep));
		}
		free(ep);
	}
#ifdef	CHECKED_BUILD
	conlPrintf(LOG_INFO,"vglCreateView(): %s OpenGL on %s ver. %s\n",
		gls->vendor,gls->renderer,gls->version);
#if	0
	conlPuts(LOG_INFO,"OpenGL extensions: \n");
	for(i=0;i<(int)gls->num_extensions;i++)
		conlPrintf(LOG_INFO,"    %s\n",gls->extensions[i]);
#endif	/* 0 */
#endif	/* CHECKED_BUILD */
	ReleaseMutex(gls->access);
	return slot_plus_1;
}

glw_t	V_GLOBAL
vglCreateWindow(char *title,u32_t orgx,u32_t orgy,u32_t extx,u32_t exty,
	void_p parent,vboolean fullscreen,u32_t bpp,u32_t refresh)
{
	return vglCreateView(title,orgx,orgy,extx,exty,(HWND)parent,
		fullscreen,bpp,refresh,false);
}

glw_t	V_GLOBAL
vglCreateOffscreenWindow(u32_t extx,u32_t exty)
{
	return vglCreateView("OffscreenWindow",0,0,extx,exty,NULL,false,24,0,
		true);
}

vboolean V_GLOBAL
vglRedirectInput(glw_t win,char *server_name)
{
	return issue_redirect_input(win,server_name);
}

void	V_GLOBAL
vglDestroyWindow(glw_t win)
{
	gl_w_settings_t *gls;
	thread_data_t *td;
	if(win==0 || win>MAX_GL_WINDOWS || gl_win[win-1]==NULL)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglDestroyWindow(): incorrect slot '%d'.\n",
			win);
#endif	/* CHECKED_BUILD */
		return;
	}
	td=GetThreadData();
	if(td->gl_current_slot_plus_1==win)
		vglMakeCurrent(0);
	gls=gl_win[win-1];
	vglDeleteContext(gls);
	issue_destroy_window(win);
	/* gl_win[win-1] is NULL now */
	free(gls);
	/*
	 *	Terminate gl_host_thread if no more windows.
	 */
	if(!gl_num_windows)
		vglTerminateHostThread();
}

/*
 *	vglMakeCurrent()
 */
vboolean	V_GLOBAL
vglMakeCurrent(glw_t win)
{
	thread_data_t *td;
	gl_w_settings_t *gls;
	if(win>MAX_GL_WINDOWS || (win && gl_win[win-1]==NULL))
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglMakeCurrent(): incorrect slot '%d'.\n",
			win);
#endif	/* CHECKED_BUILD */
		return false;
	}
	td=GetThreadData();
#ifdef	CHECKED_BUILD
	if(win && td->gl_current_slot_plus_1==win)
		return true;
#if	0
	if(td->gl_current_slot_plus_1)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_INFO,
			"vglMakeCurrent(): previous OpenGL was %s OpenGL.\n",
			gl_win[td->gl_current_slot_plus_1-1]->vendor);
#endif	/* CHECKED_BUILD */
	}
	else
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_INFO,
			"vglMakeCurrent(): was no previous context before %d.\n",win);
#endif	/* CHECKED_BUILD */
#endif	/* 0 */
#endif	/* CHECKED_BUILD */
	if(win)
	{
		gls=gl_win[win-1];
		if(wglMakeCurrent(gls->dc,gls->rc)==0)
		{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"vglMakeCurrent(): wglMakeCurrent() failed (code %d).\n",
				GetLastError());
#endif	/* CHECKED_BUILD */
			return false;
		}
		td->gl_current_slot_plus_1=win;
	}
	else
	{
		wglMakeCurrent(NULL,NULL);
#ifdef	CHECKED_BUILD
		/*
		conlPuts(LOG_INFO,
			"vglMakeCurrent(): OpenGL pipeline disabled.\n");
		*/
#endif	/* CHECKED_BUILD */
		td->gl_current_slot_plus_1=0;
	}
	return true;
}

glw_t	V_GLOBAL
vglGetCurrentWindow(void)
{
	return GetThreadData()->gl_current_slot_plus_1;
}

vboolean	V_GLOBAL
vglResizeWindow(glw_t win,u32_t width,u32_t height)
{
	return issue_resize(win,width,height);
}

vboolean	V_GLOBAL
vglMoveWindow(glw_t win,u32_t left,u32_t top)
{
	return issue_move(win,left,top);
}

vboolean	V_GLOBAL
vglSetFocus(glw_t win)
{
	return issue_set_focus(win);
}

int		V_GLOBAL
vglIterateWindows(void (V_CALL *func)(glw_t))
{
	int i, count;
	WaitForSingleObject(gl_access,INFINITE);
	if(!func)
	{
		for(count=i=0;i<MAX_GL_WINDOWS;i++)
			if(gl_win[i])
				count++;
	}
	else
	{
		for(count=i=0;i<MAX_GL_WINDOWS;i++)
			if(gl_win[i])
			{
				func(i+1);
				count++;
			}
	}
	ReleaseMutex(gl_access);
	return count;
}

void	V_GLOBAL
vglSwapBuffers(void)
{
	glw_t slot_plus_1;
	gl_w_settings_t *gls;
	slot_plus_1=GetThreadData()->gl_current_slot_plus_1;
	if(!slot_plus_1)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_WARNING,
			"vglSwapBuffers(): no one slot have made current.\n");
#endif	/* CHECKED_BUILD */
		Sleep(33);
		return;
	}
	gls=gl_win[slot_plus_1-1];
	glFlush();
	glFinish();
	SwapBuffers(gls->dc);
	/* If window is not active, reduce CPU & 3Daccel overhead */
	if(!gls->wndactive && !gls->bitmap_based)
		Sleep(33);	/* Maximum framerate limited to 30 FPS */
}

/*
 *	vglReadPixels()
 */
void_p		V_GLOBAL
vglReadPixels(u32_t *width,u32_t *height)
{
	glw_t slot_plus_1;
	gl_w_settings_t *gls;
	u8_t *pixels;
	u32_t x,y,offset,stride_adjust;
	thread_data_t *td;
	td=GetThreadData();
	WaitForSingleObject(gl_access,INFINITE);
	slot_plus_1=td->gl_current_slot_plus_1;
	if(!slot_plus_1)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_WARNING,
			"vglReadPixels(): no one slot have made current.\n");
#endif	/* CHECKED_BUILD */
		ReleaseMutex(gl_access);
		return NULL;
	}
	gls=gl_win[slot_plus_1-1];
	pixels=(u8_t*)malloc(gls->width*gls->height*4);
	stride_adjust=gls->width*3;
	while(stride_adjust & 3)
		stride_adjust++;
	stride_adjust-=gls->width*3;
	if(gls->bitmap_based)
	{
		u8_t *ptr=(u8_t*)gls->bits;
		for(y=gls->height-1;;y--)
		{
			offset=y*4*gls->width;
			for(x=0;x<gls->width;x++)
			{
				*(pixels+offset+2)=ptr[0];
				*(pixels+offset+1)=ptr[1];
				*(pixels+offset)=ptr[2];
				*(pixels+offset+3)=td->constant_alpha;
				offset+=4;
				ptr+=3;
			}
			ptr+=stride_adjust;
			if(!y)
				break;
		}
	}
	else
	{
		u32_t half;
		void *buffer;
		u8_t *upper, *lower;
		glPixelStorei(GL_UNPACK_LSB_FIRST,GL_FALSE);
		glPixelStorei(GL_UNPACK_ROW_LENGTH,0);
		glPixelStorei(GL_UNPACK_SKIP_ROWS,0);
		glPixelStorei(GL_UNPACK_SKIP_PIXELS,0);
		glPixelStorei(GL_UNPACK_ALIGNMENT,1);
		glReadPixels(0,0,gls->width,gls->height,GL_RGBA,GL_UNSIGNED_BYTE,pixels);
		buffer=malloc(gls->width*4);
		half=gls->height / 2;
		upper=pixels;
		lower=upper+(gls->height-1)*gls->width*4;
		while(half--)
		{
			v_copydw(buffer,upper,gls->width);
			v_copydw(upper,lower,gls->width);
			v_copydw(lower,buffer,gls->width);
			upper+=gls->width*4;
			lower-=gls->width*4;
		}
		free(buffer);
	}
	if(width)
		*width=gls->width;
	if(height)
		*height=gls->height;
	ReleaseMutex(gl_access);
	return pixels;
}

/*
**	======================================================================
**	FONT STUFF
**	======================================================================
*/

/*
 *	Validity check.
 */
 
#define				GLOF_MAGIC		0x474C4F46				  /* "GLOF" */

/*
**	Checks wheither passed handle is font handle.
**	Reports error if defined(CHECKED_BUILD).
*/
vboolean	V_LOCAL
#ifdef	CHECKED_BUILD
vglIsOutlineFont(gl_outline_font_t *fnt,char *caller_name)
#else	/* CHECKED_BUILD */
vglIsOutlineFont(gl_outline_font_t *fnt)
#endif	/* CHECKED_BUILD */
{
	if(!fnt || fnt->magic!=GLOF_MAGIC)
	{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_ERROR,"%s(%X): incorrect font handle%s.\n",
				caller_name,fnt,fnt ? ", magic wrong" : "");
#endif	/* CHECKED_BUILD */
		return false;
	}
	return true;
}

#ifdef	CHECKED_BUILD
#define	VGL_IS_OUTLINE_FONT(fnt,caller)	vglIsOutlineFont(fnt,caller)
#else	/* CHECKED_BUILD */
#define	VGL_IS_OUTLINE_FONT(fnt,caller)	vglIsOutlineFont(fnt)
#endif	/* CHECKED_BUILD */

/*
 *	Returns {arg | current font handle if arg is NULL | NULL if error}.
 *	If defined(CHECKED_BUILD), reports errors to console.
 */
gl_outline_font_p	V_LOCAL
#ifdef	CHECKED_BUILD
vglGetFontForOp(gl_outline_font_t *fnt,char *caller_name)
#else	/* CHECKED_BUILD */
vglGetFontForOp(gl_outline_font_t *fnt)
#endif	/* CHECKED_BUILD */
{
	if(fnt)
	{
		if(fnt->magic!=GLOF_MAGIC)
		{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_ERROR,"%s(%X): incorrect font "
				"handle, magic wrong.\n",caller_name,fnt);
#endif	/* CHECKED_BUILD */
			return NULL;
		}
	}
	else	/* !fnt */
	{
		glw_t slot=GetThreadData()->gl_current_slot_plus_1;
		if(!slot)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,"%s(): no one context "
				"have been made current.\n",caller_name);
#endif		/* CHECKED_BUILD */
			return NULL;
		}
		fnt=gl_win[slot-1]->font;
		if(!fnt)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,"%s(): no current font in context %d.\n",
				caller_name,slot-1);
#endif		/* CHECKED_BUILD */
			return NULL;
		}
	}
	return fnt;
}

#ifdef	CHECKED_BUILD
#define	VGL_GET_FONT_FOR_OP(fnt,caller)		vglGetFontForOp(fnt,caller)
#else	/* CHECKED_BUILD */
#define	VGL_GET_FONT_FOR_OP(fnt,caller)		vglGetFontForOp(fnt)
#endif	/* CHECKED_BUILD */

/*
**		Creation / destruction.
*/

/*
 *	Allocates memory for font structure
 */
gl_outline_font_p	V_LOCAL
#ifdef	CHECKED_BUILD
vglAllocFont(LOGFONT **lf,char *caller_name)
#else	/* CHECKED_BUILD */
vglAllocFont(LOGFONT **lf)
#endif	/* CHECKED_BUILD */
{
	u32_t cksize;
	gl_outline_font_t *fnt;
	cksize=sizeof(gl_outline_font_t)+256*sizeof(GLYPHMETRICSFLOAT)+
		sizeof(LOGFONT);
	fnt=(gl_outline_font_t*)malloc(cksize);
	if(!fnt)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"%s(): can't allocate %u bytes.\n",caller_name,cksize);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	v_zero(fnt,cksize);
	fnt->logfont=(void*)(fnt+1);
	*lf=(LOGFONT*)fnt->logfont;
	fnt->gm=(void*)((*lf)+1);
	return fnt;
}

static	vec_t	font_lod_deviations[GL_FONT_LOD_LEVELS]={0.001f,0.01f,0.1f};

/*
 *	Creates outlined or polygonal font.
 *	Input:
 *		face_name	- font face (may be NULL);
 *		polygonal	- if zero, creates outlined font;
 *		extrusion	- amount of extrusion;
 *		os_font		- os-dependent handle of font (may be NULL).
 *	Notes.
 *		1. Win32 only: os_font must be HFONT object that was previously
 *	created via CreateFont() / CreateFontIndirect().
 *		2. Win32 only: if os_font is non-NULL, font_face ignored.
 *		3. If both face_name and os_font are NULL, function creates
 *	font with face 'Arial'.
 *	Output:
 *		0 on error, or index+1 of font in current context's 'font' cache.
 */
gl_outline_font_p	V_GLOBAL
vglCreateFont(char *face_name,vboolean polygonal,vec_t extrusion,
	void *os_font)
{
	int i,j;
	HDC dc;				/* Current Win32 DC of RC */
	HFONT old,fnt;		/* Win32 font */
	int format;			/* Format of glyphs: either lines or polygons */
	gl_outline_font_t *f;
	LOGFONT *lf;
	u32_t slot,len;
	u32_t ctx;
	gl_w_settings_t *gls;
	/* Check if os_font is real font handle */
	if(os_font && GetObjectType(os_font)!=OBJ_FONT)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglCreateFont('%s',%d,%X): os_font isn't font handle.\n",
			dbs(face_name),polygonal,os_font);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	/* Obtain current RC's DC */
	dc=wglGetCurrentDC();
	if(!dc)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglCreateFont('%s',%d,%X): no current rendering context.\n",
			dbs(face_name),polygonal,os_font);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	/* Allocate memory for all structures */
#ifdef	CHECKED_BUILD
	f=vglAllocFont(&lf,"vglCreateFont");
#else	/* CHECKED_BUILD */
	f=vglAllocFont(&lf);
#endif	/* CHECKED_BUILD */
	if(!f)
		return NULL;
	/* Generate OpenGL display lists */
	for(i=0;i<GL_FONT_LOD_LEVELS;i++)
	{
		f->bases[i]=glGenLists(256);
		if(!f->bases[i])
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"vglCreateFont('%s',%d,%X): glGenLists() failed, GL error"
				" %d.\n",dbs(face_name),polygonal,os_font,glGetError());
#endif		/* CHECKED_BUILD */
			for(j=i-1;j>=0;j--)
				glDeleteLists(f->bases[i],256);
			free(f);
			return NULL;
		}
	}
	/* Create Win32 font and select it into DC */
	if(os_font)
	{
		/* Obtain font info from OS */
		if(GetObjectType((HGDIOBJ)os_font)!=OBJ_FONT)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"vglCreateFont('%s',%d,%X): %X isn't a font handle.\n",
				dbs(face_name),polygonal,os_font,os_font);
#endif		/* CHECKED_BUILD */
			free(f);
			return NULL;
		}
		if(GetObject((HGDIOBJ)os_font,sizeof(LOGFONT),(LPVOID)lf)!=
			sizeof(LOGFONT))
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"vglCreateFont('%s',%d,%X): GetObject(%X) failed "
				"(code %d).\n",dbs(face_name),polygonal,os_font,os_font,
				GetLastError());
#endif		/* CHECKED_BUILD */
			free(f);
			return NULL;
		}
	}
	else
	{
		/* Setup font info explicitly */
		lf->lfHeight=10;
		lf->lfWidth=0;			/* can be any value ? */
		lf->lfEscapement=0;		/* we don't use GDI transforms */
		lf->lfOrientation=0;
		lf->lfWeight=FW_NORMAL; 
		lf->lfItalic=FALSE;
		lf->lfUnderline=FALSE;
		lf->lfStrikeOut=FALSE;
#ifdef	ALWAYS_ENGLISH
		lf->lfCharSet=DEFAULT_CHARSET;
#else	/* ALWAYS_ENGLISH */
		lf->lfCharSet=RUSSIAN_CHARSET;
#endif	/* ALWAYS_ENGLISH */
		lf->lfOutPrecision=OUT_TT_PRECIS;
		lf->lfClipPrecision=CLIP_DEFAULT_PRECIS;
		lf->lfQuality=DEFAULT_QUALITY;
		lf->lfPitchAndFamily=DEFAULT_PITCH;
#ifdef	_UNICODE
#error	Implement unicode version of of vglCreateFont().
#else	/* _UNICODE */
		if(face_name)
			strCopy(lf->lfFaceName,face_name,LF_FACESIZE);
		else
			strCopy(lf->lfFaceName,"Arial",LF_FACESIZE);
#endif	/* _UNICODE */
	}
	fnt=CreateFontIndirect(lf);
	if(!fnt)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglCreateFont('%s',%d): CreateFontIndirect() failed "
			"(code %d).\n",dbs(face_name),polygonal,GetLastError());
#endif	/* CHECKED_BUILD */
		for(j=i-1;j>=0;j--)
			glDeleteLists(f->bases[i],256);
		free(f);
		return NULL;
	}
	f->hfont=fnt;
	old=SelectObject(dc,fnt);
	/* Convert quadratic B-spline glyphs to lines or polygons */
	format = polygonal ? WGL_FONT_POLYGONS : WGL_FONT_LINES;
	f->polygonal=polygonal;
	for(i=0;i<GL_FONT_LOD_LEVELS;i++)
	{
		wglUseFontOutlines(dc,0,256,f->bases[i],
			(GLfloat)font_lod_deviations[i],
			(GLfloat)extrusion,
			format,(LPGLYPHMETRICSFLOAT)f->gm);
	}
	/* Restore DC to it's previous state */
	SelectObject(dc,old);
	f->polygonal=polygonal;
	f->extrusion=extrusion;
	/* Other ops with global data */
	ctx=GetThreadData()->gl_current_slot_plus_1;
	/* Se succeded getting OpenGL DC of RC, so ctx probably not 0 */
	V_ASSERT(ctx);
	gls=gl_win[ctx-1];
	V_ASSERT(gls);
	WaitForSingleObject(gls->access,INFINITE);
	f->ctx_origin=ctx;
	f->ref=NULL;
	/* Done */
	len=vpsCount(&gls->fonts);
	for(slot=0;slot<len;slot++)
		if(!vpsGet(&gls->fonts,slot))
			break;
	if(slot<len)
		vpsSet(&gls->fonts,slot,f);
	else
		vpsPush(&gls->fonts,f);
	ReleaseMutex(gls->access);
	f->cache_index=slot;
	/* Initial drawing values for font */
	f->sp.adjustment=fa_center_center;
	vec3Set(f->sp.custom_adjustment,-0.5f,-0.5f,0);
	f->sp.scale=1;
	f->sp.fov=ANGLE_10;
	f->sp.use_fov=false;
	f->magic=GLOF_MAGIC;
	f->lod=0;	/* force LOD update */
	vglSetFontLOD(f,1);
	return f;
}

/*
 *	Selects font created in another window; returns quickly if font 
 *	already already exists or duplicates font if not exists.
 *	Assumes 'fnt' is valid font handle.
 */
gl_outline_font_p	V_LOCAL
vglDuplicateFont(gl_outline_font_t *origin)
{
	glw_t ctx;
	u32_t slot,len;
	gl_w_settings_t *gls;
	gl_outline_font_t *f;
	HFONT old,fnt;
	int i,j,format;
	LOGFONT *lf;
	ctx=GetThreadData()->gl_current_slot_plus_1;
	/* stx != origin->ctx -- we check this in vglSelectFont() */
	/* Check if font is already duplicated */
	V_ASSERT(ctx);
	gls=gl_win[ctx-1];
	V_ASSERT(gls);
	WaitForSingleObject(gls->access,INFINITE);
	len=vpsCount(&gls->fonts);
	for(slot=0;slot<len;slot++)
	{
		f=vpsGet(&gls->fonts,slot);
		if(f && f->ref==origin)
		{
			/* Already duplicated */
			ReleaseMutex(gls->access);
			return f;
		}
	}
	/* Duplicate font */
#ifdef	CHECKED_BUILD
	f=vglAllocFont(&lf,"vglCreateFont");
#else	/* CHECKED_BUILD */
	f=vglAllocFont(&lf);
#endif	/* CHECKED_BUILD */
	if(!f)
	{
		ReleaseMutex(gls->access);
		return NULL;
	}
	for(i=0;i<GL_FONT_LOD_LEVELS;i++)
	{
		f->bases[i]=glGenLists(256);
		if(!f->bases[i])
		{
			int j;
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"vglDuplicateFont(%X): glGenLists() failed, GL "
				"error %d.\n",origin,glGetError());
#endif		/* CHECKED_BUILD */
			for(j=i-1;j>=0;j--)
				glDeleteLists(f->bases[i],256);
			free(f);
			return NULL;
		}
	}
	/* Create Win32 font and select it into DC */
	v_copy(lf,origin->logfont,sizeof(LOGFONT));
	fnt=CreateFontIndirect(lf);
	if(!fnt)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglDuplicateFont(): CreateFontIndirect() failed "
			"(code %d).\n",GetLastError());
#endif	/* CHECKED_BUILD */
		for(j=i-1;j>=0;j--)
			glDeleteLists(f->bases[i],256);
		free(f);
		return NULL;
	}
	f->hfont=fnt;
	old=SelectObject(gls->dc,fnt);
	/* Convert quadratic B-spline glyphs to lines or polygons */
	format = origin->polygonal ? WGL_FONT_POLYGONS : WGL_FONT_LINES;
	f->polygonal=origin->polygonal;
	for(i=0;i<GL_FONT_LOD_LEVELS;i++)
	{
		wglUseFontOutlines(gls->dc,0,256,f->bases[i],
			(GLfloat)font_lod_deviations[i],
			(GLfloat)origin->extrusion,
			format,(LPGLYPHMETRICSFLOAT)f->gm);
	}
	/* Restore DC to it's previous state */
	SelectObject(gls->dc,old);
	f->magic=GLOF_MAGIC;
	f->extrusion=origin->extrusion;
	f->ctx_origin=ctx;
	f->ref=origin;
	f->line_width=origin->line_width;
	f->line_smooth=origin->line_smooth;
	f->lod=(1+origin->lod) % GL_FONT_LOD_LEVELS;	/* force LOD update */
	vglSetFontLOD(f,origin->lod);
	len=vpsCount(&gls->fonts);
	for(slot=0;slot<len;slot++)
		if(!vpsGet(&gls->fonts,slot))
			break;
	if(slot<len)
		vpsSet(&gls->fonts,slot,f);
	else
		vpsPush(&gls->fonts,f);
	f->cache_index=slot;
	ReleaseMutex(gls->access);
	return f;
}

/*
 *	Releases all objects and frees memory, but not updates state of 
 *	context.
 *	Assumes 'fnt' is valid font handle.
 */
void	V_LOCAL
vglReleaseFont(gl_outline_font_t *fnt)
{
	int i;
	if(!fnt)
		return;
	for(i=0;i<GL_FONT_LOD_LEVELS;i++)
		glDeleteLists(fnt->bases[i],256);
	if(fnt->hfont)
		DeleteObject(fnt->hfont);
	free(fnt);
}

/*
 *	Destroys given font.
 *	BUGBUG: there must be method to set to all references to 'fnt' in
 *	other fonts to NULL. When (you ?) implement such method, you can
 *	enable block which destroys all parent except root.
 */
void	V_GLOBAL
vglDeleteFont(gl_outline_font_t *fnt)
{
	u32_t i,len,ctx;
	gl_w_settings_t *gls;
	/* gl_outline_font_t *origin; */
	if(!VGL_IS_OUTLINE_FONT(fnt,"vglDeleteFont"))
		return;
	/* origin=fnt->ref; */
	ctx=fnt->ctx_origin;
	gls=gl_win[ctx-1];
	/* Was context already destroyed ? If yes, font handle is INVALID ! */
	if(!gls)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_WARNING,
			"vglDeleteFont(%X): parent context %d already destroyed.\n",
			fnt,ctx-1);
#endif	/* CHECKED_BUILD */
		return;
	}
	WaitForSingleObject(gls->access,INFINITE);
	len=vpsCount(&gls->fonts);
	for(i=0;i<len;i++)
		if(vpsGet(&gls->fonts,i)==fnt)
		{
			vpsSet(&gls->fonts,i,NULL);
			break;
		}
	ReleaseMutex(gls->access);
	if(i==len)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglDeleteFont(%X): font belongs to renrering context other than"
			" (%X)->ctx_origin.\n",fnt,fnt);
#endif	/* CHECKED_BUILD */
		return;
	}
	vglReleaseFont(fnt);
#if	0
	while(origin->ref)
	{
		fnt=origin;
		ctx=fnt->ctx_origin;
		gls=gl_win[ctx-1];
		if(!gls)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_WARNING,
				"vglDeleteFont(%X): parent context %d already destroyed.\n",
				fnt,ctx-1);
#endif	/* CHECKED_BUILD */
			return;
		}
		WaitForSingleObject(gls->access,INFINITE);
		len=vpsCount(&gls->fonts);
		for(i=0;i<len;i++)
			if(vpsGet(&gls->fonts,i)==fnt)
			{
				vpsSet(&gls->fonts,i,NULL);
				break;
			}
		ReleaseMutex(gls->access);
		if(i==len)
		{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"vglDeleteFont(%X): font belongs to renrering context "
				"other than (%X)->ctx_origin.\n",fnt,fnt);
#endif		/* CHECKED_BUILD */
				return;
		}
		vglReleaseFont(fnt);
		origin=fnt->ref;
	}
#endif	/* 0 */
	return;
}	

/*
 *	Adjust OpenGL state depending on font parameters.
 *	Assumes 'fnt' is valid.
 */
void	V_LOCAL
vglRealizeFont(gl_outline_font_t *fnt)
{
	glListBase(fnt->bases[fnt->lod]);
	if(!fnt->polygonal)
	{
		glLineWidth(fnt->line_width);
		if(fnt->line_smooth)
		{
			glEnable(GL_LINE_SMOOTH);
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		}
		else
		{
			glDisable(GL_BLEND);
			glDisable(GL_LINE_SMOOTH);
		}
	}
}

/*
 *	Selects given font. Returns true on success.
 *	Fonts created in other glw CAN be selected into other (also it takes
 *	very long time at first call).
 */
vboolean	V_GLOBAL
vglSelectFont(gl_outline_font_t *fnt)
{
	glw_t slot;
	if(!VGL_IS_OUTLINE_FONT(fnt,"vglSelectFont"))
		return false;
	if(!fnt->ctx_origin)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"vglSelectFont(): font was not properly initialized.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	slot=GetThreadData()->gl_current_slot_plus_1;
	if(!slot)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"vglSelectFont(): no one context have been made current.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	if(slot!=fnt->ctx_origin)
	{
		/* TODO: provide mechanizm to aquire reference font.
		** For now, vglDuplicateFont() will be stupid called again
		** and again. Also 2nd and subsequent calls will return
		** quickly, this way isn't too good.
		*/
		/*fnt=*/vglDuplicateFont(fnt);
		/*if(!fnt)
			return false;*/
	}
	/* WaitForSingleObject(); */
	gl_win[slot-1]->font=fnt;
	/* ReleaseMutex(); */
	vglRealizeFont(fnt);
	return true;
}

/*
 *	Returns currently selected font or NULL.
 */
gl_outline_font_p	V_GLOBAL
vglGetCurrentFont(void)
{
	glw_t slot;
	gl_outline_font_p result;
	slot=GetThreadData()->gl_current_slot_plus_1;
	if(!slot)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"vglGetCurrentFont(): current context is NULL.\n");
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	/* WaitForSingleObject(); */
	result=gl_win[slot-1]->font;
	/* ReleaseMutex(); */
	return result;
}

/*
 *	Returns text x-extent and height from the base line.
 *	If fnt==NULL returns extents for current selected font.
 */
void	V_GLOBAL
vglGetTextExtents(char *str,vec_t *width,vec_t *height)
{
	LPGLYPHMETRICSFLOAT gmf;
	vec_t w,h,hc;
	unsigned char *ptr;
	gl_outline_font_t *fnt=vglGetCurrentFont();
	if(!fnt)
		return;
	if(!str)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglGetTextExtents(%X,[NULL]): incorrect parameter.\n",fnt);
#endif	/* CHECKED_BUILD */
		return;
	}
	gmf=(LPGLYPHMETRICSFLOAT)fnt->gm;
	w=h=0;
	for(ptr=(unsigned char*)str;*ptr;ptr++)
	{
		hc=gmf[*ptr].gmfptGlyphOrigin.y;
		if(h<hc)
			h=hc;
		w+=gmf[*ptr].gmfCellIncX;
	}
	if(width)
		*width=w;
	if(height)
		*height=h;
}

/*
 *	Prints string.
 */
void	V_GLOBAL
vglPuts(char *str)
{
	if(str)
		glCallLists(strlen(str),GL_UNSIGNED_BYTE,str);
#ifdef	CHECKED_BUILD
	else
		conlPuts(LOG_ERROR,
			"vglPuts([NULL]): incorrect parameter.\n");
#endif	/* CHECKED_BUILD */
}

/*
 *	OpenGL "printf()".
 */
void	V_GLOBAL_VA
vglPrintf(const char *fmt,...)
{
	va_list marker;
	char 	*ptr;
	char 	fmt_buffer[MAX_FORMAT_BUFFER];
	if(fmt)
	{
		ptr=fmt_buffer;
		va_start(marker,fmt);
		vsprintf(ptr,fmt,marker);
		va_end(marker);
		strStripTrail(ptr);
		glCallLists(strlen(ptr),GL_UNSIGNED_BYTE,ptr);
	}
#ifdef	CHECKED_BUILD
	else
		conlPuts(LOG_ERROR,
			"vglPrintf([NULL]): incorrect parameter.\n");
#endif	/* CHECKED_BUILD */
}

/*
 *	Adjusts font LOD depending character size.
 */
void	V_GLOBAL
vglSetFontLOD(gl_outline_font_t *fnt,u32_t level_of_detail)
{
	fnt=VGL_GET_FONT_FOR_OP(fnt,"vglSetFontLOD");
	if(!fnt)
		return;
	if(level_of_detail>(u32_t)(GL_FONT_LOD_LEVELS-1))
		level_of_detail=(u32_t)(GL_FONT_LOD_LEVELS-1);
	if(fnt->lod!=level_of_detail)
	{
		fnt->lod=level_of_detail;
		switch(level_of_detail)
		{
			case 0:
				fnt->line_smooth=1;
				fnt->line_width=3.0;
				break;
			case 1:
				fnt->line_smooth=1;
				fnt->line_width=2.0;
				break;
			case 2:
				fnt->line_smooth=0;
				fnt->line_width=1.0;
				break;
		}
	}
	vglRealizeFont(fnt);
}

/*
 *	There are 9 useful pivot point positions for text:
 *			left+top;			center+top;				right+top;
 *			left+center;		center+center;			right+center;
 *			left+bottom;		center+bottom;			right+bottom.
 */
V_LOCAL	vec3_t	label_origins[9]=
{
	{0,-1,0},	{-0.5,-1,0},	{-1,-1,0},
	{0,-0.5,0},	{-0.5,-0.5,0},	{-1,-0.5,0},
	{0,0,0},	{-0.5,0,0},		{-1,0,0}
};


/*
	TODO: vglFontRasterQuality / vglFontMetafileQuality ...
*/
/*
V_LOCAL	vec_t	ref_flod_limits[GL_FONT_LOD_LEVELS]={0.25f,0.1f,0};
*/
V_LOCAL	vec_t	wmf_flod_limits[GL_FONT_LOD_LEVELS]={0.05f,0.02f,0};

V_LOCAL	vec_p	flod_limits=wmf_flod_limits;

/*
 *	Sets LOD based on scale and view size (font height is about 1.0).
 *	Used in vglDrawString() / vglDrawText().
 */
void	V_LOCAL
vglSetFontLODFromScale(gl_outline_font_t *fnt,vec_t scale)
{
	int i;
	scale=v_abs(scale);
	for(i=0;i<GL_FONT_LOD_LEVELS;i++)
		if(scale>flod_limits[i])
		{
			vglSetFontLOD(fnt,i);
			return;
		}
}

/*
 *	Loads appropriate matrix on modelview TOS.
 *	Assumes all args are valid.
 */
void	V_LOCAL
vglSetupStringTransform(
	gl_outline_font_t *fnt,
	vec3_t origin,
	cam_t *cam,
	char *text)
{
	vec3_t t,v;
	vec_t w,h,scale,lod_scale,t_len;
	glLoadIdentity();
	mat3x4VecMul(t,cam->transform,origin);
	t_len=vec3Abs(t);
	if(!IsZero(t_len))
	{
    	if(fnt->sp.use_fov)
    	{
    		lod_scale=(vec_t)tan(fnt->sp.fov * .5);
    		scale = lod_scale * t_len;
    	}
    	else
    	{
    		scale=fnt->sp.scale;
    		lod_scale = scale / t_len;
    	}
	}
	vglGetTextExtents(text,&w,&h);
	if((unsigned)(fnt->sp.adjustment) > (unsigned)fa_right_bottom)
		vec3Negate(v,fnt->sp.custom_adjustment);
	else
		vec3Copy(v,label_origins[fnt->sp.adjustment]);
	v[0]*=w;
	v[1]*=h;
	v[2]=fnt->extrusion * .5f;
	vec3Scale(v,v,scale);
	vglSetFontLODFromScale(fnt,lod_scale);
	vec3Add(t,t,v);
	glTranslatef(t[0],t[1],t[2]);
	glScalef(scale,scale,scale);
}

/*
 *		vglSetFontAdjustment
 */
void	V_GLOBAL
vglSetFontAdjustment(gl_outline_font_t *fnt,gl_font_adjustment_t adjustment)
{
	if(!VGL_IS_OUTLINE_FONT(fnt,"vglSetFontAdjustment"))
		return;
	fnt->sp.adjustment=adjustment;
}

/*
 *	custom_adjustment contain coordinates for text pivot point;
 *	{0,0} mean "left,bottom", {1,1} mean "right,top".
 */
void	V_GLOBAL
vglSetFontCustomAdjustment(gl_outline_font_t *fnt,vec3_t custom_adjustment)
{
	if(!VGL_IS_OUTLINE_FONT(fnt,"vglSetFontCustomAdjustment"))
		return;
	if(!custom_adjustment)
	{
		return;
	}
	fnt->sp.adjustment=fa_custom;
	vec3Copy(fnt->sp.custom_adjustment,custom_adjustment);
}

/*
 *	vglSetFontScale()
 */
void	V_GLOBAL
vglSetFontScale(gl_outline_font_t *fnt,vec_t scale)
{
	if(!VGL_IS_OUTLINE_FONT(fnt,"vglSetFontScale"))
		return;
	fnt->sp.use_fov=false;
	fnt->sp.scale=scale;
}

/*
 *	vglSetFontFOV
 */
void	V_GLOBAL
vglSetFontFOV(gl_outline_font_t *fnt,vec_t fov)
{
	if(!VGL_IS_OUTLINE_FONT(fnt,"vglSetFontFOV"))
		return;
	fnt->sp.use_fov=true;
	fnt->sp.fov=fov;
}

/*
 *	vglDrawString()
 */
void	V_GLOBAL
vglDrawString(
	vec3_t origin,
	cam_t *cam,
	char *text)
{
	gl_outline_font_t *fnt=vglGetCurrentFont();
	if(!fnt)
	{
		/* this place can produce a lot of errors, so we set log level
		 * to LOG_TRACE only */
#ifdef	CHECKED_BUILD
		conlPuts(LOG_TRACE,"vglDrawString(): no current font.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	glPushMatrix();
	vglSetupStringTransform(fnt,origin,cam,text);
	vglPuts(text);
	glPopMatrix();
}

/*</pre></strong></body></html>*/
/*EOF*/
