#ifndef	V_MATRIX_HPP
#define	V_MATRIX_HPP

#ifndef		__cplusplus
#error This file contains C++ code !
#else

class matrix_c
{
  private:
  matrix_p		m;
  public:
				matrix_c()
				{
					m=NULL;
				}
				matrix_c(int rows,int cols)
				{
					m=matAlloc(rows,cols);
				}
				matrix_c(matrix_t *m)
				{
					this->m=matDuplicate(m,true);
				}
				matrix_c(const matrix_c &m)
				{
					this->m=matDuplicate(m.m,true);
				}
				matrix_c(int rows,int cols,mvec_t *data)
				{
					m=matAlloc(rows,cols);
					v_copydw(m->data,data,rows*cols*(sizeof(mvec_t)>>2));
				}
				~matrix_c()
				{
					if(m)
						matFree(m);
				}
	int			rows()
				{
					return m->rows;
				}
	int			cols()
				{
					return m->cols;
				}
	mvec_t		*operator[](int r)
				{
					return m->cell[r];
				}
	void		rearrange()
				{
					m=matDuplicate(m,false);
				}
	void		recreate(int r,int c)
				{
					if(m)
						matFree(m);
					m=matAlloc(r,c);
				}
	operator	mvec_t**()
				{
					return m->cell;
				}
	operator	mvec_t*()
				{
					return m->data;
				}
	void		copy(matrix_t *from)
				{
					matCopy(m,from);
				}
	void		copy(matrix_t &from)
				{
					matCopy(m,&from);
				}
	void		copy(const matrix_c &from)
				{
					matCopy(m,from.m);
				}
	void		copy(mvec_t *data)
				{
					v_copydw(m->data,data,
							 m->rows*m->cols*sizeof(mvec_t)>>2);
				}
	void		zero()
				{
					matZero(m);
				}
	void        unit()
				{
					matUnit(m);
				}
	void		add(matrix_t *m)
				{
					matAdd(this->m,this->m,m);
				}
	void		add(matrix_t &m)
				{
					matAdd(this->m,this->m,&m);
				}
	void		add(const matrix_c &m)
				{
					matAdd(this->m,this->m,m.m);
				}
	void		sub(matrix_t *m)
				{
					matSub(this->m,this->m,m);
				}
	void		sub(matrix_t &m)
				{
					matSub(this->m,this->m,&m);
				}
	void		sub(const matrix_c &m)
				{
					matSub(this->m,this->m,m.m);
				}
	void		negate()
				{
					matNegate(this->m,this->m);
				}
	void		scale(mvec_t x)
				{
					matScale(m,m,x);
				}
	vector_c	operator*(const vector_c &v)
				{
					return vector_c(matByVec(NULL,this->m,v.v));
				}
	matrix_c	operator*(const matrix_c &m)
				{
					return matrix_c(matByMat(NULL,this->m,m.m));
				}
	mvec_t		det()
				{
					return matDeterm(m,true);
				}
	void		inverse()
				{
					m=matInverse(m,false);
				}
	matrix_c	operator!()
				{
					return matrix_c(matInverse(m,true));
				}
	vector_c	solve(vector_c &v)
				{
					return vector_c(matSolve(m,v.v,true));
				}
	vboolean	compare(matrix_t *m)
				{
					return matCompare(this->m,m);
				}
	vboolean	compare(matrix_t &m)
				{
					return matCompare(this->m,&m);
				}
	vboolean	compare(const matrix_c &m)
				{
					return matCompare(this->m,m.m);
				}
	void		minelem(int *row,int *col)
				{
					matMin(m,row,col);
				}
	void		maxelem(int *row,int *col)
				{
					matMax(m,row,col);
				}
	matrix_c	operator=(matrix_c &m)
				{
					return matrix_c(m);
				}
	matrix_c	operator-()
				{
					matrix_c m1=*this;m1.negate();return m1;
				}
	vboolean	operator==(matrix_c &m)
				{
					return matCompare(this->m,m.m);
				}
	vboolean	operator!=(const matrix_c &m)
				{
					return !matCompare(this->m,m.m);
				}
	matrix_c&	operator+=(const matrix_c &m)
				{
					add(m);return *this;
				}
	matrix_c&	operator-=(const matrix_c &m)
				{
					sub(m);return *this;
				}
	matrix_c	operator+(const matrix_c &m)
				{
					return matrix_c(matAdd(
							matDuplicate(this->m,true),this->m,m.m));
				}
	matrix_c	operator-(const  matrix_c &m) 
				{
					return matrix_c(matSub(
							matDuplicate(this->m,true),this->m,m.m));
				}
	matrix_c&	operator*=(mvec_t x)
				{
					matScale(m,m,x);return *this;
				}
	matrix_c	operator*(mvec_t x)
				{
					matrix_c m1(*this);
					matScale(m1.m,m1.m,x);
					return m1;
				}
	matrix_c&	operator/=(mvec_t x)
				{
					VASSERT(!IsZero(x));
					matScale(m,m,1/x);
					return *this;
				}
	matrix_c	operator/(mvec_t x)
				{
					VASSERT(!IsZero(x));
					matrix_c m1(*this);
					matScale(m1.m,m1.m,1/x);
					return m1;
				}
	void		con_print()
				{
					for(int j,i=0;i<m->rows;i++)
					{
						for(j=0;j<m->cols;j++)
					  	con_printf("%g\t",m->cell[i][j]);
						con_puts("\n");
					}
				}
};

#endif	/* __cpluplus */

#endif	/* V_MATRIX_HPP */
