#ifdef __cplusplus

class matrix_c;

class vector_c
{
  protected:
  vector_t        *v;
  friend          class matrix_c;
  public:
                  vector_c() {v=NULL;}
                  vector_c(int size) {v=vecAlloc(size);}
                  vector_c(const vector_c &v) {this->v=vecDuplicate(v.v,true);}
                  vector_c(vector_t *v) {this->v=v;}
                  vector_c(int size, vec_t *data)
                  {
                    v=vecAlloc(size);
                    v_copydw(v->elements,data,size*(sizeof(vec_t)>>2));
                  }
                  ~vector_c() {vecFree(v);}
  void            recreate(int size) {if(v) vecFree(v);v=vecAlloc(size);}
  void            copy(vector_t *from) {vecCopy(v,from);}
  void            copy(vector_t &from) {vecCopy(v,&from);}
  void            copy(const vector_c &from) {vecCopy(v,from.v);}
  void            copy(vec_t *data)
                  {
                    v_copydw(v->elements,data,v->size*sizeof(vec_t)>>2);
                  }
  void            zero() {vecZero(v);}                
  vboolean        unit() {return vecUnit(v);}
  vec_t        abs()  {return vecAbs(v);}
  void            negate() {vecNegate(v);}
  void            add(vector_t *v) {vecAdd(this->v,this->v,v);}
  void            add(vector_t &v) {vecAdd(this->v,this->v,&v);}
  void            add(const vector_c &v) {vecAdd(this->v,this->v,v.v);}
  void            sub(vector_t *v) {vecSub(this->v,this->v,v);}
  void            sub(vector_t &v) {vecSub(this->v,this->v,&v);}
  void            sub(const vector_c &v) {vecSub(this->v,this->v,v.v);}
  void            scale(vec_t x) {vecScale(v,v,x);}
  vec_t        dot(vector_t *v) {return vecDot(this->v,v);}
  vec_t        dot(vector_t &v) {return vecDot(this->v,&v);}
  vec_t        dot(const vector_c &v) {return vecDot(this->v,v.v);}
  vboolean        compare(vector_t *v) {return vecCompare(this->v,v);}
  vboolean        compare(vector_t &v) {return vecCompare(this->v,&v);}
  vboolean        compare(const vector_c &v) {return vecCompare(this->v,v.v);}
  int             minelem()  {return vecMin(v);}
  int             maxelem()  {return vecMax(v);}
  vec_t        &operator[](int i) {return v->elements[i];}
  int             size()  {return v->size;}
  int             operator!()  {return size();}
  operator        vec_t*()  {return v->elements; }
  vector_c        operator=(vector_c &v)  {return vector_c(v);}
//  void            operator=(vector_c &v)  {vecCopy(this->v,v.v);}
  vector_c        operator-() {vector_c v1=*this;v1.negate();return v1;}
  vboolean        operator==(const vector_c &v) {return vecCompare(this->v,v.v);}
  vboolean        operator!=(const vector_c &v) {return !vecCompare(this->v,v.v);}
  vector_c&       operator+=(const vector_c &v) {add(v);return *this;}
  vector_c&       operator-=(const vector_c &v) {sub(v);return *this;}
  vector_c        operator+(const vector_c &v) 
                  {
                    return vector_c(vecAdd(vecDuplicate(this->v,true),this->v,v.v));
                  }
  vector_c        operator-( vector_c &v) 
                  {
                    return vector_c(vecSub(vecDuplicate(this->v,true),this->v,v.v));
                  }
  vec_t        operator*(vector_c &v) {return dot(v);}
  vector_c&       operator*=(vec_t x) {vecScale(v,v,x);return *this;}
  vector_c        operator*(vec_t x) {vector_c v1(*this);vecScale(v1.v,v1.v,x);return v1;}
  vector_c&       operator/=(vec_t x) {V_ASSERT(!IsZero(x));vecScale(v,v,1/x);return *this;}
  vector_c        operator/(vec_t x) {V_ASSERT(!IsZero(x));vector_c v1(*this);vecScale(v1.v,v1.v,1/x);return v1;}
#ifdef DEBUG
  void            show() {for(int i=0;i<v->size;printf("%g ",this->v->elements[i++]));puts("");}
#endif
};  

